# (c) Copyright 2007-2011 Synapse Wireless, Inc.
"""
Serial port utility functions for getting lists of specific devices
connected to the system.
"""

import logging, sys, datetime, ctypes, re, os
log = logging.getLogger("snaplib.%s" % (__name__))
if sys.platform == 'win32':
    from ctypes import wintypes

# If your getting an error at the following lines, make sure your PYTHONPATH is set
from apy import PostThread
from apy import EventScheduler

if sys.platform.startswith("win"):
    import usbxserialutil, usbxserialwin32
import ftd2xxserialwin32, ftd2xxserialutil
import PyserialDriver, serial

PROBE_VER_SNAP1 = 1
PROBE_VER_SNAP2 = 2
PROBE_VER_SNYF1 = 1000

PROBE_RESULT_ERROR = 0 #General error
PROBE_RESULT_OPEN_ERROR = 1
PROBE_RESULT_SEND_ERROR = 2
PROBE_RESULT_READ_ERROR = 3
PROBE_RESULT_PERMISSIONS_ERROR = 4
PROBE_RESULT_FOUND = 100

class NoMorePortsError(Exception):
    """Raised by the ComPortScanner when the end of the list is reached"""
    pass

def getRs232Ports(maxPortNum=256):
    """
    Get the list of RS232 ports
    
    This function returns the list of RS232 ports available on the
    system.
    
    Args:
      maxPortNum: [256] The maximum number of ports to return
      
    Returns:
      A list of RS232 port names
    """
    if sys.platform == 'win32':
        port_nums = range(maxPortNum)
        QueryDosDevice = ctypes.windll.kernel32.QueryDosDeviceW
        LPTSTR = wintypes.LPWSTR
        ucchMax = 65535
        lpTargetPath = ctypes.create_unicode_buffer(ucchMax)

        QueryDosDevice.argtypes = [LPTSTR, LPTSTR, wintypes.DWORD]
        QueryDosDevice.restype = wintypes.DWORD

        numRtnChars = QueryDosDevice(None, lpTargetPath, ucchMax)

        devices = lpTargetPath[:numRtnChars].split('\x00')
        port_nums = [int(dev[3:])-1 for dev in devices if len(dev) > 3 and dev[:3] == 'COM']
        port_nums.sort()
    else:
        port_nums = []
        for filename in os.listdir('/dev'):
            if (filename.startswith('ttyS') or 
                filename.startswith('ttyUSB') or 
                filename.startswith('tty.SLAB_USBtoUART') or
                filename.startswith('tty.usbserial')):
                port_nums.append("/dev/%s" % (filename))        
    return port_nums

def getUsbPorts(pathToLibrary=None):
    """
    (Windows only) Get the list of USB ports
    
    This function returns the list of USB ports on a windows platform. For all other
    platform it returns an empty list.
    
    Args:
      pathToLibrary: [None] path to the windows DLLs
      
    Returns:
      List of windows USB ports
    """
    if sys.platform.startswith("win"):
        return range(usbxserialutil.UsbxSerialBase(pathToLibrary=pathToLibrary).numDevices)
    return []

def getSnapStickPorts(pathToLibrary=None):
    """
    Get the list of SS200s
    
    This function returns the list of SS200s plugged into the system.
    
    Args:
      pathToLibrary: [None] path to the windows DLLs
      
    Returns:
      List of SS200s
    """
    return range(ftd2xxserialutil.Ftd2xxSerialBase(pathToLibrary=pathToLibrary, pid=0x9090).numDevices)

def getTwrRfSnapPorts(pathToLibrary=None):
    """
    Get the list of TWR-RF_SNAPs
    
    This function returns the list of Freescale RF Wireless Mesh Networking Tower System Modules 
    plugged into the system.
    
    Args:
      pathToLibrary: [None] path to the windows DLLs
      
    Returns:
      List of towers
    """
    return range(ftd2xxserialutil.Ftd2xxSerialBase(pathToLibrary=pathToLibrary, pid=0x9094).numDevices)

def getMB851SnapPorts(pathToLibrary=None):
    """
    Get the list of DiZiC Eval Boards
    
    This function returns the list of DiZiC boards plugged into the system.
    
    Args:
      pathToLibrary: [None] path to the windows DLLs
      
    Returns:
      List of DiZiC eval boards
    """
    return range(ftd2xxserialutil.Ftd2xxSerialBase(pathToLibrary=pathToLibrary, pid=0x6001, requiredDesc='MB851').numDevices)


class ComPortScanner(object):
    """
    Scans the system for serial ports
    
    This class works with the apy event scheduler to scan for serial ports.    
    """
    class ProbeResult(object):
        """
        The result from one ComPortScanner probe attempt
        """
        def __init__(self, type, port, probeVer, result=PROBE_RESULT_ERROR):
            """
            Create the probe result.
            Args:
              type: the port type
              port: the port
              probeVer: the probe version (snap version)
              result: [PROBE_RESULT_ERROR] the result of the probe
            """
            self.type = type
            self.port = port
            self.result = result
            self.portStr = None
            self.probeVer = probeVer
            self.isLast = False

        def __str__(self):
            portStr = self.portStr
            errStr = "General Error"
            if self.result == PROBE_RESULT_OPEN_ERROR:
                errStr = "Unable to open"
            elif self.result == PROBE_RESULT_SEND_ERROR:
                errStr = "Unable to send"
            elif self.result == PROBE_RESULT_READ_ERROR:
                errStr = "Read error"
            elif self.result == PROBE_RESULT_FOUND:
                errStr = "Found"

            if portStr is None:
                portStr = "%i-%i" % (self.type, self.port)

            return "%s: %s" % (portStr, errStr)

    def __init__(self, evScheduler, pathToUsbLibrary='.'):
        """
        Create a ComPortScanner
        
        This method creates a new ComPortScanner object.
        Args:
          evScheduler: the apy scheduler that will make calls on the object
          pathToUsbLibrary: [.] the path to the windows DLLs
        """
        #assert isinstance(evScheduler, EventScheduler.EventScheduler)
        self.evScheduler = evScheduler
        self.pathToUsbLibrary = pathToUsbLibrary
        if self.pathToUsbLibrary is None:
            self.pathToUsbLibrary = os.path.dirname( os.path.realpath( __file__ ) )
        self._portsToScan = []
        self._postThread = PostThread.PostThread()
        self._resultCallbacks = []
        self.probe1Expr = re.compile(r'\s*ATNA\?\s+([A-F0-9]{1,5})\s+OK')
        self.probe2Expr = re.compile("\\xc0\\x01\\xff\\xc1")
        self.snyf1Expr = re.compile(r'\xc9\x01\xff\xc1')

    def callbackProbe(self, probeVer, type, port, timeout=1000):
        """
        This method posts the probe to the thread.
        Args:
          probeVer: snap version of the probe
          type: the probe type
          port: the port
          timeout: [1000] ms to wait
        """
        self._postThread.post(self.evScheduler, self.probe, [probeVer, type, port]).addCallbacks(self.onResult)

    def next(self):
        try:
            #We could have just posted everything to the thread since it queues but
            #this way enables us to control if we to continue
            (probeVer, type, portNum) = self._portsToScan.pop(0)
            self.callbackProbe(probeVer, type, portNum)
        except IndexError:
            raise NoMorePortsError("No more ports to scan")
        except:
            log.exception("Error when trying to scan next port:")

    def onResult(self, result):
        log.info(result)
        if len(self._portsToScan) == 0:
            result.isLast = True

        for callback in self._resultCallbacks:
            try:
                callback(result)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("An error occurred when notifying a callback of a probe result")

    def probe(self, probeVer, type, port, timeout=1000):
        """Sends probe to the specified port, blocks until it has as a result"""
        result = ComPortScanner.ProbeResult(type, port, probeVer)

        if type == PyserialDriver.PyserialWrapper.TYPE_USBXSERIAL:
            _serial = usbxserialwin32.UsbxSerial(pathToLibrary=self.pathToUsbLibrary)
        elif type == PyserialDriver.PyserialWrapper.TYPE_PYSERIAL:
            _serial = serial.Serial()
            timeout = timeout/1000
        elif type == PyserialDriver.PyserialWrapper.TYPE_SNAPSTICK:
            _serial = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=self.pathToUsbLibrary, pid=0x9090)
        elif type == PyserialDriver.PyserialWrapper.TYPE_TWR_RF_SNAP:
            _serial = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=self.pathToUsbLibrary, pid=0x9094)
        elif type == PyserialDriver.PyserialWrapper.TYPE_DIZIC_MB851:
            _serial = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=self.pathToUsbLibrary, pid=0x6001, requiredDesc='MB851')
        else:
            result.result = PROBE_RESULT_ERROR
            return result

        if probeVer == PROBE_VER_SNAP1:
            sendCmd = "+++AT\rATNET0\rATNA?\r"
            expr = self.probe1Expr
            readLen = 50
            _serial.rtscts = 1
            _serial.baudrate = 9600
        elif probeVer == PROBE_VER_SNAP2 or probeVer == PROBE_VER_SNYF1:
            sendCmd = "\xc0\x01\x00\xff\xc1"
            expr = self.probe2Expr
            readLen = 4
            _serial.rtscts = 0
            if probeVer == PROBE_VER_SNYF1:
                _serial.baudrate = 115200
            else:
                _serial.baudrate = 38400
        else:
            result.result = PROBE_RESULT_ERROR
            return result

        try:
            _serial.port = port
            _serial.writeTimeout = timeout

            if __debug__:
                log.debug("probing V%i %s... type: %s port: %s" % (probeVer, _serial.portstr, type, port))

            result.portStr = _serial.portstr

            if (type == PyserialDriver.PyserialWrapper.TYPE_PYSERIAL and
                (not sys.platform.startswith("win") and 
                (not os.access(port, os.R_OK) or not os.access(port, os.W_OK)))):
                result.result = PROBE_RESULT_PERMISSIONS_ERROR
                return result

            _serial.open()
            _serial.timeout = timeout
            _serial.flushInput()
            _serial.flushOutput()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self._safeClose(_serial)
            result.result = PROBE_RESULT_OPEN_ERROR
            return result

        try:
            bytesSent = _serial.write(sendCmd)
            #log.debug("Sent SNAP2 probe") 
            #if bytesSent != len(sendCmd):
                #raise Exception("Did not send all bytes")
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self._safeClose(_serial)
            result.result = PROBE_RESULT_SEND_ERROR
            return result

        try:
            resp = _serial.read(readLen)
            if not expr.search(resp):
                raise Exception("Did not receive expected response")
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self._safeClose(_serial)
            result.result = PROBE_RESULT_READ_ERROR
            return result

        self._safeClose(_serial)
        result.result = PROBE_RESULT_FOUND
        return result

    def registerResultCallback(self, callback):
        """
        Register for callbacks when the the PyserialDriver is closed

        callback -- A callable object expecting a ProbeResult parameter
        """
        if callable(callback):
            self._resultCallbacks.append(callback)

    def _safeClose(self, _serial):
        try:
            _serial.close()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            log.exception("Close Exception:")

    def start(self, 
              portTypes=[PyserialDriver.PyserialWrapper.TYPE_DIZIC_MB851,
                         PyserialDriver.PyserialWrapper.TYPE_USBXSERIAL, 
                         PyserialDriver.PyserialWrapper.TYPE_SNAPSTICK, 
                         PyserialDriver.PyserialWrapper.TYPE_PYSERIAL,
                         PyserialDriver.PyserialWrapper.TYPE_TWR_RF_SNAP], 
              probeVer=PROBE_VER_SNAP2):
        if probeVer not in [PROBE_VER_SNAP1, PROBE_VER_SNAP2, PROBE_VER_SNYF1]:
            raise Exception("Unknown probe version")

        self._portsToScan = []
        for type in portTypes:
            if type == PyserialDriver.PyserialWrapper.TYPE_USBXSERIAL:
                for portNum in getUsbPorts(self.pathToUsbLibrary):
                    self._portsToScan.append((probeVer, type, portNum))
            elif type == PyserialDriver.PyserialWrapper.TYPE_PYSERIAL:
                for portNum in getRs232Ports():
                    self._portsToScan.append((probeVer, type, portNum))
            elif type == PyserialDriver.PyserialWrapper.TYPE_SNAPSTICK:
                for portNum in getSnapStickPorts(self.pathToUsbLibrary):
                    self._portsToScan.append((probeVer, type, portNum))
            elif type == PyserialDriver.PyserialWrapper.TYPE_TWR_RF_SNAP:
                for portNum in getTwrRfSnapPorts(self.pathToUsbLibrary):
                    self._portsToScan.append((probeVer, type, portNum))
            elif type == PyserialDriver.PyserialWrapper.TYPE_DIZIC_MB851:
                for portNum in getMB851SnapPorts(self.pathToUsbLibrary):
                    self._portsToScan.append((probeVer, type, portNum))

        numPorts = len(self._portsToScan)
        try:
            self.next()
        except NoMorePortsError:
            raise
        return numPorts

    def _testCallback(self, result):
        self.next()

if  __name__ == '__main__':
    import time

    logging.basicConfig(level=logging.DEBUG,
                      format='%(asctime)s:%(msecs)03d %(levelname)-8s %(name)-8s %(message)s',
                      datefmt='%H:%M:%S')
    evQ = EventScheduler.EventScheduler()

    scanner = ComPortScanner(evQ)
    scanner.registerResultCallback(scanner._testCallback)
    scanner.start()

    while True:
        evQ.poll()
        time.sleep(0.1)
