# (c) Copyright 2007-2013 Synapse Wireless, Inc.


from __future__ import with_statement

import logging
import sys
import os
import time
import timeit
import threading
import ctypes
#import pprint

try:
    import serial
    from serial import serialutil
except (ImportError, SystemError):
    serial = None

import ftd2xxserialwin32
if sys.platform.startswith("win"):
  import usbxserialwin32
  #import pywintypes
elif sys.platform.startswith("linux") or sys.platform == 'darwin':
  import select, errno

import btle_serial

log = logging.getLogger("snaplib.%s" % (__name__))

MODULE_PATH = os.path.dirname(__file__) # The path where this module resides


# Monkey Patch pyserial to not error on when it didn't write all the bytes
def write(self, data):
  """Output the given string over the serial port."""
  if not self.hComPort: raise serial.portNotOpenError
  #~ if not isinstance(data, (bytes, bytearray)):
      #~ raise TypeError('expected %s or bytearray, got %s' % (bytes, type(data)))
  # convert data (needed in case of memoryview instance: Py 3.1 io lib), ctypes doesn't like memoryview
  data = serialutil.to_bytes(data)
  if data:
    #~ win32event.ResetEvent(self._overlappedWrite.hEvent)
    n = serial.win32.DWORD()
    err = serial.win32.WriteFile(self.hComPort, data, len(data), ctypes.byref(n), self._overlappedWrite)
    if not err and serial.win32.GetLastError() != serial.win32.ERROR_IO_PENDING:
      raise serial.SerialException("WriteFile failed (%r)" % ctypes.WinError())
    # Wait for the write to complete.
    #~ win32.WaitForSingleObject(self._overlappedWrite.hEvent, win32.INFINITE)
    err = serial.win32.GetOverlappedResult(self.hComPort, self._overlappedWrite, ctypes.byref(n), True)
    #if n.value != len(data):
        #raise writeTimeoutError
    return n.value
  else:
    return 0

def writePosix(self, data):
    """Output the given string over the serial port."""
    if not self._isOpen: raise serial.portNotOpenError
    d = serialutil.to_bytes(data)
    tx_len = len(d)

    try:
        if self._writeTimeout is not None and self._writeTimeout > 0:
            _,ready,_ = select.select([],[self.fd],[], self._writeTimeout)
            if not ready:
                return 0
        return os.write(self.fd, data)
    except OSError, v:
      if v.errno != errno.EAGAIN:
        raise serial.SerialException('write failed: %s' % (v,))
    return 0

# Monkey Patch pyserial to ensure that it cleans up better (having problems when system goes into standby)
def close(self):
  """Close port"""
  if self._isOpen:
    if self.hComPort:
      try:
        # Restore original timeout values:
        serial.win32.SetCommTimeouts(self.hComPort, self._orgTimeouts)
      except (KeyboardInterrupt, SystemExit):
        raise
      except:
        # ignore errors. can happen for unplugged USB serial devices
        pass

      try:
        # Close COM-Port:
        serial.win32.CloseHandle(self.hComPort)
        if self._overlappedRead is not None:
          serial.win32.CloseHandle(self._overlappedRead.hEvent)
          self._overlappedRead = None
        if self._overlappedWrite is not None:
          serial.win32.CloseHandle(self._overlappedWrite.hEvent)
          self._overlappedWrite = None
        self.hComPort = None
      except (KeyboardInterrupt, SystemExit):
        raise
      except:
        pass
      self.hComPort = None
    self._isOpen = False

# Monkey Patch pyserial read to catch windows specific error
def read(self, size=1):
  try:
    return serial.serialwin32.Serial._read(self, size)
  except pywintypes.error, err:
    if err.args[0] != 995: # ERROR_OPERATION_ABORTED
      raise err
    return ''

if serial:
    if sys.platform.startswith("win"):
      serial.serialwin32.Serial.close = close
      serial.serialwin32.Serial.write = write
      #serial.serialwin32.Serial._read = serial.serialwin32.Serial.read
      #serial.serialwin32.Serial.read = read
    elif sys.platform.startswith("linux") or sys.platform == 'darwin':
      serial.serialposix.Serial.write = writePosix


class SerialOpenException(Exception):
  pass


class PyserialWrapper(object):
  """
  Wraps around a pyserial like object to drive reading and writting
  """
  TYPE_SIMULATOR = 0
  TYPE_PYSERIAL = 1
  TYPE_USBXSERIAL = 2
  TYPE_LOOPBACK = 3
  TYPE_GATEWAY = 4
  TYPE_SNAPSTICK = 5
  TYPE_TWR_RF_SNAP = 6
  TYPE_DIZIC_MB851 = 7
  TYPE_BTLE = 8
  MAX_READ_BYTES = 256
  BAUDRATE = 38400 #The default baudrate if None is provided

  def __init__(self, outputType=None, port=0, openCallback=None, dllPath=MODULE_PATH, baudrate=None):
    """
    outputType -- The underlying pyserial like object to create (default TYPE_SIMULATOR)
    port -- The port identifier to pass to the serial driver (default 0)
    openCallback -- A callable object with a no parameters to be called when the Output Driver is opened (default None)
    dllPath -- The path to the dlls used for library wraping (i.e. USBX)
    """
    self.serial = None
    self.rxRegistry = []
    self.openRegistry = []
    self.closeRegistry = []
    self.registerOpenCallback(openCallback)
    self.writeBuffer = ""
    self.isOpen = False
    self.dllPath = dllPath
    self.baudrate = baudrate


    if outputType != None:
      self.setOutputType(outputType, port)
    else:
      self._outputType = None
    self.read_looping = False
    self.read_lock = threading.Lock()
    self.read_queue = []

  def close(self):
    wasOpen = self.isOpen
    try:
      if self.serial:
        self.serial.close()
      self.isOpen = False
    except:
      log.exception("Close Exception:")

    if self.serial and wasOpen:
      for callback in self.closeRegistry:
        try:
          callback()
        except:
          log.exception("Close Callback Exception:")

  def flushWriteBuf(self, *args, **kwargs):
    self.writeBuffer = ""
    if self.isOpen:
      try:
        self.serial.flushOutput()
        #Why are we calling writePoll? There should be nothing to write
        self.writePoll()
      except:
        log.exception("Flush Write Buffer Exception:")

  def flushReadBuf(self, *args, **kwargs):
    if self.isOpen:
      try:
        self.serial.flushInput()
      except:
        log.exception("Flush Read Buffer Exception:")

  def getOutputType(self):
    return self._outputType

  def inWaiting(self):
    if self.isOpen and self.serial:
      return self.serial.inWaiting()
    return 0

  def onDeviceArrival(self):
    if self.serial and self._outputType == self.TYPE_USBXSERIAL:
      self.serial.onDeviceArrival()

  def onDeviceRemoval(self):
    if self.serial and self._outputType == self.TYPE_USBXSERIAL:
      self.serial.onDeviceRemoval()

  def portName(self, type=None):
    """Get a string representation of the specified type"""
    if not type:
      type = self._outputType
    if type == self.TYPE_SIMULATOR:
      return "SIMULATOR"
    elif type == self.TYPE_PYSERIAL:
      return "RS-232"
    elif type == self.TYPE_USBXSERIAL:
      return "USB"
    elif type == self.TYPE_GATEWAY:
      return "SNAPconnect"
    elif type == self.TYPE_SNAPSTICK:
      return "SNAPstick"
    elif type == self.TYPE_TWR_RF_SNAP:
      return "TWR-RF-SNAP"
    elif type == self.TYPE_DIZIC_MB851:
      return "MB851"
    elif type == self.TYPE_BTLE:
      return "BTLE"
    else:
      return "UNKNOWN"

  def readLoop(self):
    self.serial._reconfigurePort() #For some reason port is not configured in seperate thread
    while self.isOpen and self.read_looping:
      rxbuf = self.serial.read(1)
      if rxbuf:
        log.debug("read1: %f" % (timeit.default_timer()))
        rxbuf += self.serial.read(self.serial.inWaiting())
        log.debug("read2: %f" % (timeit.default_timer()))
        #try:
          #for cb in self.rxRegistry:
            #cb(rxbuf)
        #except:
          #log.exception('Receive Processing Exception:')
        with self.read_lock:
          self.read_queue.insert(0, rxbuf)
        #self.read_loop_queue.put_nowait(rxbuf)

  def readPoll(self, bytes=256):
    if not self.isOpen:
      return

    if self.read_looping:
      log.debug("poll1: %f" % (timeit.default_timer()))
      with self.read_lock:
        try:
          rxbuf = self.read_queue.pop()
        except IndexError:
          rxbuf = None
      log.debug("poll2: %f" % (timeit.default_timer()))
      #try:
        #rxbuf = self.read_loop_queue.get_nowait()
      #except multiprocessing.queues.Empty:
        #rxbuf = None
    else:
      try:
        rxbuf = self.serial.read(bytes)
        #if rxbuf:
          #log.debug("%s rxbuf=%s" % (self.serial.portstr, pprint.pformat(rxbuf)))
        #rxbuflen = len(rxbuf)
        #if rxbuflen:
          #log.debug("rxbuflen=%i" % (rxbuflen))
      except (KeyboardInterrupt, SystemExit):
        raise
      except:
        log.debug('Read Exception:', exc_info=1)
        self.close()
        if __debug__:
          raise
        return

    #Give the buffer to everyone who's interested
    if rxbuf:
      #log.debug("read@%f=%s" % (timeit.default_timer(), pprint.pformat(rxbuf)))
      try:
        for cb in self.rxRegistry:
          cb(rxbuf)
      except (KeyboardInterrupt, SystemExit):
        raise
      except:
        log.exception('Receive Processing Exception:')
        if __debug__:
          raise
      return True

    return False

  def registerCloseCallback(self, callback):
    """
    Register for callbacks when the the PyserialDriver is closed

    callback -- A callable object with a no parameters
    """
    if callable(callback):
      self.closeRegistry.append(callback)

  def registerOpenCallback(self, callback):
    """
    Register for callbacks when the the PyserialDriver is opened

    callback -- A callable object with a no parameters
    """
    if callable(callback):
      self.openRegistry.append(callback)

  def registerRxCallback(self, callback):
    """
    Register for callbacks when characters are read for the serial driver

    callback -- A callable object with a single parameter for the received characters
    """
    if callable(callback):
      self.rxRegistry.append(callback)

  def setOutputType(self, type, port=0, blocking=False):
    if type not in [self.TYPE_SIMULATOR,
                    self.TYPE_PYSERIAL,
                    self.TYPE_USBXSERIAL,
                    self.TYPE_LOOPBACK,
                    self.TYPE_SNAPSTICK,
                    self.TYPE_TWR_RF_SNAP,
                    self.TYPE_DIZIC_MB851,
                    self.TYPE_BTLE]:
      raise TypeError("Unknown serial driver type")
    self._outputType = type
    self.port_num = port
    self.setupSerialDriver(port, blocking=blocking)

  def setupSerialDriver(self, port=0, blocking=False):
    self.close()
    try:
      if self._outputType == self.TYPE_SIMULATOR:
        raise TypeError("Unsupported output type")
      elif self._outputType == self.TYPE_PYSERIAL:
        self.serial = serial.Serial()
        self.serial.rtscts = 0
        self.serial.writeTimeout = 0.03 #Trying to force nonblocking on pyserial
        self.serial.port = port
        self.serial.open()
        if not blocking:
          self.serial.timeout = 0   # set nonblocking
        else:
          self.serial.timeout = 1
          self.read_looping = True
        pass
      elif self._outputType == self.TYPE_USBXSERIAL:
        if not sys.platform.startswith("win"):
          raise SerialOpenException("SNAP USB devices are not currently supported on this platform")
        self.serial = usbxserialwin32.UsbxSerial(pathToLibrary=self.dllPath)
        self.serial.rtscts = 0
        self.serial.writeTimeout = 0
        self.serial.port = port
        self.serial.open()
        if not blocking:
          self.serial.timeout = 0   # set nonblocking
        else:
          self.serial.timeout = 1000 #ms
          self.read_looping = True
        pass
      elif self._outputType == self.TYPE_LOOPBACK:
        self.serial = SerialLoopback()
      elif self._outputType == self.TYPE_BTLE:
        self.serial = btle_serial.BtleSerial(port)
      elif self._outputType in [self.TYPE_SNAPSTICK, self.TYPE_TWR_RF_SNAP, self.TYPE_DIZIC_MB851]:
        if self._outputType == self.TYPE_SNAPSTICK:
          self.serial = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=self.dllPath, pid=0x9090)
        elif self._outputType == self.TYPE_DIZIC_MB851:
          self.serial = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=self.dllPath, pid=0x6001, requiredDesc='MB851')
        else:
          self.serial = ftd2xxserialwin32.Ftd2xxSerial(pathToLibrary=self.dllPath, pid=0x9094)
        #Defaults to no flow control
        self.serial.writeTimeout = 100
        self.serial.port = port
        self.serial.open()
        if not blocking:
          self.serial.timeout = 1   # set nonblocking
        else:
          self.serial.timeout = 1000 #ms
          self.read_looping = True

      if self.baudrate: 
        #If we provided a specific value to the constructor, use that
        self.serial.baudrate = self.baudrate
      else:
        #If we did not provide a value to the constructor, use the class variable instead
        self.serial.baudrate = self.BAUDRATE
        
      self.isOpen = True
      self.flushWriteBuf()
      self.flushReadBuf()
      for callback in self.openRegistry:
        try:
          callback(self.serial.portstr, self.serial.baudrate)
        except:
          log.exception("Open Callback Exception:")
    except Exception, ex:
      log.debug(self.dllPath)
      log.exception("An error occured while setting up the serial port")
      self.serial = None
      self.isOpen = False
      raise SerialOpenException("An error occured while setting up the serial port: %s" % (str(ex)))

  def unregisterCloseCallback(self, callback):
    """
    Unregister a callable object from being notified of the PyserialDriver closing

    callback -- The callable object to unregister
    """
    try:
      self.closeRegistry.remove(callback)
    except KeyError:
      pass #If it's not in there than it's removed!

  def unregisterOpenCallback(self, callback):
    """
    Unregister a callable object from being notified of the PyserialDriver opening

    callback -- The callable object to unregister
    """
    try:
      self.closeRegistry.remove(callback)
    except KeyError:
      pass #If it's not in there than it's removed!

  def unregisterRxCallback(self, callback):
    """
    Unregister a callable object from being notified of characters being received

    callback -- The callable object to unregister
    """
    try:
      self.rxRegistry.remove(callback)
    except KeyError:
      pass #If it's not in there than it's removed!

  def write(self, string, *args):
    """
    Adds the characters of a string to the bytes to be written to the serial driver

    string -- The string to write to the serial driver
    """
    if string is not None:
      writeNow = (len(self.writeBuffer) == 0)

      self.writeBuffer += string

      if writeNow:
        self.writePoll()

  def writePoll(self):
    if self.isOpen and self.writeBuffer:
      try:
        bytesSent = self.serial.write(self.writeBuffer)
        if bytesSent > -1:
          #log.debug("%s tx=%s" % (self.serial.portstr, pprint.pformat(self.writeBuffer[:bytesSent])))
          self.writeBuffer = self.writeBuffer[bytesSent:]
        #else:
          #self.writeBuffer = ""
        return bytesSent #Return bytes sent in case someone cares
      except:
        log.debug('Write Exception:', exc_info=1)
        self.close()
    return False

  outputType = property(getOutputType, setOutputType)

class SerialLoopback(object):
  def __init__(self, replay=True):
    self.portstr = "SIMUALTOR"
    self.baudrate = "0" #Used in simulator as delay, eventually...
    self.replay = replay
    self.replayBuf = ""

  def close(self):
    self.replayBuf = ""

  def read(self, size=1):
    rtnVal = 0
    if self.replay:
      rtnVal = self.replayBuf[:size]
      self.replayBuf = self.replayBuf[size:]
    return rtnVal

  def write(self, string):
    if self.replay:
      self.replayBuf += string
    return len(string)

  def open(self):
    return

  def flushOutput(self):
    return

  def flushInput(self):
    return
## -------------------- test program --------------------
def _test_readCallback(buf):
  import random

  global drv
  global expectedBuf
  if expectedBuf and buf:
    checkBuf = expectedBuf[:len(buf)]
    if checkBuf != buf:
      raise Exception("Buffers do not match")
    else:
      print "Buffers Match!!!"
    expectedBuf = expectedBuf[len(buf):]
  if not expectedBuf:
    for i in range(PyserialWrapper.MAX_READ_BYTES-1):
      expectedBuf += random.sample('abcdefghijklmnopqrstuvwxyz', 1)[0]
    drv.write(expectedBuf)

def _test(N=2000):
  global drv
  global expectedBuf
  drv = PyserialWrapper(outputType=PyserialWrapper.TYPE_LOOPBACK)
  expectedBuf = ""
  drv.registerRxCallback(_test_readCallback)
  _test_readCallback(buf="")

  for i in range(N):
    drv.readPoll()
    drv.writePoll()
  drv.close()


if __name__ == '__main__':
  _test()
