# (c) Copyright 2007, Synapse


import os
from ctypes import *
from synapseGuiLib import GUID
from cp210x import *


class SynapseUsbProgrammer(object):
  def __init__(self, pathToLibrary=None):
    self.handle = c_ulong()
    self.isOpen = False
    self.cp210xDll = None

    self.pathToLibrary = pathToLibrary
    if not pathToLibrary:
      self.pathToLibrary = os.getcwd()

  def programDefaults(self):
    if self.setPid() and self.setMaxPower() and self.setProdStr() and self.setSerialNum() and self.setSelfPowered():
      return True
    else:
      return False

  def open(self, deviceNumber=0):
    self.cp210xDll = windll.LoadLibrary(os.path.join(self.pathToLibrary, CP210x_LIBRARY_NAME+'.dll'))
    if CP210x_STATUS[self.cp210xDll.CP210x_Open(c_ulong(deviceNumber), byref(self.handle))] == "CP210x_SUCCESS":
      self.isOpen = True
    return self.isOpen

  def close(self):
    self.isOpen = False
    windll.kernel32.FreeLibrary(self.cp210xDll._handle)
    self.cp210xDll._handle = None
    self.handle = c_ulong(0)
    return self.isOpen

  def reset(self):
    """Resets the USB interface and closes"""
    if self.isOpen and CP210x_STATUS[self.cp210xDll.CP210x_Reset(self.handle)] == "CP210x_SUCCESS":
      self.close()
      return True
    else:
      self.close()
      return False

  def getNumDevices(self):
    cp210xDll = windll.LoadLibrary(self.pathToLibrary + os.sep + CP210x_LIBRARY_NAME)
    _numDevices = c_ulong()
    if CP210x_STATUS[cp210xDll.CP210x_GetNumDevices(byref(_numDevices))] == "CP210x_SUCCESS":
      windll.kernel32.FreeLibrary(cp210xDll._handle)
      return _numDevices.value
    else:
      windll.kernel32.FreeLibrary(self.cp210xDll._handle)
      return 0

  numDevices = property(getNumDevices, doc="Number of SNAP USB devices detected")

  def getPid(self):
    _pid = c_ulong(0)
    if self.isOpen:
      windll.CP210xManufacturing.CP210x_GetDevicePid(self.handle, byref(_pid))
    return _pid.value

  def setPid(self, value=SYNAPSE_PID):
    _pid = c_ulong(value)
    if self.isOpen and CP210x_STATUS[self.cp210xDll.CP210x_SetPid(self.handle, _pid)] == "CP210x_SUCCESS":
      return True
    else:
      return False

  pid = property(getPid, setPid, doc="The Product ID field")

  def getMaxPower(self):
    _maxPower = c_ulong(0)
    if self.isOpen:
      windll.CP210xManufacturing.CP210x_GetMaxPower(self.handle, byref(_maxPower))
    return _maxPower.value

  def setMaxPower(self, value=SYNAPSE_MAXPOWER):
    _maxPower = c_ulong(value)
    if self.isOpen and CP210x_STATUS[self.cp210xDll.CP210x_SetMaxPower(self.handle, _maxPower)] == "CP210x_SUCCESS":
      return True
    else:
      return False

  maxPower = property(getMaxPower, setMaxPower, doc="The Max Power field in 2mA units")

  def getSerialNum(self):
    _serialNum = create_string_buffer(CP210x_MAX_SERIAL_STRLEN)
    _strLen = c_ulong()
    if self.isOpen:
      self.cp210xDll.CP210x_GetDeviceSerialNumber(self.handle, byref(_serialNum), byref(_strLen), True)
    return _serialNum.value

  def setSerialNum(self, serialNumString=None):
    if serialNumString is None:
      serialNumString = str(GUID.GUID().create_new())
    _serialNum = create_string_buffer(serialNumString)
    
    if self.isOpen and CP210x_STATUS[self.cp210xDll.CP210x_SetSerialNumber(self.handle, _serialNum, len(_serialNum), True)] == "CP210x_SUCCESS":
      return True
    else:
      return False

  serialNumber = property(getSerialNum, setSerialNum, doc="The serial number of the device")

  def getProdStr(self):
    _serialNum = create_string_buffer(CP210x_MAX_SERIAL_STRLEN)
    _strLen = c_byte()
    if self.isOpen:
      self.cp210xDll.CP210x_GetDeviceProductString(self.handle, byref(_serialNum), byref(_strLen), True)
    return _serialNum.value

  def setProdStr(self, newProductString=SYNAPSE_PRODUCT_STRING):
    _serialNum = create_string_buffer(newProductString)
    if self.isOpen and CP210x_STATUS[self.cp210xDll.CP210x_SetProductString(self.handle, _serialNum, len(_serialNum), True)] == "CP210x_SUCCESS":
      return True
    else:
      return False

  productString = property(getProdStr, setProdStr, doc="The product string of the device")

  def setSelfPowered(self, enabled=SYNAPSE_IS_SELF_POWERED):
    if self.isOpen and CP210x_STATUS[self.cp210xDll.CP210x_SetSelfPower(self.handle, int(enabled))] == "CP210x_SUCCESS":
      return True
    else:
      return False

  def getSelfPowered(self):
    _powered = c_byte()
    if self.isOpen:
      self.cp210xDll.CP210x_GetSelfPower(self.handle, byref(_powered))
    else:
      return None
    return bool(_powered.value)

  isSelfPowered = property(getSelfPowered, setSelfPowered, doc="If the device is self powered")


if __name__ == '__main__':
  import time

  prog = SynapseUsbProgrammer()
  print "Number of devices found: %s" % prog.numDevices
  prog.open(0)
  print "Current values:"
  print "PID: %s" % prog.pid
  print "Max Power: %s" % prog.maxPower
  print "Self Powered: %s" % prog.isSelfPowered
  print "Serial Number: %s" % prog.serialNumber
  print "Product String: %s" % prog.productString

  print "\nProgramming Defaults..."
  prog.programDefaults()
  print "Resetting USB interface...\n"
  prog.reset()
  time.sleep(2)
  prog.open(0)

  print "New values:"
  print "PID: %s" % prog.pid
  print "Max Power: %s" % prog.maxPower
  print "Self Powered: %s" % prog.isSelfPowered
  print "Serial Number: %s" % prog.serialNumber
  print "Product String: %s" % prog.productString
