# (c) Copyright 2014, Synapse Wireless, Inc.
"""Serial message protocol over bluetooth low energy

This layer abstracts away the hardware interface. A hardware interface providing the methods
described below needs to be implemented for any platform needing bluetooth low energy support.

The bluetooth low energy serial protocol passes data by manipulating attributes on the BTLE Peripheral.

Attribute TX:
    Read/Notify
    The Central subscribes to updates for this value
    The Peripheral sets this value in chunks to transmit data with notifications

Attribute RX:
    Write Only
    The Central writes data to this value in chunks to transmit data
"""
import binascii
import sys

# Service UUIDs
BTLE_SERIAL_SERVICE_UUID = "b9be524f-c963-11e3-bb27-3c77e6ec5dea"
BTLE_SERIAL_TX_UUID = "b9884e2e-c963-11e3-bdf4-3c77e6ec5deb"
BTLE_SERIAL_RX_UUID = "b9be524f-c963-11e3-bb27-3c77e6ec5dec"

PACKET_SIZE = 20
MAX_READ_SIZE = 4096

REALM_BLUETOOTH = "bluetooth" #Realm used for bluetooth authentication

class BtleSerialHardwareInterface(object):
    """Interface to access BTLE capabilities on each platform.

    BtleSerial.btle_interface should be set to an object of this type before attempting to
    perform any bluetooth serial operations.

    Note:
        A single hardware interface is shared between all Bluetooth serial connections.
        When multiple connections exist, they are referenced by an ID that is passed to the
        BTLE Serial interface as the port. On the Central side, the ID is the hexlified
        address of the connected peripheral. On the peripheral side, the ID is the local
        address prefixed by 'peripheral'. Ex: 'FF01BC' / 'peripheralAB23FA'

    Warning:
        All functions should be threadsafe!
    """
    # Functions common to Central and Peripheral operation
    def __init__(self, scheduler, local_name, service_uuid, tx_uuid, rx_uuid):
        """Construct a BTLE Hardware Interface"""
        raise NotImplementedError

    def supports_central(self):
        """Returns True if the central role is supported on this platform"""
        return False

    def supports_peripheral(self):
        """Returns True if the peripheral role is supported on this platform"""
        return False

    def tx(self, id, data):
        """Transmit data and return the amount written"""
        raise NotImplementedError

    def rx(self, id, size):
        """Return received data from the btle buffer or an empty string if there is no data in the buffer"""
        raise NotImplementedError

    def flush_output(self, id):
        """Dump the contents of the TX buffer"""
        raise NotImplementedError

    def flush_input(self, id):
        """Dump the contents of the RX buffer"""
        raise NotImplementedError

    def is_closed(self, id):
        """Return True if the connection has been closed"""
        raise NotImplementedError

    def set_closed_callback(self, callback):
        """Set a threadsafe callback for when a connection is closed.

        The callback accepts an address and an error message if there was a problem."""
        raise NotImplementedError

    def set_opened_callback(self, callback):
        """Set a threadsafe callback for when a connection is opened.

        The callback accepts an address and an error message if there was a problem."""
        raise NotImplementedError

    def set_authentication_callback(self, id, callback):
        """Set a threadsafe callback for when a device has successfully authenticated"""
        raise NotImplementedError

    # Functions required for Peripherals
    def start_advertising(self):
        """Start advertising the serial over gatt service"""
        raise NotImplementedError

    def stop_advertising(self):
        """Stop advertising the serial over gatt service. This closes existing connections."""
        raise NotImplementedError

    def set_server_auth(self, auth):
        """Set the authentication function. Accepts a realm,username and returns a password.

        The realm will be set to bluetooth"""
        raise NotImplementedError

    # Functions required for Centrals
    def set_client_auth(self, auth):
        """Set the authentication function. Accepts a realm returns a (username, password).

        The realm will be set to bluetooth"""
        raise NotImplementedError

    def set_discovery_callback(self, callback):
        """Set a callback for BTLE peripheral discovery. The callback accepts the node address and link quality in db as arguments"""
        raise NotImplementedError

    def start_listening(self):
        """Start listening for bluetooth peripherals with the GATT serial service

        After a connection is made listening is temporarily disabled."""
        raise NotImplementedError

    def stop_listening(self):
        """Stop listening for bluetooth peripherals with the GATT serial service

        Existing connections are closed when listening is stopped"""
        raise NotImplementedError

    def connect(self, id):
        """Connect to the given peripheral"""
        raise NotImplementedError

    def disconnect(self, id):
        """Disconnect from the given peripheral"""
        raise NotImplementedError
#################################################

class BtleSerial(object):
    btle_interface = None #BtleSerialHardwareInterface shared between all BTLE serial connections

    def __init__(self, id):
        if not self.btle_interface:
            raise ImportError("Unable to find BTLE library required for BTLE Serial")

        self.id = id

        if 'peripheral' in id:
            self.portstr = "peripheral %s"%(binascii.hexlify(id[-3:]))
        else:
            self.portstr = "central %s"%binascii.hexlify(id)

        self.baudrate = None # Not used by this serial type

    def open(self):
        pass

    def close(self):
        pass

    def read(self, size=MAX_READ_SIZE):
        if self.btle_interface.is_closed(self.id):
            raise IOError("The BTLE interface for %r has closed"%self.id)

        result = self.btle_interface.rx(self.id, size)
        return result

    def write(self, string):
        if self.btle_interface.is_closed(self.id):
            raise IOError("The BTLE interface for %r has closed"%self.id)

        result = self.btle_interface.tx(self.id, string[:PACKET_SIZE])
        return result

    def flushOutput(self):
        self.btle_interface.flush_output(self.id)

    def flushInput(self):
        self.btle_interface.flush_input(self.id)

    def inWaiting(self):
        raise NotImplemented

    def _reconfigurePort(self):
        raise NotImplemented


