""" Interface to the FTD2XX.DLL """

# (c) Copyright 2008-2011 Synapse Wireless, Inc.

from ctypes import *
import os, sys

FT_DEVICE_STATUS = ["FT_OK",
                    "FT_INVALID_HANDLE",
                    "FT_DEVICE_NOT_FOUND",
                    "FT_DEVICE_NOT_OPENED,"
                    "FT_IO_ERROR",
                    "FT_INSUFFICIENT_RESOURCES",
                    "FT_INVALID_PARAMETER",
                    "FT_INVALID_BAUD_RATE",
                    "FT_DEVICE_NOT_OPENED_FOR_ERASE",
                    "FT_DEVICE_NOT_OPENED_FOR_WRITE",
                    "FT_FAILED_TO_WRITE_DEVICE",
                    "FT_EEPROM_READ_FAILED",
                    "FT_EEPROM_WRITE_FAILED",
                    "FT_EEPROM_ERASE_FAILED",
                    "FT_EEPROM_NOT_PRESENT",
                    "FT_EEPROM_NOT_PROGRAMMED",
                    "FT_INVALID_ARGS",
                    "FT_NOT_SUPPORTED",
                    "FT_OTHER_ERROR",
                    "FT_DEVICE_LIST_NOT_READY"]

if sys.platform.startswith("win"):
    from ctypes import wintypes
    FT_LIBRARY_NAME = 'ftd2xx.dll'
    FT_HANDLE = DWORD = wintypes.DWORD
elif sys.platform.startswith("darwin"):
    FT_LIBRARY_NAME = 'libftd2xx.dylib'
    FT_HANDLE = DWORD = c_ulong
else:
    FT_LIBRARY_NAME = 'libftd2xx.so'
    FT_HANDLE = DWORD = c_ulong

FT_PURGE_RX = 1
FT_PURGE_TX = 2

FT_STOP_BITS_1 = c_ubyte(0)
FT_STOP_BITS_1_5 = c_ubyte(1)
FT_STOP_BITS_2 = c_ubyte(2)

FT_PARITY_NONE = c_ubyte(0)
FT_PARITY_ODD = c_ubyte(1)
FT_PARITY_EVEN = c_ubyte(2)
FT_PARITY_MARK = c_ubyte(3)
FT_PARITY_SPACE = c_ubyte(4)

FT_FLOW_NONE = 0x0000
FT_FLOW_RTS_CTS = 0x0100
FT_FLOW_DTR_DSR = 0x0200
FT_FLOW_XON_XOFF = 0x0400

FT_OPEN_BY_SERIAL_NUMBER = 1

PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = 'N', 'E', 'O', 'M', 'S'

SNAPSTICK_PID = 0x9090
TWR_RF_SNAP_PID = 0x9094
DIZIC_MB851_PID = 0x6001

## This class is the baseclass for FTDI exceptions
class Ftd2xxSerialException(Exception):
    """Base class for serial port related exceptions."""

portNotOpenError = Ftd2xxSerialException('Port not open')

## This abstract class provides file-like operations for derived classes
#  that implement "read" and "write".
#
#  The "read" and "write" methods are abstract and must be overriden by
#  by the derived class.
class FileLike(object):
    """An abstract file like class.

    This class implements readline and readlines based on read and
    writelines based on write.
    This class is used to provide the above functions for to Serial
    port objects.

    Note that when the serial port was opened with _NO_ timeout that
    readline blocks until it sees a newline (or the specified size is
    reached) and that readlines would never return and therefore
    refuses to work (it raises an exception in this case)!
    """

    def read(self, size): 
        raise NotImplementedError
    
    def write(self, s): 
        raise NotImplementedError

    def readline(self, size=None, eol='\n'):
        """read a line which is terminated with end-of-line (eol) character
        ('\n' by default) or until timeout"""
        line = ''
        while 1:
            c = self.read(1)
            if c:
                line += c   #not very efficient but lines are usually not that long
                if c == eol:
                    break
                if size is not None and len(line) >= size:
                    break
            else:
                break
        return line

    def readlines(self, sizehint=None, eol='\n'):
        """read a list of lines, until timeout
        sizehint is ignored"""
        if self.timeout is None:
            raise ValueError, "Serial port MUST have enabled timeout for this function!"
        lines = []
        while 1:
            line = self.readline(eol=eol)
            if line:
                lines.append(line)
                if line[-1] != eol:    #was the line received with a timeout?
                    break
            else:
                break
        return lines

    def xreadlines(self, sizehint=None):
        """just call readlines - here for compatibility"""
        return self.readlines()

    def writelines(self, sequence):
        for line in sequence:
            self.write(line)

    def flush(self):
        """flush of file like objects"""
        pass


class Ftd2xxSerialBase(FileLike):
    """Serial port base class. Provides __init__ function and properties to
       get/set port settings."""

    #default values, may be overriden in subclasses that do not support all values
    BAUDRATES = (50,75,110,134,150,200,300,600,1200,1800,2400,4800,9600,
                 19200,38400,57600,115200,230400,460800,500000,576000,921600,
                 1000000,1152000,1500000,2000000,2500000,3000000,3500000,4000000)
    PARITIES  = (PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE)

    def __init__(self,
                 port = None,          #number of device, numbering starts at zero.
                 baudrate=9600,        #baudrate
                 bytesize=8,           #number of databits
                 parity=PARITY_NONE,   #enable parity checking
                 stopbits=1,           #number of stopbits
                 timeout=100,          #set a timeout value, None to wait forever
                 xonxoff=0,            #enable software flow control
                 rtscts=0,             #enable RTS/CTS flow control
                 writeTimeout=100,     #set a timeout for writes
                 dsrdtr=None,          #None: use rtscts setting, dsrdtr override if true or false
                 pathToLibrary=None,
                 vid=0x0403,
                 pid=SNAPSTICK_PID,
                 requiredDesc = ''     # Normally we do not care, but there are some devices that we can only distinguish via USB "
                 ):
        """Initialize comm port object. If a port is given, then the port will be
           opened immediately. Otherwise a Serial port object in closed state
           is returned."""

        self.pathToLibrary = pathToLibrary
        if pathToLibrary:
            self.libraryName = pathToLibrary + os.sep + FT_LIBRARY_NAME
        else:
            self.libraryName = FT_LIBRARY_NAME
            self.pathToLibrary = os.getcwd()
        self.ftd2xxDll = None

        self.vid = vid
        self.pid = pid
        self.requiredDesc = requiredDesc

        self._numDevices = DWORD(0)
        self._isOpen   = False
        self._port     = None           #correct value is assigned below trough properties
        self._baudrate = None           #correct value is assigned below trough properties
        self._bytesize = None           #correct value is assigned below trough properties
        self._parity   = None           #correct value is assigned below trough properties
        self._stopbits = None           #correct value is assigned below trough properties
        self._timeout  = c_ulong(100)            #correct value is assigned below trough properties
        self._writeTimeout = c_ulong(100)       #correct value is assigned below trough properties
        self._xonxoff  = None           #correct value is assigned below trough properties
        self._rtscts   = None           #correct value is assigned below trough properties
        self._dsrdtr   = None           #correct value is assigned below trough properties
        self._rtsState = None
        self._dtrState = None

        #assign values using get/set methods using the properties feature
        self.port     = port
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity   = parity
        self.stopbits = stopbits
        self.timeout  = timeout
        self.writeTimeout = writeTimeout
        self.xonxoff  = xonxoff
        self.rtscts   = rtscts
        self.dsrdtr   = dsrdtr

        self.devNotifyHandle = None

        if port is not None:
            self.open()

    def isOpen(self):
        """Check if the port is opened."""
        return self._isOpen

    def setPort(self, port):
        """Change the port."""

        was_open = self._isOpen
        if was_open: self.close()

        if isinstance(port, str):
            rez = port.upper().split("USB", 2)
            if rez and len(rez) == 2:
                self._port = int(rez[1])
                self.portstr = port
            else:
                raise ValueError("Not a valid port: %s" % str(port))
        elif port is not None:
            self._port = port
            if self.pid == SNAPSTICK_PID:
                self.portstr = "SNAPstick%i" % port
            elif self.pid == TWR_RF_SNAP_PID:
                self.portstr = "TWR-RF-SNAP%i" % port
            elif self.pid == DIZIC_MB851_PID:
                self.portstr = "MB851-%s" % port
            else:
                self.portstr = "FTUSB%s" % port
        else:
            self.portstr = None
        if was_open: self.open()
    
    def getPort(self):
        """Get the current port setting. The value that was passed on init or using
           setPort() is passed back."""
        return self._port

    port = property(getPort, setPort, doc="Port setting")

    def setBaudrate(self, baudrate):
        """Change baudrate. It raises a ValueError if the port is open and the
        baudrate is not possible. If the port is closed, then tha value is
        accepted and the exception is raised when the port is opened."""
        #~ if baudrate not in self.BAUDRATES: raise ValueError("Not a valid baudrate: %r" % baudrate)
        try:
            self._baudrate = c_ulong(baudrate)
        except TypeError:
            raise ValueError("Not a valid baudrate: %r" % baudrate)
        else:
            if self._isOpen:  self._reconfigurePort()
        
    def getBaudrate(self):
        """Get the current baudrate setting."""
        return self._baudrate.value
        
    baudrate = property(getBaudrate, setBaudrate, doc="Baudrate setting")

    def setByteSize(self, bytesize):
        """Change byte size."""
        if bytesize not in [5, 6, 7, 8]: raise ValueError("Not a valid byte size: %r" % bytesize)
        self._bytesize = bytesize
        if self._isOpen: self._reconfigurePort()
    
    def getByteSize(self):
        """Get the current byte size setting."""
        return self._bytesize
    
    bytesize = property(getByteSize, setByteSize, doc="Byte size setting")

    def setParity(self, parity):
        """Change parity setting."""
        if parity not in self.PARITIES: raise ValueError("Not a valid parity: %r" % parity)
        self._parity = parity
        if self._isOpen: self._reconfigurePort()
    
    def getParity(self):
        """Get the current parity setting."""
        return self._parity
    
    parity = property(getParity, setParity, doc="Parity setting")

    def setStopbits(self, stopbits):
        """Change stopbits size."""
        if stopbits not in [1, 1.5, 2]: raise ValueError("Not a valid stopbit size: %r" % stopbits)
        self._stopbits = stopbits
        if self._isOpen: self._reconfigurePort()
    
    def getStopbits(self):
        """Get the current stopbits setting."""
        return self._stopbits
    
    stopbits = property(getStopbits, setStopbits, doc="Stopbits setting")

    def setTimeout(self, timeout):
        """Change timeout setting."""
        if timeout is not None:
            if timeout < 0: raise ValueError("Not a valid timeout: %r" % timeout)
            try:
                timeout + 1     #test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: %r" % timeout)

        self._timeout = DWORD(int(timeout))
        if self._isOpen: self._reconfigurePort()
    
    def getTimeout(self):
        """Get the current timeout setting."""
        return self._timeout.value
    
    timeout = property(getTimeout, setTimeout, doc="Timeout setting for read()")

    def setWriteTimeout(self, timeout):
        """Change timeout setting."""
        if timeout is not None:
            if timeout < 0: raise ValueError("Not a valid timeout: %r" % timeout)
            try:
                timeout + 1     #test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: %r" % timeout)

        self._writeTimeout = DWORD(timeout)
        if self._isOpen: self._reconfigurePort()
    
    def getWriteTimeout(self):
        """Get the current timeout setting."""
        return self._writeTimeout.value
    
    writeTimeout = property(getWriteTimeout, setWriteTimeout, doc="Timeout setting for write()")

    def setXonXoff(self, xonxoff):
        """Change XonXoff setting."""
        self._xonxoff = xonxoff
        if self._isOpen: self._reconfigurePort()
    
    def getXonXoff(self):
        """Get the current XonXoff setting."""
        return self._xonxoff
    
    xonxoff = property(getXonXoff, setXonXoff, doc="Xon/Xoff setting")

    def setRtsCts(self, rtscts):
        """Change RtsCts flow control setting."""
        self._rtscts = rtscts
        if self._isOpen: self._reconfigurePort()
    
    def getRtsCts(self):
        """Get the current RtsCts flow control setting."""
        return self._rtscts
    
    rtscts = property(getRtsCts, setRtsCts, doc="RTS/CTS flow control setting")

    def setDsrDtr(self, dsrdtr=None):
        """Change DsrDtr flow control setting."""
        if dsrdtr is None:
            #if not set, keep backwards compatibility and follow rtscts setting
            self._dsrdtr = self._rtscts
        else:
            #if defined independently, follow its value
            self._dsrdtr = dsrdtr
        if self._isOpen: self._reconfigurePort()
    
    def getDsrDtr(self):
        """Get the current DsrDtr flow control setting."""
        return self._dsrdtr
    
    dsrdtr = property(getDsrDtr, setDsrDtr, "DSR/DTR flow control setting")

    def setNumDevices(self):
        """Set the current number of devices."""
        raise Ftd2xxSerialException("Cannot set number of devices")

    # The ftd2xxDll.FT_GetDeviceInfoDetail() call returns the description, but as a "false unicode" string
    # This routine does the necessary decoding to get the "real" ASCII description.
    def convertDescription(self, desc):
        result = ''
        count = len(desc)
        index = 0
        while index < count:
            temp = desc[index]
            temp = ord(temp)
            lsb = temp & 0xFF
            if lsb == 0:
                return result
            else:
                result += chr(lsb)
            msb = (temp >> 8) & 0xFF
            if msb == 0:
                return result
            else:
                result += chr(msb)
            index += 1
        return result    
    
    def getNumDevices(self):
        """Get the current number of devices."""
        try:
            if sys.platform.startswith("win32"):
                ftd2xxDll = windll.LoadLibrary(self.libraryName)
            else:
                ftd2xxDll = cdll.LoadLibrary(self.libraryName)
                result = ftd2xxDll.FT_SetVIDPID(self.vid, self.pid)

            if FT_DEVICE_STATUS[ftd2xxDll.FT_CreateDeviceInfoList(byref(self._numDevices))] != "FT_OK":
                self._numDevices = DWORD(0)
            else:
                correct_pids = 0
                for i in range(0, self._numDevices.value):
                    flags = DWORD()
                    device_type = DWORD()
                    device_id = DWORD()
                    location_id = DWORD()
                    serial_num = create_unicode_buffer(16)
                    desc = create_unicode_buffer(64)
                    handle = FT_HANDLE()
                    result = ftd2xxDll.FT_GetDeviceInfoDetail(i,
                                                              byref(flags),
                                                              byref(device_type),
                                                              byref(device_id),
                                                              byref(location_id),
                                                              serial_num,
                                                              desc,
                                                              byref(handle))
                    if result == 0 and device_id.value == ((self.vid<<16)+self.pid):
                        if self.requiredDesc == '': # The normal case - no further qualification required
                            correct_pids += 1
                        else: # A matching Product Description has also been specified as a requirement for a match
                            temp_str = self.convertDescription(desc.value) # undo unicode foolishness
                            if self.requiredDesc == temp_str:
                                correct_pids += 1
                self._numDevices =  DWORD(correct_pids)

            if sys.platform.startswith("win32"):
                windll.kernel32.FreeLibrary(ftd2xxDll._handle)
            else:
                ftd2xxDll.dlclose(ftd2xxDll._handle)
            ftd2xxDll._handle = None

            return self._numDevices.value
        except:
            return 0

    numDevices = property(getNumDevices, setNumDevices, "Number of USB devices")
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    def __repr__(self):
        """String representation of the current port settings and its state."""
        return "%s<id=0x%x, open=%s>(port=%r, baudrate=%r, bytesize=%r, parity=%r, stopbits=%r, timeout=%r, xonxoff=%r, rtscts=%r, dsrdtr=%r)" % (
            self.__class__.__name__,
            id(self),
            self._isOpen,
            self.portstr,
            self.baudrate,
            self.bytesize,
            self.parity,
            self.stopbits,
            self.timeout,
            self.xonxoff,
            self.rtscts,
            self.dsrdtr,
        )


if __name__ == '__main__':
    s = Ftd2xxSerialBase()
    print s.portstr
    print s.getSupportedBaudrates()
    print s.getSupportedByteSizes()
    print s.getSupportedParities()
    print s.getSupportedStopbits()
    print s
