# (c) Copyright 2007-2010, Synapse Wireless Inc.


from ftd2xxserialutil import *
from ctypes import *
import os, time


class Ftd2xxSerial(Ftd2xxSerialBase):
    """USB Express serial port implemenation for Win32. This implemenatation requires a 
       FTDI D2XX driver installation."""

    BAUDRATES = (50,75,110,134,150,200,300,600,1200,1800,2400,4800,9600,
                 19200,38400,57600,115200)

    def open(self):
        """Open port with current settings. This may throw a SerialException
           if the port cannot be opened."""
        if self._port is None:
            raise Ftd2xxSerialException("Port must be configured before it can be used.")
        self.handle = DWORD()
        try:
            if sys.platform.startswith("win32"):
                self.ftd2xxDll = windll.LoadLibrary(self.libraryName)
            else:
                self.ftd2xxDll = cdll.LoadLibrary(self.libraryName)
                self.ftd2xxDll.FT_SetVIDPID(self.vid, self.pid)

            # Unfortunately FTDI doesn't let you open based on VID, PID so we have to go find the right port
            ftdi_serial_num = None
            if FT_DEVICE_STATUS[self.ftd2xxDll.FT_CreateDeviceInfoList(byref(self._numDevices))] != "FT_OK":
                raise Ftd2xxSerialException("Could not find devices to open: %s" % rtnCode)
            else:
                correct_pids = 0
                for i in range(0, self._numDevices.value):
                    flags = DWORD()
                    device_type = DWORD()
                    device_id = DWORD()
                    location_id = DWORD()
                    serial_num = create_unicode_buffer(16)
                    desc = create_unicode_buffer(64)
                    handle = FT_HANDLE()
                    result = self.ftd2xxDll.FT_GetDeviceInfoDetail(i,
                                                                   byref(flags),
                                                                   byref(device_type),
                                                                   byref(device_id),
                                                                   byref(location_id),
                                                                   serial_num,
                                                                   desc,
                                                                   byref(handle))
                    if result == 0 and device_id.value == ((self.vid<<16)+self.pid):
                        if correct_pids == self._port:
                            ftdi_serial_num = serial_num
                            break
                        correct_pids += 1

            if ftdi_serial_num is None:
                raise Ftd2xxSerialException("Could not find port")

            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_OpenEx(ftdi_serial_num, DWORD(FT_OPEN_BY_SERIAL_NUMBER), byref(self.handle))]
            if rtnCode != "FT_OK":
                self._closeDll()
                raise Ftd2xxSerialException("Could not open device: %s" % rtnCode)

            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetLatencyTimer(self.handle, 1)]
            if rtnCode != "FT_OK":
                self._closeDll()
                raise Ftd2xxSerialException("Could not set latency: %s" % rtnCode)
        except Exception, msg:
            self.handle = c_void_p()   #'cause __del__ is called anyway
            self._port = None
            raise Ftd2xxSerialException("Could not open device: %s" % msg)

        self._isOpen = True

        # Clear buffers:
        # Remove anything that was there (read & write)
        self.ftd2xxDll.FT_Purge(self.handle, FT_PURGE_RX|FT_PURGE_TX)

        self._reconfigurePort()

    def _closeDll(self):
        if self.ftd2xxDll:
            if sys.platform.startswith("win32"):
                windll.kernel32.FreeLibrary(self.ftd2xxDll._handle)
            elif self.ftd2xxDll._handle is not None:
                self.ftd2xxDll.dlclose(self.ftd2xxDll._handle)
            self.ftd2xxDll._handle = None
        self.handle = DWORD()
        self._isOpen = False

    def _reconfigurePort(self):
        """Set commuication parameters on opened port."""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        try:
            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetTimeouts(self.handle, self._timeout, self._writeTimeout)] 
            if rtnCode != "FT_OK":
                raise Ftd2xxSerialException("Could not set timeouts: %s" % rtnCode)

            # Setup the connection info.
            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetBaudRate(self.handle, self._baudrate)]
            if rtnCode != "FT_OK":
                raise Ftd2xxSerialException("Could not set baud rate: %s" % rtnCode)

            stopbits = FT_STOP_BITS_1
            if self.stopbits == 1.5:
                stopbits = FT_STOP_BITS_1_5
            elif self.stopbits == 2:
                stopbits = FT_STOP_BITS_2

            parity = FT_PARITY_NONE
            if self.parity == PARITY_ODD:
                parity = FT_PARITY_ODD
            elif self.parity == PARITY_EVEN:
                parity = FT_PARITY_EVEN
            elif self.parity == PARITY_MARK:
                parity = FT_PARITY_MARK
            elif self.parity == PARITY_SPACE:
                parity = FT_PARITY_SPACE

            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetDataCharacteristics(self.handle, c_byte(self._bytesize), stopbits, parity)]
            if rtnCode != "FT_OK":
                raise Ftd2xxSerialException("Could not set stop bits, parity, and/or bits per word: %s" % rtnCode)

            self._setFlowControl()

        except Exception, msg:
            raise Ftd2xxSerialException("Cannot configure port, some setting was wrong. Original message: %s" % msg)

    def close(self):
        """Close port"""
        if hasattr(self, 'handle') and self.handle:
            self.ftd2xxDll.FT_Close(self.handle)
            self.handle = DWORD()
        self._closeDll()

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    def inWaiting(self):
        """Return the number of characters currently in the input buffer."""
        numBytesInQueue = DWORD()
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_GetQueueStatus(self.handle, byref(numBytesInQueue))]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("An error occured while checking rx queue: %s" % rtnCode)

        return numBytesInQueue.value

    def read(self, size=256):
        """Read size bytes from the port. If a timeout is set it may
           return less characters as requested. With no timeout it will block
           until the requested number of bytes is read."""
        #return ''
        if not self._isOpen: raise portNotOpenError
        if size > 0:
            if self._timeout == 0:
                bytesInBuffer = self.inWaiting()
                if bytesInBuffer == 0:
                    return ''         
                numToRead = min(bytesInBuffer, size)
            else:
                numToRead = size

            readBuffer = create_string_buffer(numToRead)
            bytesRead = DWORD()
            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_Read(self.handle, readBuffer, numToRead, byref(bytesRead))]

            if rtnCode != "FT_OK":
                raise Ftd2xxSerialException("An error occured while reading: %s" % rtnCode)

            return readBuffer.raw[:bytesRead.value]
        else:
            return ''

    def write(self, string):
        """Output the given string over the port."""
        if not self._isOpen: raise portNotOpenError
        if string:
            writeBuffer = create_string_buffer(string)
            bytesWritten = DWORD()
            #remember to remove null terminator
            rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_Write(self.handle, writeBuffer, len(writeBuffer)-1, byref(bytesWritten))]
            #print bytesWritten.value
            if rtnCode != "FT_OK":
                raise Ftd2xxSerialException("An error occured while writting: %s" % rtnCode)
            return bytesWritten.value
        else:
            return 0

    def flushInput(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self._isOpen: raise portNotOpenError
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_Purge(self.handle, FT_PURGE_RX)]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("An error occured while flushing the input buffer: %s" % rtnCode)

    def flushOutput(self):
        """Clear output buffer, aborting the current output and discarding all that is in the buffer."""
        if not self._isOpen: raise portNotOpenError
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_Purge(self.handle, FT_PURGE_TX)]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("An error occured while flushing the input buffer: %s" % rtnCode)

    def sendBreak(self):
        """Send break condition."""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetBreakOn(self.handle)]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not set break: %s" % rtnCode)
        #TODO (from pyserial): how to set the correct duration??
        time.sleep(0.020)
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetBreakOff(self.handle)]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not reset break: %s" % rtnCode)

    def setRTS(self,level=1):
        """Set terminal status line: Request To Send"""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")
        if level:
            if FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetRts(self.handle)] != "FT_OK":
                raise Ftd2xxSerialException("An error occured while setting RTS: %s" % rtnCode)
        elif FT_DEVICE_STATUS[self.ftd2xxDll.FT_ClrRts(self.handle)] != "FT_OK":
                raise Ftd2xxSerialException("An error occured while clearing RTS: %s" % rtnCode)

    def setDTR(self,level=1):
        """Set terminal status line: Data Terminal Ready"""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")
        if level:
            if FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetDtr(self.handle)] != "FT_OK":
                raise Ftd2xxSerialException("An error occured while setting RTS: %s" % rtnCode)
        elif FT_DEVICE_STATUS[self.ftd2xxDll.FT_ClrDtr(self.handle)] != "FT_OK":
                raise Ftd2xxSerialException("An error occured while clearing RTS: %s" % rtnCode)

    def getCTS(self):
        """Read terminal status line: Clear To Send"""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        status = DWORD()
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 0x10

    def getDSR(self):
        """Read terminal status line: Data Set Ready"""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        status = DWORD()
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not get DSR state: " % rtnCode)
        return status.value >> 0x20

    def getRI(self):
        """Read terminal status line: Ring Indicator"""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        status = DWORD()
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not get RI state: " % rtnCode)
        return status.value >> 0x40

    def getCD(self):
        """Read terminal status line: Carrier Detect"""
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        status = DWORD()
        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not get CD state: " % rtnCode)
        return status.value >> 0x80

    def getDTR(self):
        """Read terminal status line: Data Terminal Ready"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 1

    def getRTS(self):
        """Read terminal status line: Ready To Send"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 2

    # - - platform specific - - - -
#    def setXON(self, level=True):
#        """Platform specific - set flow state."""
#        if not self.hComPort: raise portNotOpenError	
#        if level:
#            win32file.EscapeCommFunction(self.hComPort, win32file.SETXON)
#        else:
#            win32file.EscapeCommFunction(self.hComPort, win32file.SETXOFF)

    def _setFlowControl(self):
        if not self._isOpen:
            raise Ftd2xxSerialException("Can only operate on an open port")

        flow = FT_FLOW_NONE
        xon = c_byte()
        xoff = c_byte()
        if self._rtscts:
            flow = FT_FLOW_RTS_CTS
        elif self._dsrdtr:
            flow = FT_FLOW_DTR_DSR
        elif self._xonxoff:
            flow = FT_FLOW_XON_XOFF
            #TODO: Get xon xoff characters

        rtnCode = FT_DEVICE_STATUS[self.ftd2xxDll.FT_SetFlowControl(self.handle, flow, xon, xoff)]
        if rtnCode != "FT_OK":
            raise Ftd2xxSerialException("Could not set flow control: %s" % rtnCode)

    def onDeviceArrival(self):
        pass  #No extra stuff to do right now

    def onDeviceRemoval(self):
        pass  #We are closing the connection from the main frame

    def setBitMode(self, mask, mode):
        """FTDI specific bit bang operation"""
        if not self._isOpen: raise portNotOpenError

        if self.ftd2xxDll.FT_SetBitMode(self.handle, mask, mode) == 0:
            #print "%02X" % (mask)
            return True
        return False


#Nur Testfunktion!!
if __name__ == '__main__':
    print __name__
    s = Ftd2xxSerial()
    print s

    s = Ftd2xxSerial(0)
    print s

    s.baudrate = 9600
    s.close()
    s.open()
    print s
