# (c) Copyright 2007, Synapse


from ctypes import *
import os, sys


SI_MAX_WRITE_SIZE = 4096
SI_MAX_READ_SIZE = SI_MAX_WRITE_SIZE*16

SI_RETURN_CODES = {0:"SI_SUCCESS", 
                   1:"SI_INVALID_HANDLE", 
                   2:"SI_READ_ERROR",
                   3:"SI_RX_QUEUE_NOT_READY",
                   4:"SI_WRITE_ERROR", 
                   5:"SI_RESET_ERROR", 
                   6:"SI_INVALID_PARAMETER", 
                   7:"SI_INVALID_REQUEST_LENGTH", 
                   8:"SI_DEVICE_IO_FAILED",
                   9:"SI_INVALID_BAUDRATE", 
                   10:"SI_FUNCTION_NOT_SUPPORTED", 
                   11:"SI_GLOBAL_DATA_ERROR", 
                   12:"SI_SYSTEM_ERROR_CODE",
                   13:"SI_READ_TIMED_OUT", 
                   14:"SI_WRITE_TIMED_OUT", 
                   15:"SI_IO_PENDING",
                   255:"SI_DEVICE_NOT_FOUND"}

SI_QUEUE_STATUS_CODES = ("SI_RX_NO_OVERRUN && SI_RX_EMPTY", "SI_RX_OVERRUN", "SI_RX_READY")
SI_RX_NO_OVERRUN, SI_RX_EMPTY, SI_RX_OVERRUN, SI_RX_READY = (0,0,1,2)
SI_SUCCESS, SI_SYSTEM_ERROR_CODE, SI_READ_TIMED_OUT, SI_WRITE_TIMED_OUT = (0,12,13,14)
SI_HELD_INACTIVE, SI_HELD_ACTIVE, SI_FIRMWARE_CONTROLLED, SI_RECEIVE_FLOW_CONTROL, SI_TRANSMIT_ACTIVE_SIGNAL, SI_STATUS_INPUT, SI_HANDSHAKE_LINE = (0,1,2,2,3,0,1)

USBX_GUID = "{3c5e1462-5695-4e18-876b-f3f3d08aaf18}"
USBX_LIBRARY_NAME = 'SiUSBXp.dll'

PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE = 'N', 'E', 'O', 'M', 'S'




class UsbxSerialException(Exception):
    """Base class for serial port related exceptions."""

portNotOpenError = UsbxSerialException('Port not open')

#class SerialTimeoutException(SerialException):
#    """Write timeouts give an exception"""
#
#writeTimeoutError = SerialTimeoutException("Write timeout")

class FileLike(object):
    """An abstract file like class.

    This class implements readline and readlines based on read and
    writelines based on write.
    This class is used to provide the above functions for to Serial
    port objects.
    
    Note that when the serial port was opened with _NO_ timeout that
    readline blocks until it sees a newline (or the specified size is
    reached) and that readlines would never return and therefore
    refuses to work (it raises an exception in this case)!
    """

    def read(self, size): raise NotImplementedError
    def write(self, s): raise NotImplementedError

    def readline(self, size=None, eol='\n'):
        """read a line which is terminated with end-of-line (eol) character
        ('\n' by default) or until timeout"""
        line = ''
        while 1:
            c = self.read(1)
            if c:
                line += c   #not very efficient but lines are usually not that long
                if c == eol:
                    break
                if size is not None and len(line) >= size:
                    break
            else:
                break
        return line

    def readlines(self, sizehint=None, eol='\n'):
        """read a list of lines, until timeout
        sizehint is ignored"""
        if self.timeout is None:
            raise ValueError, "Serial port MUST have enabled timeout for this function!"
        lines = []
        while 1:
            line = self.readline(eol=eol)
            if line:
                lines.append(line)
                if line[-1] != eol:    #was the line received with a timeout?
                    break
            else:
                break
        return lines

    def xreadlines(self, sizehint=None):
        """just call readlines - here for compatibility"""
        return self.readlines()

    def writelines(self, sequence):
        for line in sequence:
            self.write(line)

    def flush(self):
        """flush of file like objects"""
        pass


class UsbxSerialBase(FileLike):
    """Serial port base class. Provides __init__ function and properties to
       get/set port settings."""

    #default values, may be overriden in subclasses that do not support all values
    BAUDRATES = (50,75,110,134,150,200,300,600,1200,1800,2400,4800,9600,
                 19200,38400,57600,115200,230400,460800,500000,576000,921600,
                 1000000,1152000,1500000,2000000,2500000,3000000,3500000,4000000)
    PARITIES  = (PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE)

    def __init__(self,
                 port = None,          #number of device, numbering starts at zero.
                 baudrate=9600,        #baudrate
                 bytesize=8,           #number of databits
                 parity=PARITY_NONE,   #enable parity checking
                 stopbits=1,           #number of stopbits
                 timeout=100,          #set a timeout value, None to wait forever
                 xonxoff=0,            #enable software flow control
                 rtscts=0,             #enable RTS/CTS flow control
                 writeTimeout=100,     #set a timeout for writes
                 dsrdtr=None,          #None: use rtscts setting, dsrdtr override if true or false
                 pathToLibrary=None
                 ):
        """Initialize comm port object. If a port is given, then the port will be
           opened immediately. Otherwise a Serial port object in closed state
           is returned."""

        # This is a HACK, to compensate for the USBx driver generating a WRITE_TIMEOUT (reporting zero bytes sent, when the data really _was_ sent) on slow PCs
        self.MIN_WRITE_TIME = 30   # ms

        self.pathToLibrary = pathToLibrary
        if pathToLibrary:
            self.libraryName = pathToLibrary + os.sep + USBX_LIBRARY_NAME
        else:
            self.libraryName = USBX_LIBRARY_NAME
            self.pathToLibrary = os.getcwd()
        self.siusbxpDll = None

        self._numDevices = c_ulong(0)
        self._isOpen   = False
        self._port     = None           #correct value is assigned below trough properties
        self._baudrate = None           #correct value is assigned below trough properties
        self._bytesize = None           #correct value is assigned below trough properties
        self._parity   = None           #correct value is assigned below trough properties
        self._stopbits = None           #correct value is assigned below trough properties
        self._timeout  = c_ulong(100)            #correct value is assigned below trough properties
        self._writeTimeout = c_ulong(100)       #correct value is assigned below trough properties
        self._xonxoff  = None           #correct value is assigned below trough properties
        self._rtscts   = None           #correct value is assigned below trough properties
        self._dsrdtr   = None           #correct value is assigned below trough properties
        self._rtsState = None
        self._dtrState = None

        #assign values using get/set methods using the properties feature
        self.port     = port
        self.baudrate = baudrate
        self.bytesize = bytesize
        self.parity   = parity
        self.stopbits = stopbits
        self.timeout  = timeout
        self.writeTimeout = writeTimeout
        self.xonxoff  = xonxoff
        self.rtscts   = rtscts
        self.dsrdtr   = dsrdtr

        self.devNotifyHandle = None

        if port is not None:
            self.open()

    def isOpen(self):
        """Check if the port is opened."""
        return self._isOpen

    def setPort(self, port):
        """Change the port."""

        was_open = self._isOpen
        if was_open: self.close()

        if isinstance(port, str):
            rez = port.upper().split("USB", 2)
            if rez and len(rez) == 2:
                self._port = int(rez[1])
                self.portstr = port
            else:
                raise ValueError("Not a valid port: %s" % str(port))
        elif port is not None:
            self._port = port
            self.portstr = "USB%s" % port
        else:
            self.portstr = None
        if was_open: self.open()
    
    def getPort(self):
        """Get the current port setting. The value that was passed on init or using
           setPort() is passed back."""
        return self._port

    port = property(getPort, setPort, doc="Port setting")

    def setBaudrate(self, baudrate):
        """Change baudrate. It raises a ValueError if the port is open and the
        baudrate is not possible. If the port is closed, then tha value is
        accepted and the exception is raised when the port is opened."""
        #~ if baudrate not in self.BAUDRATES: raise ValueError("Not a valid baudrate: %r" % baudrate)
        try:
            self._baudrate = c_ulong(baudrate)
        except TypeError:
            raise ValueError("Not a valid baudrate: %r" % baudrate)
        else:
            if self._isOpen:  self._reconfigurePort()
        
    def getBaudrate(self):
        """Get the current baudrate setting."""
        return self._baudrate.value
        
    baudrate = property(getBaudrate, setBaudrate, doc="Baudrate setting")

    def setByteSize(self, bytesize):
        """Change byte size."""
        if bytesize not in [5, 6, 7, 8]: raise ValueError("Not a valid byte size: %r" % bytesize)
        self._bytesize = bytesize
        if self._isOpen: self._reconfigurePort()
    
    def getByteSize(self):
        """Get the current byte size setting."""
        return self._bytesize
    
    bytesize = property(getByteSize, setByteSize, doc="Byte size setting")

    def setParity(self, parity):
        """Change parity setting."""
        if parity not in self.PARITIES: raise ValueError("Not a valid parity: %r" % parity)
        self._parity = parity
        if self._isOpen: self._reconfigurePort()
    
    def getParity(self):
        """Get the current parity setting."""
        return self._parity
    
    parity = property(getParity, setParity, doc="Parity setting")

    def setStopbits(self, stopbits):
        """Change stopbits size."""
        if stopbits not in [1, 1.5, 2]: raise ValueError("Not a valid stopbit size: %r" % stopbits)
        self._stopbits = stopbits
        if self._isOpen: self._reconfigurePort()
    
    def getStopbits(self):
        """Get the current stopbits setting."""
        return self._stopbits
    
    stopbits = property(getStopbits, setStopbits, doc="Stopbits setting")

    def setTimeout(self, timeout):
        """Change timeout setting."""
        if timeout is not None:
            if timeout < 0: raise ValueError("Not a valid timeout: %r" % timeout)
            try:
                timeout + 1     #test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: %r" % timeout)
        
        self._timeout = c_ulong(int(timeout))
        if self._isOpen: self._reconfigurePort()
    
    def getTimeout(self):
        """Get the current timeout setting."""
        return self._timeout.value
    
    timeout = property(getTimeout, setTimeout, doc="Timeout setting for read()")

    def setWriteTimeout(self, timeout):
        """Change timeout setting."""
        if timeout is not None:
            if timeout < 0: raise ValueError("Not a valid timeout: %r" % timeout)
            try:
                timeout + 1     #test if it's a number, will throw a TypeError if not...
            except TypeError:
                raise ValueError("Not a valid timeout: %r" % timeout)
            
        # Enforce a minimum timeout - USBX doesn't always respond well to zero timeout, setting 10ms to be safe (1ms may work as well...)
        if timeout < self.MIN_WRITE_TIME:
            timeout = self.MIN_WRITE_TIME
            
        self._writeTimeout = c_ulong(timeout)
        if self._isOpen: self._reconfigurePort()
    
    def getWriteTimeout(self):
        """Get the current timeout setting."""
        return self._writeTimeout.value
    
    writeTimeout = property(getWriteTimeout, setWriteTimeout, doc="Timeout setting for write()")

    def setXonXoff(self, xonxoff):
        """Change XonXoff setting."""
        self._xonxoff = xonxoff
        if self._isOpen: self._reconfigurePort()
    
    def getXonXoff(self):
        """Get the current XonXoff setting."""
        return self._xonxoff
    
    xonxoff = property(getXonXoff, setXonXoff, doc="Xon/Xoff setting")

    def setRtsCts(self, rtscts):
        """Change RtsCts flow control setting."""
        self._rtscts = rtscts
        if self._isOpen: self._reconfigurePort()
    
    def getRtsCts(self):
        """Get the current RtsCts flow control setting."""
        return self._rtscts
    
    rtscts = property(getRtsCts, setRtsCts, doc="RTS/CTS flow control setting")

    def setDsrDtr(self, dsrdtr=None):
        """Change DsrDtr flow control setting."""
        if dsrdtr is None:
            #if not set, keep backwards compatibility and follow rtscts setting
            self._dsrdtr = self._rtscts
        else:
            #if defined independently, follow its value
            self._dsrdtr = dsrdtr
        if self._isOpen: self._reconfigurePort()
    
    def getDsrDtr(self):
        """Get the current DsrDtr flow control setting."""
        return self._dsrdtr
    
    dsrdtr = property(getDsrDtr, setDsrDtr, "DSR/DTR flow control setting")

    def setNumDevices(self):
        """Set the current number of devices."""
        raise UsbxSerialException("Cannot set number of devices")

    def getNumDevices(self):
        """Get the current number of devices."""
        try:
            siusbxpDll = windll.LoadLibrary(self.libraryName)
    
            if SI_RETURN_CODES[siusbxpDll.SI_GetNumDevices(byref(self._numDevices))] != "SI_SUCCESS":
                self._numDevices = c_ulong(0)
    
            windll.kernel32.FreeLibrary(siusbxpDll._handle)
            #windll.SiUSBXp._handle = None
            siusbxpDll._handle = None
    
            return self._numDevices.value
        except:
            return 0

    numDevices = property(getNumDevices, setNumDevices, "Number of USB devices")
    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    def __repr__(self):
        """String representation of the current port settings and its state."""
        return "%s<id=0x%x, open=%s>(port=%r, baudrate=%r, bytesize=%r, parity=%r, stopbits=%r, timeout=%r, xonxoff=%r, rtscts=%r, dsrdtr=%r)" % (
            self.__class__.__name__,
            id(self),
            self._isOpen,
            self.portstr,
            self.baudrate,
            self.bytesize,
            self.parity,
            self.stopbits,
            self.timeout,
            self.xonxoff,
            self.rtscts,
            self.dsrdtr,
        )


if __name__ == '__main__':
    s = SerialBase()
    print s.portstr
    print s.getSupportedBaudrates()
    print s.getSupportedByteSizes()
    print s.getSupportedParities()
    print s.getSupportedStopbits()
    print s
