# (c) Copyright 2007, Synapse


from usbxserialutil import *
from ctypes import *
import cp210x
import os, time


CP210x_LIBRARY_NAME = 'CP210xManufacturing'


class UsbxSerial(UsbxSerialBase):
    """USB Express serial port implemenation for Win32. This implemenatation requires a
       cp201x driver installation."""

    BAUDRATES = (50,75,110,134,150,200,300,600,1200,1800,2400,4800,9600,
                 19200,38400,57600,115200)

    def open(self):
        """Open port with current settings. This may throw a SerialException
           if the port cannot be opened."""
        if self._port is None:
            raise UsbxSerialException("Port must be configured before it can be used.")
        self.handle = c_ulong()
        try:
            self.siusbxpDll = windll.LoadLibrary(self.libraryName)

            #Verify that this CP210x device has our pid
            deviceString = create_string_buffer(cp210x.CP210x_MAX_DEVICE_STRLEN)
            result = SI_RETURN_CODES[self.siusbxpDll.SI_GetProductString(self._port, deviceString, cp210x.CP210x_RETURN_FULL_PATH)]
            if result != "SI_SUCCESS":
                self.close()
                raise UsbxSerialException("Unable to verify device PID: %s" % str(result))
            elif deviceString.value.find(cp210x.SYNAPSE_PID_DEVICE_STRING) == -1:
                self.close()
                raise UsbxSerialException("USB device was not found to be a SNAP USB Device")

            rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_Open(self._port, byref(self.handle))]
            if rtnCode != "SI_SUCCESS":
                raise UsbxSerialException("Could not open device: %s" % rtnCode)
        except Exception, msg:
            self.handle = c_ulong()    #'cause __del__ is called anyway
            self._port = None
            raise UsbxSerialException("Could not open device: %s" % msg)

        self._isOpen = True

        # Clear buffers:
        # Remove anything that was there (read & write)
        self.siusbxpDll.SI_FlushBuffers(self.handle, 1, 1)

        self._reconfigurePort()

    def closeCP210xManufacturingDll(self, cp210xDll):
        #pass
        windll.kernel32.FreeLibrary(cp210xDll._handle)
        #windll.CP210xManufacturing._handle = None
        cp210xDll._handle = None

    def _reconfigurePort(self):
        """Set commuication parameters on opened port."""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        try:
            rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_SetTimeouts(self._timeout, self._writeTimeout)]
            if rtnCode != "SI_SUCCESS":
                raise UsbxSerialException("Could not set timeouts: %s" % rtnCode)

            # Setup the connection info.
            rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_SetBaudRate(self.handle, self._baudrate)]
            if rtnCode != "SI_SUCCESS":
                raise UsbxSerialException("Could not set baud rate: %s" % rtnCode)

            lineControlVal = 0
            if self.stopbits == 1.5:
                lineControlVal += 1
            elif self.stopbits == 2:
                lineControlVal += 2

            if self.parity == PARITY_ODD:
                lineControlVal += 16
            elif self.parity == PARITY_EVEN:
                lineControlVal += 32
            elif self.parity == PARITY_MARK:
                lineControlVal += 48
            elif self.parity == PARITY_SPACE:
                lineControlVal += 64

            if self.bytesize == 5:
                lineControlVal += 1280
            elif self.bytesize == 6:
                lineControlVal += 1536
            elif self.bytesize == 7:
                lineControlVal += 1792
            else:
                lineControlVal += 2048

            lineControl = c_ushort(lineControlVal)
            rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_SetLineControl(self.handle, lineControl)]
            if rtnCode != "SI_SUCCESS":
                raise UsbxSerialException("Could not set stop bits, parity, and/or bits per word: %s" % rtnCode)

            self._setFlowControl()

        except Exception, msg:
            raise UsbxSerialException("Cannot configure port, some setting was wrong. Original message: %s" % msg)

    #~ def __del__(self):
        #~ self.close()

    def close(self):
        """Close port"""
        if hasattr(self, 'handle') and self.handle:
          self.siusbxpDll.SI_Close(self.handle)
          self.handle = c_ulong()
        if self.siusbxpDll:
          if isinstance(self.siusbxpDll._handle,int):
            dll_handle = self.siusbxpDll._handle
          else:
            #Some python implementations (PyPy) store an object in _handle that can not be
            #passed to FreeLibrary, so we ask kernel32 for the correct handle here
            dll_handle = windll.kernel32.GetModuleHandleA(self.siusbxpDll._name)
          windll.kernel32.FreeLibrary(dll_handle)
          self.siusbxpDll._handle = None
        self._isOpen = False

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -
    def inWaiting(self):
        """Return the number of characters currently in the input buffer."""
        numBytesInQueue = c_ulong()
        queueStatus = c_ulong()
        rtnCode = self.siusbxpDll.SI_CheckRXQueue(self.handle, byref(numBytesInQueue), byref(queueStatus))
        queueRtnCode = queueStatus.value
        if rtnCode != SI_SUCCESS:
          raise UsbxSerialException("An error occured while checking rx queue: %s" % SI_RETURN_CODES[rtnCode])

        if queueRtnCode == SI_RX_READY:
          return numBytesInQueue.value
        elif queueRtnCode == SI_RX_EMPTY:
          return 0
        else:  # SI_RX_OVERRUN
          # Flush all buffers - per AN169
          self.siusbxpDll.SI_FlushBuffers(self.handle, 1, 1)
          return 0

    def read(self, size=SI_MAX_READ_SIZE):
        """Read size bytes from the port. If a timeout is set it may
           return less characters as requested. With no timeout it will block
           until the requested number of bytes is read."""
        #return ''
        if not self._isOpen: raise portNotOpenError
        if size > 0:
            if self._timeout.value == 0:
                bytesInBuffer = self.inWaiting()
                if bytesInBuffer == 0:
                    return ''
                numToRead = min(bytesInBuffer, size)
            else:
                numToRead = size

            readBuffer = create_string_buffer(numToRead)
            bytesRead = c_ulong()
            rtnCode = self.siusbxpDll.SI_Read(self.handle, readBuffer, sizeof(readBuffer), byref(bytesRead), 0)

            if rtnCode == SI_READ_TIMED_OUT:
                return ''
            elif rtnCode != SI_SUCCESS:
                raise UsbxSerialException("An error occured while reading: %s" % SI_RETURN_CODES[rtnCode])

            return readBuffer.raw
        else:
            return ''

    def write(self, string):
        """Output the given string over the port."""
        if not self._isOpen: raise portNotOpenError
        if string:
            writeBuffer = create_string_buffer(string)
            bytesWritten = c_ulong()
            #remember to remove null terminator
            rtnCode = self.siusbxpDll.SI_Write(self.handle, writeBuffer, len(writeBuffer)-1, byref(bytesWritten), 0)
            if rtnCode == SI_WRITE_TIMED_OUT:
                if self.numDevices == 0:  #We can get a write timeout when someone unplugs the device
                    raise UsbxSerialException("Write time out occurred and there are no USB devices")
            elif rtnCode == SI_SYSTEM_ERROR_CODE:
                raise UsbxSerialException("A system error occured while writting: %s" % windll.kernel32.GetLastError())
            elif rtnCode != SI_SUCCESS:
                raise UsbxSerialException("An error occured while writting: %s" % SI_RETURN_CODES[rtnCode])
            return bytesWritten.value
        else:
            return 0

    def flushInput(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self._isOpen: raise portNotOpenError
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_FlushBuffers(self.handle, 0, 1)]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("An error occured while flushing the input buffer: %s" % rtnCode)


    def flushOutput(self):
        """Clear output buffer, aborting the current output and
        discarding all that is in the buffer."""
        if not self._isOpen: raise portNotOpenError
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_FlushBuffers(self.handle, 1, 0)]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("An error occured while flushing the output buffer: %s" % rtnCode)

    def sendBreak(self):
        """Send break condition."""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_SetBreak(self.handle, 1)]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not set stop bits, parity, and/or bits per word: %s" % rtnCode)
        #TODO (from pyserial): how to set the correct duration??
        time.sleep(0.020)
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_SetBreak(self.handle, 0)]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not set stop bits, parity, and/or bits per word: %s" % rtnCode)

    def setRTS(self,level=1):
        """Set terminal status line: Request To Send"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")
        if level:
            self._rtsState = SI_HELD_ACTIVE
        else:
            self._rtsState = SI_HELD_INACTIVE
        self._setFlowControl()

    def setDTR(self,level=1):
        """Set terminal status line: Data Terminal Ready"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")
        if level:
            self._dtrState = SI_HELD_ACTIVE
        else:
            self._dtrState = SI_HELD_INACTIVE
        self._setFlowControl()

    def getCTS(self):
        """Read terminal status line: Clear To Send"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 5

    def getDSR(self):
        """Read terminal status line: Data Set Ready"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 6

    def getRI(self):
        """Read terminal status line: Ring Indicator"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 7

    def getCD(self):
        """Read terminal status line: Carrier Detect"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 8

    def getDTR(self):
        """Read terminal status line: Data Terminal Ready"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 1

    def getRTS(self):
        """Read terminal status line: Ready To Send"""
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        status = c_byte()
        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_GetModemStatus(self.handle, byref(status))]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not get CTS state: " % rtnCode)
        return status.value >> 2

    # - - platform specific - - - -
#    def setXON(self, level=True):
#        """Platform specific - set flow state."""
#        if not self.hComPort: raise portNotOpenError
#        if level:
#            win32file.EscapeCommFunction(self.hComPort, win32file.SETXON)
#        else:
#            win32file.EscapeCommFunction(self.hComPort, win32file.SETXOFF)

    def _setFlowControl(self):
        if not self._isOpen:
            raise UsbxSerialException("Can only operate on an open port")

        if self._rtscts:
            cts = SI_HANDSHAKE_LINE
        else:
            cts = SI_STATUS_INPUT

        if self._rtsState == None:
            self._rtsState = SI_HELD_ACTIVE
        if self._dtrState == None:
            self._dtrState = SI_HELD_ACTIVE

        rtnCode = SI_RETURN_CODES[self.siusbxpDll.SI_SetFlowControl(self.handle,
                                                                  cts,
                                                                  self._rtsState,
                                                                  self._dtrState,
                                                                  SI_STATUS_INPUT,
                                                                  SI_STATUS_INPUT,
                                                                  0)]
        if rtnCode != "SI_SUCCESS":
            raise UsbxSerialException("Could not set flow control: %s" % rtnCode)

    def onDeviceArrival(self):
        pass  #No extra stuff to do right now

    def onDeviceRemoval(self):
        pass  #We are closing the connection from the main frame

#Nur Testfunktion!!
if __name__ == '__main__':
    print __name__
    s = UsbxSerial()
    print s

    s = UsbxSerial(0)
    print s

    s.baudrate = 9600
    s.close()
    s.open()
    print s
