# (c) Copyright 2014, Synapse Wireless, Inc.
"""BTLE Peripheral implementation of SNAP Serial over GATT for iOS

This requires the pyobjus library to run.
"""
import logging
import base64
import random
import string
import hashlib
import binascii

from pyobjus import autoclass, protocol, objc_str, objc_arr, objc_dict, objc_b

######################################
# Constants/Objective C Enumerations
NSUTF8StringEncoding = 4
MAX_WRITE_SIZE = 20

CBCentralManagerStateUnknown = 0
CBCentralManagerStateResetting = 1
CBCentralManagerStateUnsupported = 2
CBCentralManagerStateUnauthorized = 3
CBCentralManagerStatePoweredOff = 4
CBCentralManagerStatePoweredOn = 5

BLUETOOTH_REALM = 'bluetooth'

def protocol_exception_catcher(fn):
    # The protocol functions ignore exceptions and mask them from showing up in logs
    # This decorator catches the exception before the protocol manager and logs it
    def new_func(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception as e:
            logging.exception("Protocol Exception %s: %r", fn.__name__, e)
            raise
    new_func.__name__ = fn.__name__
    return new_func

#######################
# GATT Serial Central
class GattSerialCentral(object):
    def __init__(self, scheduler, local_name, service_uuid, tx_uuid, rx_uuid):
        """Create a GATT Serial central in iOS.

        References:
            This implements all of the methods required to be a peripheral for
            btle_serial.py. See class BtleSerialHardwareInterface.

        Args:
            scheduler : The SNAP Connect instance scheduler
            local_name : Name to advertise. Should be address in format 'FF0102'
            service_uuid : UUID for the GATT Serial Service
            tx_uuid : UUID for the TX characteristic
            rx_uuid : UUID for the RX characteristic
        """
        self.log = logging.getLogger('GattSerialCentral')
        self.verbose_logging = False #Enable for even more verbose logging

        self.scheduler = scheduler
        self.local_name = local_name
        self.service_uuid = service_uuid
        self.tx_uuid = tx_uuid
        self.rx_uuid = rx_uuid

        #User callbacks
        self.user_on_close_callback = None
        self.user_on_open_callback = None
        self.user_on_discover_callback = None
        self.user_on_authentication_callbacks = {}

        self.client_auth = None

        #Objective C Classes
        self.CBUUID = autoclass('CBUUID')
        self.CBCentralManager = autoclass('CBCentralManager')
        self.NSData = autoclass('NSData')

        #BTLE State
        self.rx_buffer = {} # Maps from ID to a string for the RX buffer
        self.connections_closed = {} # Maps from ID to True if this connection has closed
        self.peripherals = {} # {UUID: (addr, peripheral)
        self.snap_service = {} # {addr/id: [peripheral, tx_characteristic, rx_characteristic]}
        self.current_state = None # Current state of the bluetooth hardware
        self.scanning = False #Are we currently scanning for peripherals
        self.is_authenticated = {} #Maps froM addr/id to a bool for authentication

        #Start the objective C service
        self.central_manager = self.CBCentralManager.alloc().initWithDelegate_queue_(self, None)

    ####################
    # Helper functions
    def uuids_are_equal(self, a, b):
        #For some reason, UUID equality like this isn't working:
        #characteristic.UUID.isEqual_(self.CBUUID.UUIDWithString_(objc_str(BTLE_SERIAL_RX_UUID)))
        #It always returns True. This works around the problem, but it shouldn't be needed.
        if isinstance(a, str):
            a = self.CBUUID.UUIDWithString_(objc_str(b))

        if isinstance(b, str):
            b = self.CBUUID.UUIDWithString_(objc_str(b))

        a = a.description().cString()
        b = b.description().cString()

        return a == b

    ############################
    # Callbacks
    def on_close_callback(self, id, error):
        self.connections_closed[id] = True

        if error:
            logging.error("An error occurred while closing BTLE connection %r: %s", id, error)

        if callable(self.user_on_close_callback):
            self.user_on_close_callback(id, error)

    def on_open_callback(self, id, error):
        self.connections_closed[id] = False

        if error:
            logging.error("An error occurred while opening BTLE connection %r: %s", id, error)

        if callable(self.user_on_open_callback):
            self.user_on_open_callback(id, error)

    ###############################
    # Central Manager Delegate
    @protocol('CBCentralManagerDelegate')
    @protocol_exception_catcher
    def centralManagerDidUpdateState_(self, central):
        state = central.state
        self.current_state = state

        if state == CBCentralManagerStatePoweredOn:
            logging.info('CentralManager: Bluetooth is powered on')
            if self.scanning:
                self.start_listening()
        else:
            if state == CBCentralManagerStateUnknown:
                logging.info('CentralManager: Unknown state')
            elif state == CBCentralManagerStatePoweredOff:
                logging.info('CentralManager: Bluetooth is powered off')
            elif state == CBCentralManagerStateUnauthorized:
                logging.info('CentralManager: The application is not authorized to use BLE')
            elif state == CBCentralManagerStateUnsupported:
                logging.info('CentralManager: This hardware doesnt support BLE')

    @protocol('CBCentralManagerDelegate')
    @protocol_exception_catcher
    def centralManager_didDiscoverPeripheral_advertisementData_RSSI_(self, central, peripheral, data, rssi):
        #Store the peripheral for later
        uuid = peripheral.identifier.UUIDString().cString()

        addr = data.objectForKey_(objc_str('kCBAdvDataLocalName'))
        states = ('Disconnected', 'Connecting', 'Connected')
        state = states[peripheral.state]
        name = ''
        if peripheral.name:
            name = peripheral.name.cString()

        if addr:
            addr = addr.cString()

        rssi = peripheral.readRSSI() if state == 'Connected' else rssi.integerValue()

        self.peripherals[uuid] = (addr, peripheral)
        self.snap_service[addr] = [peripheral, None, None]

        if callable(self.user_on_discover_callback):
            def discovery_callback():
                self.user_on_discover_callback(uuid, name, addr, rssi)
            self.scheduler.schedule(0, discovery_callback)

    @protocol('CBCentralManagerDelegate')
    @protocol_exception_catcher
    def centralManager_didFailToConnectPeripheral_error_(self, central_manager, peripheral, error):
        uuid = peripheral.identifier.UUIDString().cString()
        addr = self.peripherals[uuid][0]
        error = error.localizedDescription().cString()

        logging.info("Failed to connect to peripheral %s. %s", uuid, error)

        self.on_open_callback(addr, error)

        self.disconnect(addr)

    @protocol('CBCentralManagerDelegate')
    @protocol_exception_catcher
    def centralManager_didConnectPeripheral_(self, central_manager, peripheral):
        logging.info("Connected to peripheral")
        uuid = peripheral.identifier.UUIDString().cString()
        addr = self.peripherals[uuid][0]
        peripheral.delegate = self
        services = objc_arr(self.CBUUID.UUIDWithString_(objc_str(self.service_uuid)))
        peripheral.discoverServices_(services)

        self.is_authenticated[addr] = False
        self.rx_buffer[addr] = "" #Initialize the receive buffer
        self.on_open_callback(addr, None)


    @protocol('CBCentralManagerDelegate')
    @protocol_exception_catcher
    def centralManager_didDisconnectPeripheral_error_(self, central_manager, peripheral, error):
        uuid = peripheral.identifier.UUIDString().cString()

        if error:
            error = error.localizedDescription().cString()

        uuid = peripheral.identifier.UUIDString().cString()
        addr = self.peripherals[uuid][0]

        self.on_close_callback(addr, error)

    ###############################
    # Peripheral Delegate
    @protocol('CBPeripheralDelegate')
    @protocol_exception_catcher
    def peripheral_didDiscoverServices_(self, peripheral, error):
        if error:
            logging.error("Error discovering services : %s", error.localizedDescription().cString())
            self.central_manager.cancelPeripheralConnection_(peripheral)
        else:
            service_count = peripheral.services.count()
            logging.info("Discovered %d services", service_count)
            for i in xrange(service_count):
                service = peripheral.services.objectAtIndex_(i)
                characteristics = objc_arr(self.CBUUID.UUIDWithString_(objc_str(self.tx_uuid)),
                                           self.CBUUID.UUIDWithString_(objc_str(self.rx_uuid)))
                peripheral.discoverCharacteristics_forService_(characteristics, service)

    @protocol('CBPeripheralDelegate')
    @protocol_exception_catcher
    def peripheral_didDiscoverCharacteristicsForService_error_(self, peripheral, service, error):
        if error:
            logging.error("Error discovering characteristics : %s", error.localizedDescription().cString())
            self.central_manager.cancelPeripheralConnection_(peripheral)
        else:
            uuid = peripheral.identifier.UUIDString().cString()
            addr = self.peripherals[uuid][0]
            characteristic_count = service.characteristics.count()
            logging.info("Discovered %d characteristics for %s", characteristic_count, addr)
            for i in xrange(characteristic_count):
                characteristic = service.characteristics.objectAtIndex_(i)

                if self.uuids_are_equal(characteristic.UUID, self.tx_uuid):
                    logging.info("Starting notification for TX Characteristic for %s", addr)
                    peripheral.setNotifyValue_forCharacteristic_(True, characteristic)
                    self.snap_service[addr][1] = characteristic
                elif self.uuids_are_equal(characteristic.UUID, self.rx_uuid):
                    logging.info("Found an RX characteristic for %s", addr)
                    self.snap_service[addr][2] = characteristic

    @protocol('CBPeripheralDelegate')
    @protocol_exception_catcher
    def peripheral_didUpdateValueForCharacteristic_error_(self, peripheral, characteristic, error):
        if error:
            logging.error("Error updating value for characteristic: %s", error.localizedDescription().cString())
        else:
            # If we try to get a string from the data, we truncate at \x00 and fail to decode other characters
            # Using base64 encoding from the NSData and then decoding in Python gets us the correct data, but there
            # may be a better way to do this.
            data = base64.b64decode(characteristic.value.base64EncodedStringWithOptions_(0).cString())

            uuid = peripheral.identifier.UUIDString().cString()
            addr = self.peripherals[uuid][0]

            if self.verbose_logging:
                logging.debug("BTLE %r Received %r", addr, data)

            if self.is_authenticated[addr]:
                self.rx_buffer[addr] += data
            elif data == "ACCEPTED":
                self.is_authenticated[addr] = True
                if callable(self.user_on_authentication_callbacks.get(addr,None)):
                    self.user_on_authentication_callbacks[addr]()
            else:
                #We're receiving data and we're not authenticated, this must be the key
                server_key = data
                client_key = ''.join(random.Random().sample(string.letters + string.digits, 8))
                user_name, password = self.client_auth(BLUETOOTH_REALM)

                #Send over the username
                self.tx(addr, user_name)

                #Send over the client key
                self.tx(addr, client_key)

                #Send over the hash value
                unhashed = "%s:%s:%s:%s:%s"%(server_key, client_key, user_name, BLUETOOTH_REALM, password)
                self.tx(addr, binascii.unhexlify(hashlib.md5(unhashed).hexdigest()))


    @protocol('CBPeripheralDelegate')
    @protocol_exception_catcher
    def peripheral_didUpdateNotificationStateForCharacteristic_error_(self, peripheral, characteristic, error):
        if error:
            logging.error("Error changing notification state: %s", error.localizedDescription().cString())
        else:

            #Exit if it's not the transfer characteristic
            if not self.uuids_are_equal(characteristic.UUID, self.tx_uuid):
                return

            uuid = peripheral.identifier.UUIDString().cString()
            addr = self.peripherals[uuid][0]

            #Notification has started
            if characteristic.isNotifying:
                logging.info("Notifications started")

                #After our connection is opened, and everything is read, start authentication
                logging.info("Sending Connect Message...")
                self.tx(addr, 'CONNECT%s'%self.local_name)
            else:
                logging.info("Notifications stopped! Disconnecting from peripheral.")
                self.disconnect(addr)

    @protocol('CBPeripheralDelegate')
    @protocol_exception_catcher
    def peripheral_didWriteValueForCharacteristic_error_(self, peripheral, characteristic, error):
        if error:
            logging.error("Error writing characteristic: %s", error.localizedDescription().cString())
        else:
            logging.info("Wrote value for characteristic")

    ###############################
    # SNAP Connect BTLE Interface
    def supports_central(self):
        """Returns True if the central role is supported on this platform"""
        return True

    def supports_peripheral(self):
        """Returns True if the peripheral role is supported on this platform"""
        return False

    def tx(self, id, data):
        data = data[:MAX_WRITE_SIZE]

        if self.verbose_logging:
            logging.debug("BTLE Transmitting %r", data)

        data_length = len(data)
        data = base64.b64encode(data)
        data = self.NSData.alloc().initWithBase64EncodedString_options_(objc_str(data), 0)
        peripheral, tx, rx = self.snap_service[id]

        if not rx:
            logging.debug("BTLE could not transmit because RX characteristic was not discovered")
            return 0

        characteristic_type = 0 #0 - with response or 1 - without response

        peripheral.writeValue_forCharacteristic_type_(data, rx, characteristic_type)

        data.release()

        return data_length

    def rx(self, id, size):
        """Return received data from the btle buffer or an empty string if there is no data in the buffer"""
        result, self.rx_buffer[id] = self.rx_buffer[id][:size], self.rx_buffer[id][size:]
        return result

    def flush_output(self, id):
        """Dump the contents of the TX buffer"""
        pass

    def flush_input(self, id):
        """Dump the contents of the RX buffer"""
        self.rx_buffer[id] = ""

    def is_closed(self, id):
        """Return True if the connection has been closed"""
        return self.connections_closed.get(id, False)

    def set_closed_callback(self, callback):
        """Set a threadsafe callback for when a connection is closed"""
        self.user_on_close_callback = callback

    def set_opened_callback(self, callback):
        """Set a threadsafe callback for when a connection is opened.

        The callback accepts an address and an error message if there was a problem."""
        self.user_on_open_callback = callback

    def set_authentication_callback(self, id, callback):
        self.user_on_authentication_callbacks[id] = callback

    #######################################S
    # Functions required for Peripherals
    def start_advertising(self):
        """Start advertising the serial over gatt service"""
        raise NotImplementedError

    def stop_advertising(self):
        """Stop advertising the serial over gatt service. This closes existing connections."""
        raise NotImplementedError

    #######################################
    # Functions required for Centrals
    def set_client_auth(self, auth):
        self.client_auth = auth

    def set_discovery_callback(self, callback):
        """Set a callback for BTLE peripheral discovery. The callback accepts the node address and link quality in db as arguments"""
        self.user_on_discover_callback = callback

    def start_listening(self):
        """Start listening for bluetooth peripherals with the GATT serial service

        After a connection is made listening is temporarily disabled."""
        logging.info("Attempting to start scanning...")
        if self.scanning:
            logging.info("Already scanning")
            return
        else:
            self.scanning = True

        if self.current_state == CBCentralManagerStatePoweredOn:
            services = objc_arr(self.CBUUID.UUIDWithString_(objc_str(self.service_uuid)))
            options = objc_dict({'kCBScanOptionAllowDuplicates': objc_b(True)})
            self.central_manager.scanForPeripheralsWithServices_options_(services, options)
            logging.info("Scanning Started")
        else:
            logging.info("Waiting on PoweredOn to start scanning")

    def stop_listening(self):
        """Stop listening for bluetooth peripherals with the GATT serial service"""
        logging.info("Attempting to stop scanning...")
        if not self.scanning:
            logging.info("Not currently scanning")
            return
        else:
            self.scanning = False

        if self.current_state == CBCentralManagerStatePoweredOn:
            self.central_manager.stopScan()
            logging.info("Scanning stopped")
        else:
            logging.info("Device not currently PoweredOn")

    def connect(self, id):
        """Connect to the given peripheral"""
        self.central_manager.connectPeripheral_options_(self.snap_service[id][0], None)

    def disconnect(self, id):
        """Disconnect from the given peripheral"""
        self.central_manager.cancelPeripheralConnection_(self.snap_service[id][0])
