# (c) Copyright 2014, Synapse Wireless, Inc.
"""Bluetooth Low Energy peripheral interface for linux.

This communicates with an application that launches a gatt server.

References:
    https://www.bluetooth.org/en-us/specification/adopted-specifications
        Core_V4.0: Volume 3 - Part F ATTRIBUTE PROTOCOL (ATT)

    https://developer.bluetooth.org/gatt/descriptors/Pages/DescriptorsHomePage.aspx
    https://developer.bluetooth.org/gatt/services/Pages/ServicesHome.aspx
    https://developer.bluetooth.org/gatt/characteristics/Pages/CharacteristicsHome.aspx

    https://developer.apple.com/library/ios/documentation/CoreBluetooth/Reference/CBPeripheralDelegate_Protocol/translated_content/CBPeripheralDelegate.html
    https://developer.apple.com/library/ios/documentation/CoreBluetooth/Reference/CBPeripheralManagerDelegate_Protocol/Reference/CBPeripheralManagerDelegate.html

    https://github.com/sandeepmistry/bleno - A similar project in Node.JS
"""
from btle_constants import *
from btle_service import BtleService, BtleDescriptor, BtleCharacteristic, BtleCharacteristicValue

import os
import logging
import asyncore
import subprocess
import signal
import re
import binascii

def reverse_uuid(uuid):
    """Reverse bytes in a UUID for sending over the air"""
    return ''.join(reversed([uuid[i:i+2] for i in range(0, len(uuid), 2)]))

class BtleDispatcher(asyncore.file_dispatcher):
    """This class interacts with BTLE socket file for sending and receiving data asynchronously"""
    def __init__(self, file_name, on_data):
        """Create a BtleDispatcher that can talk to the BTLE Socket application

        Args:
            file_name : Executable file to run. Uses stdin, stdout, and signals for communications
                        If arguments are passed, this should be a list. Ex: 'my_file' or ['my_file', 'arg1', 'arg2']
            on_data : Callback for when data is received from the application
        """
        if isinstance(file_name, list):
            log_name = file_name[0]
        else:
            log_name = file_name

        self.log = logging.getLogger('BtleDispatcher')

        self.log.info("Starting BTLE Dispatcher")
        self.proc = subprocess.Popen(file_name, stdout=subprocess.PIPE, stdin=subprocess.PIPE)
        asyncore.file_dispatcher.__init__(self, self.proc.stdout)
        self.on_data_callback = on_data

    def handle_read(self):
        data = self.recv(8192)
        self.log.debug("Reading: %r", data)
        self.on_data_callback(data)

    def write(self, data):
        self.log.debug("Writing: %r", data)
        self.proc.stdin.write(data + "\n")

    def signal(self, signal):
        self.log.debug("Signaling %d"%signal)
        self.proc.send_signal(signal)

class BtlePeripheral(object):
    def __init__(self, packet_size = 20, device_id=0, on_close_callback=None, on_open_callback=None, l2cap_path=None):
        """Create a gatt_serial peripheral

        Args:
            packet_size : Size of data to transmit on a read
            device_id : The device ID. See 'hciconfig'
            on_close_callback : Callable for when the peripheral connection is closed
            l2cap_path : Full path to the l2cap_socket application. If None, the path of this python file will be used.
        """
        self.log = logging.getLogger('BtlePeripheral')
        self.log.info("Starting GATT Serial Peripheral")
        self.verbose_logging = False

        self.device_id = device_id
        self.device_name = "hci%d"%device_id
        self.state = "Not Set"

        self.packet_size = packet_size

        #Bluetooth State
        self.mtu = DEFAULT_MTU
        self.security = None

        #Dispatchers - Talk to C applications
        if not l2cap_path:
            l2cap_path = os.path.join(os.path.dirname(__file__), 'l2cap_socket')

        self.l2cap = BtleDispatcher([l2cap_path, str(self.device_id)], self.on_l2cap_data)
        self.l2cap_buffer = ""

        self.handles = [] # handles[handle_id] = BtleService/BtleCharacteristic/BtleDescriptor/BtleCharacteristicValue
        self.subscriptions = [] # A list of handles that are subscribed to

        self.advertising_data = None # ([UUID, UUID, UUID], local_name) - Used for advertising after powerup

        self.connected = False

        #Callbacks
        self.on_close_callback = on_close_callback
        self.on_open_callback = on_open_callback

        self.kill_on_disconnect = False #If True, kill the process after a disconnect occurs

    def notify_subscription(self, handle, data):
        notification = "%c%c%c%s"%(ATT_OP_HANDLE_NOTIFY,
                                  handle & 0xFF,
                                  (handle << 8) & 0xFF,
                                  data[:self.mtu-3])
        self.l2cap.write(binascii.hexlify(notification))

    ####################
    # Receive from apps
    def on_l2cap_data(self, data):
        self.l2cap_buffer += data
        lines = self.l2cap_buffer.split('\n')
        self.l2cap_buffer = lines[-1]

        for line in lines[:-1]:
            if self.verbose_logging:
                logging.debug("l2cap Buffer Received %r",line)

            match = re.match('^data (.*)$', line)
            if match:
                line_data = match.group(1)
                self.l2cap_handle_request(line_data)
                continue

            match = re.match('^accept (.*)$', line)
            if match:
                client_addr = match.group(1)
                self.mtu = DEFAULT_MTU
                self.on_accept(client_addr)
                continue

            match = re.match('^disconnect (.*)$', line)
            if match:
                client_addr = match.group(1)
                self.on_disconnect(client_addr)
                continue

            match = re.match('^state (.*)$', line)
            if match:
                adapter_state = match.group(1)
                self.on_state_change(adapter_state)

            match = re.match('^security (.*)$', line)
            if match:
                # 1 low, 2 medium, 3 high
                self.security = int(match.group(1))
                self.log.info("Setting security %r",self.security)
                continue

    ######################
    # l2cap Request handling
    def get_handle(self, handle):
        try:
            return self.handles[handle-1]
        except IndexError:
            return None

    def set_services(self, services = []):
        # Reset the handles
        self.handles = []
        handle = 0

        for service in services:
            handle += 1
            service.start_handle = handle
            self.handles.append(service)

            for characteristic in service.characteristics:
                handle += 1
                characteristic.handle = handle
                self.handles.append(characteristic)

                handle += 1
                characteristic.value.handle = handle
                self.handles.append(characteristic.value)

                for descriptor in characteristic.descriptors:
                    handle += 1
                    descriptor.handle = handle
                    self.handles.append(descriptor)

            service.end_handle = handle

        if self.verbose_logging:
            logging.info("Handles added:")
            for handle in self.handles:
                logging.info("%s",handle)

    def l2cap_error_response(self, opcode, handle, status):
        self.log.debug("Sending error response. opcode=0x%x handle=%r status=%r", opcode, handle, status)
        return binascii.hexlify("%c%c%c%c%c"%(ATT_OP_ERROR, opcode, handle&0xFF, (handle >> 8)&0xFF, status))

    def l2cap_handle_request(self, request):
        request = binascii.unhexlify(request)
        request_type = ord(request[0])
        response = None

        if request_type == ATT_OP_MTU_REQ:
            response = self.handle_mtu_request(request)

        elif request_type == ATT_OP_FIND_INFO_REQ:
            response = self.handle_find_info_request(request)

        elif request_type == ATT_OP_FIND_BY_TYPE_REQ:
            response = self.handle_find_by_type_request(request)

        elif request_type == ATT_OP_READ_BY_TYPE_REQ:
            response = self.handle_read_by_type_request(request)

        elif request_type in [ATT_OP_READ_REQ, ATT_OP_READ_BLOB_REQ]:
            response = self.handle_read_request(request)

        elif request_type == ATT_OP_READ_BY_GROUP_REQ:
            response = self.handle_read_by_group_request(request)

        elif request_type in [ATT_OP_WRITE_REQ, ATT_OP_WRITE_CMD]:
            response = self.handle_write_request(request)

        else:
            response = self.l2cap_error_response(request_type, 0x0000, ATT_ECODE_REQ_NOT_SUPP)

        if response:
            self.l2cap.write(response)

    # Request Handling
    def handle_mtu_request(self, request):
        request = request[1:] #Remove the request code
        self.mtu = (ord(request[1]) << 8) + ord(request[0])
        self.log.info("Setting MTU to %d",self.mtu)
        return binascii.hexlify("%c%s"%(ATT_OP_MTU_RESP, request))

    def handle_find_info_request(self, request):
        response = None
        request = request[1:] #Remove the request code
        start_handle = (ord(request[1]) << 8) + ord(request[0])
        end_handle = (ord(request[3]) << 8) + ord(request[2])

        self.log.info("Find Info Request: 0x%04x - 0x%04x", start_handle, end_handle)

        uuid_size = None
        attribute_data_list = ""
        attributes_added = 0
        for handle in self.handles[start_handle-1:end_handle]:
            if isinstance(handle, BtleService):
                uuid = '2800' #Primary Service
            elif isinstance(handle, BtleCharacteristic):
                uuid = '2803' #Characteristic
            elif isinstance(handle, BtleCharacteristicValue):
                uuid = self.handles[handle.handle - 1].uuid #Get uuid of Characteristic
            elif isinstance(handle, BtleDescriptor):
                uuid = handle.uuid
            else:
                continue

            if not uuid_size:
                uuid_size = len(uuid)/2
                length_per_info = uuid_size + 2
                max_attributes = (self.mtu - 2) // length_per_info
            elif len(uuid)/2 != uuid_size:
                break

            attributes_added += 1
            if attributes_added == max_attributes:
                break

            uuid = binascii.unhexlify(reverse_uuid(uuid))
            attribute_data_list = "%s%c%c%s"%(attribute_data_list,
                                              handle.handle & 0xFF,
                                              (handle.handle << 8) & 0xFF,
                                              uuid)

        if attribute_data_list:
            format = 0x01 if uuid_size == 2 else 0x02
            response = binascii.hexlify("%c%c%s"%(ATT_OP_FIND_INFO_RESP, format, attribute_data_list))
        else:
            response = self.l2cap_error_response(ATT_OP_FIND_INFO_REQ, start_handle, ATT_ECODE_ATTR_NOT_FOUND)

        return response

    def handle_find_by_type_request(self, request):
        self.log.info("Calling unimplemented handle_find_by_type_request")

    def handle_read_by_type_request(self, request):
        response = None
        request = request[1:] #Remove the request code
        start_handle = (ord(request[1]) << 8) + ord(request[0])
        end_handle = (ord(request[3]) << 8) + ord(request[2])
        request = binascii.hexlify(request[4:])
        uuid = reverse_uuid(request)

        self.log.info("Read By Type Request: 0x%04x - 0x%04x uuid:%s", start_handle, end_handle, uuid)

        if '2803' == uuid: #Characteristic
            characteristics = filter(lambda h: isinstance(h, BtleCharacteristic), self.handles[start_handle-1:end_handle])

            if characteristics:
                uuid_size = len(characteristics[0].uuid) / 2
                characteristic_count = 0

                for characteristic in characteristics:
                    if len(characteristic.uuid)/2 != uuid_size:
                        break
                    characteristic_count += 1

                characteristic_length = (uuid_size + 5)
                max_characteristics = (self.mtu - 2) // characteristic_length

                attribute_data_list = ""
                for characteristic in characteristics[:min(characteristic_count, max_characteristics)]:
                    characteristic_uuid = binascii.unhexlify(reverse_uuid(characteristic.uuid))
                    attribute_data_list = "%s%c%c%c%c%c%s"%(attribute_data_list,
                                                          characteristic.handle & 0xFF,
                                                          (characteristic.handle << 8) & 0xFF,
                                                          characteristic.properties,
                                                          characteristic.value.handle & 0xFF,
                                                          (characteristic.value.handle << 8) & 0xFF,
                                                          characteristic_uuid)

                response = binascii.hexlify("%c%c%s"%(ATT_OP_READ_BY_TYPE_RESP, characteristic_length, attribute_data_list))
            else:
                response = self.l2cap_error_response(ATT_OP_READ_BY_TYPE_REQ, start_handle, ATT_ECODE_ATTR_NOT_FOUND)
        else:
            value_handle = None
            is_secure = False

            for handle in self.handles[start_handle-1:end_handle]:
                if handle.uuid == uuid:
                    if isinstance(handle, BtleCharacteristic):
                        value_handle = handle.value
                        is_secure = handle.secure & BTLE_READ
                        break
                    elif isinstance(handle, BtleDescriptor):
                        value_handle = handle
                        is_secure = handle.secure & BTLE_READ
                        break

            if is_secure and (self.security != BTLE_SECURITY_MEDIUM and self.security != BTLE_SECURITY_HIGH):
                response = self.l2cap_error_response(ATT_OP_READ_BY_TYPE_REQ, start_handle, ATT_ECODE_AUTHENTICATION)
            elif value_handle:
                data = value_handle.value

                data_length = min(len(data), self.mtu - 4)
                response = binascii.hexlify("%c%c%c%c%s"%(ATT_OP_READ_BY_TYPE_RESP,
                                         len(data) + 2,
                                         value_handle.handle & 0xFF,
                                         (value_handle.handle << 8) & 0xFF,
                                         data))
            else:
                response = self.l2cap_error_response(ATT_OP_READ_BY_TYPE_REQ, start_handle, ATT_ECODE_ATTR_NOT_FOUND)

        return response

    def handle_read_request(self, request):
        response = None

        request_type = ord(request[0])
        value_handle = (ord(request[2]) << 8) + ord(request[1])

        if request_type == ATT_OP_READ_BLOB_REQ:
            offset = (ord(request[4]) << 8) + ord(request[3])
            self.log.info("Read Blob Request: handle:0x%04x offset:%d", value_handle, offset)
        else:
            offset = 0
            self.log.info("Read Request: handle:0x%04x", value_handle)

        handle = self.get_handle(value_handle)

        if handle:
            result = None
            data = None

            if isinstance(handle, BtleService):
                result = ATT_ECODE_SUCCESS
                data = binascii.unhexlify(reverse_uuid(handle.uuid))

            elif isinstance(handle, BtleCharacteristic):
                result = ATT_ECODE_SUCCESS
                uuid = binascii.unhexlify(reverse_uuid(handle.uuid))
                data = "%c%c%c%s"%(handle.properties,
                                 handle.value.handle & 0xFF,
                                 (handle.value.handle << 8) & 0xFF,
                                 uuid)
            elif isinstance(handle, BtleCharacteristicValue) or isinstance(handle, BtleDescriptor):
                if isinstance(handle, BtleCharacteristicValue):
                    #Get the characteristic handle for the value
                    read_handle = self.get_handle(value_handle - 1)
                else:
                    #Descriptors have all the information we need
                    read_handle = handle

                if read_handle.properties & BTLE_READ:
                    if read_handle.secure & BTLE_READ and (self.security != BTLE_SECURITY_MEDIUM and self.security != BTLE_SECURITY_HIGH):
                        result = ATT_ECODE_AUTHENTICATION
                    else:
                        result = ATT_ECODE_SUCCESS
                        if read_handle.on_read:
                            data = read_handle.on_read()
                        else:
                            data = handle.value
                else:
                    result = ATT_ECODE_READ_NOT_PERM

            if result == ATT_ECODE_SUCCESS and data and offset:
                if len(data) < offset:
                    errorCode = ATT_ECODE_INVALID_OFFSET
                    data = None
                else:
                    data = data[offset:]

            if result == ATT_ECODE_SUCCESS:
                data_length = min(len(data), self.mtu - 1)
                response_type = ATT_OP_READ_BLOB_RESP if request_type == ATT_OP_READ_BLOB_REQ else ATT_OP_READ_RESP
                response = binascii.hexlify("%c%s"%(response_type, data))
            else:
                response = self.l2cap_error_response(request_type, value_handle, result)
        else:
            response = self.l2cap_error_response(request_type, value_handle, ATT_ECODE_INVALID_HANDLE)

        return response

    def handle_read_by_group_request(self, request):
        response = None
        request = request[1:] #Remove the request code
        start_handle = (ord(request[1]) << 8) + ord(request[0])
        end_handle = (ord(request[3]) << 8) + ord(request[2])
        request = binascii.hexlify(request[4:])
        uuid = reverse_uuid(request)

        self.log.info("Read By Group Request: 0x%04x - 0x%04x uuid:%s", start_handle, end_handle, uuid)

        if '2800' == uuid or '2802' == uuid: #Primary Service / Include
            services = []

            if uuid == '2800': #Primary Service
                services = filter(lambda h: isinstance(h, BtleService), self.handles[start_handle-1:end_handle])

            if services:
                uuid_size = len(services[0].uuid) / 2
                service_count = 0

                for service in services:
                    if len(service.uuid)/2 != uuid_size:
                        break
                    service_count += 1

                max_services = (self.mtu - 2) // (uuid_size + 4)

                attribute_data_list = ""
                for service in services[:min(service_count, max_services)]:
                    service_uuid = binascii.unhexlify(reverse_uuid(service.uuid))
                    attribute_data_list = "%s%c%c%c%c%s"%(attribute_data_list,
                                                          service.start_handle & 0xFF,
                                                          (service.start_handle << 8) & 0xFF,
                                                          service.end_handle & 0xFF,
                                                          (service.end_handle << 8) & 0xFF,
                                                          service_uuid)



                response = binascii.hexlify("%c%c%s"%(ATT_OP_READ_BY_GROUP_RESP, len(attribute_data_list), attribute_data_list))
            else:
                response = self.l2cap_error_response(ATT_OP_READ_BY_GROUP_REQ, start_handle, ATT_ECODE_ATTR_NOT_FOUND)
        else:
            response = self.l2cap_error_response(ATT_OP_READ_BY_GROUP_REQ, start_handle, ATT_ECODE_UNSUPP_GRP_TYPE)

        return response

    def handle_write_request(self, request):
        response = None
        request_type = ord(request[0])
        value_handle = (ord(request[2]) << 8) + ord(request[1])

        data = request[3:]

        handle = self.get_handle(value_handle)

        if request_type == ATT_OP_WRITE_CMD:
            self.log.info("Write Command: 0x%04x value:%r", value_handle, data)
        elif request_type == ATT_OP_WRITE_REQ:
            self.log.info("Write Request: 0x%04x value:%r", value_handle, data)

        if handle:
            if isinstance(handle, BtleCharacteristicValue):
                #Get the characteristic handle for the value
                write_handle = self.get_handle(value_handle - 1)
            else:
                write_handle = handle

            if request_type == ATT_OP_WRITE_CMD:
                can_write = write_handle.properties & BTLE_WRITE_WO_RESPONSE
                is_secure = write_handle.secure & BTLE_WRITE_WO_RESPONSE
            elif request_type == ATT_OP_WRITE_REQ:
                can_write = write_handle.properties & BTLE_WRITE
                is_secure = write_handle.secure & BTLE_WRITE
            else:
                can_write = False

            if can_write:
                if is_secure and (self.security != BTLE_SECURITY_MEDIUM and self.security != BTLE_SECURITY_HIGH):
                    response = self.l2cap_error_response(request_type, value_handle, ATT_ECODE_AUTHENTICATION)
                else:
                    result = None

                    if isinstance(handle, BtleDescriptor):
                        if len(data) != 2:
                            result = ATT_ECODE_INVAL_ATTR_VALUE_LEN
                        else:
                            value = (ord(data[1]) << 8) + ord(data[0])

                            handle.value = data
                            if write_handle.on_write:
                                write_handle.on_write(data)

                            if handle.uuid == '2902':
                                #Client Characteristic Configuration - Used to subscribe/unsubscribe
                                characteristic_handle = value_handle - 1
                                if value & 0x0001: #Subscribing to data
                                    self.log.info("Subscription started for handle %s", characteristic_handle)
                                    self.subscriptions.append(characteristic_handle)
                                else:
                                    self.log.info("Subscription stopped for handle %s", characteristic_handle)
                                    self.subscriptions.remove(characteristic_handle)

                            result = ATT_ECODE_SUCCESS
                    else:
                        #Regular Write to characteristic value
                        handle.value = data
                        if write_handle.on_write:
                            write_handle.on_write(data)
                            result = ATT_ECODE_SUCCESS

                    if request_type == ATT_OP_WRITE_REQ:
                        if result == ATT_ECODE_SUCCESS:
                            response = binascii.hexlify("%c"%ATT_OP_WRITE_RESP)
                        else:
                            response = self.l2cap_error_response(request_type, value_handle, result)
            else:
                response = self.l2cap_error_response(request_type, value_handle, ATT_ECODE_WRITE_NOT_PERM)
        else:
            response = self.l2cap_error_response(request_type, value_handle, ATT_ECODE_INVALID_HANDLE)

        return response

    #######################################################
    # Bluetooth Hooks
    def on_accept(self, addr):
        self.log.info("Connection accepted %r",addr)
        self.connected = True

        if callable(self.on_open_callback):
            self.on_open_callback()

    def on_disconnect(self, addr):
        self.log.info("Connection disconnected %r", addr)
        self.connected = False

        self.restart_advertising()

        if callable(self.on_close_callback):
            self.on_close_callback()

        if self.kill_on_disconnect:
            self.kill_server()

    def on_advertising_start(self, result):
        self.log.info("Advertising started %s", result)

    def on_advertising_stop(self):
        self.log.info("Advertising stopped")

    def on_state_change(self,state):
        self.log.info("State Changed to %r", state)
        self.state = state

        if state == 'poweredOn':
            if self.advertising_data:
                self.start_advertising(*self.advertising_data)
        else:
            self._no_leadv()

    ###################################
    # Bluetooth Hardware Interface
    def stop_peripheral(self):
        """Completely stop the peripheral application. Terminates advertising and connections."""
        self.stop_advertising()

        if self.connected:
            self.kill_on_disconnect = True
            self.disconnect()
        else:
            self.kill_server()

    def kill_server(self):
        self.l2cap.signal(signal.SIGKILL)

    def disconnect(self):
        self.l2cap.signal(signal.SIGHUP)

    def start_advertising(self, service_uuids, local_name=None):
        # References: Bluetooth Specification Version 4.0 [ Vol 2 ] 7.8.7 LE Set Advertising Data Command
        if self.state != "poweredOn":
            #We will automatically start advertising when the the device is ready
            self.advertising_data = (service_uuids, local_name)
            return

        error = None
        service_uuids_128bit = ""
        service_uuids_16bit = ""

        for service in service_uuids:
            #Change the byte order
            service = binascii.unhexlify(reverse_uuid(service.replace("-","")))

            if len(service) == 2:
                service_uuids_16bit+=(service)
            elif len(service) == 16:
                service_uuids_128bit+=(service)

        #Each block is formatted as: "%c%c%s"%(len(data+1),data_type,data)
        advertisement_data = ""
        advertisement_data = "%s%c%c%c"%(advertisement_data, 0x02, 0x01, 0x05) #Flags: Limited Discoverable Mode | BR/EDR Not supported

        if local_name:
            data_type = 0x09 #Complete local name
            advertisement_data = "%s%c%c%s"%(advertisement_data, len(local_name) + 1, data_type, local_name)

        if service_uuids_16bit:
            data_type = 0x03 #Complete list of 16-but UUIDs available
            advertisement_data = "%s%c%c%s"%(advertisement_data, len(service_uuids_16bit) + 1, data_type, service_uuids_16bit)

        if service_uuids_128bit:
            data_type = 0x06 #More 128-bit UUIDs available
            advertisement_data = "%s%c%c%s"%(advertisement_data, len(service_uuids_128bit) + 1, data_type, service_uuids_128bit)

        if len(advertisement_data) > 31:
            error = "Bluetooth advertisement must be a maximum of 31 bytes"

        #Convert to format for hcitool (\xaa\xbb\xcc -> "aa bb cc 00 00 00 ... 00")
        advertisement = "%c%s"%(len(advertisement_data), advertisement_data.ljust(31, '\x00'))
        advertisement = binascii.hexlify(advertisement)
        advertisement = ' '.join(a+b for a,b in zip(advertisement[::2], advertisement[1::2]))

        advertise_command = ['hcitool', '-i', self.device_name, 'cmd','0x08', '0x0008'] + advertisement.split()
        self.log.debug("Starting advertisement with %s",' '.join(advertise_command))
        #Must call leadv to start advertising BEFORE setting advertising data
        # Otherwise, leadv will overwrite whatever is configured
        try:
            subprocess.check_output(['hciconfig', self.device_name, 'leadv'])
        except subprocess.CalledProcessError:
            pass
        subprocess.check_output(advertise_command)

        self.on_advertising_start(error)

    def _no_leadv(self):
        """Execute the noleadv command to stop advertising"""
        try:
            subprocess.check_output(['hciconfig', self.device_name, 'noleadv'])
        except subprocess.CalledProcessError:
            #noleadv fails if we are already stopped
            pass

    def restart_advertising(self):
        """If already advertising, restart the advertisement"""
        if self.advertising_data:
            self._no_leadv()
            self.start_advertising(*self.advertising_data)

    def stop_advertising(self):
        """Stop advertising and prevent power on from starting advertising"""
        self._no_leadv()
        self.advertising_data = None
