# (c) Copyright 2014, Synapse Wireless, Inc.
"""blte_services provides classes used to define services and characteristics for
use with bluetooth low energy peripherals
"""
from btle_constants import *

class BtleService(object):
    def __init__(self, uuid, characteristics):
        """Represents a BTLE Service

        Args:
            uuid : A string UUID without dashes
            characteristics : List of BtleCharacteristics
        """
        self.uuid = uuid
        self.characteristics = characteristics

        self.start_handle = None
        self.end_handle = None

    def __str__(self):
        return "<BtleService 0x%04x-0x%04x %s>"%(self.start_handle, self.end_handle, self.uuid)

class BtleDescriptor(object):
    def __init__(self, uuid, value, properties=BTLE_NONE, secure=BTLE_NONE, on_write=None, on_read=None):
        """Represents a BTLE Descriptor

        Args:
            uuid : A string UUID without dashes
            value : The descriptor value
            properties : Flags for property values (ex: BTLE_READ | BTLE_WRITE | BTLE_NOTIFY)
            secure : Flags for security values (ex: BTLE_READ | BTLE_WRITE | BTLE_NOTIFY)
            on_write : Callable that accepts a value to be written
            on_read : Callable that returns the read value
        """
        self.uuid = uuid
        self.value = value
        self.properties = properties
        self.secure = secure
        self.handle = None
        self.on_read = on_read
        self.on_write = on_write

    def __str__(self):
        return "<BtleDescriptor 0x%04x %s>"%(self.handle, self.uuid)

class BtleCharacteristic(object):
    def __init__(self, uuid, properties, secure=BTLE_NONE, value=None, descriptors=None, on_write=None, on_read=None):
        """Represents a BTLE Characteristic

        Args:
            uuid : A string UUID without dashes
            properties : Flags for property values (ex: BTLE_READ | BTLE_WRITE | BTLE_NOTIFY)
            secure : Flags for security values (ex: BTLE_READ | BTLE_WRITE | BTLE_NOTIFY)
            value : Static value for the characteristic
            descriptors : A List of BtleDescriptor objects
            on_write : Callable that accepts a value to be written
            on_read : Callable that returns the read value
        """
        if not descriptors:
            descriptors = []

        self.uuid = uuid
        self.properties = properties
        self.secure = secure & properties
        self.value = BtleCharacteristicValue(None, value)
        self.descriptors = descriptors
        self.handle = None
        self.on_write = on_write
        self.on_read = on_read

        if BTLE_NOTIFY & self.properties:
            # We need to add a Client Characteristic Configuration Descriptor if we allow notifications
            descriptor = BtleDescriptor('2902',
                                        value = '\x00\x00', #Zero out as the default
                                        properties = BTLE_WRITE|BTLE_READ,
                                        secure = BTLE_WRITE|BTLE_READ if BTLE_NOTIFY & self.secure else BTLE_NONE)
            self.descriptors.insert(0, descriptor)


    def __str__(self):
        return "<BtleCharacteristic 0x%04x %s>"%(self.handle, self.uuid)

class BtleCharacteristicValue(object):
    def __init__(self, handle, value):
        """Represents a BTLE Characteristic Value

        Args:
            handle: Bluetooth handle for the characteristic
            value : Value for the characteristic
        """
        self.uuid = None #Values don't have associated UUIDs
        self.handle = handle
        self.value = value

    def __str__(self):
        return "<BtleCharacteristicValue 0x%04x>"%(self.handle)