# (c) Copyright 2014, Synapse Wireless, Inc.
"""BTLE Peripheral implementation of SNAP Serial over GATT"""
from btle_constants import *
from btle_service import BtleService, BtleDescriptor, BtleCharacteristic, BtleCharacteristicValue
import btle_peripheral

import binascii
import logging
import platform
import random
import string
import hashlib
import os

MAX_NOTIFICATION_SIZE = 20
AUTHENTICATION_TIMEOUT = 2 #Time in seconds to authenticate

#Authentication States
WAITING_ON_CONNECT = "waiting on connect"
WAITING_ON_USERNAME = "waiting on username"
WAITING_ON_USER_KEY = "waiting on user key"
WAITING_ON_HASH = "waiting on hash"

BLUETOOTH_REALM = 'bluetooth'

class GattSerialPeripheral(object):
    def __init__(self, scheduler, local_name, service_uuid, tx_uuid, rx_uuid, device_id=0):
        """Create a GATT Serial peripheral.

        References:
            This implements all of the methods required to be a peripheral for
            btle_serial.py. See class BtleSerialHardwareInterface.

        Args:
            scheduler : The SNAP Connect instance scheduler
            local_name : Name to advertise. Should be address in format 'FF0102'
            service_uuid : UUID for the GATT Serial Service
            tx_uuid : UUID for the TX characteristic
            rx_uuid : UUID for the RX characteristic
            device_id : The device ID. See 'hciconfig'
        """
        self.log = logging.getLogger('GattSerialPeripheral')

        self.scheduler = scheduler
        self.local_name = local_name
        self.device_id = device_id
        self.rx_buffer = ""

        #Remove dashes from the UUIDs
        self.service_uuid = service_uuid.replace('-','')
        self.tx_uuid = tx_uuid.replace('-','')
        self.rx_uuid = rx_uuid.replace('-','')

        #Generic Access Service
        self.device_name_characteristic = BtleCharacteristic('2a00', BTLE_READ, value=platform.uname()[1])
        self.appearance_characteristic = BtleCharacteristic('2a01', BTLE_READ, value='\x80\x00') #Generic Computer
        self.generic_service = BtleService('1800', [self.device_name_characteristic, self.appearance_characteristic])

        #Serial Service
        self.tx_characteristic = BtleCharacteristic(self.tx_uuid, BTLE_NOTIFY)
        self.rx_characteristic = BtleCharacteristic(self.rx_uuid, BTLE_WRITE, on_write=self.on_rx_write)
        self.serial_service = BtleService(self.service_uuid, [self.tx_characteristic, self.rx_characteristic])
        self.services = [self.generic_service, self.serial_service]

        self.peripheral = None #BtlePeripheral object
        self.central_addr = None #Address of the central who has connected

        self.connection_closed = False

        self.user_on_close_callback = None
        self.user_on_open_callback = None

        # Authentication Values
        self.user_on_authentication_callback = None
        self.server_auth = None #Authentication function. Realm is 'bluetooth'
        self.authenticated = False #True if the central has authenticated itself
        self.timeout_event = None #scheduler event for authentication timeout
        self.auth_key = ""
        self.auth_state = None
        self.user_name = None
        self.user_key = None

        self.verbose_logging = False #Enable for even more verbose logging!

        #Path to the l2cap_socket application
        self.l2cap_path = os.path.join(os.path.dirname(__file__), 'l2cap_socket')

    ############################
    # Characteristic Callbacks
    def on_rx_write(self, data):
        """Called when the central writes the RX characteristic.

        This function adds the new characters to the receive buffer
        """
        if self.authenticated:
            if self.verbose_logging:
                self.log.debug("Received Data: %r", data)
            #If we're authenticated, this is serial traffic
            self.rx_buffer += data

        elif self.auth_state == WAITING_ON_CONNECT and data.startswith("CONNECT"):
            #We're not authenticated and receive a connect message
            # Parse the address out of the message
            if len(data) == 13:
                self.central_addr = binascii.unhexlify(data[7:])

            self.log.info("Received BTLE connection request from %s", data[7:])
            #Respond to the connect with an auth key
            self.auth_key = ''.join(random.Random().sample(string.letters + string.digits, 16))
            self.tx(None, self.auth_key)
            self.auth_state = WAITING_ON_USERNAME

        elif self.auth_state == WAITING_ON_USERNAME:
            self.log.info("Received username: %r", data)
            self.user_name = data
            self.auth_state = WAITING_ON_USER_KEY

        elif self.auth_state == WAITING_ON_USER_KEY:
            self.log.info("Received user key: %r", data)
            self.user_key = data
            self.auth_state = WAITING_ON_HASH

        elif self.auth_state == WAITING_ON_HASH:
            self.log.info("Received user hash: %r", data)

            password = self.server_auth(BLUETOOTH_REALM, self.user_name)

            unhashed = "%s:%s:%s:%s:%s"%(self.auth_key, self.user_key, self.user_name, BLUETOOTH_REALM, password)
            expected_value = binascii.unhexlify(hashlib.md5(unhashed).hexdigest())

            #Disable the timeout, we've either connected or failed at this point
            self.timeout_event.Stop()
            self.timeout_event = None

            if data == expected_value:
                self.log.info("Authentication successful")
                self.auth_state = None
                self.authenticated = True
                error = None
                self.tx(None, "ACCEPTED")
            else:
                self.log.info("Authentication failed")
                error = "invalid password"
                self.peripheral.disconnect()

            if callable(self.user_on_open_callback):
                self.user_on_open_callback(self.central_addr, error)

            if not error:
                self.user_on_authentication_callback()
        else:
            self.log.info("Received data with invalid auth state: %r", data)
    def on_close_callback(self):
        self.connection_closed = True
        self.authenticated = False
        self.auth_state = None
        self.user_name = None
        self.user_key = None

        if callable(self.user_on_close_callback):
            self.user_on_close_callback(self.central_addr, None)

        self.central_addr = None

    def on_open_callback(self):
        self.connection_closed = False
        self.authenticated = False
        self.central_addr = None
        self.auth_state = WAITING_ON_CONNECT
        self.user_name = None
        self.user_key = None

        #Start a timeout for authentication
        self.timeout_event = self.scheduler.schedule(AUTHENTICATION_TIMEOUT, self.on_authentication_timeout)

    def on_authentication_timeout(self):
        """When authentication times out, we need to close the connection and call the open callback with an error"""
        self.timeout_event = None

        if callable(self.user_on_open_callback):
            self.user_on_open_callback(self.central_addr, "authentication was not provided before timeout")

        self.peripheral.disconnect()

    ###############################
    # SNAP Connect BTLE Interface
    def supports_central(self):
        """Returns True if the central role is supported on this platform"""
        return False

    def supports_peripheral(self):
        """Returns True if the peripheral role is supported on this platform"""
        if os.path.exists(self.l2cap_path):
            return True
        else:
            return False

    def tx(self, id, data):
        """Transmit data and return the amount written"""
        if self.peripheral:
            handle = self.tx_characteristic.value.handle
            if self.peripheral.connected:
                if handle in self.peripheral.subscriptions:
                    data = data[:MAX_NOTIFICATION_SIZE] #Max notification size
                    if self.verbose_logging:
                        self.log.debug("Sending Data: %r", data)
                    self.peripheral.notify_subscription(handle, data)
                    return len(data)
        else:
            self.log.debug("Unable to send BTLE serial data. The peripheral has not been created.")
        return 0

    def rx(self, id, size):
        """Return received data from the btle buffer or an empty string if there is no data in the buffer"""
        result, self.rx_buffer = self.rx_buffer[:size], self.rx_buffer[size:]
        return result

    def flush_output(self, id):
        """Dump the contents of the TX buffer"""
        pass

    def flush_input(self, id):
        """Dump the contents of the RX buffer"""
        self.rx_buffer = ""

    def is_closed(self, id):
        """Return True if the connection has been closed"""

        if self.peripheral:
            return self.connection_closed
        else:
            return True

    def set_closed_callback(self, callback):
        """Set a threadsafe callback for when a connection is closed"""
        self.user_on_close_callback = callback

    def set_opened_callback(self, callback):
        """Set a threadsafe callback for when a connection is opened"""
        self.user_on_open_callback = callback

    def set_authentication_callback(self, id, callback):
        self.user_on_authentication_callback = callback

    #######################################S
    # Functions required for Peripherals
    def start_advertising(self):
        """Start advertising the serial over gatt service"""
        if not self.peripheral:
            self.peripheral = btle_peripheral.BtlePeripheral(on_close_callback = self.on_close_callback,
                                                             on_open_callback = self.on_open_callback,
                                                             l2cap_path = self.l2cap_path)
            self.peripheral.set_services(self.services)
            self.peripheral.start_advertising([self.service_uuid], self.local_name)
            self.connection_closed = False

    def stop_advertising(self):
        """Stop advertising the serial over gatt service. This closes existing connections."""
        if self.peripheral:
            self.peripheral.stop_peripheral()
            self.peripheral = None

    def set_server_auth(self, auth):
        """Set the authentication function. Accepts a realm,username and returns a password.

        The realm will be set to bluetooth"""
        self.server_auth = auth

    #######################################
    # Functions required for Centrals
    def set_discovery_callback(self, callback):
        """Set a callback for BTLE peripheral discovery. The callback accepts the node address and link quality in db as arguments"""
        raise NotImplementedError

    def start_listening(self):
        """Start listening for bluetooth peripherals with the GATT serial service

        After a connection is made listening is temporarily disabled."""
        raise NotImplementedError

    def stop_listening(self):
        """Stop listening for bluetooth peripherals with the GATT serial service

        Existing connections are closed when listening is stopped"""
        raise NotImplementedError

    def connect(self, id):
        """Connect to the given peripheral"""
        raise NotImplementedError

    def disconnect(self, id):
        """Disconnect from the given peripheral"""
        raise NotImplementedError

if __name__ == "__main__":
    import asyncore
    from threading import Timer
    logging.basicConfig(level=logging.DEBUG)

    BTLE_SERIAL_SERVICE_UUID = "b9be524f-c963-11e3-bb27-3c77e6ec5dea"
    BTLE_SERIAL_TX_UUID = "b9884e2e-c963-11e3-bdf4-3c77e6ec5deb"
    BTLE_SERIAL_RX_UUID = "b9be524f-c963-11e3-bb27-3c77e6ec5dec"
    name = 'FF0001'

    gatt_serial = GattSerialPeripheral(name, BTLE_SERIAL_SERVICE_UUID, BTLE_SERIAL_TX_UUID, BTLE_SERIAL_RX_UUID)
    gatt_serial.start_advertising()

    def test_send():
        gatt_serial.tx("peripheralFF0001","Sending Data")
        Timer(10.0, test_send).start()
    Timer(10.0, test_send).start()

    asyncore.loop()
