# (c) Copyright 2007, Synapse
"""Used to interface with snaplib PacketQueue"""
# Disable "Used * or ** magic"
# pylint: disable-msg=W0142


import logging
log = logging.getLogger("snap.%s" % (__name__))
import time

from snaplib import SnapConstantsHelper

from Responses import EventResponse


class CallProxy(object):
    """A proxy object for functions being called by SNAP nodes"""
    def __init__(self, callSink, netAddr, methodName, snapNetAddr, multicast):
        """Initializes an instance of a CallProxy"""
        assert isinstance(callSink, CallSink)
        self.__callSink = callSink
        self.methodName = methodName
        self.netAddr = netAddr
        self.timestamp = time.time()
        self.snapNetAddr = snapNetAddr
        self.multicast = multicast

    def __call__(self, *args):
        """When the proxy object gets called store the information"""
        response = EventResponse(self.netAddr, self.methodName, self.timestamp, *args)
        if self.multicast:
            self.__callSink.addMulticastResponse(response, self.netAddr)
        else:
            self.__callSink.addResponse(response, self.snapNetAddr)


class CallSink(object):
    """Called by PacketQueue on RPC and data mode events"""
    def __init__(self):
        """Initializes an instance of a CallSink"""
        self.eventResponses = {}
        self._addResponseCallbacks = []

    def __getattribute__(self, name):
        """Overrides the default object version so we can RPC arbitrary functions"""
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            #remoteAddr, rpcDescriptor, and portalNetAddr are set by the calling function
            return CallProxy(self, self.remoteAddr, name, self.portalNetAddr, self.rpcDescriptor.isMulticast)

    def addResponse(self, response, snapNetAddr='\x00\x00\x01'):
        """Add a response from a snap node to the list for the sepcified snapNetAddr
           to be retreived by those waiting

           response -- Data to be sent to waiting clients
           snapNetAddr -- The SNAP network address this response is for
        """
        if not self.eventResponses.has_key(snapNetAddr):
            self.eventResponses[snapNetAddr] = []
        self.eventResponses[snapNetAddr].append(response)

        for callback in self._addResponseCallbacks:
            try:
                callback(snapNetAddr)
            except:
                log.exception("An error occurred while notifying a callback of an event response")

    def addMulticastResponse(self, response, source_addr):
        """Add a multicast from a local client

           response -- Data to be sent to waiting clients
           source_addr -- The originator for this multicast
        """
        for (addr, event_queue) in self.eventResponses.iteritems():
            if addr != source_addr:
                event_queue.append(response)
                for callback in self._addResponseCallbacks:
                    try:
                        callback(addr)
                    except:
                        log.exception("An error occurred while notifying a callback of an event response")

    def onDataModePkt(self, pkt):
        """Called when a data (transparent) mode packet arrives.
           The packet gets encapsulated and added to the response list
        """
        src = pkt[SnapConstantsHelper.DMPTP_ORIG_SRC_OFS : SnapConstantsHelper.DMPTP_ORIG_SRC_OFS + 3]
        dst = pkt[SnapConstantsHelper.DMPTP_FINALDST_OFS : SnapConstantsHelper.DMPTP_FINALDST_OFS + 3]
        data = pkt[SnapConstantsHelper.DMPTP_ORIG_SRC_OFS + 3:]
        self.addResponse(EventResponse(src, 'dataMode', time.time(), data), dst)

    def registerAddResponseCallback(self, callback):
        """
        Register for callbacks when a response gets added

        callback -- A callable object expecting a single parameter of a netaddr
        """
        assert callable(callback)
        if callable(callback):
            self._addResponseCallbacks.append(callback)

    def unregisterAddResponseCallback(self, callback):
        """
        Unregister a callable object from being notified of reponses

        callback -- The callable object to unregister
        """
        try:
            self._addResponseCallbacks.remove(callback)
        except KeyError:
            return  # If it's not in there than it's removed!
