# (c) Copyright 2007-2011, Synapse Wireless Inc.
"""An XML-RPC Server to allow access to a SNAP network"""
__docformat__ = "plaintext en"
# Disable "Used * or ** magic"
# pylint: disable-msg=W0142


import logging, logging.handlers, binascii, asyncore, socket, os, optparse, sys, threading, Queue, time
log = logging.getLogger('SNAPconnect')

START_PATH = str(sys.path[0][:]) #Copy the string of the directory this module was started from

import snap
import snaptcp

from apy import Deferred, AsyncXmlRpc, xmlrpclib

from snaplib import SnappyUploader
from snaplib import RpcCodec
from snaplib import ScriptsManager
from snaplib import SnapConstantsHelper
from snaplib.snappy import hooks

from PacketSinks import CallSink
from Responses import EventResponse, InfoResponse
from Responses import SYNAPSE_XMLRPC_CONNECTION_CLOSED_FAULTCODE, SYNAPSE_XMLRPC_MULTIPE_WAITS_FAULTCODE
import LicenseManager


VERSION = snap.VERSION
SYNAPSE_MAX_LOG_SIZE = 500000 #bytes
SYNAPSE_MAX_LOG_FILES = 5

PORTAL_ADDR_RANGE = range(1, 0x10)


class SnapXmlRpcRequestHandler(object):
    """A SNAPconnect server to a SNAP network that talks via XML-RPC"""
    def __init__(self, closeMethod, translate_portal_addrs=False,snapcomPort=snaptcp.SNAP_IP_PORT):
        """Initializes an instance of a SNAPconnectServer"""
        self.main_snapcom = snap.Snap(rpc_handler=self._on_rpc)
        self.evScheduler = self.main_snapcom.scheduler
        assert isinstance(self.main_snapcom, snap.SNAPcom)
        self.deferreds = {}
        self.uploadDeferreds = {}
        self.callSink = CallSink()
        self.callSink.registerAddResponseCallback(self._onResponseAdd)
        self.callSink.eventResponses[self.main_snapcom.local_addr()] = []
        self.uploadMgr = SnappyUploader.SnappyUploadMgr(self.evScheduler, self.rpc, False)
        self._closeMethod = closeMethod
        self.main_snapcom.set_hook(snap.hooks.HOOK_RPC_SENT, lambda rpc_id, success: self._on_rpc_sent(self.main_snapcom, rpc_id, success))
        self.main_snapcom.set_hook(snap.hooks.HOOK_STDIN, lambda data: self._on_stdin(self.main_snapcom, data))
        self.main_snapcom.set_hook(hooks.HOOK_SERIAL_CLOSE, self._onSerialClose)
        self._snapcomPort = snapcomPort
        self.main_snapcom.accept_tcp(port=snapcomPort)
        self.main_snapcom._outstanding_rpcs = {}
        self._port_type = None
        self._port_num = None
        self.snapcoms = {self.main_snapcom.local_addr(): self.main_snapcom}
        self._pkt_seqs = {}
        self.translate_portal_addrs = translate_portal_addrs
        self.portal_addrs = {} # { snapcom_addr: portal_addr }
        self.snapcom_addrs = {} # { portal_addr: snapcom_addr  }

    def _get_snapcom(self, addr):
        snapcom = self.snapcoms.get(addr, None)
        if snapcom is None:
            if self.translate_portal_addrs:
                int_addr = int(binascii.hexlify(addr), 16)
                if int_addr in PORTAL_ADDR_RANGE:
                    snapcom_addr = self.snapcom_addrs.get(addr, None)
                    if snapcom_addr is None:
                        if len(self.main_snapcom.license.addr) >= len(self.portal_addrs)+1:
                            snapcom_addr = self.main_snapcom.license.addr[len(self.portal_addrs)]
                            self.portal_addrs[snapcom_addr] = addr
                            self.snapcom_addrs[addr] = snapcom_addr
                            return self._get_snapcom(snapcom_addr)
                    else:
                        return self._get_snapcom(snapcom_addr)
            snapcom = snap.SNAPcom(scheduler=self.evScheduler, addr=addr, rpc_handler=self._on_rpc)
            log.debug("Unknown address: %s" % (binascii.hexlify(addr)))
            snapcom.connect_tcp('127.0.0.1',port=self._snapcomPort)
            snapcom.set_hook(snap.hooks.HOOK_RPC_SENT, lambda rpc_id, success: self._on_rpc_sent(snapcom, rpc_id, success))
            snapcom.set_hook(snap.hooks.HOOK_STDIN, lambda data: self._on_stdin(snapcom, data))
            snapcom._outstanding_rpcs = {}
            def add():
                self.snapcoms[addr] = snapcom
            self.evScheduler.schedule(0, add)
        return snapcom

    def __iter__(self):
        return self.callSink.eventResponses.iterkeys()

    def _listMethods(self):
        """Returns the list of methods that are available to be called"""
        return ['addSpy', 'clearQueuedResponses', 'connectSerial', 'delSpy',
                'disconnect', 'exit', 'gatewayInfo', 'mcastRpc', 'rpc',
                'sendDataModePkt', 'uploadSpy', 'waitOnEvent', 'waitOnXmlMethod']

    def _onResponseAdd(self, snapNetAddr):
        """
        Is called when an EventResponse object is added to the queue

        snapNetAddr -- The network address of the new EventResponse object
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data

        deferred_addr = snapNetAddr
        if self.translate_portal_addrs:
            portal_addr = self.portal_addrs.get(snapNetAddr, None)
            if portal_addr:
                deferred_addr = portal_addr

        defferred = self.deferreds.pop(deferred_addr, 0)
        if isinstance(defferred, Deferred.Deferred):
            self._rtnResponse(defferred, snapNetAddr)

    def _onDeferredSocketClose(self, deferred):
        """Called when a socket is closed that had a deferred associated with it"""
        if hasattr(deferred, 'snapNetAddr'):
            if self.deferreds.pop(deferred.snapNetAddr, None) is None:
                # I've commented these lines out because I don't think this is correct
                # when we create the upload deferred we use the address we are uploading to
                # Here we are checking the source addr not the destination addr

                #upload_deferred = self.uploadDeferreds.pop(deferred.snapNetAddr, None)
                #if upload_deferred:
                    #upload_deferred(False)
                pass
            elif not deferred.called and not deferred.errored:
                log.info("Socket unexpectedly closed on wait: %s" % (binascii.hexlify(deferred.snapNetAddr)))
        elif __debug__ and not deferred.called and not deferred.errored:
            log.debug("Socket unexpectedly closed")

    def _on_rpc(self, descriptor):
        prev_seqs = [self._pkt_seqs.get(descriptor.srcAddr, None)]
        if descriptor.isMulticast and prev_seqs[0] is not None:
            prev_seqs = [i % 2**7 for i in range(prev_seqs[0],
                                                 prev_seqs[0]-SnapConstantsHelper.XCAST_DUPLICATE_SEQUENCE_WINDOW-1,
                                                 -1)]
        if descriptor.seq not in prev_seqs:
            if len(self.uploadDeferreds) != 0 and descriptor:
                assert isinstance(descriptor, RpcCodec.RpcDescriptor)
                if descriptor.funcName == 'tellVmStat' and len(descriptor.args) == 2:
                    self.uploadMgr.onTellVmStat(descriptor.srcAddr, descriptor.args[0], descriptor.args[1])
                elif descriptor.funcName == 'su_recvd_reboot' and len(descriptor.args) == 1:
                    self.uploadMgr.on_recvd_reboot(descriptor.srcAddr)
            self._pkt_seqs[descriptor.srcAddr] = descriptor.seq

            #Use function names instead of function IDs
            arg_list = list(descriptor.args)
            list_updated = False
            for i,arg in enumerate(arg_list):
                if isinstance(arg, RpcCodec.RpcFunc):
                    arg_list[i] = arg.func_name
                    list_updated = True

            if list_updated:
                descriptor.args = tuple(arg_list)

            descriptor(self.callSink)

    def _on_rpc_sent(self, snapcom, rpc_id, success):
        deferred = snapcom._outstanding_rpcs.pop(rpc_id, None)
        if deferred:
            if success > 0:
                deferred(True)
            else:
                deferred(False)

    def _on_stdin(self, snapcom, data):
        self.callSink.addResponse(EventResponse(snapcom.rpc_source_addr(), 'dataMode', time.time(), data), snapcom.local_addr())

    def _onSerialClose(self, port_type, port_num):
        """Called when underlying packet serial connection is closed"""
        if port_type == self._port_type and port_num == self._port_num:
            for snapNetAddr in self.deferreds.keys():
                defferred = self.deferreds.pop(snapNetAddr, 0)
                if isinstance(defferred, Deferred.Deferred):
                    defferred(xmlrpclib.Fault(SYNAPSE_XMLRPC_CONNECTION_CLOSED_FAULTCODE, "The underlying connection was closed"),
                              methodresponse=False)

    def _onUploadComplete(self, upload, result):
        """
        Is called when an upload has completed in error or success

        upload -- SnappyUpload object of caller
        result -- Result code
        """
        assert isinstance(upload, SnappyUploader.SnappyUpload)

        _deferred = self.uploadDeferreds.pop(upload.netAddr, None)
        if isinstance(_deferred, Deferred.Deferred):
            if result == SnappyUploader.SNAPPY_PROGRESS_COMPLETE:
                _deferred(True)
            else:
                _deferred(False)

        if len(self.uploadDeferreds) == 0:
            self.deferreds.pop(self.main_snapcom.local_addr(), 0)
            self.callSink.eventResponses[self.main_snapcom.local_addr()] = []

    def _rtnResponse(self, defferred, snapNetAddr):
        """
        Check if there is an EventResponse for the specified network address

        defferred -- The deferred object waiting for an EventResponse
        snapNetAddr -- The network address to see if there are pending Event responses for
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        try:
            er = self.callSink.eventResponses[snapNetAddr].pop(0)
            #assert isinstance(er, EventResponse)
            try:
                if defferred.sendXml:
                    defferred(xmlrpclib.dumps(er.parameters, methodname=er.methodName))
                else:
                    defferred(er)
                    if defferred.event is not None:
                        defferred.event.Stop()
            except AttributeError:
                defferred(er)
            if __debug__:
                log.debug("Sent response to: %s" % (binascii.hexlify(snapNetAddr)))
        except KeyError:
            #Send an rpc to let the directly connected node know that we are there
            self.callSink.eventResponses[snapNetAddr] = []
            self.mcastRpc(snapNetAddr, 1, 1, '_', [])
            return True
        except IndexError:
            return True

    def addSpy(self, name, data):
        """
        Add a SPY file to SNAPconnect's local file system that can then be uploaded
        to remote SNAP nodes.

        A SPY file is an exported SNAPpy script created by Portal.  These exported
        SNAPpy scripts can be loaded onto SNAPconnect using this function. The
        file will be saved with the name provided in the working directory that
        SNAPconnect was originally started from.

        If a file with the same name already exists it will be overwritten.

        name : string -- The full name of the file as it should be saved
        data  : base64 -- The data of the SPY file to write in the form of a binary
                          byte array
        """
        if os.path.splitext(name)[1] != '.spy':
            raise Exception("Filename must have a SPY extension")

        f = None
        if isinstance(data, xmlrpclib.Binary):
            data = data.data

        try:
            f = open(os.path.join(START_PATH, os.path.basename(name)), 'wb')
            f.write(data)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            log.exception("Unable to write SPY file: %s" % name)
            return False
        finally:
            if f is not None:
                f.close()

        return True

    def clearQueuedResponses(self, snapNetAddr):
        """
        Clears any EventResponse objects queued on the server, for the specified
        network address.

        This method allows you to clear any pending EventResponse objects that are
        currently queued for the specified network address.

        Warning: Verify that the specified address is associated with the correct
        client application (i.e. your own).  Specifying an incorrect address could
        lead to the clearing of queues belonging to other client applications.
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        self.callSink.eventResponses.pop(snapNetAddr, 0)
        return True

    def connectSerial(self, portType, portNum, reconnect=False):
        """
        Connect SNAPconnect to the bridge node at the specified interface

        For SNAPconnect to communicate with SNAP nodes on your network, it must
        be connected to a SNAP bridge via a serial port.  Currently, you can connect
        to a bridge node located on an RS-232 port or to a Synapse USB device.

        SNAPconnect can be connected to any RS-232 port that has been recognized by
        the Operating System of the host computer. It will not scan the port to see if
        there is an active node. This means SNAPconnect can successfully connect to
        an RS-232 port even if there is no SNAP node present.

        Choosing the reconnect option will forcibly close an existing connection and
        open a new connection to the specified port.  If you specify a connection that
        is already open and choose not to reconnect, the connection will NOT be closed.

        portType : int  -- The type of port to use for the connection
                           (RS232=1, USB=2, SNAPSTICK=3)
        portNum : int -- The port number of the specified type starting at 1. For USB
                         ports, a value of 1 is associated with USB0 on a PC. For
                         RS-232 ports, a value of 1 is associated with Com1 on a PC.
        reconnect : boolean -- Optional; If set, SNAPconnect will close the current
                               connection before reopening it. By default, if the
                               serial connection is already configured to the specified
                               values (portType, portNum) SNAPconnect will not reconnect
                               (default False)
        """
        if not portType:
            return False

        _port = portNum
        if not isinstance(_port, str):
            _port -= 1

        _dll_path = START_PATH
        if not hasattr(sys, 'frozen'):
            _dll_path = os.path.join(START_PATH, 'serialwrapper')
        self.main_snapcom.open_serial(portType, _port, reconnect, dll_path=_dll_path)
        self._port_type = portType
        self._port_num = _port

        # We return true because if it fails setOutputType throws an exception
        return True

    def delSpy(self, name):
        """
        Delete the specified SPY file from SNAPconnect's local file system

        This method removes the specified file from the working directory
        of SNAPconnect.

        If any path information was included in the name, it will be ignored.

        The file extension (.spy) is required as part of the filename.

        name : string -- The full name of the file that will be deleted
        """
        if os.path.splitext(name)[1] != '.spy':
            raise Exception("Filename must have a SPY extension")

        try:
            os.remove(os.path.join(START_PATH, os.path.basename(name)))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            log.exception("An error occurred while removing the file: %s" % name)
            return False
        return True

    def disconnect(self, snapNetAddr, clear=True):
        """
        Disconnects from SNAPconnect, returning a value of None to any outstanding
        calls to the waitOnEvent() function associated with client applications on the
        specified address.

        When a client is no longer interested in listening for events it should call
        disconnect with the clear argument set to True.

        Warning: Verify that the specified address is associated with the correct client
        application (i.e. your own). Specifying an incorrect address could mean other
        client applications are unwittingly disconnected from SNAPconnect.

        snapNetAddr : base64 -- The network address to disconnect
        clear : boolean -- Optional; clear any queued responses and stop responding to
                           the specified network address (default True)
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        deferred = self.deferreds.pop(snapNetAddr, None)
        if deferred is not None:
            deferred(None)

        if clear:
            self.clearQueuedResponses(snapNetAddr)
            snapcom = self.snapcoms.get(snapNetAddr, None)
            if snapcom and snapcom != self.main_snapcom:
                snapcom.disconnect_tcp('127.0.0.1')
                snapcom = self.snapcoms.pop(snapNetAddr)
                del snapcom

        return True

    def exit(self):
        """
        Exits SNAPconnect closing all connections.

        This method can be invoking even after the calling application has previously
        used the disconnect() method to disconnect from SNAPconnect

        Warning: Verify that no other clients are connected to SNAPconnect before invoking
        this method. It will close SNAPconnect server for all connected clients.
        """
        #Delay the close so that we can send the response
        self.evScheduler.schedule(0.1, self._closeMethod)
        return True

    def gatewayInfo(self):
        """
        Returns an InfoResponse structure with information about
        the status of this SNAPconnect

        See the InfoResponse structure for details regarding what information is
        available.
        """
        if self._port_type:
            return InfoResponse(VERSION, True, self._port_type, self._port_num)
        else:
            return InfoResponse(VERSION, False, None, None)

    def mcastRpc(self, snapNetAddr, group, ttl, funcName, args):
        """
        Sends a multicast RPC to the specified multicast group

        This method will block until it has successfully sent the multicast RPC to the
        SNAP bridge node or a packet serial timeout occurs.

        This method will return false if the SNAP bridge node connected to SNAPconnect
        cannot transmit the RPC.

        This will occur if SNAPconnect cannot communicate with the bridge node
        (ex. RS-232 cable has been disconnected) or the transmit queue of the bridge
        node is full.

        snapNetAddr : base64 -- The source network address of the multicast RPC
        group : int -- The multicast group to send to
        ttl : int -- The Time-To-Live of the multicast RPC
        funcName : string -- The function name the multicast RPC is invoking
        args : array -- Any arguments for the function specified by funcName
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        if isinstance(group, int):
            group = binascii.unhexlify("%04X" % group)
        if isinstance(group, xmlrpclib.Binary):
            group = group.data
        fixedArgs = []
        for arg in args:
            if isinstance(arg, xmlrpclib.Binary):
                arg = arg.data
            fixedArgs.append(arg)

        ## Send this multicast to other clients connected to this SNAPconnect
        #response = EventResponse(snapNetAddr, funcName, time.time(), *args)
        #self.callSink.addMulticastResponse(response, snapNetAddr)

        #return getattr(self.virtualEndpoint.multicast(group, ttl, srcAddr=snapNetAddr), funcName)(*fixedArgs)
        rpc_id = self._get_snapcom(snapNetAddr).mcast_rpc(group, ttl, funcName, *fixedArgs)
        deferred = Deferred.Deferred()
        self._get_snapcom(snapNetAddr)._outstanding_rpcs[rpc_id] = deferred
        return deferred

    def rpc(self, snapNetAddr, remoteNetAddr, funcName, args):
        """
        Sends a unicast RPC to the specified network address

        This method will block until it has successfully sent the RPC to the SNAP bridge node
        or a packet serial timeout occurs.

        This method will return false if the SNAP bridge node connected to SNAPconnect cannot
        transmit the RPC.

        This will occur if SNAPconnect cannot communicate with the bridge node (ex. RS-232
        cable has been disconnected) or the transmit queue of the bridge node is full.

        snapNetAddr : base64 -- The source network address of the unicast RPC
        remoteNetAddr : base64 -- The remote network address the RPC is destined for
        funcName : string -- The function name the RPC is invoking
        args : array -- Any arguments for the function specified by funcName
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        if isinstance(remoteNetAddr, xmlrpclib.Binary):
            remoteNetAddr = remoteNetAddr.data

        fixedArgs = []
        for arg in args:
            if isinstance(arg, xmlrpclib.Binary):
                arg = arg.data
            fixedArgs.append(arg)

        rpc_id = self._get_snapcom(snapNetAddr).rpc(remoteNetAddr, funcName, *fixedArgs)
        deferred = Deferred.Deferred()
        self._get_snapcom(snapNetAddr)._outstanding_rpcs[rpc_id] = deferred
        return deferred

    def sendDataModePkt(self, snapNetAddr, remoteNetAddr, data, mcastTtl = 0):
        """
        Sends a transparent (aka data) mode packet to the specified network address.

        The data should be sent to SNAPconnect as an XML-RPC binary type when the
        data contains non-printable ASCII characters or carriage returns.

        snapNetAddr : base64 -- The source network address of the data mode packet
        remoteNetAddr : base64 -- The remote network address the data mode packet is destined for
        data  : non-specific -- The data to send; has no specific data type
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        if isinstance(remoteNetAddr, xmlrpclib.Binary):
            remoteNetAddr = remoteNetAddr.data
        if isinstance(data, xmlrpclib.Binary):
            data = data.data

        if mcastTtl > 0:
            rpc_id = self._get_snapcom(snapNetAddr).mcast_data_mode(remoteNetAddr, mcastTtl, data)
        else:
            rpc_id = self._get_snapcom(snapNetAddr).data_mode(remoteNetAddr, data)

        deferred = Deferred.Deferred()
        self._get_snapcom(snapNetAddr)._outstanding_rpcs[rpc_id] = deferred
        return deferred

    def uploadSpy(self, remoteNetAddr, name):
        """
        Uploads the specified SNAPpy image to the specified network address.

        This method does not return until the upload is complete.

        The SNAPpy image must have been previously added to SNAPconnect's local file
        system using the addSpy() function

        remoteNetAddr : base64 -- The network address to upload the SNAPpy image to
        name : string -- The name of the image to upload
        """
        if isinstance(remoteNetAddr, xmlrpclib.Binary):
            remoteNetAddr = remoteNetAddr.data

        #remove any path information they may have sent
        name = os.path.join(START_PATH, os.path.basename(name))

        if not os.path.isfile(name):
            raise IOError("Not a regular file")

        f = None
        uploadImage = None
        if os.path.splitext(name)[1] == ScriptsManager.SNAPPY_EXPORT_EXTENSION:
            try:
                f = file(name, 'rb')
                uploadImage = ScriptsManager.getSnappyStringFromExport(f.read())
            except:
                raise IOError("Unable to read the selected SNAPpy image")
            finally:
                if f:
                    f.close()
        if not uploadImage: raise IOError("SNAPpy Script Error")

        func = lambda remoteNetAddr, funcName, *args: self.rpc(self.main_snapcom.local_addr(), remoteNetAddr, funcName, args)
        upload = self.uploadMgr.startUpload(remoteNetAddr, uploadImage, rpcFunc=func)
        assert isinstance(upload, SnappyUploader.SnappyUpload)
        upload.registerFinishedCallback(self._onUploadComplete)

        _deferred = Deferred.Deferred()
        _deferred.snapNetAddr = remoteNetAddr
        self.uploadDeferreds[remoteNetAddr] = _deferred
        return _deferred

    def waitOnEvent(self, snapNetAddr=None, flush=False, portType=None, portNum=None, timeout=-1.0):
        """
        Wait for an EventResponse object

        This method blocks until an EventResponse object is available for the
        specified network address or the specified timeout occurs.

        If the flush parameter is set to True then the disconnect() method is
        automatically called before the wait process begins.

        If a portType and portNum are passed as parameters, connectSerial() is
        automatically called using the same portType and portNum for its parameters.

        The timeout period is specified in seconds. A fraction of a second can be used
        by passing a floating point value (ex. 0.1 would represent 100 milliseconds)

        snapNetAddr : base64 -- The network address to wait for EventResponse objects
        flush : boolean -- Optional; invokes the disconnect() function before initiating
                           the wait process (default False)
        portType : int -- Optional; parameter to pass to connectSerial() (default None)
        portNum : int -- Optional; parameter to pass to connectSerial() (default None)
        timeout : int -- Optional; the maximum timeout period, in seconds, to wait for
                         an event
        """
        if isinstance(snapNetAddr, xmlrpclib.Binary):
            snapNetAddr = snapNetAddr.data
        self._get_snapcom(snapNetAddr)

        if portType is not None and portNum is not None:
            self.connectSerial(portType, portNum)

        if self.deferreds.has_key(snapNetAddr) and isinstance(self.deferreds[snapNetAddr], Deferred.Deferred):
            raise xmlrpclib.Fault(SYNAPSE_XMLRPC_MULTIPE_WAITS_FAULTCODE,
                                  "Cannot have multiple outsanding calls to waitOnEvent")
        elif flush:
            self.disconnect(snapNetAddr)

        if not self.callSink.eventResponses.has_key(snapNetAddr):
            self.callSink.eventResponses[snapNetAddr] = []
        _deferred = Deferred.Deferred()
        _deferred.event = None

        deferred_addr = snapNetAddr
        if self.translate_portal_addrs:
            portal_addr = self.portal_addrs.get(snapNetAddr, None)
            if portal_addr:
                deferred_addr = portal_addr

        if len(self.callSink.eventResponses[deferred_addr]):
            self.evScheduler.scheduleEvent(self._onResponseAdd, [snapNetAddr])
            #return self.callSink.eventResponses.get(self.snapcom.local_addr())
        elif timeout == 0:
            return None
        elif timeout > 0:
            _deferred.event = self.evScheduler.schedule(timeout, self.callSink.addResponse, None, snapNetAddr)

        _deferred.sendXml = False
        _deferred.snapNetAddr = snapNetAddr
        self.deferreds[snapNetAddr] = _deferred

        return _deferred

    def waitOnXmlMethod(self, snapNetAddr=None, flush=False, portType=None, portNum=None, timeout=-1.0):
        """
        Wait for an XML-RPC method structure

        Similar to the waitOnEvent() method but instead of returning an EventResponse
        object, returns an xml-encoded method call. If the associated XML-RPC library
        supports it, this may be passed to a handler which un-marshals the call and
        invokes the related function.

        If the flush parameter is set to True then the disconnect() method is
        automatically called before the wait process begins.

        If a portType and portNum are passed as parameters, connectSerial() is
        automatically called using the same portType and portNum for its parameters.

        The timeout period is specified in seconds. A fraction of a second can be
        used by passing a floating point value (ex. 0.1 would represent 100 milliseconds)

        snapNetAddr : base64 -- The network address to wait for EventResponse objects
        flush : boolean -- Optional; also invokes the disconnect() method if set to True
                           (default False)
        portType : int -- Optional; parameter to pass to the connectSerial() (default None)
        portNum : int -- Optional; parameter to pass to the connectSerial() (default None)
        timeout : int -- Optional; the maximum timeout period, in seconds, to wait
                         for an event
        """
        _deferred = self.waitOnEvent(snapNetAddr, flush, portType, portNum, timeout)
        if _deferred is not None:
            _deferred.sendXml = True
        return _deferred


class SnapXmlRpcServer(object):
    """Setups and controls the server"""
    def __init__(self,
                 libraryPath=START_PATH,
                 portNum=8080,
                 logHttp=True,
                 logResponses=True,
                 logRequests=True,
                 statusReport=False,
                 portType=None,
                 serial_port_num=0,
                 timeout_interval=0.00,
                 translate_addrs=False,
                 snapcomPort=snaptcp.SNAP_IP_PORT):
        #self.evScheduler = self.snap.scheduler
        self.callSink = CallSink()
        self.app = None
        self.closeEvent = threading.Event()
        self.pyserial_close_event = threading.Event()
        self.task_queue = Queue.Queue()
        self.done_queue = Queue.Queue()
        self.statusEnabled = statusReport
        self.portNum = portNum
        self.pollCallbacks = []
        self.timeout_interval = timeout_interval

        # Start listening for RPC connections
        try:
            self.server = SnapXmlRpcRequestHandler(self.close, translate_addrs, snapcomPort)
            if portType is not None:
                self.server.connectSerial(portType, serial_port_num)
            self.rpcSrv = AsyncXmlRpc.HttpXmlRpcServer(('', portNum), logHttp, logResponses, logRequests, statusReport)
            self.rpcSrv.xmlRpcDispatcher.registerCloseCallbacks(self.server._onDeferredSocketClose)
            self.rpcSrv.xmlRpcDispatcher.register_introspection_functions()
            self.rpcSrv.xmlRpcDispatcher.register_multicall_functions()
            self.rpcSrv.xmlRpcDispatcher.register_instance(self.server)
            log.info("Now listening for XML-RPC connections on port %d" % portNum)
        except socket.error:
            self.closeEvent.set()
            log.exception("Unable to start XML-RPC server on port %d" % portNum)

    def close(self):
        """Closes everything and ends the main loop, thus exiting the program"""
        #If you change stuff in here make sure when running in GUI mode that the icon gets removed from the tray
        if self.app is not None and self.app.frame:
            #We are not calling self.closeEvent.set() here because we need the GUI code to clean up first
            self.app.frame.ThreadClose()
        else:
            self.closeEvent.set()

        log.info("Closing")

    def getSerialConnInfo(self):
        """Get the current connection from the pyserial driver"""
        if snap._serials:
            desc = []
            for listener in snap._serials.itervalues():
                desc.append(listener.serial_driver.serial.portstr)
            return desc
        else:
            return ['Not Connected']

    def loop(self):
        """The main event loop"""
        timeout = self.timeout_interval

        try:
            scheduler_poll = self.server.evScheduler.poll
            my_poll = self.poll
            while not self.closeEvent.isSet():
                try:
                    asyncore.poll(timeout)
                except ValueError:
                    pass # too many file descriptors in select()
                scheduler_poll()

                if not self.task_queue.empty():
                    task = self.task_queue.get_nowait()
                    try:
                        self.done_queue.put(getattr(self, task)())
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        log.exception("An error occurred while running a task:")

                my_poll()
        except KeyboardInterrupt:
            asyncore.close_all()

    def poll(self):
        """Periodic event timer call"""
        if self.pyserial_close_event.isSet():
            self.server.main_snapcom.close_all_serial()
            self.pyserial_close_event.clear()
            return True

        try:
            for cb in self.pollCallbacks: cb()
            for snapcom in self.server.snapcoms.itervalues():
                snapcom.poll_internals()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            log.exception("A poll error occurred:")
        return True

    def registerPollCallback(self, callback):
        """
        Register for callbacks for object who want to get polled

        callback -- A callable object with a no parameters
        """
        if callable(callback):
            self.pollCallbacks.append(callback)

    def unregisterPollCallback(self, callback):
        """
        Unregister a callable object from being polled

        callback -- The callable object to unregister
        """
        try:
            self.pollCallbacks.remove(callback)
        except KeyError:
            return #If it's not in there than it's removed!


def Main(port=8080,snapcomPort=snaptcp.SNAP_IP_PORT):
    """Starts up an instance of SNAPconnect"""
    serport_type = "int"
    serport_default = 0
    if sys.platform.startswith("linux") or sys.platform == 'darwin':
        serport_type = "string"
        serport_default = "/dev/ttyS1"

    # Parse command-line options
    parser = optparse.OptionParser()
    parser.add_option("-a", "--portaladdrs", action="store_true", dest="portaladdrs", default=False, help="Enable support for Portal address ranges")
    parser.add_option("-c", "--consolelogging", action="store_false", dest="logConsole", default=True, help="Disable logging to console")
    parser.add_option("-i", "--interval", type="int", dest="timeout_interval", metavar="timeout_interval", default=1, help="The socket polling timeout interval in milliseconds")
    parser.add_option("-m", "--httplogging", action="store_false", dest="logHttp", default=True, help="Disable logging of HTTP server messages")
    parser.add_option("-n", "--serial_port_number", type=serport_type, dest="serial_port_num", metavar="serial_port_num", default=serport_default, help="Open a connection to the specified port number")
    parser.add_option("-p", type="int", dest="portNum", metavar="portNum", default=port, help="TCP port number to listen for XML-RPC requests on")
    parser.add_option("-q", "--requestlogging", action="store_false", dest="logReq", default=True, help="Disable logging of SNAPconnect XML-RPC requests")
    parser.add_option("-r", "--responselogging", action="store_false", dest="logResp", default=True, help="Disable logging of SNAPconnect XML-RPC responses")
    parser.add_option("-s", "--status", action="store_true", dest="status", default=False, help="Enable HTTP status report @ /status")
    parser.add_option("-t", "--serial_port_type", type="int", dest="portType", metavar="portType", default=None, help="Open a connection to the specified port type")
    options = parser.parse_args()[0]

    if options.logConsole:
        logging.basicConfig(level=logging.DEBUG, format='%(asctime)s:%(msecs)03d %(levelname)-8s %(name)-8s %(message)s', datefmt='%H:%M:%S')

    if __debug__ and not hasattr(sys, 'frozen'):
        rotatingFileLogHandler = None
    else:
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.DEBUG)

        traceFilePath = os.path.join(START_PATH, 'SNAPconnectTrace')

        # Log to rotating files
        try:
            rotatingFileLogHandler = logging.handlers.RotatingFileHandler(traceFilePath, maxBytes=SYNAPSE_MAX_LOG_SIZE, backupCount=SYNAPSE_MAX_LOG_FILES)
        except IOError:
            log.error("Unable to open log file")
        else:
            logFormat = logging.Formatter('%(asctime)s:%(msecs)03d %(levelname)-8s %(name)-8s %(message)s', datefmt='%H:%M:%S')
            rotatingFileLogHandler.setFormatter(logFormat)
            rotatingFileLogHandler.setLevel(logging.DEBUG)
            rootLogger.addHandler(rotatingFileLogHandler)

    logging.raiseExceptions = __debug__
    log.debug('******** Synapse SNAPconnect Debug Log - ' + VERSION + ' ********')

    try:
        gw = SnapXmlRpcServer(START_PATH, options.portNum, options.logHttp, options.logResp, options.logReq, options.status, options.portType, options.serial_port_num, options.timeout_interval*0.001, options.portaladdrs, snapcomPort)
    except LicenseManager.LicenseError, le:
        log.error(le.message)
        return le
    except Exception, ex:
        log.exception("An error occurred while trying to start the server")
        return ex

    return gw


if __name__ == '__main__':
    srv = Main()

    if isinstance(srv, SnapXmlRpcServer) and hasattr(sys, 'frozen'):
        srv.loop()
    elif isinstance(srv, SnapXmlRpcServer):
        import cProfile

        cProfile.run('srv.loop()', os.path.abspath(os.path.curdir) + os.sep + 'profile.tmp')
