# (c) Copyright 2009-2012 Synapse Wireless, Inc.
"""
Dispatchers for SNAP Connect to use when deciding where to send packets
"""


import logging
import binascii
import time
import timeit
import random

import snap

from snaplib import SnapConstantsHelper
from snaplib import MeshCodec, RpcCodec, TraceRouteCodec, DataModeCodec
from snaplib import EventCallbacks
from snaplib.snappy import hooks

from apy import Deferred

if __debug__:
    from snaplib import PacketSerialProtocol
    import mesh
    from apy import EventScheduler


SEQ_CHECK_TIMEOUT = 5  # seconds
SEQ_CHECK_LIMIT = 5  # number of previous seq nums to store
SNIFFER_VERSION = 1 # Version of sniffer. For future use if remote sniffer meta data format changes

log = logging.getLogger("snap.dispatchers")

class LocalDispatcher(object):
    """Dispatches RPCs to local functions"""
    def __init__(self, dispatcher, funcs, scheduler):
        """
        Initializes an instance of LocalDispatcher

        Args:
            dispatcher : A PktDispatcher instance
            funcs : A dictionary of function names to function objects
            scheduler : The main SNAP connect scheduler in use
        """
        self.funcs = funcs
        self.internal_funcs = {} #Functions used internally. These are passed the rpc source as the first argument
        assert isinstance(self.funcs, dict), "funcs must be a dictionary"
        assert self.funcs.has_key('rpc'), "required RPC function NOT defined"
        assert self.funcs.has_key('load_nv_param'), "required load_nv_param function NOT defined"
        self.funcs['vmStat'] = self.vm_stat
        self.funcs['callback'] = self.callback
        self.name = "SNAPcom" #Device name used for vm_stat
        self.dispatcher = dispatcher
        assert isinstance(self.dispatcher, PktDispatcher), "dispatcher is not an instance of PktDispatcher"
        self.scheduler = scheduler  # Could we get this via a singleton?
        self.verbose = True

    def add_func(self, func_name, func):
        """
        Adds an existing function to the function dictionary

        Args:
            func_name : The function name as called in an RPC
            func : The function object to call
        """
        if self.funcs.has_key(func_name):
            return False
        self.funcs[func_name] = func
        return True

    def callout(self, dst_addr, callout_func_name, func_name, *args):
        """
        Like callback() but final result is reported to the specified dst_addr

        See SNAP Reference Manual for more details
        """
        descriptor = RpcCodec.RpcDescriptor(crypto=self.dispatcher.crypto)
        descriptor.funcName = func_name
        descriptor.args = args
        rtn_val = self.on_rpc(descriptor)

        self.funcs['rpc'](dst_addr, callout_func_name, rtn_val)

    def callback(self, callback, func_name, *args):
        """Like rpc() but also invokes specified callback routine with the result of the remoteFnObj routine"""
        if isinstance(func_name, RpcCodec.RpcFunc):
            func_name = func_name.func_name

        if self.funcs.get(func_name, None):
            descriptor = RpcCodec.RpcDescriptor(crypto=self.dispatcher.crypto)
            descriptor.funcName = func_name
            descriptor.args = args
            rtn_val = self.on_rpc(descriptor)
            self.funcs['rpc'](self.dispatcher.remote_addr, callback, rtn_val)

    def on_rpc(self, descriptor):
        """
        Called upon receipt of an RPC and then determines if we have a function to call

        Args:
            descriptor : A snaplib.RpcCodec object
        """
        assert isinstance(descriptor, RpcCodec.RpcDescriptor), "Did NOT receive an RPC descriptor object"

        #Check if we have any internal functionality that goes along with this RPC
        #For backwords compatibility, we'll still allow a user to act on that RPC as well,
        #but this makes sure we still get the RPC call where we need it for things like OTA upgrade
        func = self.internal_funcs.get(descriptor.funcName, None)
        if callable(func):
            try:
                #Don't return here, we shouldn't interrupt the user function
                func(descriptor.srcAddr, *descriptor.args)
            except TypeError:
                log.exception("Unable to call function %s:" % (descriptor.funcName,))

        #Check for user defined functions
        func = self.funcs.get(descriptor.funcName, None)
        if callable(func):
            try:
                return func(*descriptor.args)
            except TypeError:
                log.exception("Unable to call function %s:" % (descriptor.funcName,))
        elif self.verbose:
            log.debug("Unable to find callable function %s" % (descriptor.funcName))

    def set_name(self, name):
        """Set the node name used by vm_stat"""
        if name is None:
            self.name = "SNAPcom"
        else:
            assert isinstance(name, str), "Node name must be a string"
            self.name = name

    def vm_stat(self, status, arg=None, delay=0):
        """
        Just like embedded SNAP nodes we want to respond to vmStat calls

        See the SNAP Reference Manual for argument meanings
        """
        rpc_args = None

        if status == SnapConstantsHelper.VM_INFO:
            # Script name
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_INFO, '')
        elif status == SnapConstantsHelper.VM_VER:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_VER, snap.VERSION)
        elif status == SnapConstantsHelper.VM_NET:
            # We return 0 for channel and network id
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', 0, 0)
        elif status == SnapConstantsHelper.VM_NVREAD:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', (arg << 8) + SnapConstantsHelper.VM_NVREAD, self.funcs['load_nv_param'](arg))
        elif status == SnapConstantsHelper.VM_NAME:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_NAME, self.name)
        elif status == SnapConstantsHelper.VM_SPACE:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_SPACE, 0)
        elif status == SnapConstantsHelper.VM_DATA_SPACE:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_DATA_SPACE, 0)
        elif status == SnapConstantsHelper.VM_IMG_ERASE:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_IMG_ERASE, SnapConstantsHelper.TELL_VM_WR_BLK_UNSUPPORTED)
        elif status == SnapConstantsHelper.VM_WR_BLK:
            rpc_args = (self.dispatcher.remote_addr, 'tellVmStat', SnapConstantsHelper.VM_WR_BLK, SnapConstantsHelper.TELL_VM_WR_BLK_UNSUPPORTED)

        if rpc_args:
            def delayed_write(*rpc_args):
                self.funcs['rpc'](*rpc_args)
            self.scheduler.schedule(delay*random.random(), delayed_write, *rpc_args)


class PktDispatcher(EventCallbacks.Registrar):
    """Dispatches packets to the appropriate interfaces"""
    def __init__(self, crypto, mesh_handler, local_addr, scheduler, forwarded_group=0x01, processed_groups=0x01, alternate_crypto=None):
        """
        Initializes an instance of PktDispatcher

        Args:
            crypto : An instance of the decrypter to use
            mesh_handler : This SNAP Connect's instance of mesh.MeshHandler to use
            local_addr : This SNAP Connect's instance SNAP address
            scheduler : This SNAP Connect's instance scheduler to use
            forwarded_group : The multicast forward groups that should be used
            processed_groups : Which multicast groups we should process and dispatch locally
            alternate_crypto : An instance of the alternate decrypter to use
        """
        EventCallbacks.Registrar.__init__(self, ['broadcasts', hooks.HOOK_RPC_SENT, hooks.HOOK_STDIN])
        self.interested = {}  # {pid: []}
        self._local_sniffers = [] # [Callable(descriptor)] used for locally sniffing all traffic being sent and received by this instance
        self._remote_sniffers = [] #[Callable(raw_pkt_data)] used for remotely sniffing all traffic being sent and received by this instance
        self.whitelisted_funcs = [] # functions allowed even if received with alternate encryption
        self.alternate_key_used = False # <- RX direction
        self.send_seq = 0 #Used to tie return value of rpc to the value passed to HOOK_RPC_SENT. NOT embedded in sent packets!
        self.mesh_handler = mesh_handler
        #assert isinstance(self.mesh_handler, mesh.MeshHandler)
        self._waiting = {}
        self.local_addr = local_addr
        self.remote_addr = None
        self.incoming_intf = None
        self.scheduler = scheduler
        self._pkt_seqs = {}  # {addr: (timestamp, [seq, seq, ...])}
        self.forwarded_group = forwarded_group
        self.processed_groups = processed_groups
        self.crypto = crypto
        self.alternate_crypto = alternate_crypto
        self.verbose = True
        self._traceroute_hook = None

        # Used for getInfo calls
        self.last_pkt_is_multicast = False
        self.last_pkt_remaining_ttl = 0

    def add_sniffer(self, sniffer_function, raw_data):
        """Add a sniffer to all incoming and outgoing descriptors

        Args:
            sniffer_function: The function to be called for each packet
            raw_data: If True, pass the raw packet string instead of a descriptor

        Returns:
            True if the sniffer was added
            False if we already have a sniffer
        """
        assert callable(sniffer_function)

        if not raw_data:
            if sniffer_function not in self._local_sniffers:
                self._local_sniffers.append(sniffer_function)
                return True
            else:
                return False
        else:
            if sniffer_function not in self._remote_sniffers:
                self._remote_sniffers.append(sniffer_function)
                return True
            else:
                return False

    def add_whitelisted_func(self, func_name):
        """
        Adds a function name to that can be called using the alternate crypto key

        Args:
            func_name : The function name to add
        """
        if func_name not in self.whitelisted_funcs:
            self.whitelisted_funcs.append(func_name)

    def _delayed_write(self, write_func, data, seq=None):
        """
        Called when it is time to write the data out an interface

        Args:
            write_func : The function object to call to write the data
            data : The data to write
            seq : An optional internal sequence number to use
        """
        if seq is None:
            seq = self.send_seq
        success = write_func(data)
        if isinstance(success, Deferred.Deferred):
            success.addCallbacks(lambda result: self._notify_callbacks(hooks.HOOK_RPC_SENT, seq, result))
        else:
            self._notify_callbacks(hooks.HOOK_RPC_SENT, seq, success is not False)
        return None

    def is_valid_seq(self, seq_num, addr):
        """Is this a valid sequence number for this address"""
        (last_seq_timestamp, seq_nums) = self._pkt_seqs.get(addr, (None, None))
        if last_seq_timestamp is None:
            # We have not heard from this node before
            return True
        elif time.time() - last_seq_timestamp > SEQ_CHECK_TIMEOUT:
            #log.debug("Timing out previous sequence numbers from node %s" % (binascii.hexlify(descriptor.srcAddr)))
            self._pkt_seqs[addr] = (time.time(), [])
            return True
        elif seq_num not in seq_nums:
            return True
        return False

    #
    # The packet was not for us. See if we know a "next hop" that will eventually reach the
    # correct address. If so, send it along.
    #
    def _forward_packet(self, descriptor):
        """
        See if we can get mesh to forward this packet

        Args:
            descriptor : The packet descriptor to forward
        """
        (addr, intf) = self.mesh_handler.get_next_hop(descriptor.final_dst_addr, None)
        if not addr:
            # Even if we didn't get an addr put one in because the get_intf.write call will send the RERR for us
            addr = self.local_addr
        descriptor.dstAddr = addr  # <- This actually changes the encoded_data...

        self.mesh_handler.get_intf(descriptor.final_dst_addr).write(descriptor.encoded_data)

    def _on_data_mode(self, pkt, interface):
        """
        Called on the receipt of a data mode packet
        Decodes the packet and determines if it needs to be forwarded

        Args:
            pkt : The raw bytes of the data mode packet
            interface : The interface this packet was received on
        """
        descriptor = DataModeCodec.decode(pkt, crypto=self.crypto)
        assert isinstance(descriptor, DataModeCodec.DataModeDescriptor)

        if self.is_valid_seq(ord(descriptor.seq), descriptor.srcAddr):
            log.debug("dm %s: %s" % (binascii.hexlify(descriptor.srcAddr), descriptor.data))
            self.update_seq(ord(descriptor.seq), descriptor.srcAddr)
            self.remote_addr = descriptor.srcAddr
            self.incoming_intf = interface

            descriptor.last_hop_addr = self.local_addr
            if descriptor.isMulticast:
                if descriptor.srcAddr == self.local_addr:
                    log.debug("multi-cast data mode originated from us")
                else:
                    descriptor.ttl -= 1
                    if descriptor.ttl > 0:
                        if self.verbose:
                            log.debug("Rebroadcasting data mode pkt to %s" % (binascii.hexlify(descriptor.dstAddr)))
                        self._notify_callbacks('broadcasts', descriptor.encoded_data, interface, int(binascii.hexlify(descriptor.dstAddr), 16))
                    elif __debug__ and self.verbose:
                        log.debug("Data mode broadcast TTL too low to re-broadcast")
            elif descriptor.final_dst_addr != self.local_addr:  # if the packet was not for me...
                self._forward_packet(descriptor)  # ...send it on to it's final destination
                descriptor = None  # Set to None so we don't notify local code about packets that aren't for us
        else:
            if __debug__ and self.verbose:
                log.debug("Received data mode packet with same sequence number from %s" % (binascii.hexlify(descriptor.srcAddr)))
            descriptor = None

        return descriptor

    def get_trace_route_timestamp(self):
        """
        The embedded SNAP Nodes that originally implemented TRACE ROUTE used their internal
        "16-bit unsigned millisecond ticks" as a timestamp source. This routine emulates
        having that same sort of "tick counter".
        """
        current_time = timeit.default_timer()
        # We want to discard anything less than milliseconds
        current_time *= 1000.0
        current_time = int(current_time)  # No floating point in embedded SNAP Nodes
        # Force it to roll-over the same way it does in the embedded SNAP Nodes
        current_time &= 0xffff
        return current_time

    def get_trace_route_elapsed(self, start_time):
        """
        Determine how many milliseconds have elapsed for a trace route

        Args:
            start_time : A trace route timestamp
        """
        end_time = self.get_trace_route_timestamp()
        if end_time >= start_time:
            elapsed_time = end_time - start_time
        else:
            # account for the rollover at 65535
            elapsed_time = (65536 - start_time) + end_time
        return elapsed_time

    def _decode_pkt(self, pkt):
        """Decode a packet into a descriptor, and leave it untouched otherwise

        Eventually, this code should be be called in on_decoded_pkt and
        register_pkt_type should be used to perform additional transformations
        """
        descriptor = None
        if len(pkt) > SnapConstantsHelper.RPC_PID_OFS:
            # Remove the encryption bit
            pid = ord(pkt[SnapConstantsHelper.RPC_PID_OFS]) & ~SnapConstantsHelper.PID_ENCRYPTION_MASK

            if pid in (SnapConstantsHelper.PID_MESH_BCAST, SnapConstantsHelper.PID_MESH_PTP):
                descriptor = MeshCodec.MeshCodec.decode(pkt, crypto=self.crypto)

            elif pid in (SnapConstantsHelper.PID_RPC, SnapConstantsHelper.PID_MULTICAST_RPC):
                #RPC packets need to be checked against both crypto types
                descriptor = RpcCodec.decode(pkt, crypto=self.crypto)
                if descriptor is None:
                    if (self.alternate_crypto is not None):
                        descriptor = RpcCodec.decode(pkt, crypto=self.alternate_crypto)

            elif pid in (SnapConstantsHelper.PID_DM_PTP, SnapConstantsHelper.PID_DM_BCAST):
                descriptor = DataModeCodec.decode(pkt, crypto=self.crypto)

            elif pid in (SnapConstantsHelper.PID_TRACE_ROUTE,):
                descriptor = TraceRouteCodec.decode(pkt, crypto=self.crypto)

        if descriptor is None:
            descriptor = RawDescriptor()

        return descriptor

    def on_decoded_pkt(self, pkt, interface):
        """
        Called by a packet listener to determine what type of packet it is
        and where to send it to get decoded and dealt with

        Args:
            pkt : The raw bytes of a packet
            interface : The interface the packet was received on
        """
        process_packet = self.sniff_packet(interface, pkt, True)

        if len(pkt) > SnapConstantsHelper.RPC_PID_OFS and process_packet:
            pid = ord(pkt[SnapConstantsHelper.RPC_PID_OFS])
            descriptor = None
            self.last_pkt_is_multicast = False
            self.last_pkt_remaining_ttl = 0
            self.alternate_key_used = False # We will override this if we learn differently

            if (self.crypto and
                self.crypto.has_valid_key and
                not pid & SnapConstantsHelper.PID_ENCRYPTION_MASK and
                pid != SnapConstantsHelper.PID_MESH_PTP): # <- this exception is made because some older SNAP cores did not correctly apply the encryption
                log.debug("Received regular packet while encryption was on")
            elif pid in (SnapConstantsHelper.PID_MESH_BCAST,
                         SnapConstantsHelper.PID_MESH_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                         SnapConstantsHelper.PID_MESH_PTP,
                         SnapConstantsHelper.PID_MESH_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK):
                descriptor = MeshCodec.MeshCodec.decode(pkt, crypto=self.crypto)
                if descriptor:
                    descriptor.incoming_interface = interface
            elif pid in (SnapConstantsHelper.PID_RPC,
                         SnapConstantsHelper.PID_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                         SnapConstantsHelper.PID_MULTICAST_RPC,
                         SnapConstantsHelper.PID_MULTICAST_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK):
                descriptor = self._on_rpc(pkt, interface)
                if descriptor:
                    self.last_pkt_is_multicast = descriptor.isMulticast
                    self.last_pkt_remaining_ttl = descriptor.ttl
            elif pid in (SnapConstantsHelper.PID_DM_PTP,
                         SnapConstantsHelper.PID_DM_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                         SnapConstantsHelper.PID_DM_BCAST,
                         SnapConstantsHelper.PID_DM_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK):
                descriptor = self._on_data_mode(pkt, interface)
                if descriptor:
                    self.last_pkt_is_multicast = descriptor.isMulticast
                    self.last_pkt_remaining_ttl = descriptor.ttl
                    if descriptor.data:
                        self._notify_callbacks(hooks.HOOK_STDIN, descriptor.data)
            elif pid in (SnapConstantsHelper.PID_TRACE_ROUTE,
                         SnapConstantsHelper.PID_TRACE_ROUTE | SnapConstantsHelper.PID_ENCRYPTION_MASK):
                descriptor = TraceRouteCodec.decode(pkt, crypto=self.crypto)
                if descriptor:
                    if descriptor.final_dst_addr == self.local_addr:  # is it for me?
                        # This packet could be a TRACE ROUTE REQUEST (TR_REQ) or a REPLY (TR_REP)
                        # If it is a REQUEST, and it was addressed to us, then WE are the "turn-around point"
                        if descriptor.msgId == TraceRouteCodec.SnapConstantsHelper.TR_REQ:
                            descriptor.srcAddr = self.local_addr
                            descriptor.final_dst_addr = descriptor.orig_src_addr
                            descriptor.orig_src_addr = self.local_addr
                            descriptor.msgId = TraceRouteCodec.SnapConstantsHelper.TR_REP
                            self.send(descriptor)
                        else:  # If a REPLY comes in addressed to us, then we must have originated the TRACE ROUTE
                            if self._traceroute_hook:
                                self.remote_addr = descriptor.srcAddr
                                self.incoming_intf = interface
                                self._traceroute_hook(descriptor.orig_src_addr, descriptor.elapsed, [(record.address, record.linkQuality) for record in descriptor.records])
                    else:  # not for me, forward it onwards
                        # The decoder automatically adds the last hop to the list of records

                        # Special cases:
                        # 1: I am the FIRST node to see this TRACE ROUTE (after the originator)
                        if len(descriptor.records) == 1:
                            # Save the START time in what will ULTIMATELY be the ELAPSED time field
                            descriptor.elapsed = self.get_trace_route_timestamp()
                        # 2: I am the LAST node to see this TRACE ROUTE (before the originator)
                        elif descriptor.records[1].address == self.local_addr:
                            # Compute the round trip ELAPSED time using the START time that was saved up above
                            descriptor.elapsed = self.get_trace_route_elapsed(descriptor.elapsed)

                        descriptor.srcAddr = self.local_addr
                        self.send(descriptor)
                    descriptor = None
            else:
                pass

            if descriptor:
                for callback in self.interested.get(pid, []):
                    callback(descriptor)
            self.remote_addr = None
            self.incoming_intf = None

    def on_intf_closed(self, intf):
        """Called when an interface is closed"""
        addrs = self.mesh_handler.get_addrs(intf)
        [self._pkt_seqs.pop(addr, 0) for addr in addrs]  # Not using map here because I want to pass pop a 0

    def _on_next_hop(self, descriptor, addr, intf):
        """Called when a route lookup is completed"""
        success = False
        if addr and intf:
            descriptor.dstAddr = addr
            descriptor.srcAddr = self.local_addr
            encoded_data = descriptor.encoded_data
            self.sniff_packet(intf, encoded_data, rx=False)
            success = intf.write(encoded_data)
        try:
            seq = self._waiting.pop(descriptor, None)
            if seq is not None:
                self._notify_callbacks(hooks.HOOK_RPC_SENT, seq, success is not False)
        except ValueError:
            pass

    def on_packet_serial_opened(self, packet_serial):
        """
        Called when a new packet serial connection is opened

        Args:
            packet_serial : The instance of the new packet serial connection object
        """
        assert isinstance(packet_serial, PacketSerialProtocol.PacketSerialProtocol)
        packet_serial.registerCallback("rx", self.on_decoded_pkt)
        #TODO: I don't think we need to know when they are closed but do we get a circular reference?

    def _on_rpc(self, pkt, interface):
        """
        Called on the receipt of an RPC packet
        Decodes the packet and determines if it needs to be forwarded

        Args:
            pkt : The raw bytes of the data mode packet
            interface : The interface this packet was received on
        """
        descriptor = RpcCodec.decode(pkt, crypto=self.crypto)

        # Support the "RPC whitelist" feature
        if descriptor is None: # If the packet did not decode...
            if (self.alternate_crypto is not None) and (len(self.whitelisted_funcs) != 0): # Any chance it might be ALLOWED using the ALTERNATE key?
                descriptor = RpcCodec.decode(pkt, crypto=self.alternate_crypto) # ...see if we can decode it using the ALTERNATE key
                if descriptor is not None: # Yes, we were able to decode it using the ALTERNATE key
                    if descriptor.funcName in self.whitelisted_funcs: # Only explicitly "whitelisted" RPC calls are allowed to do this!
                        self.alternate_key_used = True # make a note that this came in as a special case
                    else:
                        descriptor = None # otherwise we throw it away, even though we WERE able to decode it

        if descriptor:
            assert isinstance(descriptor, RpcCodec.RpcDescriptor), "Did NOT receive an RPC descriptor object"

            if self.is_valid_seq(descriptor.seq, descriptor.srcAddr):
                log.debug("rpc: %s.%s%s" % (binascii.hexlify(descriptor.srcAddr), descriptor.funcName, descriptor.args))
                self.update_seq(descriptor.seq, descriptor.srcAddr)
                self.remote_addr = descriptor.srcAddr
                self.incoming_intf = interface

                descriptor.last_hop_addr = self.local_addr
                if descriptor.isMulticast:
                    if descriptor.srcAddr == self.local_addr:
                        descriptor = None
                        if self.verbose:
                            log.debug("multi-cast RPC originated from us")
                    else:
                        descriptor.ttl -= 1
                        if descriptor.ttl > 0:
                            if int(binascii.hexlify(descriptor.dstAddr), 16) & self.forwarded_group:
                                if self.verbose:
                                    log.debug("Rebroadcasting RPC to group %s" % (binascii.hexlify(descriptor.dstAddr)))
                                self._notify_callbacks('broadcasts', descriptor.encoded_data, interface, int(binascii.hexlify(descriptor.dstAddr), 16))
                            elif __debug__ and self.verbose:
                                log.debug("Group was not a member of forwarding groups")
                        elif __debug__ and self.verbose:
                            log.debug("Multicast RPC TTL too low to re-broadcast")

                        if not (int(binascii.hexlify(descriptor.dstAddr), 16) & self.processed_groups):
                            descriptor = None
                            if __debug__ and self.verbose:
                                log.debug("Group was not a member of processing groups")
                elif descriptor.final_dst_addr != self.local_addr:  # if the packet was not for me...
                    self._forward_packet(descriptor)  # ...send it on to it's final destination
                    descriptor = None  # Set to None so we don't notify local code about packets that aren't for us
            else:
                if __debug__ and self.verbose:
                    multicast = "multi-cast " if descriptor.isMulticast else ""
                    log.debug("Received %sRPC from %s with same duplicate number %02X" % (multicast, binascii.hexlify(descriptor.srcAddr), descriptor.seq,))
                descriptor = None

        return descriptor

    def register_pkt_type(self, pid, callback):
        """
        Register for notifications when a packet is received with specified pid

        Args:
            pid : A valid SNAP PID
            callback : The function object to be notified
        """
        if not self.interested.has_key(pid):
            self.interested[pid] = []

        if callback not in self.interested[pid] and callable(callback):
            self.interested[pid].append(callback)

    def remove_sniffer(self, sniffer_function, raw_data):
        """Remove the local sniffer

        Args:
            sniffer_function: The function to be removed or None if removing a local sniffer
            raw_data: If this sniffer is being used for descriptors or raw data

        Returns:
            True if the sniffer was removed
            False if we didn't have a sniffer
        """
        assert callable(sniffer_function)

        if not raw_data:
            if sniffer_function in self._local_sniffers:
                self._local_sniffers.remove(sniffer_function)
                return True
            else:
                return False
        else:
            if sniffer_function in self._remote_sniffers:
                self._remote_sniffers.remove(sniffer_function)
                return True
            else:
                return False

    def send(self, descriptor, conn=None):
        """
        Sends the specified descriptor, preforming a route lookup if necessary

        Args:
            descriptor : The descriptor object to send
            conn : A connection object to use to send the descriptor otherwise to a lookup
        """
        self.send_seq += 1

        if (descriptor and
            isinstance(descriptor, RpcCodec.RpcDescriptor) or
            isinstance(descriptor, TraceRouteCodec.TraceRouteDescriptor) or
            isinstance(descriptor, DataModeCodec.DataModeDescriptor)):
            log.debug("send: %s" % (str(descriptor)))

            if descriptor.isMulticast:
                descriptor.srcAddr = self.local_addr
                if conn is None:
                    #self._notify_callbacks('broadcasts', descriptor.encoded_data, None) # DON'T DO THIS
                    self.scheduler.schedule(0, self._notify_callbacks, 'broadcasts', descriptor.encoded_data, None, int(binascii.hexlify(descriptor.dstAddr), 16))
                    self.scheduler.schedule(0, self._notifyCallbacks, hooks.HOOK_RPC_SENT, self.send_seq, True)
                else:
                    #conn.intf_write(descriptor.encoded_data) # DON'T DO THIS
                    encoded_data = descriptor.encoded_data
                    self.scheduler.schedule(0, self._delayed_write, conn.intf_write, encoded_data, self.send_seq)
                    log.debug("Directing multi-cast to %s" % (conn.interface))
                    self.sniff_packet(conn.interface, encoded_data, rx=False)
            else:
                assert conn is None, "We only can send directed multi-casts"
                (addr, intf) = self.mesh_handler.get_next_hop(descriptor.final_dst_addr, lambda addr, intf: self._on_next_hop(descriptor, addr, intf))

                if addr is None:
                    self._waiting[descriptor] = self.send_seq
                else:
                    descriptor.dstAddr = addr
                    encoded_data = descriptor.encoded_data
                    self.scheduler.schedule(0, self._delayed_write, intf.write, encoded_data, self.send_seq)
                    self.sniff_packet(intf, encoded_data, rx=False)
        else:
            raise TypeError("Unknown descriptor type")

        return self.send_seq

    def set_crypto(self, crypto):
        """Sets the current crypto to use"""
        self.crypto = crypto

    def set_alternate_crypto(self, alternate_crypto):
        """Sets the alternate crypto to use"""
        self.alternate_crypto = alternate_crypto

    def set_mcast_groups(self, forwarding_groups, processed_groups):
        """
        Allows dynamic changing of the multicast groups

        Args:
            forwarded_group : The multicast forward groups that should be used
            processed_groups : Which multicast groups we should process and dispatch locally
        """
        assert isinstance(forwarding_groups, int)
        assert isinstance(processed_groups, int)
        self.forwarded_group = forwarding_groups
        self.processed_groups = processed_groups

    def sniff_packet(self, interface, pkt, rx):
        """Pass a packet to a local or remote sniffer

        Packets received on an interface with a sniffer_callback will have extra
        meta data sent from the other end.  The following describes how this information is encoded:

        pkt[0] is sniffer version

        SNIFFER_VERSION 1:
            pkt[1] is T for transmit or R for received
            pkt[2:] is the original packet

        Args:
            interface: Interface the packet was received on
            pkt: The raw packet data
            rx: If True, this packet was being, if False, this packet was being transmitted

        Returns:
            True if the packet should be processed
            False if the packet should not be processed
        """
        process_packet = True
        if rx:
            sniffer_callback = getattr(interface, "sniffer_callback", None)
            if sniffer_callback:
                #Handle packets received on a sniffer interface
                #These are sent from a remote_sniffer on the other side.
                #These packets have extra meta data attached
                if ord(pkt[0]) != SNIFFER_VERSION:
                    log.info("Attempting to remotely sniff packets from a different sniffer version: %d"%(ord(pkt[0])))
                    return False
                direction_code = pkt[1]
                pkt = pkt[2:]
                descriptor = self._decode_pkt(pkt)
                descriptor.rx = (direction_code == "R")
                descriptor.tx = (direction_code == "T")
                descriptor.raw = pkt
                sniffer_callback(descriptor)
                return False #Do not process packets on a sniffer interface

        if self._local_sniffers:
            descriptor = self._decode_pkt(pkt)
            descriptor.rx = rx
            descriptor.tx = not rx
            descriptor.raw = pkt

            for sniffer in self._local_sniffers:
                try:
                    sniffer(descriptor) #Local sniffers receive descriptor objects
                except Exception, ex:
                    log.exception("An error occurred while processing a packet for a local sniffer")

        #Forward packets to other SNAP Connect instances with sniffers
        #These will be receieved on the other side by an interface with 'sniffer_callback' set
        if self._remote_sniffers:
            #If we have remote_sniffers we may receive a message from the server, and we shouldn't process them
            process_packet = not getattr(interface, "is_sniffer_server", False)
            direction_code = "R" if rx else "T"
            for sniffer in self._remote_sniffers:
                try:
                    sniffer(chr(SNIFFER_VERSION)+direction_code+pkt)
                except Exception, es:
                    log.exception("An error occurred while processing an RX packet for a remote sniffer")

        return process_packet

    def update_seq(self, seq_num, addr):
        """Updates the seq number list for the given node address"""
        (last_seq_timestamp, seq_nums) = self._pkt_seqs.get(addr, (None, []))
        if last_seq_timestamp is None:
            self._pkt_seqs[addr] = (time.time(), seq_nums)
        elif len(seq_nums) >= SEQ_CHECK_LIMIT:
            seq_nums.pop(0)

        last_seq_timestamp = time.time()
        seq_nums.append(seq_num)


class RawDescriptor(object):
    """Represent a raw packet that could not be decoded for some reason"""
    def __init__(self):
        pass
