# (c) Copyright 2009-2010, Synapse Wireless, Inc.
"""SNAP protocol listeners"""

import sys
import asyncore
import socket
import collections
import logging

import snaptcp

from apy import multicast, Deferred

from snaplib import PacketSerialProtocol, interfaces, SnapConstantsHelper

from serialwrapper import PyserialDriver

log = logging.getLogger("snap.%s" % (__name__))


class LocalListener(multicast.MulticastListener, PacketSerialProtocol.PacketSerialProtocol):
    """Start a listener for multicast SNAP messages"""
    def __init__(self, mcast_addr=snaptcp.SNAP_IP_ADDR, port=snaptcp.SNAP_IP_PORT):
        """
        Initializes and instance of a LocalListener object

        Args:
            mcast_addr : The multicast IP address to listen on
            port : The TCP port number to listen on
        """
        multicast.MulticastListener.__init__(self, mcast_addr, port)
        PacketSerialProtocol.PacketSerialProtocol.__init__(self)
        # PacketSerialProtocol is already a Registrar
        assert not self._registeredEvents.has_key('connect')
        self._registeredEvents['connect'] = []
        self.verboseLogging = True
        self.decode = True
        self.neverAck = True

        # Not sure why I can't send on the socket I'm listening on
        self._sending_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self._sending_socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 20)

    def broadcast(self, data, intf):
        """
        Sends the given data out all interfaces except the one provided

        Args:
            data : The data to send
            intf : The interface this data came from which we will not resend to
        """
        from_eth = isinstance(intf, interfaces.TcpIpInterface)
        if (from_eth and not intf.multicast) or not from_eth:
            self.send_data(data)

    def flush_write_buff(self):
        pass  # This implementation does nothing when called

    def get_intf_name(self):
        """Returns the interface identifier string"""
        return "LocalListener"

    def handle_read(self):
        """Called upon receiving a multi-cast packet"""
        #(src_addr, src_port) = self.socket.getsockname()
        data, (src_addr, src_port) = self.socket.recvfrom(8192)

        # Check to make sure this did not originate from us
        if src_addr != socket.gethostbyname(socket.gethostname()):
            # I think we want to notify of the connection before notifying of the data to possibly setup a new connection
            self._notify_callbacks('connect', src_addr)
            # Send on to PacketSerial
            self.on_data(data, interfaces.TcpIpInterface(src_addr, src_port, multicast=True))

    def ready_to_send(self):
        return True  # We are always ready!

    def write(self, data):
        """
        Sends the data out the multicast socket

        Args:
            data : The data bytes to send
        """
        self._sending_socket.sendto(data, (snaptcp.SNAP_IP_ADDR, snaptcp.SNAP_IP_PORT))


class SerialListener(PacketSerialProtocol.PacketSerialProtocol):
    """A listener for SNAP messages on a serial port"""
    def __init__(self, serial_type, port_num, dll_path=PyserialDriver.MODULE_PATH, forward_groups=0x01, baudrate=None):
        """
        Initializes an instance of a SerialListener object

        Args:
            serial_type : PyserialDriver type to instantiate
            port_num : The port number of the specified type to open
            dll_path : The full path to the location of the USB DLLs
            forward_groups : The multicast forwarding groups to use on this interface
            baudrate : The baudrate for the serial connection
        """
        PacketSerialProtocol.PacketSerialProtocol.__init__(self)
        #self.verboseLogging = __debug__
        self.serial_driver = PyserialDriver.PyserialWrapper(serial_type, port_num, dllPath=dll_path, baudrate=baudrate)
        if serial_type == PyserialDriver.PyserialWrapper.TYPE_USBXSERIAL:
            self.intf = interfaces.UsbInterface(port_num, SnapConstantsHelper.INTF_TYPE_SILABS_USB)
        elif serial_type == PyserialDriver.PyserialWrapper.TYPE_SNAPSTICK:
            self.intf = interfaces.UsbInterface(port_num, SnapConstantsHelper.INTF_TYPE_SNAPSTICK)
        elif serial_type == PyserialDriver.PyserialWrapper.TYPE_PYSERIAL:
            self.intf = interfaces.Rs232Interface(port_num)
        elif serial_type == PyserialDriver.PyserialWrapper.TYPE_BTLE:
            self.neverAck = True # Because of the large number of packet fragments, disable this
            self.intf = interfaces.BtleSerialInterface(port_num)
        else:
            raise ValueError("Unknown serial type")
        assert isinstance(self.intf, interfaces.Interface)
        self.intf.write = self.intf_write
        self.serial_driver.registerRxCallback(lambda data: self.on_data(data, self.intf))
        self.serial_driver.registerCloseCallback(self._on_serial_driver_close)
        self.register_callback('tx', self._on_pkt_tx)
        self._pkt_queue = collections.deque()
        self.remote_addr = None
        self.initiator = True
        self.interface = self.intf
        self.forward_groups = forward_groups

    def close(self):
        """Closes this instance"""
        self.serial_driver.close()

    def flush_write_buff(self):
        """Flushes the underlying serial interface write buffer"""
        return self.serial_driver.flushWriteBuf()

    def get_intf_name(self):
        """Returns a meaningful name for this interface"""
        return self.serial_driver.serial.portstr

    def intf_write(self, data):
        """
        Writes the provided data out this serial interface

        Args:
            data : The data to write
        """
        if self.ready_to_send():
            return self.send_data(data)
        else:
            deferred = Deferred.Deferred()
            self._pkt_queue.append((data, deferred))
            return deferred

    def on_broadcast(self, data, intf, group=0x01):
        """
        Called when a broadcast should be sent out this interface

        Args:
            data : The data to send
            intf : The source interface of this data
            group : The multicast group this data is destined for
        """
        if intf != self.intf:
            if group & self.forward_groups:
                self.intf.write(data)
            else:
                log.debug("Broadcast filtered to group %02X" % (group))
        #else:
            #log.debug("Broadcast filtered since it originated from same interface" % (group))

    def _on_pkt_tx(self, success, tell=False):
        """
        Called when the serial interface has sent or stopped trying to send a packet

        Args:
            success : The result of the last packet send attempt
            tell : [No longer used]
        """
        try:
            (data, deferred) = self._pkt_queue.popleft()
            self.send_data(data)
            deferred(True)
        except IndexError:
            pass

    def _on_serial_driver_close(self):
        """Called when the underlying serial driver is closed and notifies interested parties the packets were not sent"""
        for (data, deferred) in self._pkt_queue:
            deferred(False)
        self._pkt_queue.clear()

    def poll(self):
        """Called by a scheduler when we can service the underlying serial connection"""
        read = bytes_written = 0
        try:
            read = self.serial_driver.readPoll()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            log.exception("Exception in readPoll:")
        if self.serial_driver.isOpen:
            #TODO: Remove isopen check since all 3 of these methods are checking it
            try:
                self._poll()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("Exception in serial_driver poll:")
        try:
            bytes_written = self.serial_driver.writePoll()
        except (KeyboardInterrupt, SystemExit):
                raise
        except:
            log.exception("Exception in writePoll:")
        return (read and bytes_written)

    def ready_to_send(self):
        """Returns if this interface is ready to send data"""
        if self.serial_driver.isOpen and self.curTxPkt is None:
            return True
        return False

    def reconnect(self):
        """Close and reopen the serial port"""
        if self.serial_driver.getOutputType():
            self.serial_driver.setOutputType(self.serial_driver.getOutputType(), self.serial_driver.port_num)
            return True
        return False

    def write(self, data, *args):
        """Write the data out the underlying serial interface"""
        self.serial_driver.write(data)
        return True  # TODO: We probably should return true once it is actually written but we dont know when that is

    def _get_forward_groups(self):
        """Get the forward groups for this interface"""
        if hasattr(self._forward_groups, 'forward_groups'):
            return self._forward_groups.forward_groups
        else:
            return self._forward_groups

    def _set_forward_groups(self, forward_groups):
        """Set the forward groups used for this interface

        This is used by the forward_groups property. forward_groups may either be an
        integer reprensenting the forwarding groups or an object that has an attribute
        named 'forward_groups'.
        """
        self._forward_groups = forward_groups

    forward_groups = property(_get_forward_groups,_set_forward_groups)


def main():
    while True:
        asyncore.poll(0.001)


if __name__ == '__main__':
    main()