# (c) Copyright 2009-2010, Synapse Wireless, Inc.
"""A Python implemenation of SNAP mesh"""


import logging
import binascii
import timeit

from snaplib import MeshCodec
from snaplib import SnapConstantsHelper
from snaplib import EventCallbacks
from snaplib import interfaces

log = logging.getLogger("snap.mesh")


NEW_UNUSED_ROUTE = 0
YOUNG_UNUSED_ROUTE = 1
USED_ROUTE = 2
DEAD_ROUTE_WALKING = 3

DEFAULT_INITIAL_HOP_LIMIT = 2

class MeshProxyInterface(interfaces.Interface):
    """
    This proxy interface is returned when someone is looking for
    the interface that an address can be reached from
    """
    def __init__(self, addr, mesh_handler):
        """
        Initializes a MeshProxyInterface instance

        addr : The SNAP address being queried for
        mesh_handler : The MeshHandler instance that created this object
        """
        interfaces.Interface.__init__(self, SnapConstantsHelper.INTF_TYPE_PROXY)
        assert len(addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        self.addr = addr
        self.mesh_handler = mesh_handler
        assert isinstance(self.mesh_handler, MeshHandler)

    def write(self, data):
        """
        Called when the route lookup has finished

        data : The data to send
        """
        route_entry = self.mesh_handler.route_table.get(self.addr, None)
        if route_entry:
            assert isinstance(route_entry, MeshRouteEntry)
            log.debug("Updating timeout for %s" % (binascii.hexlify(self.addr)))
            route_entry.timeout = self.mesh_handler.config.used_timeout + timeit.default_timer()
            self.mesh_handler._notify_callbacks('unicasts', data, route_entry.intf)
            return route_entry.intf.write(data)
        else:
            self.mesh_handler.send_route_error(self.addr)
        return False  # TODO: Is this really what we should return?


class MeshConfig(object):
    """Describes a SNAPpy mesh configuration"""
    def __init__(self):
        self.override = MeshCodec.MESH_OVERRIDE
        """determines if we are participating in the mesh"""
        self._min_timeout = MeshCodec.MIN_TIMEOUT
        """the minimum time a route will be kept"""
        self._new_timeout = MeshCodec.NEW_TIMEOUT
        """a grace period that a newly discovered route will be kept, even if it is never actually used"""
        self._used_timeout = MeshCodec.USED_TIMEOUT
        """how much additional time a route gets whenever it is used"""
        self._delete_timeout = MeshCodec.DELETE_TIEMOUT
        """The timeout before deleteing a route after going threw other states"""
        self._max_delete_timeout = MeshCodec.MAX_TIMEOUT
        """the maximum time a route can "live" """
        self._rreq_timeout = MeshCodec.RREQ_WAIT_TIMEOUT
        """how long a node will wait for a response to a Route Request (RREQ) before trying again"""
        self._initial_hop_limit = DEFAULT_INITIAL_HOP_LIMIT  # We use 2 here because all our links are P2P not shared like radios
        """how far the initial "discovery broadcast" message is propagated across the mesh"""
        self.max_hop_limit = MeshCodec.MAX_HOP_LIMIT
        """the TTL to use for subsequent RREQs"""
        self.rreq_retries = MeshCodec.RREQ_RETRIES
        """the total number of retries that will be made when attempting to "discover" a route over the mesh"""
        self.initial_seq = 1
        """the starting mesh sequence number for mesh routing packets"""
        self.max_rreq_replies = 0
        """the maximum number of times to respond to a unique RREQ"""

    # Do not allow the initial hop limit to be 0 (References #2965)
    initial_hop_limit = property(lambda self: self._initial_hop_limit if self._initial_hop_limit > 0 else DEFAULT_INITIAL_HOP_LIMIT,
                                 lambda self, value: setattr(self, '_initial_hop_limit', value))

    # Automatically convert to seconds
    min_timeout = property(lambda self: self._min_timeout * 0.001, lambda self, value: setattr(self, '_min_timeout', value))
    used_timeout = property(lambda self: self._used_timeout * 0.001, lambda self, value: setattr(self, '_used_timeout', value))
    delete_timeout = property(lambda self: self._delete_timeout * 0.001, lambda self, value: setattr(self, '_delete_timeout', value))
    max_delete_timeout = property(lambda self: self._max_delete_timeout * 0.001, lambda self, value: setattr(self, '_max_delete_timeout', value))
    rreq_timeout = property(lambda self: self._rreq_timeout * 0.001, lambda self, value: setattr(self, '_rreq_timeout', value))
    new_timeout = property(lambda self: self._new_timeout * 0.001, lambda self, value: setattr(self, '_new_timeout', value))


class MeshRouteEntry(object):
    """Describes an entry in the SNAPpy mesh route table"""
    def __init__(self, seq_num, next_hop_addr, distance, intf, config):
        """
        Initializes a MeshRouteEntry instance

        seq_num : The starting mesh sequence number
        next_hop_addr : The next hop SNAP address for this route
        distance : The mesh "distance" this hop was calculated as
        intf : The interface this route uses
        config : The MeshConfig used when the route was formed
        """
        assert isinstance(config, MeshConfig)
        # route addr is stored in container
        self.seq_num = seq_num
        self.next_hop_addr = next_hop_addr
        assert len(self.next_hop_addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        self.broken = False
        self.distance = distance
        self.state = NEW_UNUSED_ROUTE
        if config.new_timeout > config.min_timeout:
            self.timeout = config.new_timeout + timeit.default_timer()
        else:
            self.timeout = config.min_timeout + timeit.default_timer()
        self.max_delete_timeout = config.max_delete_timeout + timeit.default_timer()
        self.intf = intf
        assert isinstance(self.intf, interfaces.Interface)


class MeshRreqEntry(object):
    """Describes an entry in the SNAPpy mesh oustanding RREQ table"""
    def __init__(self, descriptor, timeout):
        """
        Initializes a MeshRouteEntry instance

        descriptor : The snaplib.MeshCodec descriptor for this RREQ
        timeout : The current timeout value for this RREQ
        """
        self.retries = 1
        self.timeout = timeout + timeit.default_timer()
        self.callbacks = []
        self.descriptor = descriptor
        assert isinstance(self.descriptor, MeshCodec.MeshDescriptor)


class MeshHandler(EventCallbacks.Registrar):
    """Handles looking up routes to SNAP nodes responding to route requests"""
    def __init__(self, local_addr, crypto, config=MeshConfig()):
        """
        Initalizes an instance of a MeshHandler

        Args:
            local_addr : This SNAP Connect instance SNAP address
            crypto : The decrypter to use
            config : A MeshConfig instance to use for this handler

        Notifications:
            broadcasts: Called when a packet should be broadcast with (data, interface)
            unicasts: Called after a unicast packet is sent with (data, interface)
        """
        self.local_addr = local_addr
        assert len(self.local_addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        EventCallbacks.Registrar.__init__(self, ('broadcasts', 'unicasts'))
        self.route_table = {}  # { addr: MeshRouteEntry }
        self.config = config
        assert isinstance(self.config, MeshConfig)
        self._rreq_table = {}  # { addr: MeshRreqEntry }
        MeshCodec.MeshCodec.sequence_number = self.config.initial_seq
        self.crypto = crypto
        self._rreq_seqs = {}  # {addr: (seq_num, count)}

    def _create_route_entry(self, descriptor):
        """
        Given a descriptor create a MeshRouteEntry and add it to the table

        descriptor : A MeshCodec.MeshDescriptor object
        """
        assert isinstance(descriptor, MeshCodec.MeshDescriptor)
        assert not self.route_table.has_key(descriptor.originator)
        entry = MeshRouteEntry(descriptor.originatorSequence,
                               descriptor.source,
                               descriptor.originatorDistance + 1,
                               descriptor.incoming_interface,
                               self.config,
                               )
        self.route_table[descriptor.originator] = entry
        log.debug("Added route to %s" % (binascii.hexlify(descriptor.originator)))
        rreq_entry = self._rreq_table.pop(descriptor.originator, None)
        if rreq_entry:
            assert isinstance(rreq_entry, MeshRreqEntry)
            for callback in rreq_entry.callbacks:
                callback(entry.next_hop_addr, entry.intf)
        return entry

    def find_route(self, addr):
        """
        Given an address get the route if known

        addr : The SNAP address to find a route for
        """
        assert len(addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        return self.route_table.get(addr, None)

    def get_addrs(self, intf):
        """Get all addresses for a specific interface"""
        return [addr for addr, entry in self.route_table.iteritems() if entry.intf == intf]

    def get_intf(self, addr):
        """Returns a MeshProxyInterface instance"""
        return MeshProxyInterface(addr, self)

    def get_next_hop(self, addr, callback):
        """Get the route to the specified address calling the callback function with the next hop address"""
        route = self.find_route(addr)
        if route is None:
            entry = self._rreq_table.get(addr, None)
            if callable(callback):
                if entry is None:
                    # No point in sending the RREQ if no one to callback
                    descriptor = MeshCodec.MeshCodec.build_rreq(addr, self.local_addr, self.config.initial_hop_limit, crypto=self.crypto)
                    try:
                        self._notify_callbacks('broadcasts', descriptor.encoded_data, None)
                        log.debug("Sent RREQ for %s" % (binascii.hexlify(addr)))
                    except MeshCodec.MeshError:
                        log.exception("Unable to send RREQ for %s" % (binascii.hexlify(addr)))

                    entry = MeshRreqEntry(descriptor, self.config.rreq_timeout)
                    self._rreq_table[addr] = entry

                entry.callbacks.append(callback)
            return (None, None)

        route.timeout = self.config.used_timeout + timeit.default_timer()
        return (route.next_hop_addr, route.intf)

    def on_descriptor(self, descriptor):
        """Called when a mesh paket is received and ready for processing"""
        assert isinstance(descriptor, MeshCodec.MeshDescriptor)
        if descriptor.msgId == MeshCodec.MESH_RREQ:
            if self.config.override:
                return

            if descriptor.originator != self.local_addr and self.local_addr not in descriptor.additionalAddresses:
                #Don't respond if you sent the original packet or have already re-sent this packet
                if descriptor.target == self.local_addr:
                    route_entry = self.route_table.get(descriptor.originator, None)
                    if route_entry:
                        if (route_entry.distance < descriptor.originatorDistance + 1 or
                            route_entry.intf != descriptor.incoming_interface):
                            self._update_route(descriptor)
                    else:
                        self._create_route_entry(descriptor)

                    (last_seq, seq_count) = self._rreq_seqs.get(descriptor.originator, (None, 0))
                    if last_seq != descriptor.originatorSequence or seq_count < self.config.max_rreq_replies:
                        MeshCodec.MeshCodec.adjustSequenceNumber(descriptor)
                        rrep_descriptor = MeshCodec.MeshCodec.build_rrep(descriptor)
                        log.debug("Sending route reply to %s:%i" % (binascii.hexlify(rrep_descriptor.target), descriptor.originatorSequence))

                        encoded_data = MeshCodec.MeshCodec.encode(rrep_descriptor, crypto=self.crypto)
                        self._notify_callbacks('unicasts', encoded_data, descriptor.incoming_interface)
                        descriptor.incoming_interface.write(encoded_data)
                        if last_seq != descriptor.originatorSequence:
                            seq_count = 0
                        self._rreq_seqs[descriptor.originator] = (descriptor.originatorSequence, seq_count + 1)
                    else:
                        log.debug("RREQ max replies was reached for %s:%i" % (binascii.hexlify(descriptor.originator), descriptor.originatorSequence))
                elif descriptor.hopLimit > 1 and self.config.override == 0:
                    if not self.route_table.has_key(descriptor.originator):
                        self._create_route_entry(descriptor)
                    descriptor.hopLimit -= 1

                    max_rreq_addrs = (SnapConstantsHelper.MAX_PKT_SIZE - SnapConstantsHelper.MBST_EXTRA_OFS) // SnapConstantsHelper.SNAP2_ADDR_LEN

                    if len(descriptor.additionalAddresses) >= max_rreq_addrs:
                        #If we have too many additional addresses, truncate here and then append our address
                        descriptor.additionalAddresses = descriptor.additionalAddresses[:max_rreq_addrs-1]

                    descriptor.additionalAddresses.append(self.local_addr)
                    descriptor.originatorDistance += 1
                    descriptor.source = self.local_addr
                    try:
                        self._notify_callbacks('broadcasts', descriptor.encoded_data, descriptor.incoming_interface)
                        log.debug("Rebroadcasting RREQ")
                    except MeshCodec.MeshError:
                        log.exception("Unable rebroadcast RREQ")
                else:
                    log.debug("RREQ reached hop limit, or mesh override is on")
            else:
                log.debug("RREQ was sent by us or we have already seen")
        elif descriptor.msgId == MeshCodec.MESH_RREP:
            self._process_message(descriptor)
            if descriptor.hopLimit > 1 and descriptor.target != self.local_addr:
                descriptor.hopLimit -= 1
                route_entry = self.find_route(descriptor.target)
                if route_entry and not route_entry.broken:
                    assert isinstance(route_entry, MeshRouteEntry)
                    descriptor.source = self.local_addr
                    descriptor.destination = route_entry.next_hop_addr
                    try:
                        encoded_data = descriptor.encoded_data
                        self._notify_callbacks('unicasts', encoded_data, route_entry.intf)
                        route_entry.intf.write(encoded_data)
                        log.debug("Sending RREP to %s about %s" % (binascii.hexlify(descriptor.originator),
                                                                   binascii.hexlify(descriptor.target)))
                    except MeshCodec.MeshError:
                        log.exception("Unable to send RREP to %s about %s" % (binascii.hexlify(descriptor.originator),
                                                                              binascii.hexlify(descriptor.target)))
            else:
                log.debug("RREP hop limit reached or was for our address")
        elif descriptor.msgId == MeshCodec.MESH_RERR:
            resend = False
            addrs = descriptor.additionalAddresses[:]  # Copy addrs
            seqs = descriptor.additionalSeqs[:]
            descriptor.additionalAddresses = []
            descriptor.additionalSeqs = []

            #We only rebroadcast the RERR if we know about this route
            route = self.route_table.get(descriptor.target, None)
            if route:
                resend = True
                if (descriptor.targetSequence == 0 or
                    route.seq_num == 0 or
                    route.seq_num - descriptor.targetSequence <= 0):
                    orig_source = descriptor.source
                    descriptor.source = self.local_addr  # Get descriptor ready to resend
                    self.route_table.pop(descriptor.target, None)
                    log.debug("Removed route to %s because of RERR from %s" % (binascii.hexlify(descriptor.target), binascii.hexlify(orig_source)))

                for i in range(0, len(addrs)):
                    route = self.route_table.get(addrs[i], None)
                    if route:
                        if (seqs[i] == 0 or
                            route.seq_num == 0 or
                            route.seq_num - seqs[i] <= 0):
                            descriptor.additionalAddresses.append(addrs[i])
                            descriptor.additionalSeqs.append(seqs[i])
                            self.route_table.pop(addrs[i], None)
                            log.debug("Removed route to %s because of RERR" % (binascii.hexlify(addrs[i])))

            if resend and descriptor.hopLimit > 1:
                descriptor.hopLimit -= 1
                try:
                    self._notify_callbacks('broadcasts', descriptor.encoded_data, None)
                    log.debug("Rebroadcasting RERR")
                except MeshCodec.MeshError:
                    log.exception("Unabled to rebroadcast RERR")
        else:
            log.info("Received unknown mesh routing message")

    def on_next_hop_addr(self, addr, interface):
        """
        Called when an interface determines the SNAP address of his directly connected neighbor

        addr : The directly connect SNAP address
        interface : The interface the address is connected to
        """
        if addr:
            entry_remove = []
            for (entry_addr, entry) in self.route_table.iteritems():
                if entry.next_hop_addr == addr and entry.intf != interface:
                    entry_remove.append(entry_addr)

            for entry_addr in entry_remove:
                log.debug("Removing route entry for %s via %s due to %s CHANGING" % (binascii.hexlify(entry_addr),
                                                                                     binascii.hexlify(addr),
                                                                                     interface))
                self.route_table.pop(entry_addr, None)

            self.route_table[addr] = MeshRouteEntry(0,
                                                    addr,
                                                    1,
                                                    interface,
                                                    self.config
                                                   )

    def on_intf_closed(self, intf):
        """
        Called when an interface is closed so we can clear all the routes we know out that interface

        intf : The interface that was just closed
        """
        entry_remove = []
        for (addr, entry) in self.route_table.iteritems():
            if entry.intf == intf:
                entry_remove.append(addr)

        for addr in entry_remove:
            log.debug("Removing route entry for %s due to interface closure" % (binascii.hexlify(addr)))
            self.route_table.pop(addr, None)

    def poll(self):
        """Call this method frequenty to help prune routes for tables"""
        rreq_remove = []
        entry_remove = []
        now = timeit.default_timer()
        work_done = False

        for (addr, entry) in self._rreq_table.iteritems():
            work_done = True
            assert isinstance(entry, MeshRreqEntry)
            if now > entry.timeout:
                if entry.retries < self.config.rreq_retries:
                    entry.descriptor.hopLimit = self.config.max_hop_limit
                    try:
                        self._notify_callbacks('broadcasts', entry.descriptor.encoded_data, None)
                        log.debug("RREQ retry #%i for %s" % (entry.retries, binascii.hexlify(addr)))
                    except MeshCodec.MeshError:
                        log.exception("Unable to send RREQ retry #%i for %s" % (entry.retries, binascii.hexlify(addr)))
                    entry.retries += 1
                    entry.timeout = self.config.rreq_timeout * entry.retries + timeit.default_timer()
                else:
                    rreq_remove.append(addr)

        for addr in rreq_remove:
            entry = self._rreq_table.pop(addr, None)
            if entry:
                for callback in entry.callbacks:
                    log.debug("RREQ timeout for %s" % (binascii.hexlify(addr)))
                    try:
                        callback(None, None)
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        log.exception("An error occurred when notifying a callback of a RREQ timeout")

        if self.config.max_delete_timeout > 0:
            # I wonder if its faster to get the copy and just delete or the iterator and delete later?
            for (addr, entry) in self.route_table.iteritems():
                work_done = True
                assert isinstance(entry, MeshRouteEntry)
                if now > entry.max_delete_timeout:
                    entry_remove.append(addr)
                    log.debug("Route entry for %s removed because MAX timeout reached" % (binascii.hexlify(addr)))
                elif now > entry.timeout:
                    if entry.state == DEAD_ROUTE_WALKING:
                        entry_remove.append(addr)
                        log.debug("Route entry for %s removed because timeout reached" % (binascii.hexlify(addr)))
                    else:
                        entry.state = DEAD_ROUTE_WALKING
                        entry.timeout = self.config.delete_timeout + timeit.default_timer()

            for addr in entry_remove:
                log.debug("Removing route entry for %s" % (binascii.hexlify(addr)))
                self.route_table.pop(addr, None)

        return work_done

    def print_route_table(self):
        """A helper function to print the route table in a some what useful format"""
        for addr, entry in self.route_table.iteritems():
            log.debug("%s|%02X|%s|%s|%i|%i|%f|%f|%s" % (binascii.hexlify(addr),
                                                        entry.seq_num,
                                                        binascii.hexlify(entry.next_hop_addr),
                                                        entry.broken,
                                                        entry.distance,
                                                        entry.state,
                                                        entry.timeout,
                                                        entry.max_delete_timeout,
                                                        entry.intf))

        for addr, entry in self._rreq_table.iteritems():
            log.debug("%s|%i|%f" % (binascii.hexlify(addr), entry.retries, entry.timeout))

    def _process_message(self, descriptor):
        """This method modifies the given descriptor to make it ready to send"""
        assert isinstance(descriptor, MeshCodec.MeshDescriptor)
        route_entry = self.find_route(descriptor.originator)
        if route_entry:
            assert isinstance(route_entry, MeshRouteEntry)
            #We don't deal with link qualities so we currently only base it off hops
            if route_entry.broken or route_entry.distance < descriptor.hopLimit + 1:
                self._update_route(descriptor)
            else:
                log.debug("Route is not better")
        else:
            self._create_route_entry(descriptor)

    def remove_next_hop(self, addr, interface):
        """
        Remove all entries in the route table that use the specified address as their next hop
        """
        if addr:
            entry_remove = []
            for (entry_addr, entry) in self.route_table.iteritems():
                if entry.next_hop_addr == addr and entry.intf == interface:
                    entry_remove.append(entry_addr)

            for entry_addr in entry_remove:
                log.debug("Removing route entry for %s via %s due to %s going AWAY" % (binascii.hexlify(entry_addr),
                                                                                       binascii.hexlify(addr),
                                                                                       interface))
                self.route_table.pop(entry_addr, None)

    def send_route_error(self, addr):
        """
        Send a RERR packet for the specified address

        addr : The SNAP address this RERR packet is about
        """
        assert len(addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        descriptor = MeshCodec.MeshDescriptor(crypto=self.crypto)
        descriptor.msgId = MeshCodec.MESH_RERR
        descriptor.hopLimit = self.config.max_hop_limit
        descriptor.target = addr

        route = self.find_route(addr)
        if route:
            descriptor.targetSequence = route.seq_num
        else:
            descriptor.targetSequence = 0

        descriptor.originator = descriptor.source = self.local_addr

        for (entry_addr, entry) in self.route_table.iteritems():
            assert isinstance(entry, MeshRouteEntry)
            if entry.next_hop_addr == addr:
                descriptor.additionalAddresses.append(entry_addr)
                descriptor.additionalSeqs.append(entry.seq_num)

        # ROUTE ERRORS have unreachable pairs of address and sequence number after the unreachable sequence number
        max_rerr_addrs = (SnapConstantsHelper.MAX_PKT_SIZE - SnapConstantsHelper.MBST_TGT_SEQ_OFS - 2) // (SnapConstantsHelper.SNAP2_ADDR_LEN + 2)

        if len(descriptor.additionalAddresses) > max_rerr_addrs:
            #Truncate if we have too many addresses to fit into a packet
            descriptor.additionalAddresses = descriptor.additionalAddresses[:max_rerr_addrs]
            descriptor.additionalSeqs = descriptor.additionalSeqs[:max_rerr_addrs]

        try:
            self._notify_callbacks('broadcasts', descriptor.encoded_data, None)
            log.debug("Sending RERR for %s" % (binascii.hexlify(addr)))
        except MeshCodec.MeshError:
            log.exception("Unable to send RERR for %s" % (binascii.hexlify(addr)))

    def set_config(self, config):
        """Sets the configuration to use for new routes"""
        assert isinstance(config, MeshConfig)
        self.config = config

    def set_crypto(self, crypto):
        """Set the current crypto to use"""
        self.crypto = crypto

    def _update_route(self, descriptor):
        """Update the currently stored route with new information"""
        assert isinstance(descriptor, MeshCodec.MeshDescriptor)
        assert self.route_table.has_key(descriptor.originator)
        self.route_table.pop(descriptor.originator, None)
        return self._create_route_entry(descriptor)
