"""SNAP Connect over the air upgrader for SNAP node firmware"""
import logging
log = logging.getLogger("snapconnect.%s" % (__name__))

import snap

from snaplib.upgrader import *
from snaplib import upgrader
from snaplib import SnapConstantsHelper
from snaplayout import MC1322x_Layout
from snaplayout import STM32W108xB_Layout
from snaplayout import ATmega128RFA1_Layout

class SnapConnectOtaUpgrader(OtaUpgrader):
    """Class used to perform OTA Firmware upgrades with SNAP Connect"""

    def __init__(self, scheduler, rpc_func):
        super(SnapConnectOtaUpgrader, self).__init__(scheduler, rpc_func)

        #Supported Layouts (CPU, Layout Class)
        self.register_layout(SnapConstantsHelper.CPU_ARM7, MC1322x_Layout.MC1322xLayout)
        self.register_layout(SnapConstantsHelper.CPU_ARM_CORTEX_M3, STM32W108xB_Layout.STM32W108xB_Layout)
        self.register_layout(SnapConstantsHelper.CPU_ATMEGA128RFA1, ATmega128RFA1_Layout.ATmega128RFA1_Layout)
        self.register_layout(SnapConstantsHelper.CPU_ATMEGA128RFR2, ATmega128RFA1_Layout.ATmega128RFA1_Layout)
        self.register_layout(SnapConstantsHelper.CPU_ATMEGA1284RFR2, ATmega128RFA1_Layout.ATmega128RFA1_Layout)

        self.hook_upgrade_complete = None # callable(addr, status, message)
        self.hook_upgrade_status = None # callable(addr, percent_complete)

        self._max_progress = {} # {addr: max_progress}

    def register_complete_hook(self, hook):
        """Add HOOK_OTA_UPGRADE_COMPLETE(addr, status, message) hook"""
        self.hook_upgrade_complete = hook

    def register_status_hook(self, hook):
        """Add HOOK_OTA_UPGRADE_STATUS(addr, percent_complete) hook"""
        self.hook_upgrade_status = hook

    def handle_error(self, addr, error_type, arg=None):
        """Handle OTA errors. See upgrader.py for details"""
        #Remove the upgrade from the upgrade list
        if addr in snap._upgrades:
            del snap._upgrades[addr]

        #Set the message for each error type
        message = None

        if error_type == OTA_ERROR_BAD_FILENAME:
            message = 'The image filename %r must include the platform (like RF200 or SS200)'%(arg)

        elif error_type == OTA_ERROR_IMCOMPATIBLE_IMAGE:
            message = 'The image file is not compatible with the selected node, please select an appropriate image file.'

        elif error_type == OTA_ERROR_UNABLE_TO_OPEN:
            log.exception('An error occurred while opening the firmware image')
            message = 'An error occurred while opening the firmware image'

        elif error_type == OTA_ERROR_FLASHER:
            log.exception("An error occurred in the upgrader:")
            message = 'An error occurred in the upgrader, please check file type and connection'

        else:
            # Log any unexpected calls to handle_error
            log.exception("An unexpected error occurred in the flasher: %r" % error_type)
            message = 'An exception has occured during an over the air upgrade'

        if self.hook_upgrade_complete:
            self.hook_upgrade_complete(addr, OTA_PROGRESS_ERROR, message)

    def start_upload(self, addr, layout):
        """SNAP Connect just kicks off the upload without performing any additional startup"""
        return self.startOtaUpgrade(addr, layout, meta=None) #SNAP Connect has no extra metadata to add

    def upload_complete(self, addr, result, meta=None):
        #Remove the upgrade from the upgrade list
        if addr in snap._upgrades:
            del snap._upgrades[addr]

        #Set the message for each error type
        message = None
        if result == OTA_PROGRESS_TIMEOUT:
            message = "A timeout occurred while uploading the firmware image"

        elif result == OTA_PROGRESS_ERROR:
            message = "An error occurred while uploading the firmware image, please try again"

        elif result == OTA_PROGRESS_WRITE_ERROR:
            message =  "An error occurred while writing the firmware image"

        elif result == OTA_PROGRESS_WRITE_REFUSED:
            message = "An error occurred while writing the firmware image, operating voltage too low"

        elif result == OTA_PROGRESS_UNSUPPORTED:
            message = "An error occurred while writing the firmware image, uploading unsupported"

        elif result == OTA_PROGRESS_COMPLETE:
            pass

        if self.hook_upgrade_complete:
            self.hook_upgrade_complete(addr, result, message)


    def upload_progress(self, addr, value, meta=None, max_progress=None):
        progress = 0
        if max_progress:
            self._max_progress[addr] = max_progress
        else:
            max_progress = self._max_progress.get(addr,None)

        if max_progress:
            progress = 100.0 * value / max_progress

        self.hook_upgrade_status(addr, progress)
