# (c) Copyright 2009-2012 Synapse Wireless, Inc.
"""A SNAP implementation in Python"""
__docformat__ = "plaintext en"


import sys
import platform
import asyncore
import logging
import inspect
import os
import errno
import cPickle
import binascii
import types
import random
import string
import time
import socket

import dispatchers
import listeners
import mesh
import snaptcp
import snaptcp_asyncore
import LicenseManager
import AutoPickler
import ota_upgrader

from snaplib import SnapConstants
from snaplib import SnapConstantsHelper
from snaplib import RpcCodec
from snaplib import DataModeCodec
from snaplib.snappy import hooks
try:
    from snaplib import aes
except (ImportError, SystemError):
    aes = None
from snaplib import pme
from snaplib import Snap2Decoder
from snaplib import TraceRouteCodec
from snaplib import PacketSerialProtocol
from snaplib import EventCallbacks
from snaplib import SnappyUploader

# Disable pylint complaining about imports not being used, we added them for user convenience
# pylint: disable-msg=W0611
from snaplib.SnapConstantsHelper import (NV_MAC_ADDR_ID, NV_NETWORK_ID, NV_CHANNEL_ID, NV_GROUP_INTEREST_MASK_ID, NV_GROUP_FORWARDING_MASK_ID,
                                         NV_MFG_DATE_ID, NV_DEVICE_NAME_ID, NV_SYSTEM_ERROR_ID, NV_DEVICE_TYPE_ID, NV_FEATURE_BITS_ID,
                                         NV_DEFAULT_UART_ID, NV_UART_DM_TIMEOUT_ID, NV_UART_DM_THRESHOLD_ID, NV_UART_DM_INTERCHAR_ID,
                                         NV_CARRIER_SENSE_ID, NV_COLLISION_DETECT_ID, NV_COLLISION_AVOIDANCE_ID, NV_SNAP_MAX_RETRIES_ID,
                                         NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID, NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID, NV_MESH_ROUTE_NEW_TIMEOUT_ID,
                                         NV_MESH_ROUTE_USED_TIMEOUT_ID, NV_MESH_ROUTE_DELETE_TIMEOUT_ID, NV_MESH_RREQ_TRIES_ID,
                                         NV_MESH_RREQ_WAIT_TIME_ID, NV_MESH_INITIAL_HOPLIMIT_ID, NV_MESH_MAX_HOPLIMIT_ID, NV_MESH_SEQUENCE_NUMBER_ID,
                                         NV_MESH_OVERRIDE_ID, NV_MESH_PENALTY_LQ_THRESHOLD_ID, NV_MESH_REJECTION_LQ_THRESHOLD_ID, NV_CS_CD_LEVEL_ID,
                                         NV_SNAP_LQ_THRESHOLD_ID, NV_SNAPPY_CRC_ID, NV_SYS_PLATFORM_ID, NV_CRYPTO_TYPE, NV_AES128_ENABLE_ID,
                                         NV_CRYPTO_KEY, NV_AES128_KEY_ID, NV_LOCKDOWN_FLAGS_ID, NV_MAX_LOYALTY_ID, NV_ALT_CRYPTO_KEY, NV_ALTERNATE_KEY_ID,
                                         NV_LAST_VERSION_BOOTED_ID, NV_REBOOTS_REMAINING_ID, NV_ALT_RADIO_TRIM_ID, NV_VENDOR_SETTINGS_ID,
                                         NV_CLOCK_REGULATOR_ID, NV_RADIO_CALIBRATION_ID, NV_TX_POWER_LIMIT_ID,
                                         ENCRYPTION_TYPE_AES128, ENCRYPTION_TYPE_BASIC, ENCRYPTION_TYPE_NONE)
from snaplib.SnapConstantsHelper import (INTF_TYPE_UNKNOWN, INTF_TYPE_802154, INTF_TYPE_SERIAL,
                                         INTF_TYPE_SILABS_USB, INTF_TYPE_ETH, INTF_TYPE_PROXY, INTF_TYPE_SNAPSTICK)
from snaplib.SnappyUploader import (SNAPPY_PROGRESS_ERASE, SNAPPY_PROGRESS_UPLOAD, SNAPPY_PROGRESS_COMPLETE,
                                    SNAPPY_PROGRESS_ERROR, SNAPPY_PROGRESS_CANCELED, SNAPPY_PROGRESS_TIMEOUT,
                                    SNAPPY_PROGRESS_WRITE_ERROR, SNAPPY_PROGRESS_WRITE_REFUSED, SNAPPY_PROGRESS_UNSUPPORTED)
from snaplib.upgrader import (OTA_PROGRESS_ERASE, OTA_PROGRESS_UPLOAD, OTA_PROGRESS_COMPLETE, OTA_PROGRESS_ERROR,
                              OTA_PROGRESS_CANCELED, OTA_PROGRESS_TIMEOUT, OTA_PROGRESS_WRITE_ERROR,
                              OTA_PROGRESS_WRITE_REFUSED, OTA_PROGRESS_UNSUPPORTED)

# Add SNAP Connect only hooks
hooks.HOOK_OTA_UPGRADE_COMPLETE = "HOOK_OTA_UPGRADE_COMPLETE"   #(addr, status, message)
hooks.HOOK_OTA_UPGRADE_STATUS = "HOOK_OTA_UPGRADE_STATUS"       #(addr, percent_complete)
hooks.HOOK_BTLE_OPENED = "HOOK_BTLE_OPENED" #(addr, error or None)
hooks.HOOK_BTLE_CLOSED = "HOOK_BTLE_CLOSED" #(addr, error or None)

# Add descriptors for use by the sniffer
from snaplib.DataModeCodec import DataModeDescriptor
from snaplib.MeshCodec import MeshDescriptor
from snaplib.RpcCodec import RpcDescriptor
from snaplib.TraceRouteCodec import TraceRouteDescriptor
from dispatchers import RawDescriptor

from apy import Deferred
from apy import EventScheduler

from serialwrapper.PyserialDriver import PyserialWrapper, MODULE_PATH


# Add realms used for authentication
from snaptcp import REALM_SNIFFER, REALM_SNAP
from serialwrapper.btle_serial import REALM_BLUETOOTH

# Tornado Support
use_tornado_ioloop = False  #  Use the Tornado IO Loop instead of Asyncore
SUPPORTS_TORNADO_IOLOOP = False  #  True if using the Tornado IO Loop is supported by SNAP Connect
try:
    import snaptcp_tornado
    import tornado.ioloop
    from apy import ioloop_scheduler
    SUPPORTS_TORNADO_IOLOOP = True
except (ImportError, SystemError, SyntaxError):
    pass


# Bluetooth Support
try:
    if sys.platform.startswith("linux"):
        from snap_btle.linux_peripheral.gatt_serial_service import GattSerialPeripheral as BtleInterface
    elif sys.platform == 'darwin' and platform.machine().startswith("iP"): #includes iPad, iPod, iPhone
        from snap_btle.ios_central.gatt_serial_service import GattSerialCentral as BtleInterface
    else:
        BtleInterface = None
except (ImportError, SystemError):
    BtleInterface = None

from serialwrapper import btle_serial

# Serial Enumerations - Used with open_serial / close_serial
SERIAL_TYPE_SNAPSTICK100 = PyserialWrapper.TYPE_USBXSERIAL
SERIAL_TYPE_SNAPSTICK200 = PyserialWrapper.TYPE_SNAPSTICK
SERIAL_TYPE_RS232 = PyserialWrapper.TYPE_PYSERIAL
SERIAL_TYPE_BTLE = PyserialWrapper.TYPE_BTLE

def client_auth(realm):
    """
    An example client authorization function

    Args:
        realm : The realm specified by the remote server

    Returns:
        A tuple holding the username "public" and password "public" to the remote server
        to use as its authorization
    """
    return ("public", "public")


def server_auth(realm, username):
    """
    An example sever authentication function

    Args:
        realm : This server's realm
        username : The username specified by the remote server

    Returns:
        The password for the specified username in the realm
    """
    if username == "public":
        return "public"

# Version Information
VERSION = "3.2.0" # Version of SNAP Connect
RELEASE_TYPE = "Release Candidate 10" # "Development" / "Release Candidate N"

START_PATH = os.path.abspath(str(sys.path[0]))  # Copy the string of the directory this module was started from

SLEEP_TIME = 0.001  # seconds

NV_PARAMS_FILENAME = 'nvparams.dat'
NV_PARAMS_VERISON = 1

MAC_PREFIX = "\x00\x1c\x2c\x00\x00"

log = logging.getLogger("snap")

_upgrades = {} # {addr: upgrade_object} for upgrades in progress.  Shared across ALL SNAP Connect instances.
_serials = {}
share_serial_connections = True  # Toggle sharing of serial connections between multiple SNAP Connect instances

class Snap(EventCallbacks.Registrar):
    """Creates a SNAP Connect instance"""
    def __init__(self, license_file=None, nvparams_file=None, funcs=None, scheduler=None, addr=None, rpc_handler=None):
        """
        Initializes a SNAP Connect instance

        Args:
            license_file : The full path to a SNAP Connect license file
                           (default License.dat)
            nvparams_file : The full path to a SNAP Connect NV parameters file
                            (default nvparams.dat)
            funcs : The callable functions for this instance to expose to the
                    SNAP network (default None)
            scheduler : Internally used by SNAP Connect
            addr : The SNAP address to use from the license file (default None)
            rpc_handler : Internally used by SNAP Connect
        """
        EventCallbacks.Registrar.__init__(self, ('next_hop_addr',))
        if use_tornado_ioloop and not SUPPORTS_TORNADO_IOLOOP:
            raise RuntimeError("Tornado ioloop not supported")
        if license_file is None:
            license_file = os.path.join(START_PATH, LicenseManager.default_filename)
        try:
            self.license = LicenseManager.check_license(filename=license_file)
        except LicenseManager.LicenseError, le:
            log.error(le.message)
            raise
        except Exception:
            log.exception("An error occurred while trying to read the license:")
            raise
        if isinstance(self.license, LicenseManager.License):
            if addr is None:
                snap_addr = self.license.addr[0]
            elif addr in self.license.addr:
                snap_addr = addr
            else:
                raise RuntimeError("Non-licensed address provided")
        else:
            raise RuntimeError("Invalid license found")
        assert len(snap_addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        self.licensed_mac = "%s%s" % (MAC_PREFIX, snap_addr)

        # Max packet size changes depending on the packet CRC feature bit
        self._original_max_pkt_size = SnapConstants.MAX_PKT_SIZE

        mesh_config = mesh.MeshConfig()

        if nvparams_file is None:
            nvparams_file = os.path.join(START_PATH, NV_PARAMS_FILENAME)
        try:
            self.nv_params = AutoPickler.loadSaver(nvparams_file)
        except (EOFError, IOError), err:
            # ENOENT = No such file or directory
            # EOFError can get raised by the pickler if it is a bad file
            if isinstance(err, EOFError) or err.errno == errno.ENOENT:
                try:
                    # We didn't find the file or was invalid so create one
                    self.nv_params = AutoPickler.AutoSaver(nvparams_file)

                    self.nv_params.params = {SnapConstantsHelper.NV_MAC_ADDR_ID: self.licensed_mac,
                                             SnapConstantsHelper.NV_NETWORK_ID: 0,
                                             SnapConstantsHelper.NV_CHANNEL_ID: 0,
                                             SnapConstantsHelper.NV_GROUP_FORWARDING_MASK_ID: 1,
                                             SnapConstantsHelper.NV_GROUP_INTEREST_MASK_ID: 1,
                                             SnapConstantsHelper.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID: mesh_config._max_delete_timeout,
                                             SnapConstantsHelper.NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID: mesh_config._min_timeout,
                                             SnapConstantsHelper.NV_MESH_ROUTE_NEW_TIMEOUT_ID: mesh_config._new_timeout,
                                             SnapConstantsHelper.NV_MESH_ROUTE_USED_TIMEOUT_ID: mesh_config._used_timeout,
                                             SnapConstantsHelper.NV_MESH_ROUTE_DELETE_TIMEOUT_ID: mesh_config._delete_timeout,
                                             SnapConstantsHelper.NV_MESH_RREQ_TRIES_ID: mesh_config.rreq_retries,
                                             SnapConstantsHelper.NV_MESH_RREQ_WAIT_TIME_ID: mesh_config._rreq_timeout,
                                             SnapConstantsHelper.NV_MESH_INITIAL_HOPLIMIT_ID: mesh_config.initial_hop_limit,
                                             SnapConstantsHelper.NV_MESH_MAX_HOPLIMIT_ID: mesh_config.max_hop_limit,
                                             SnapConstantsHelper.NV_MESH_SEQUENCE_NUMBER_ID: 1,
                                             SnapConstantsHelper.NV_MESH_OVERRIDE_ID: 0,
                                             SnapConstantsHelper.NV_SNAP_MAX_RETRIES_ID: 8,
                                             SnapConstantsHelper.NV_LOCKDOWN_FLAGS_ID: 0,
                                             SnapConstantsHelper.NV_FEATURE_BITS_ID: 0,
                                             SnapConstantsHelper.NV_AES128_ENABLE_ID: SnapConstantsHelper.ENCRYPTION_TYPE_NONE,
                                            }
                    self.nv_params.version = NV_PARAMS_VERISON
                    log.info("Created new NV parameters file")
                except:
                    log.exception("An error occurred while trying to create the NV parameters file")
                    raise
            else:
                raise
        except:
            raise IOError("Unable to load NV params file: %s" % (nvparams_file))
        assert isinstance(self.nv_params.params, dict)
        assert self.nv_params.version == NV_PARAMS_VERISON

        self.apply_feature_bits()

        #mac = self.load_nv_param(SnapConstantsHelper.NV_MAC_ADDR_ID)
        #if mac != self.licensed_mac:
            #log.error("The licensed MAC address does not match the NV parameters")
            #self.save_nv_param(SnapConstantsHelper.NV_MAC_ADDR_ID, self.licensed_mac)

        if rpc_handler is None:
            if funcs is None:
                prev_frame = inspect.currentframe().f_back
                if prev_frame and prev_frame.f_globals:
                    funcs = prev_frame.f_globals
                else:
                    raise RuntimeError("Unable to determine callable functions")
            elif isinstance(funcs, dict):
                pass
            elif isinstance(funcs, types.ObjectType):
                funcs_dict = {}
                for attr in dir(funcs):
                    if attr[0] != '_':
                        funcs_dict[attr] = getattr(funcs, attr)
                funcs = funcs_dict

            builtins = {'mcastRpc': self.mcast_rpc,
                        'rpc': self.rpc,
                        'localAddr': self.local_addr,
                        'loadNvParam': self.load_nv_param,
                        'load_nv_param': self.load_nv_param,
                        'saveNvParam': self.save_nv_param,
                        'save_nv_param': self.save_nv_param,
                        'getInfo': self.get_info,
                        }
            for (name, func) in builtins.iteritems():
                if name not in funcs:
                    funcs[name] = func

        # Alternate Encryption Ket Parameters
        self.use_alternate_key = False  # <- TX direction
        self.alternate_crypto = None

        # Apply the initial NV parameters
        self._apply_nv_params(initial_apply=True)

        # Update mesh config with loaded NV params
        self._apply_nv_params_to_mesh(mesh_config)


        self.mesh_handler = mesh.MeshHandler(snap_addr, self.crypto, mesh_config)
        self.register_callback('next_hop_addr', self.mesh_handler.on_next_hop_addr)

        if scheduler is None:
            if use_tornado_ioloop:
                scheduler = ioloop_scheduler.IOLoopScheduler()
            else:
                scheduler = EventScheduler.EventScheduler()
        self.scheduler = scheduler

        self.packet_dispatcher = dispatchers.PktDispatcher(self.crypto,
                                                           self.mesh_handler,
                                                           snap_addr,
                                                           scheduler,
                                                           self.nv_params.params[SnapConstantsHelper.NV_GROUP_FORWARDING_MASK_ID],
                                                           self.nv_params.params[SnapConstantsHelper.NV_GROUP_INTEREST_MASK_ID],
                                                           self.alternate_crypto
                                                           )

        if rpc_handler is None:
            self.spy_upload_mgr = SnappyUploader.SnappyUploadMgr(self.scheduler, self._deferred_rpc)
            self.ota_upgrade_mgr = ota_upgrader.SnapConnectOtaUpgrader(self.scheduler, self._deferred_rpc)
            self.local_dispatcher = dispatchers.LocalDispatcher(self.packet_dispatcher, funcs, scheduler)
            rpc_handler = self.local_dispatcher.on_rpc

            self.local_dispatcher.internal_funcs['tellVmStat'] = self.ota_upgrade_mgr.onTellVmStat
            self.local_dispatcher.internal_funcs['getBankSpace'] = self.ota_upgrade_mgr.onGetBankSpace
            self.local_dispatcher.set_name(self.nv_params.params.get(SnapConstantsHelper.NV_DEVICE_NAME_ID,None))
        else:
            self.spy_upload_mgr = None
            self.ota_upgrade_mgr = None

        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_MESH_BCAST, self.mesh_handler.on_descriptor)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_MESH_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                                                 self.mesh_handler.on_descriptor)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_MESH_PTP, self.mesh_handler.on_descriptor)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_MESH_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                                                 self.mesh_handler.on_descriptor)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_RPC, rpc_handler)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                                                 rpc_handler)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_MULTICAST_RPC, rpc_handler)
        self.packet_dispatcher.register_pkt_type(SnapConstantsHelper.PID_MULTICAST_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                                                 rpc_handler)

        self.packet_dispatcher.register_callback(hooks.HOOK_RPC_SENT, self._on_rpc_sent)

        self.local_listener = None
        self.snaptcp_server = None

        self.poll_needed = [self.mesh_handler]
        self._snapcom_opened_hooks = []
        self._snapcom_closed_hooks = []
        self._serial_opened_hooks = []
        self._serial_closed_hooks = []
        self._rpc_sent_ids_to_deferreds = {}

        if share_serial_connections:
            for serial in _serials.itervalues():
                self.mesh_handler.register_callback('broadcasts', serial.on_broadcast)
                self.packet_dispatcher.register_callback('broadcasts', serial.on_broadcast)
                serial.registerCallback("rx", self.packet_dispatcher.on_decoded_pkt)
                serial.serial_driver.registerCloseCallback(lambda: self.packet_dispatcher.on_intf_closed(serial.intf))
                serial.serial_driver.registerCloseCallback(lambda: self._notify_serial_closed_hooks(serial.serial_driver.getOutputType(), serial.serial_driver.port_num))
                serial.serial_driver.registerCloseCallback(lambda: self.mesh_handler.on_intf_closed(serial.intf))

        #Setup the sniffer
        self.local_sniffer = None #callable(descriptor) used for local packet sniffers
        self.packet_dispatcher.register_callback('broadcasts', lambda data, interface, group=None: self.packet_dispatcher.sniff_packet(interface, data, False))
        self.mesh_handler.register_callback('broadcasts', lambda data, interface, group=None: self.packet_dispatcher.sniff_packet(interface, data, False))
        self.mesh_handler.register_callback('unicasts', lambda data, interface, group=None: self.packet_dispatcher.sniff_packet(interface, data, False))

        # For SNAPpy backwards compatibility; Added here to prevent epydoc from seeing
        self.setHook = self.set_hook
        self.mcastRpc = self.mcast_rpc
        self.mcastSerial = self.mcast_data_mode
        self.localAddr = self.local_addr
        self.loadNvParam = self.load_nv_param
        self.saveNvParam = self.save_nv_param
        self.getInfo = self.get_info
        self.rpcSourceAddr = self.rpc_source_addr
        self.rpcAlternateKeyUsed = self.rpc_alternate_key_used  # This is how you ask if an INCOMING packet used the alternate key
        self.rpcUseAlternateKey = self.rpc_use_alternate_key  # This is how you tell subsequent packets to be SENT using the alternate key

        # Bluetooth
        if BtleInterface:
            # A single instance of this interface is shared between all BTLE Serial connections

            if btle_serial.BtleSerial.btle_interface:
                # If a SNAP Connect instance has already created a BTLE Interface, use that one
                self.btle_serial_interface = btle_serial.BtleSerial.btle_interface
            else:
                # Create a new BTLE interface
                try:
                    self.btle_serial_interface = BtleInterface(self.scheduler,
                                                        binascii.hexlify(self.local_addr()),
                                                        btle_serial.BTLE_SERIAL_SERVICE_UUID,
                                                        btle_serial.BTLE_SERIAL_TX_UUID,
                                                        btle_serial.BTLE_SERIAL_RX_UUID)
                except:
                    logging.exception("Unable to initialize the BTLE Serial interface")
                    self.btle_serial_interface = None

                btle_serial.BtleSerial.btle_interface = self.btle_serial_interface
        else:
            self.btle_serial_interface = None

    def accept_btle_serial(self, auth_info=server_auth, forward_groups=None):
        """Start accepting bluetooth low energy serial connections

        Note:
            This runs your device as a bluetooth low energy peripheral.
            BLE peripherals only support a connection to a single BLE Central at a time.
            This is currently only supported on E20 Linux Devices.

        Args:
            auth_info : The function to call when authenticating remote credentials
                        (defaults to public credentials)
            forward_groups : Multi-cast groups to forward through; defaults to using NV param
        """
        if not self.btle_serial_interface or not self.btle_serial_interface.supports_peripheral():
            raise NotImplementedError("BTLE Peripheral mode is not supported on this device")

        self.btle_serial_interface.start_advertising()

        btle_id = 'peripheral%s'%binascii.hexlify(self.local_addr())

        def on_authentication():
            self.open_serial(SERIAL_TYPE_BTLE, btle_id, forward_groups=forward_groups, reconnect=True)

        self.btle_serial_interface.set_authentication_callback(btle_id, on_authentication)
        self.btle_serial_interface.set_server_auth(auth_info)

    def accept_sniffer(self):
        """Start allowing remote sniffer connections over TCP

        A client will connect to this instance by calling 'connect_tcp' and
        providing a 'snifer_callback'. The regular TCP auth_info will be used
        for remote sniffer connections, but the passed in realm will be set to
        snap.REALM_SNIFFER instead of 'snapcom'
        """
        if not self.snaptcp_server:
            raise RuntimeError("You must be accepting TCP connections to call accept_sniffer()")

        self.snaptcp_server.accept_sniffer()

    def accept_tcp(self, auth_info=server_auth, ip='', port=snaptcp.SNAP_IP_PORT, tcp_keepalives=False, forward_groups=None):
        """
        Start listening for and accepting remote IP connections

        Args:
            auth_info : The function to call when authenticating remote credentials
                        (defaults to public credentials)
            ip : The IP address to listen on for remote connections (default all addresses)
            port : The IP port number to listen on for remote connections (default 48625)
            tcp_keepalives : Enable or disable TCP keepalives (default False)
            forward_groups : Multi-cast groups to forward through; defaults to using NV param
        """
        if self.snaptcp_server is None:
            self._create_tcp_server()

        if forward_groups is None:
            forward_groups = self #Use the SNAP Connect copy of forward_groups

        self.snaptcp_server.start_listening(auth_info, ip=ip, port=port, tcp_keepalives=tcp_keepalives, forward_groups=forward_groups)

        if not callable(auth_info):
            raise ValueError("auth_info is not callable")

        #if multicast:
            #self.local_listener = listeners.LocalListener()
            #self.local_listener.registerCallback("rx", self.packet_dispatcher.on_decoded_pkt)
            #self.packet_dispatcher.register_callback('broadcasts', self.local_listener.broadcast)
            #def auto_connect(ip):
                #"""Automatically connect to the specified IP address"""
                #if not self.snaptcp_server.has_connection(ip):
                    #self.connect_tcp(ip, auth_info)
            #self.local_listener.register_callback('connect', auto_connect)

    def add_rpc_func(self, rpc_func_name, rpc_func, allow_alternate_key=False):
        """
        Add an existing function to be called remotely via RPC

        Args:
            rpc_func_name : The public RPC function name
            rpc_func : The callable to be executed via RPC
            allow_alternate_key : if regular decryption fails, try the packet again with the alternate key
        """
        result = self.local_dispatcher.add_func(rpc_func_name, rpc_func)
        if result and allow_alternate_key:
            self.packet_dispatcher.add_whitelisted_func(rpc_func_name)
        return result

    @staticmethod
    def allow_serial_sharing(enabled=True):
        """Enable or disable sharing of open serial connections between SNAP Connect instances

        Calling this function impacts all SNAP Connect instances running in the current interpreter.

        Serial connections are only shared when a SNAP Connect instance is constructed, and only open
        connections are shared.

        Args:
            enabled: Enable sharing of open serial connections when a SNAP Connect instance is created
        """
        global share_serial_connections
        share_serial_connections = enabled

    def apply_feature_bits(self):
        """Apply any Feature Bit settings to the running code"""
        # 0x0100 bit - Enable RPC_CRCs
        RpcCodec.appendCrc = False
        RpcCodec.validateCrc = False
        SnapConstants.MAX_PKT_SIZE = self._original_max_pkt_size
        SnapConstantsHelper.MAX_PKT_SIZE = self._original_max_pkt_size

        featureBits = self.load_nv_param(SnapConstantsHelper.NV_FEATURE_BITS_ID)
        if featureBits and featureBits & 0x0100:
            RpcCodec.appendCrc = True
            RpcCodec.validateCrc = True

        # 0x0400 bit - Enable RF packet CRC
        # This extra CRC is stripped before it reaches SNAP Connect, but we need to take the extra 2 bytes
        # of data it will use into account when generating packets. The extra CRC is appended by the radio
        # before transmit, and is NOT added to serial or TCP communications.
        if featureBits and featureBits & 0x0400:
            #Right now, only SnapConstantsHelper is being used in the codecs to determine length
            #However, changing both protects us in the case where someone is using the other constant value
            SnapConstants.MAX_PKT_SIZE = self._original_max_pkt_size - 2 #We lose 2 bytes to the extra CRC
            SnapConstantsHelper.MAX_PKT_SIZE = self._original_max_pkt_size - 2

    def _apply_nv_params(self, initial_apply=False):
        """Apply current NV params to this SNAP instance

        Args:
            initial_apply: True when this is the first application of the NV parameters
        """
        # Not currently used in SNAP Connect
        #NV_MAC_ADDR_ID
        #NV_NETWORK_ID
        #NV_CHANNEL_ID

        forward_groups = self.load_nv_param(SnapConstantsHelper.NV_GROUP_FORWARDING_MASK_ID)

        if not initial_apply:
            #During the initial apply, this object has not yet been created
            self.packet_dispatcher.set_mcast_groups(forward_groups,
                                                self.load_nv_param(SnapConstantsHelper.NV_GROUP_INTEREST_MASK_ID))

        self.forward_groups = forward_groups  #Used when forward_groups=None is used for serial or TCP connections

        # Not currently used in SNAP Connect
        #NV_MFG_DATE_ID
        if not initial_apply and hasattr(self, 'local_dispatcher'):
            # During the initial apply, this object has not yet been created
            self.local_dispatcher.set_name(self.load_nv_param(SnapConstantsHelper.NV_DEVICE_NAME_ID))

        # Not currently used in SNAP Connect
        #NV_SYSTEM_ERROR_ID
        #NV_DEVICE_TYPE_ID

        # See apply_feature_bits
        #NV_FEATURE_BITS_ID

        # Not currently used in SNAP Connect
        #NV_DEFAULT_UART_ID
        #NV_UART_DM_TIMEOUT_ID
        #NV_UART_DM_THRESHOLD_ID
        #NV_UART_DM_INTERCHAR_ID
        #NV_CARRIER_SENSE_ID
        #NV_COLLISION_DETECT_ID
        #NV_COLLISION_AVOIDANCE_ID

        PacketSerialProtocol.PacketSerialProtocol.maxRetries = self.load_nv_param(SnapConstantsHelper.NV_SNAP_MAX_RETRIES_ID)

        if not initial_apply:
            mesh_config = mesh.MeshConfig()
            self._apply_nv_params_to_mesh(mesh_config)
            self.mesh_handler.set_config(mesh_config)

        # Not currently used in SNAP Connect
        #NV_MESH_PENALTY_LQ_THRESHOLD_ID
        #NV_MESH_REJECTION_LQ_THRESHOLD_ID
        #NV_SNAP_LQ_THRESHOLD_ID
        #NV_SNAPPY_CRC_ID
        #NV_SYS_PLATFORM_ID

        #Encryption/Decryption/Security Related
        crypto_type = self.load_nv_param(SnapConstantsHelper.NV_AES128_ENABLE_ID)
        crypto_key = self.load_nv_param(SnapConstantsHelper.NV_AES128_KEY_ID)
        alternate_crypto_key = self.load_nv_param(SnapConstantsHelper.NV_ALTERNATE_KEY_ID)
        if crypto_key:
            if crypto_type == SnapConstantsHelper.ENCRYPTION_TYPE_AES128:
                if aes is None:
                    if not initial_apply:
                        raise ImportError("Unable to find PyCrypto library required for AES support")
                    else:
                        log.error("Unable to find PyCrypto library required for AES support. Defaulting to no encryption.")
                        self.crypto = Snap2Decoder.crypto = None
                        self.alternate_crypto = None
                else:
                    #Crypto is set to AES and the PyCrypto library is available
                    self.crypto = Snap2Decoder.crypto = aes.CipherStealingAES128(crypto_key)
                    if alternate_crypto_key:
                        self.alternate_crypto = aes.CipherStealingAES128(alternate_crypto_key)
            elif crypto_type == SnapConstantsHelper.ENCRYPTION_TYPE_BASIC:
                self.crypto = Snap2Decoder.crypto = pme.PoorManEncryption(crypto_key)
                if alternate_crypto_key:
                    self.alternate_crypto = pme.PoorManEncryption(alternate_crypto_key)
            elif crypto_type == SnapConstantsHelper.ENCRYPTION_TYPE_NONE:
                self.crypto = Snap2Decoder.crypto = None
                self.alternate_crypto = None
            else:
                if not initial_apply:
                    raise ValueError("Unknown encryption type specified: %s" % (self.load_nv_param(SnapConstantsHelper.NV_AES128_ENABLE_ID)))
                else:
                    log.error("Unknown encryption type specified: %s. Defaulting to no encryption." % (self.load_nv_param(SnapConstantsHelper.NV_AES128_ENABLE_ID)))
                    self.crypto = Snap2Decoder.crypto = None
                    self.alternate_crypto = None
        else:
            self.crypto = Snap2Decoder.crypto = None
            self.alternate_crypto = None

        if not initial_apply:
            #On the initial apply of NV params we still haven't constructed these objects
            self.mesh_handler.set_crypto(self.crypto)
            self.packet_dispatcher.set_crypto(self.crypto)
            self.packet_dispatcher.set_alternate_crypto(self.alternate_crypto)

        #NV_LOCKDOWN_FLAGS_ID - is checked on-the-fly to allow/disallow REMOTE NV Parameter changes

        # Not currently used in SNAP Connect
        #NV_MAX_LOYALTY_ID
        #NV_LAST_VERSION_BOOTED_ID
        #NV_REBOOTS_REMAINING_ID
        #NV_ALT_RADIO_TRIM_ID
        #NV_VENDOR_SETTINGS_ID

    def _apply_nv_params_to_mesh(self, mesh_config):
        """Apply the NV parameters to a mesh configuration instance

        Args:
            mesh_config: The MeshConfig instance to have set with NV param values
        """
        assert isinstance(mesh_config, mesh.MeshConfig)
        mesh_config.max_delete_timeout = self.load_nv_param(SnapConstantsHelper.NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID)
        mesh_config.min_timeout = self.load_nv_param(SnapConstantsHelper.NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID)
        mesh_config.new_timeout = self.load_nv_param(SnapConstantsHelper.NV_MESH_ROUTE_NEW_TIMEOUT_ID)
        mesh_config.used_timeout = self.load_nv_param(SnapConstantsHelper.NV_MESH_ROUTE_USED_TIMEOUT_ID)
        mesh_config.delete_timeout = self.load_nv_param(SnapConstantsHelper.NV_MESH_ROUTE_DELETE_TIMEOUT_ID)
        mesh_config.rreq_retries = self.load_nv_param(SnapConstantsHelper.NV_MESH_RREQ_TRIES_ID)
        mesh_config.rreq_timeout = self.load_nv_param(SnapConstantsHelper.NV_MESH_RREQ_WAIT_TIME_ID)
        mesh_config.initial_hop_limit = self.load_nv_param(SnapConstantsHelper.NV_MESH_INITIAL_HOPLIMIT_ID)
        mesh_config.max_hop_limit = self.load_nv_param(SnapConstantsHelper.NV_MESH_MAX_HOPLIMIT_ID)
        mesh_config.override = self.load_nv_param(SnapConstantsHelper.NV_MESH_OVERRIDE_ID)
        mesh_config.initial_seq = self.load_nv_param(SnapConstantsHelper.NV_MESH_SEQUENCE_NUMBER_ID)

    def cancel_upgrade(self, addr):
        """Cancel an over the air firmware upgrade

        Returns:
            True if the upgrade has been canceled
            False if the upgrade has already completed
        """
        if addr in _upgrades:
            _upgrades[addr].cancel()
            return True
        else:
            return False

    @staticmethod
    def close_all_serial():
        """Closes all open serial connections"""
        for serial in _serials.itervalues():
            serial.close()
        _serials.clear()

    def close_serial(self, serial_type, port):
        """
        Close the specified serial port if open.

        Args:
            serial_type : The type of serial interface to close
            port : The port number to close
        """
        serial = _serials.pop("%s:%s" % (serial_type, port), None)
        if serial:
            serial.close()
            self.mesh_handler.unregister_callback('broadcasts', serial.on_broadcast)
            self.packet_dispatcher.unregister_callback('broadcasts', serial.on_broadcast)
            serial.unregister_callback("rx", self.packet_dispatcher.on_decoded_pkt)
            self.poll_needed.remove(serial)
            del serial

    def connect_btle_serial(self, addr, auth_info=client_auth, forward_groups=None):
        """Connect to a bluetooth low energy serial connection

        Note:
            This runs your device as a bluetooth low energy central.
            BLE centrals may connect to multiple peripherals at a time.
            This is currently only supported on iOS 7 devices with additional
            iOS Objective C SNAP Connect libraries.

        Args:
            addr : The address of the BTLE serial device to connect to. See HOOK_BTLE_DISCOVERED.
            auth_info : A local function to call to retrieve the authorization information to use
                        (defaults to public credentials)
            forward_groups : Multi-cast groups to forward through; defaults to using NV param
        """
        if not self.btle_serial_interface or not self.btle_serial_interface.supports_central():
            raise NotImplementedError("BTLE Central mode is not supported on this device")

        self.btle_serial_interface.connect(binascii.hexlify(addr))

        def on_authentication():
            self.open_serial(SERIAL_TYPE_BTLE, binascii.hexlify(addr), forward_groups=forward_groups, reconnect=True)

        self.btle_serial_interface.set_authentication_callback(binascii.hexlify(addr), on_authentication)
        self.btle_serial_interface.set_client_auth(auth_info)

    def connect_tcp(self, host, auth_info=client_auth, port=None, retry_timeout=60, secure=False, forward_groups=None, tcp_keepalives=False, cache_dns_lookup=False, sniffer_callback=None):
        """
        Connect to another SNAP node over IP

        Args:
            host : The IP address or hostname of the other SNAP node
            auth_info : A local function to call to retrieve the authorization information to use
                        (defaults to public credentials)
            port : The IP port number to connect to (default 48625 or 443 if secure=True)
            retry_timeout : A timeout in seconds to wait before retrying to connect (default 60)
            secure : A boolean indicating if all traffic should be encrypted using SSL (default False)
            forward_groups : Multi-cast groups to forward through; defaults to using NV param
            tcp_keepalives : Enable TCP level keep alives (default False)
            cache_dns_lookup : Only preform a DNS lookup once for the host being connected to (default False)
            sniffer_callback : Callable that accepts a descriptor. Used for making remote sniffer connections.
        """
        if self.snaptcp_server is None:
            self._create_tcp_server()

        if port is None:
            port = snaptcp.SNAP_IP_PORT
            if secure:
                port = 443

        if use_tornado_ioloop:
            tcpConnectionType = snaptcp_tornado.TornadoSNAPtcpConnection
        else:
            tcpConnectionType = snaptcp_asyncore.AsyncoreSNAPtcpConnection

        if forward_groups is None:
            forward_groups = self #Use the SNAP Connect copy of forward_groups

        conn = tcpConnectionType(local_addr=self.local_addr(), server=self.snaptcp_server, conn=(host, port), retry_timeout=retry_timeout,
                                 auth_info=auth_info, secure=secure, forward_groups=forward_groups, tcp_keepalives=tcp_keepalives,
                                 cache_dns_lookup=cache_dns_lookup, sniffer_callback=sniffer_callback)

        try:
            local_info = "%s:%i" % (conn.socket.getsockname()[0], conn.socket.getsockname()[1])
        except socket.error:
            local_info = '0.0.0.0'
        log.debug("New SNAPtcpConnection %s to %s:%i" % (local_info, host, port))
        #conn.registerCallback("rx", self.packet_dispatcher.on_decoded_pkt) # This gets called when the connection is actually opened
        conn.decode = False
        conn.log_raw = False

    def _create_tcp_server(self):
        """Create the main SNAP TCP server instance"""
        assert self.snaptcp_server is None
        if use_tornado_ioloop:
            io_loop = None
            if isinstance(self.scheduler, ioloop_scheduler.IOLoopScheduler):
                io_loop = self.scheduler.io_loop
            self.snaptcp_server = snaptcp_tornado.TornadoSnapTcpServer(self.scheduler, self.local_addr(), io_loop)
        else:
            self.snaptcp_server = snaptcp_asyncore.AsyncoreSnapTcpServer(self.scheduler, self.local_addr())

        self.register_callback('next_hop_addr', self.snaptcp_server.on_next_hop_addr)
        self.mesh_handler.register_callback('broadcasts', self.snaptcp_server.broadcast)
        self.packet_dispatcher.register_callback('broadcasts', self.snaptcp_server.broadcast)
        self.snaptcp_server.registerCallback('connection_opened', self.packet_dispatcher.on_packet_serial_opened)
        self.snaptcp_server.registerCallback('connection_opened', self._notify_snapcom_opened_hooks)
        #self.snaptcp_server.registerCallback('connection_opened', self._find_next_hop) # We send over the SNAP address in authentication now
        self.snaptcp_server.registerCallback('connection_closed', self._notify_snapcom_closed_hooks)
        self.snaptcp_server.registerCallback('connection_closed', lambda handler: self.packet_dispatcher.on_intf_closed(handler.interface))
        self.snaptcp_server.registerCallback('connection_closed', lambda handler: self.mesh_handler.on_intf_closed(handler.interface))
        self.snaptcp_server.registerCallback('connection_closed', lambda handler: self.mesh_handler.remove_next_hop(handler.remote_addr, handler.interface))

    def _create_mcast_descriptor(self, group, ttl, func_name, *args):
        """Helper function to create a multi-cast RPC descriptor"""
        descriptor = RpcCodec.create_mcast_descriptor(group, ttl, func_name, *args)
        descriptor.srcAddr = self.packet_dispatcher.local_addr
        if self.use_alternate_key:
            descriptor.crypto = self.alternate_crypto
        else:
            descriptor.crypto = self.crypto
        descriptor.args = args
        return descriptor

    def data_mode(self, dst_addr, data):
        """
        Sends a transparent (aka data) mode packet to the specified network address

        Args:
            dst_addr : The Network Address of the remote node
            data : The data to send
        """
        assert len(dst_addr) == SnapConstantsHelper.SNAP2_ADDR_LEN
        descriptor = DataModeCodec.DataModeDescriptor()
        descriptor.final_dst_addr = descriptor.dstAddr = dst_addr
        descriptor.srcAddr = self.local_addr()
        descriptor.data = data
        descriptor.crypto = self.crypto
        return self.packet_dispatcher.send(descriptor)

    def _deferred_rpc(self, *args):
        """
        Send an RPC but returned a deferred instead of ID
        """
        rpc_id = self.rpc(*args)
        deferred = Deferred.Deferred()
        self._rpc_sent_ids_to_deferreds[rpc_id] = deferred
        return deferred

    def _direct_mcast(self, conn, group, ttl, func_name, *args):
        """Sends a multi-cast RPC out the specified connection"""
        return self.packet_dispatcher.send(self._create_mcast_descriptor(group, ttl, func_name, *args), conn)

    def directed_mcast_rpc(self, conn, port, group, ttl, func_name, *args):
        """Sends a multi-cast RPC out of the specified connection

        Args:
            conn : The connection type for serial or IP address for TCP
            port : The port for the connection
            group : specifies which nodes should respond to the request
            ttl : specifies the Time To Live (TTL) for the request
            func_name: The function name to be invoked
            args : Any arguments for the function specified by func_name

        Returns:
            The identifier for the packet sent
            False if the interface did not exist
        """
        send_interface = self._get_interface(conn, port)

        if send_interface:
            return self._direct_mcast(send_interface, group, ttl, func_name, *args)
        else:
            return False

    def disconnect_btle_serial(self, addr):
        """Disconnect a bluetooth low energy serial connection

        Args:
            addr : The address of the BTLE serial device to connect to. See HOOK_BTLE_DISCOVERED.
        """
        if not self.btle_serial_interface or not self.btle_serial_interface.supports_central():
            raise NotImplementedError("BTLE Central mode is not supported on this device")

        self.btle_serial_interface.disconnect(binascii.hexlify(addr))
        self.close_serial(SERIAL_TYPE_BTLE, binascii.hexlify(addr))

    def disconnect_tcp(self, host, port=snaptcp.SNAP_IP_PORT, all=False, retry=False):
        """
        Disconnect from the specified instance.

        Args:
            host : The IP address or hostname of the other instance
            port : The IP port number (default 48625)
            all : Disconnect all connections matching the criteria
            retry : Disconnect, but retry connecting to the same host and port

        Returns:
            True if the host was found.
        """
        if self.snaptcp_server:
            return self.snaptcp_server.disconnect_connection(host, port, all, retry)
        return 0

    def find_next_hop(self, conn):
        """Find the next hop address of the connection

        This is only used by serial connections. TCP Connections send the SNAP address during authentication.
        This creates a random callback and sends a directed multicast to the node at the end of the serial connection.
        When the node responds, we are able to add it to the route table and remove any previous routes based on the
        address.

        After finding the address or timing out, the HOOK_SERIAL_OPEN hook is called.
        """
        if hasattr(self, 'local_dispatcher'):
            rand_func_name = ''.join(random.Random().sample(string.letters + string.digits, 16))
            log.debug("Querying for remote addr")
            self._direct_mcast(conn, 1, 1, 'callback', rand_func_name, 'getInfo', 3)

            def timeout():
                log.debug("Timed out finding next hop for %s" % (conn.intf))
                self.local_dispatcher.funcs.pop(rand_func_name, None)
                self._notify_serial_opened_hooks(conn.serial_driver.getOutputType(), conn.serial_driver.port_num)

            _timeout = PacketSerialProtocol.PacketSerialProtocol.maxRetries*(PacketSerialProtocol.PacketSerialProtocol.sendTimeout.microseconds/1000000.0) # Need to wait at least this long since a user may connect to a device that really isn't there
            timeout_event = self.scheduler.schedule(_timeout, timeout)

            def add_remote(arg):
                """Called when the remote node responds"""
                timeout_event.Stop()
                conn.remote_addr = self.rpc_source_addr()
                del self.local_dispatcher.funcs[rand_func_name]
                log.debug("Next hop for %s is %s" % (conn.intf, binascii.hexlify(conn.remote_addr)))
                self._notify_callbacks('next_hop_addr', conn.remote_addr, conn.intf)
                self._notify_serial_opened_hooks(conn.serial_driver.getOutputType(), conn.serial_driver.port_num, conn.remote_addr)

            self.local_dispatcher.funcs[rand_func_name] = add_remote

    def get_info(self, which_info):
        """
        Get the specified system info

        Args:
            which_info : Specifies the type of information to be retrieved
        """
        info = [SnapConstantsHelper.SYNAPSE,
                SnapConstantsHelper.RADIO_NONE,
                SnapConstantsHelper.CPU_UNKNOWN,
                SnapConstantsHelper.SNAPCOM,
                int(not __debug__),
                VERSION.split('.')[0],
                VERSION.split('.')[1],
                VERSION.split('.')[2],
                1 if aes is not None else 2,
                self.packet_dispatcher.send_seq,
                self.packet_dispatcher.last_pkt_is_multicast,  # is multi-cast
                self.packet_dispatcher.last_pkt_remaining_ttl,  # remaining ttl
                None,  # remaining tiny strings
                None,  # remaining medium strings
                None,  # route table size
                len(self.mesh_handler.route_table)  # routes in route table
                ]
        if isinstance(which_info, int) and (0 <= which_info < len(info)):
            return info[which_info]
        return None

    def _get_interface(self, conn, port):
        """Return the TCP or Serial interface for the given connection and port

        Args:
            conn : The connection type for serial or the IP address for TCP
            port : The port for the connection


        Returns:
            The interface associated with conn/port or None if it could not be found
        """
        interface = None
        if isinstance(conn, int):
            try:
                # If conn is an integer, this is a serial connection
                serial_type = self._get_serial_listener_type(conn)
                key = "%s:%s" % (serial_type, port)
                interface = _serials[key]
            except KeyError:
                return None


        elif isinstance(conn, str):
            # If conn is a string, it represents an IP and this is a TCP connection
            if self.snaptcp_server is None:
                # We need a SNAP TCP Server to get this kind of interface
                return None

            for handler in self.snaptcp_server.active_handlers:
               if ((handler.initiator and handler._conn == (conn,port)) or
                   (not handler.initiator and handler.addr == (conn,port))):
                    interface = handler
                    break
        else:
            # If it's not an int or str, we can't do anything with it
            raise TypeError("Invalid connection parameter provided")

        return interface

    def _get_serial_listener_type(self, serial_type):
        """Convert from the enumerations used by SNAP Connect to those used by the serial listener

        This doesn't force users to use the same enumerations as we do

        Args:
            A SNAP Connect serial port type enumeration

        Returns:
            An enumeration we can pass to a serial_listener
        """
        if not isinstance(serial_type, int):
            raise ValueError("Serial interface type must be an integer")
        elif serial_type == SERIAL_TYPE_RS232:
            return PyserialWrapper.TYPE_PYSERIAL
        elif serial_type == SERIAL_TYPE_SNAPSTICK100:
            return PyserialWrapper.TYPE_USBXSERIAL
        elif serial_type == SERIAL_TYPE_SNAPSTICK200:
            return PyserialWrapper.TYPE_SNAPSTICK
        elif serial_type == SERIAL_TYPE_BTLE:
            return PyserialWrapper.TYPE_BTLE
        else:
            raise ValueError("Unsupported serial interface type %s" % (serial_type))

    def get_tcpip_stats(self, host, port=None):
        if self.snaptcp_server:
            return self.snaptcp_server.get_stats(host, port)
        return None

    def listen_btle_serial(self, callback):
        """Start listening for BTLE Serial Peripherals

        Args:
            callback : Callback in the form of callback(uuid, name, addr, link_quality)
        """
        if not self.btle_serial_interface or not self.btle_serial_interface.supports_central():
            raise NotImplementedError("BTLE Central mode is not supported on this device")

        self.btle_serial_interface.set_discovery_callback(callback)
        self.btle_serial_interface.start_listening()

    def load_nv_param(self, nv_param_id):
        """
        Return the indexed parameter from storage

        Args:
            nv_param_id : Specifies which "key" to retrieve from storage
        """
        if nv_param_id == SnapConstantsHelper.NV_MAC_ADDR_ID:
            return self.licensed_mac
        return self.nv_params.params.get(nv_param_id, None)

    def local_addr(self):
        """Returns the local network address of this instance."""
        return self.mesh_handler.local_addr

    def loop(self):
        """Continually loop calling poll"""
        if use_tornado_ioloop:
            callback = tornado.ioloop.PeriodicCallback(self.poll_internals, 10, io_loop=self.scheduler.io_loop)
            callback.start()
            self.scheduler.io_loop.start()
        else:
            poll_func = self.poll
            while True:
                poll_func()
                if not asyncore.socket_map:
                    time.sleep(SLEEP_TIME)
        self.stop_accepting_tcp()

    def mcast_data_mode(self, group, ttl, data):
        """
        Sends a multicast transparent (aka data) mode packet to the specified network address

        Args:
            group : specifies which nodes should respond to the request
            ttl : specifies the Time To Live (TTL) for the request
            data : The data to send
        """
        if isinstance(group, int):
            if group < 0:
                #Large numbers in SNAPpy appear as negative numbers to SNAP Connect
                #This will handle the case where we're passing one of those numbers directly down
                group += 0x10000

            group = binascii.unhexlify("%04x" % (group))

        assert len(group) == 2
        descriptor = DataModeCodec.DataModeDescriptor()
        descriptor.isMulticast = True
        descriptor.dstAddr = group
        descriptor.srcAddr = self.local_addr()
        descriptor.ttl = ttl
        descriptor.data = data
        descriptor.crypto = self.crypto
        return self.packet_dispatcher.send(descriptor)

    def mcast_rpc(self, group, ttl, func_name, *args):
        """
        Sends a multi-cast RPC

        Call a Remote Procedure (make a Remote Procedure Call, or RPC), using
        multicast messaging.  This means the message could be acted upon by
        multiple nodes.

        Args:
            group : specifies which nodes should respond to the request
            ttl : specifies the Time To Live (TTL) for the request
            func_name: The function name to be invoked
            args : Any arguments for the function specified by func_name
        """
        return self.packet_dispatcher.send(self._create_mcast_descriptor(group, ttl, func_name, *args))

    def _notify_serial_closed_hooks(self, serial_type, port):
        """
        Notify listeners when a serial connection has been closed.

        Args:
            serial_type : The type of serial interface that closed (RS-232=1, USB=2)
            port : The port number that closed
        """
        for hook in self._serial_closed_hooks:
            try:
                hook(serial_type, port)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("An error occurred while notifying a callback that a serial connection was closed")

    def _notify_serial_opened_hooks(self, serial_type, port, addr=None):
        """
        Notify listeners when a serial connection has been opened.

        Args:
            serial_type : The type of serial interface that closed (RS-232=1, USB=2)
            port : The port number that closed
            addr : The address of the node attached to the serial interface
                   If the address could not be determined, the addr will be set to None
        """
        for hook in self._serial_opened_hooks:
            try:
                hook(serial_type, port, addr)
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("An error occurred while notifying a callback that a serial connection was opened")

    def _notify_snapcom_closed_hooks(self, conn):
        """
        Notify listeners when a SNAP tcp connection has been closed.

        Args:
            conn : The SNAP TCP connection object
        """
        assert isinstance(conn, snaptcp.SNAPtcpConnection), "Not a SNAPtcpConnection"

        if getattr(conn.interface,"is_sniffer_server",False):
            try:
                self.packet_dispatcher.remove_sniffer(conn.send, raw_data=True)
            except Exception, exc:
                logging.error("Unable to remove a remote sniffer connection")

        for hook in self._snapcom_closed_hooks:
            if conn.initiator:
                hook(conn._conn, conn.remote_addr)
            else:
                hook(conn.addr, conn.remote_addr)

    def _notify_snapcom_opened_hooks(self, conn):
        """
        Notify listeners when a SNAP TCP connection has been opened.

        Args:
            conn : The SNAP TCP connection object
        """
        assert isinstance(conn, snaptcp.SNAPtcpConnection), "Not a SNAPtcpConnection"


        if getattr(conn.interface,"is_sniffer_server",False):
            self.packet_dispatcher.add_sniffer(conn.send, raw_data=True)
        else:
            # Notify about the next hop address before calling hooks
            # We don't want to know about the next hop address if this is a sniffer interface
            if not getattr(conn.interface,"is_sniffer_client",False):
                self._notify_callbacks('next_hop_addr', conn.remote_addr, conn.intf)

        for hook in self._snapcom_opened_hooks:
            if conn.initiator:
                hook(conn._conn, conn.remote_addr)
            else:
                hook(conn.addr, conn.remote_addr)

    def _on_rpc_sent(self, rpc_id, success):
        deferred = self._rpc_sent_ids_to_deferreds.pop(rpc_id, None)
        if deferred:
            deferred(success)

    def open_serial(self, serial_type, port, reconnect=False, dll_path=MODULE_PATH, forward_groups=None, baudrate=None):
        """
        Open the specified serial port for sending and receiving SNAP packets.

        Args:
            serial_type : The type of serial interface to open
            port : The port number to open
            reconnect : Close the connection and re-open (default False)
            forward_groups : Multi-cast groups to forward through; defaults to using NV param
            baudrate : The baudrate for the serial connection
        """
        # Convert from our enumerations to those used by serial listener
        serial_type = self._get_serial_listener_type(serial_type)

        key = "%s:%s" % (serial_type, port)
        if key not in _serials:
            if forward_groups is None:
                forward_groups = self #Use the SNAP Connect copy of forward_groups

            serial = listeners.SerialListener(serial_type, port, dll_path=dll_path, forward_groups=forward_groups, baudrate=baudrate)

            serial.decode = False
            serial.log_raw = False

            self.mesh_handler.register_callback('broadcasts', serial.on_broadcast)
            self.packet_dispatcher.register_callback('broadcasts', serial.on_broadcast)
            serial.registerCallback("rx", self.packet_dispatcher.on_decoded_pkt)
            serial.serial_driver.registerCloseCallback(lambda: self.packet_dispatcher.on_intf_closed(serial.intf))
            serial.serial_driver.registerCloseCallback(lambda: self.mesh_handler.on_intf_closed(serial.intf))

            self.poll_needed.append(serial)
            _serials[key] = serial

            if serial_type != PyserialWrapper.TYPE_BTLE:
                # BTLE connections don't have anyone to talk to when they're opened, and they use a different open/close hook
                serial.serial_driver.registerCloseCallback(lambda: self._notify_serial_closed_hooks(serial.serial_driver.getOutputType(), serial.serial_driver.port_num))
                self.find_next_hop(serial)

        elif reconnect:
            serial = _serials[key]
            return serial.reconnect()

    def poll(self):
        """Polls all needed components packages"""
        work_done = False
        asyncore.poll(SLEEP_TIME)
        for func in self.poll_needed:
            try:
                work_done = func.poll() or work_done
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("An error occurred while calling a functions poll method")
        work_done = self.scheduler.poll() or work_done
        #if not work_done:
            #time.sleep(SLEEP_TIME)
        return work_done

    def poll_internals(self):
        """Polls only internal components for this instance"""
        for func in self.poll_needed:
            try:
                func.poll()
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                log.exception("An error occurred while calling a functions poll method")

    def replace_rpc_func(self, rpc_func_name, rpc_func):
        """
        Replace an existing function to be called remotely via RPC

        Args:
            rpc_func_name : An existing public RPC function name
            rpc_func : The callable to be executed via RPC

        Returns:
            The original function that was replaced
            None if the function name had not been previously added
        """
        original_rpc_func = self.local_dispatcher.funcs.get(rpc_func_name,None)

        if not original_rpc_func:
            return None

        self.local_dispatcher.funcs[rpc_func_name] = rpc_func
        return original_rpc_func

    def rpc(self, dst_addr, func_name, *args):
        """
        Sends a unicast RPC

        Call a Remote Procedure (make a Remote Procedure Call, or RPC), using unicast messaging.
        A packet will be sent to the node specified by parameter address, asking that remote node
        to execute the function specified by parameter function. The specified function will be
        invoked with the parameters specified by args, if any args are present.

        Args:
            dst_addr : The Network Address of the remote node
            func_name : The function name to be invoked on the remote node
            args : Any arguments for the function specified by func_name
        """
        descriptor = RpcCodec.RpcDescriptor(crypto=self.crypto)
        descriptor.funcName = func_name
        descriptor.final_dst_addr = dst_addr
        descriptor.args = args
        descriptor.srcAddr = self.local_addr()
        descriptor.crypto = self.crypto
        return self.packet_dispatcher.send(descriptor)

    def rpc_source_addr(self):
        """
        Originating address of the current RPC context (None if called outside RPC)

        Returns:
            The Network Address of the remote node which initiated the RPC call.
        """
        return self.packet_dispatcher.remote_addr

    def rpc_source_interface(self):
        """
        Originating interface of the current RPC context (None if called outside RPC)

        Returns:
            The Interface of the remote node which initiated the RPC call.
        """
        return self.packet_dispatcher.incoming_intf

    def rpc_alternate_key_used(self):
        """
        Returns True if the packet came in with the alternate encryption key and the function was on the whitelist
        """
        return self.packet_dispatcher.alternate_key_used

    def rpc_use_alternate_key(self, enable_flag):
        """
        Set enable_flag == True when you DO want to use the alternate key, False to revert to the normal key
        When enabled, this setting persists until you disable it, and it affects ALL multicast RPC calls you send
        """
        self.use_alternate_key = enable_flag

    def save_nv_param(self, nv_param_id, obj):
        """
        Store individual objects for later access by id

        Args:
            nv_param_id : Specifies which "key" to store the obj parameter under. 0-127 are pre-assigned system IDs.
            obj : The object you would like to store for later use.
        """
        if self.rpc_source_interface() is not None and self.load_nv_param(SnapConstantsHelper.NV_LOCKDOWN_FLAGS_ID) & 2:
            return False

        try:
            # See if this object is pickleable
            cPickle.dumps(obj)
        except cPickle.PickleError:
            return False

        prev_val = self.load_nv_param(nv_param_id)

        self.nv_params.params[nv_param_id] = obj
        self.nv_params.__autosave__()  # We are calling this explicitly because dicts don't auto save yet :(

        # Real SNAP Nodes don't apply NV Parameter changes immediately, but there does not seem to be a compelling reason to ENFORCE this restriction in SNAP Connect
        if nv_param_id == SnapConstantsHelper.NV_FEATURE_BITS_ID:
            self.apply_feature_bits()
        elif prev_val != obj:
            self._apply_nv_params()

        return True

    def set_hook(self, hook, callback=None):
        """
        Set the specified SNAP hook to call the provided function

        Args:
            hook : The SNAP event hook identifier
            callback : The function to invoke when the specified event occurs
        """
        if callback is None:
            def add_hook(func):
                """Used for decorator"""
                self.set_hook(hook, func)
                return func
            return add_hook
        elif callable(callback):
            if hook == hooks.HOOK_RPC_SENT:
                self.packet_dispatcher.register_callback(hook, callback)
            elif hook == hooks.HOOK_STDIN:
                self.packet_dispatcher.register_callback(hook, callback)
            elif hook == hooks.HOOK_SNAPCOM_OPENED:
                self._snapcom_opened_hooks.append(callback)
            elif hook == hooks.HOOK_SNAPCOM_CLOSED:
                self._snapcom_closed_hooks.append(callback)
            elif hook == hooks.HOOK_TRACEROUTE:
                self.packet_dispatcher._traceroute_hook = callback
            elif hook == hooks.HOOK_SERIAL_CLOSE:
                self._serial_closed_hooks.append(callback)
            elif hook == hooks.HOOK_SERIAL_OPEN:
                self._serial_opened_hooks.append(callback)
            elif hook == hooks.HOOK_OTA_UPGRADE_COMPLETE:
                self.ota_upgrade_mgr.register_complete_hook(callback)
            elif hook == hooks.HOOK_OTA_UPGRADE_STATUS:
                self.ota_upgrade_mgr.register_status_hook(callback)
            elif hook == hooks.HOOK_BTLE_OPENED:
                if self.btle_serial_interface:
                    self.btle_serial_interface.set_opened_callback(callback)
            elif hook == hooks.HOOK_BTLE_CLOSED:
                if self.btle_serial_interface:
                    self.btle_serial_interface.set_closed_callback(callback)
            else:
                raise TypeError("Unknown hook type")
        else:
            raise TypeError("Invalid callback")

    def start_sniffer(self, callback):
        """Start a local sniffer that passes descriptors to the given callback

        Addr:
            callback: Called when packets are sent or received by this SNAP Connect instance.
                      Callable(descriptor)

        Returns:
            True if a sniffer could be started
            False if one was already running
        """

        if self.local_sniffer:
            return False
        else:
            self.local_sniffer = callback
            return self.packet_dispatcher.add_sniffer(callback, raw_data=False)

    def stop_accepting_btle_serial(self):
        """Stop accepting bluetooth low energy serial connections. Closes existing connections."""
        if not self.btle_serial_interface or not self.btle_serial_interface.supports_peripheral():
            raise NotImplementedError("BTLE Peripheral mode is not supported on this device")

        self.btle_serial_interface.stop_advertising()
        self.close_serial(SERIAL_TYPE_BTLE, 'peripheral%s'%binascii.hexlify(self.local_addr()))

    def stop_accepting_sniffer(self, close_existing=False):
        """Stop accepting remote sniffer connections over TCP

        Args:
            close_existing: If True, close all existing sniffer connections
        """
        if self.snaptcp_server:
            self.snaptcp_server.stop_accepting_sniffer(close_existing)

    def stop_accepting_tcp(self, close_existing=False):
        """Stop listening for and accepting remote IP connections

        Args:
            close_existing: If True, close all existing TCP connections
        """
        if self.snaptcp_server:
            self.snaptcp_server.stop_listening()
            if close_existing:
                self.snaptcp_server.close_all()

    def stop_listening_btle_serial(self):
        """Stop listening for BTLE Serial Peripherals"""
        if not self.btle_serial_interface or not self.btle_serial_interface.supports_central():
            raise NotImplementedError("BTLE Central mode is not supported on this device")

        self.btle_serial_interface.stop_listening()

    def stop_sniffer(self):
        """Stop a running local sniffer

        Returns:
            True if the sniffer could be stopped
            False if one was not already running
        """
        if not self.local_sniffer:
            return False
        else:
            callback = self.local_sniffer
            self.local_sniffer = None
            return self.packet_dispatcher.remove_sniffer(callback, raw_data=False)

    def traceroute(self, dst_addr):
        """
        Sends a traceroute request

        Args:
            dst_addr : The Network Address of the remote node
        """
        descriptor = TraceRouteCodec.TraceRouteDescriptor(crypto=self.crypto)
        descriptor.final_dst_addr = descriptor.dstAddr = dst_addr
        descriptor.srcAddr = descriptor.orig_src_addr = self.local_addr()
        descriptor.msgId = SnapConstantsHelper.TR_REQ
        descriptor.crypto = self.crypto
        self.packet_dispatcher.send(descriptor)

    def upgrade_firmware(self, addr, file):
        """Start an over the air firmware upgrade.

        Programming may take up to 2 minutes after the transfer completes.
        Do not remove power to the device during this time.

        Args:
            addr: The address of the node to upgrade
            file: The path to the firmware image

        Hooks:
            HOOK_OTA_UPGRADE_COMPLETE (addr, status, message)
            HOOK_OTA_UPGRADE_STATUS (addr, percent_complete)

        Returns:
            Returns True if an upgrade can be started or False if the upgrade
            can't be started for some reason. Will result in a call to the
            OTA upgrade complete hook.
        """
        if not self.ota_upgrade_mgr:
            raise Exception("OTA Firmware Upgrade manager could not be created")

        #Check if we have an upgrade already in progress
        if addr in _upgrades:
            return False

        #Start the firmware upgrade process
        upgrade = self.ota_upgrade_mgr.upgrade_firmware(addr, file)

        if upgrade:
            _upgrades[addr] = upgrade
            return True
        else:
            return False


SNAPcom = Snap

def main(setup_logging=True):
    """Initialize all needed SNAP components and start running"""
    if setup_logging:
        logging.basicConfig(level=logging.DEBUG, format='%(asctime)s:%(msecs)03d %(levelname)-8s %(name)-8s %(message)s', datefmt='%H:%M:%S')

    instance = SNAPcom()
    instance.loop()