# (c) Copyright 2009-2010, Synapse Wireless, Inc.
"""SNAP protocol over TCP"""

import asyncore
import socket
import logging
import os
import random
import string
import hashlib
import urllib2
import binascii
try:
    # pylint: disable-msg=F0401
    import ssl
except ImportError:
    ssl = None
import pprint
import sys
import re

import auth_digest

from snaplib import interfaces
from snaplib import EventCallbacks
from snaplib import PacketSerialProtocol


_auth_key = ''.join(random.Random().sample(string.letters + string.digits, 16))

SNAP_ADDR_HEADER_RE = re.compile(r'(Server|User-Agent): [^\(]+\(([^\)]+)\).+')

SNAP_IP_PORT = 48625  # Not currently assigned by IANA
SNAP_IP_ADDR = '224.0.0.101'  # In LAN usage range and not assigned

# http://tldp.org/HOWTO/html_single/TCP-Keepalive-HOWTO/
TCP_KEEPIDLE = 59  # seconds
TCP_KEEPINTVL = 15  # seconds, until reprobe
TCP_KEEPCNT = 4  # Number of times before a connection is considered broken

# Realms for authentication
REALM_SNAP = "SNAPcom"
SNAP_USER_AGENT = "SNAP Connect"

REALM_SNIFFER = "SNAPsniffer"
SNIFFER_USER_AGENT = "SNAP Sniffer"

# Based off of code from: http://twistedmatrix.com/trac/ticket/2425
def format_error(errorno):
    try:
        from ctypes import WinError
    except ImportError:
        WinError = None
    else:
        return WinError(errorno)[1]

    try:
        # Disable unable to import because this wont work in Linux buts ok we are catching it
        # pylint: disable-msg=F0401
        from win32api import FormatMessage
    except ImportError:
        FormatMessage = None
    else:
        return FormatMessage(errorno)

    try:
        from socket import errorTab
    except ImportError:
        errorTab = None
    else:
        return errorTab.get(errorno, str(errorno))

    return os.strerror(errorno)


class ConnectRequest(urllib2.Request):
    """A urllib2.Request for the HTTP CONNECT method"""
    def get_method(self):
        return "CONNECT"

    def get_selector(self):
        return "/"

# SSL asyncore code Copyright (C) 2007-2010 Giampaolo Rodola' <g.rodola@gmail.com>
class SNAPtcpConnection(PacketSerialProtocol.PacketSerialProtocol):
    """Handles sending and receiving SNAP packets over a TCP connection"""
    CONNECT_STATEMENT = "CONNECT localhost:48625 HTTP/1."
    CONNECTION_ESTABLISHED = "HTTP/1.0 200 Connection established\r\n"
    FORBIDDEN = "403 Forbidden\r\n"
    UNAUTHORIZED = "401 Unauthorized\r\n"
    AUTHENTICATE = "WWW-Authenticate: Digest"
    log = logging.getLogger('snap.SNAPtcpConnection')

    def __init__(self,
                 local_addr,
                 server=None,
                 conn=None,
                 retry_timeout=0,
                 realm=REALM_SNAP,
                 auth_info=None,
                 secure=False,
                 tcp_keepalives=False,
                 forward_groups=0x01,
                 cache_dns_lookup=False,
                 sniffer_callback=None):
        """
        Initializes an instance of a SNAPtcpConnection object

        Args:
            local_addr : The SNAP address for this interface to use
            server : The SNAPtcpServer object that instantiated this instance
            conn : An existing socket or connection pair [defaults to None]
            retry_timeout : How often to retry
            realm : The HTTP digest authentication realm
            auth_info : The authentication function to call
            secure : Use HTTPS
            tcp_keepalives : Enable TCP level keepalives
            forward_groups : The SNAP multicast forwarding groups to use for this interface
            cache_dns_lookup : Should we do a DNS lookup on every connection attempt
            sniffer_callback : The callback if this is to be used as the remote sniffer client
        """
        PacketSerialProtocol.PacketSerialProtocol.__init__(self)
        self.neverAck = True
        self.decode = False
        self.verboseLogging = False
        self.tcp_keepalives = tcp_keepalives

        self._waiting_for_header = True
        self._waiting_for_auth = False
        self._waiting_for_200 = False
        self._write_buffer = ""
        self._close_after_send = False
        self._server = server
        self.initiator = False
        self._conn = None
        self.retry_timeout = self.orig_retry_timeout = retry_timeout
        self.closed = False
        self.authenticated = False
        self.has_connected = False  # Have we ever connected (including SSL and authentication)?
        self.connecting = False
        self.disconnecting = False
        self.remote_addr = None  # Remote SNAP address
        self.forward_groups = forward_groups
        self.realm = realm
        self.auth_info = auth_info
        self.local_addr = local_addr  # Local SNAP address
        self._cached_lookup = None
        self.cache_dns_lookup = cache_dns_lookup

        self.check_connection_event = None #Holds the timeout for checking that a connection is up

        # Stats; Does NOT include SSL overhead
        self.total_bytes_sent = 0
        self.total_bytes_recv = 0

        self._ssl_accepting = False
        self._ssl_established = False
        self._ssl_closing = False
        self._ssl_connecting = False
        if ssl is None and secure:
            raise ValueError("SSL support was not found")
        else:
            self.secure = secure

        if isinstance(conn, socket.SocketType):
            self.intf = self.interface = interfaces.TcpIpInterface(*conn.getpeername())
        elif len(conn) == 2:
            self.intf = self.interface = interfaces.TcpIpInterface(*conn)

        self.interface.is_sniffer_server = False

        if sniffer_callback is not None:
            self.check_len = False #Because we may send extra meta data in packets, we don't want to restrict ourselves to the max length
            self.interface.is_sniffer_client = True #Set on the remote server client side
            self.interface.sniffer_callback = sniffer_callback
        else:
           self.interface.is_sniffer_client = False

    def disconnect(self, retry=False):
        """
        Disconnect the TCP connection and optionally retry

        Args:
            retry : After disconnecting retry to connect
        """
        if retry:
            self.disconnecting = False
        else:
            self.disconnecting = True

        self.handle_close()
        self.disconnecting = False

    def _do_ssl_handshake(self):
        """The TCP connection has been made now try to start the SSL handshake"""
        try:
            self.socket.do_handshake()
        except ssl.SSLError, err:
            if err.args[0] in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE):
                return
            elif err.args[0] == ssl.SSL_ERROR_EOF:
                return self.handle_close()
            elif err.args[0] == ssl.SSL_ERROR_SSL:
                self.handle_failed_ssl_handshake()
            else:
                raise
        else:
            self._ssl_accepting = False
            self._ssl_established = True

    def _do_ssl_shutdown(self):
        """Called when something went wrong with the SSL connection so shutdown the socket"""
        self._ssl_closing = True
        try:
            self.socket = self.socket.unwrap()
        except ssl.SSLError, err:
            if err.args[0] in (ssl.SSL_ERROR_WANT_READ,
                               ssl.SSL_ERROR_WANT_WRITE):
                return
            elif err.args[0] == ssl.SSL_ERROR_SSL:
                pass
            else:
                raise
        except socket.error, err:
            # Any "socket error" corresponds to a SSL_ERROR_SYSCALL
            # return from OpenSSL's SSL_shutdown(), corresponding to
            # a closed socket condition. See also:
            # http://www.mail-archive.com/openssl-users@openssl.org/msg60710.html
            pass
        self._ssl_closing = False
        super(SSLConnection, self).close()

    def flush_write_buff(self):
        """Flush this connections write buffer"""
        self._write_buffer = ""

    def buffered_write(self,data):
        """Add data to be written to the write buffer"""
        self._write_buffer += data

    def get_intf_name(self):
        """Returns a readable name for this interface"""
        return "%s:%s" % (self.interface.ip_addr, self.interface.port)

    def handle_close(self):
        """Handle closing this SNAP TCP connection"""
        self.close()
        self.log.debug('Connection closed to %s' % (self.interface))

        if self.check_connection_event:
            # If we're closing before the check connection event, cancel it
            # This prevents duplicate close events from triggering
            self.check_connection_event.Stop()
            self.check_connection_event = None

        # Clear states
        self.closed = True
        self.authenticated = False
        self._ssl_accepting = False
        self._ssl_established = False
        self._ssl_closing = False
        self._ssl_connecting = False
        self.connecting = False

        # Tell the SNAPtcpServer you have disconnected, he may let others know
        if self._server:
            self._server.remove_handler(self)

        if not self.disconnecting and not self.connecting and self._server:
            if self.orig_retry_timeout > 0:
                if self.retry_timeout == 0.1:
                    self.retry_timeout = self.orig_retry_timeout  # reset timeout
                elif self.has_connected:
                    self.retry_timeout = 0.1  # Try to reconnect right away first
                self._server.on_conn_err(self)

    def handle_failed_ssl_handshake(self):
        """Handles a failed SSL handshake attempt"""
        # TLS/SSL handshake failure, probably client's fault which
        # used a SSL version different from server's.
        # We can't rely on the control connection anymore so we just
        # disconnect the client without sending any response.
        ssl_version = ssl.get_protocol_name(self.socket.ssl_version)
        self.log.info("%s handshake failed." % (ssl_version,))
        self.handle_close()

    def handle_read_event(self):
        """Handles reading from the socket when in different states"""
        if self._ssl_accepting:
            self._do_ssl_handshake()
        elif self._ssl_closing:
            self._do_ssl_shutdown()
        else:
            super(SNAPtcpConnection, self).handle_read_event()

    def handle_write_event(self):
        """Handles writing to the socket when in different states"""
        if self._ssl_connecting:
            self.secure_connection(None, ssl.PROTOCOL_TLSv1)
            self._ssl_connecting = False
            # Technically we could fall into _do_ssl_handshake right after secure_connection
        elif self._ssl_accepting:
            self._do_ssl_handshake()
        elif self._ssl_closing:
            self._do_ssl_shutdown()
        else:
            super(SNAPtcpConnection, self).handle_write_event()

    def on_broadcast(self, data, group=0x01):
        """
        Called when a broadcast should be sent out this interface

        Args:
            data : The data to send
            intf : The source interface of this data
            group : The multicast group this data is destined for
        """

        if group & self.forward_groups:
            self.send(data)
        else:
            self.log.debug("Broadcast filtered to group %02X" % (group))

    def on_tcp_data(self, data):
        """
        Called when data is received on the TCP socket
        This method checks what state we are currently in and preforms the necessary actions
        such as authorizing and SSL setup before allowing arbitrary data to be sent
        """
        if data:
            #self.log.info("handle_read: %s" % (pprint.pformat(data),))
            self.total_bytes_recv += len(data)
            #print data
            #print self._waiting_for_header, self._waiting_for_auth, self._waiting_for_200, self.authenticated
            if self._waiting_for_header and self.CONNECT_STATEMENT in data:

                if SNIFFER_USER_AGENT in data:
                    self.log.info("Received connect statement from a SNAP Sniffer")

                    if self._server.accepting_sniffers:
                        # The sniffer uses a different realm from a regular connection
                        realm = REALM_SNIFFER
                        self.interface.is_sniffer_server = True
                    else:
                        self.buffered_write("%s\r\n" % (self.UNAUTHORIZED))
                        self._close_after_send = True
                        self.log.debug('Sniffer auth unsuccessful, server not accepting sniffer connection')
                        return
                else:
                    realm = self.realm

                self.buffered_write("HTTP/1.0 " + self.UNAUTHORIZED + "Server: SNAP Connect (%s)\r\nWWW-Authenticate: %s\r\n\r\n" % (binascii.hexlify(self.local_addr),
                                                                                                                                       auth_digest.www_authenticate(key=_auth_key,
                                                                                                                                                               realm=realm,
                                                                                                                                                               qop=auth_digest.qop_auth),
                                                                                                                                      ))
                self._waiting_for_auth = True
                self._waiting_for_header = False
                self.authenticated = False
                #print self.socket.getpeername()
                addr_check = SNAP_ADDR_HEADER_RE.search(data)
                if addr_check:
                    self._set_remote_addr(binascii.unhexlify(addr_check.group(2)))
                self.log.debug('Sent auth request')
            elif self._waiting_for_auth:
                loc = data.find("Authorization: ")
                #self.log.debug("In waiting for auth: %i" % (loc))
                if loc > -1 and callable(self.auth_info):
                    auth = auth_digest.HttpDigestAuthorization(data[loc + 15:], "CONNECT")
                    password = self.auth_info(auth.realm, auth.username)
                    #print password
                    if password is not None:
                        A1 = "%s:%s:%s" % (auth.username, auth.realm, password)
                        ha1 = hashlib.md5(A1).hexdigest()
                        if auth.response == auth.request_digest(ha1):
                            self.authenticated = True
                            self._waiting_for_auth = False
                            self.buffered_write("%s\r\n" % (self.CONNECTION_ESTABLISHED))
                            self.log.debug('auth successful')
                            if self._server:
                                self._server.add_handler(self)
                            return

                self.buffered_write("%s\r\n" % (self.UNAUTHORIZED))
                self._close_after_send = True
                self.log.debug('auth unsuccessful')
            elif self._waiting_for_header:
                self.buffered_write("%s\r\n" % (self.FORBIDDEN))
                self._close_after_send = True
                self.log.debug("Returning FORBIDDEN")
            elif self._waiting_for_200 and self.AUTHENTICATE in data:
                loc = data.find("Digest")

                options = urllib2.parse_keqv_list(urllib2.parse_http_list(data[loc + 7:]))
                if (options.has_key("nonce") and
                    options.has_key("qop") and
                    options.get('qop', "") == "auth" and
                    callable(self.auth_info)):

                    class Passwd(urllib2.HTTPPasswordMgr):
                        def find_user_password(self, realm, authuri):
                            return self.auth_info(realm)
                    pass_mgr = Passwd()
                    pass_mgr.auth_info = self.auth_info

                    addr_check = SNAP_ADDR_HEADER_RE.search(data)
                    if addr_check:
                        self._set_remote_addr(binascii.unhexlify(addr_check.group(2)))

                    digest_handler = urllib2.AbstractDigestAuthHandler(pass_mgr)
                    auth = digest_handler.get_authorization(ConnectRequest(""), options)
                    self.buffered_write("Authorization: Digest %s\r\n" % (auth))
            elif self._waiting_for_200 and self.CONNECTION_ESTABLISHED in data:
                self._waiting_for_200 = False
                self.authenticated = True
                self.log.debug("Authenticated connection established")
                self.has_connected = True
                if self._server:
                    self._server.add_handler(self)

                loc = data.find(self.CONNECTION_ESTABLISHED)
                self.on_data(data[loc + len(self.CONNECTION_ESTABLISHED + "\r\n"):], self.interface)
            elif self._waiting_for_200:
                self.log.debug("Closing, was waiting for 200")
                self.handle_close()
            elif self.authenticated:
                self.on_data(data, self.interface)
                #pprint.pprint(data)
            else:
                # Unknown state
                self.log.debug("Reached unknown state")
                self.handle_close()

    def readable(self):
        """Returns if this socket is currently readable"""
        return not self._close_after_send

    def ready_to_send(self):
        """Returns if this socket is ready to send"""
        # It is ok to always return true here since we don't notify others that
        # we have connected until we have authenticated which means the
        # connection is fully up. See the comment in the write() function as well
        return True

    def reconnect(self):
        """Attempts to reconnect this SNAP TCP connect"""
        assert self.initiator
        self.log.debug("Initiating connection to %s:%s" % (self._conn[0], self._conn[1]))
        self.flush_write_buff()

        if self.interface.is_sniffer_client:
            user_agent = SNIFFER_USER_AGENT
        else:
            user_agent = SNAP_USER_AGENT

        self.buffered_write(self.CONNECT_STATEMENT + "0\r\nHost: %s\r\nUser-Agent: %s (%s)\r\n\r\n" % (self._conn[0], user_agent, binascii.hexlify(self.local_addr)))
        self._waiting_for_header = False
        self._waiting_for_200 = True
        self.initiator = True
        self.connecting = True

        try:
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._setup_tcp_keepalive(self.tcp_keepalives)
            if not self.cache_dns_lookup or self._cached_lookup is None:
                addr_result = socket.getaddrinfo(self._conn[0], self._conn[1], socket.AF_INET, 0, 0, 0)  # returns [(family, socktype, proto, canonname, sockaddr)]
                try:
                    self._cached_lookup = addr_result[0][4]
                except IndexError:
                    raise socket.error("Did not recieve a valid IPv4 lookup result")
            self.connect(self._cached_lookup)
            if self.secure:
                self._ssl_connecting = True
        except socket.error:
            self.log.exception("An error ocurred while trying to connect")
            self.handle_close()  # Make sure any socket stuff is cleaned up and have it handle reconnecting

    def recv(self, buffer_size):
        """Calls the correct socket receive function depending on SSL"""
        try:
            return super(SNAPtcpConnection, self).recv(buffer_size)
        except ssl.SSLError, err:
            if err.args[0] in (ssl.SSL_ERROR_EOF, ssl.SSL_ERROR_ZERO_RETURN):
                self.handle_close()
                return ''
            if err.args[0] in (ssl.SSL_ERROR_WANT_READ,
                               ssl.SSL_ERROR_WANT_WRITE):
                return ''
            raise

    def secure_connection(self, certfile, ssl_version):
        """Setup encrypted connection."""
        self.socket = ssl.wrap_socket(self.socket, suppress_ragged_eofs=False,
                                      certfile=certfile, server_side=False,
                                      do_handshake_on_connect=False,
                                      ssl_version=ssl_version)
        self._ssl_accepting = True

    def send(self, data):
        """Calls the correct socket send function depending on SSL"""
        try:
            val = super(SSLConnection, self).send(data)
            return val
        except ssl.SSLError, err:
            if err.args[0] in (ssl.SSL_ERROR_EOF, ssl.SSL_ERROR_ZERO_RETURN):
                return 0
            raise

    def _set_remote_addr(self, addr):
        """
        Set the directly connected remote address

        Args:
            addr : The remote SNAP address
        """
        self.remote_addr = addr

    def _setup_tcp_keepalive(self, tcp_keepalives):
        """Enables/Disables TCP level keepalives"""
        try:
            if tcp_keepalives:
                # Code from: http://skytools.projects.postgresql.org/skytools-3.0/api/skytools.psycopgwrapper-pysrc.html
                # OS defaults for SO_KEEPALIVE=1:
                #- Linux: (7200, 9, 75) - can configure all.
                #- MacOS: (7200, 8, 75) - can configure only tcp_keepidle.
                #- Win32: (7200, 5|10, 1) - can configure tcp_keepidle and tcp_keepintvl.
                #  Python needs SIO_KEEPALIVE_VALS support in socket.ioctl to enable it.
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
                #if hasattr(socket, 'TCP_KEEPCNT'):
                    #self.socket.setsockopt(socket.IPPROTO_TCP, getattr(socket, 'TCP_KEEPIDLE'), TCP_KEEPIDLE)
                    #self.socket.setsockopt(socket.IPPROTO_TCP, getattr(socket, 'TCP_KEEPCNT'), TCP_KEEPCNT)
                    #self.socket.setsockopt(socket.IPPROTO_TCP, getattr(socket, 'TCP_KEEPINTVL'), TCP_KEEPINTVL)
                #elif hasattr(socket, 'TCP_KEEPALIVE'):
                    #self.socket.setsockopt(socket.IPPROTO_TCP, getattr(socket, 'TCP_KEEPALIVE'), TCP_KEEPIDLE)
                #elif sys.platform == 'darwin':
                    #TCP_KEEPALIVE = 0x10
                    #self.socket.setsockopt(socket.IPPROTO_TCP, TCP_KEEPALIVE, TCP_KEEPIDLE)
                #elif sys.platform == 'win32' and hasattr(socket, 'SIO_KEEPALIVE_VALS'):
                    #self.socket.ioctl(SIO_KEEPALIVE_VALS, (1, TCP_KEEPIDLE*1000, TCP_KEEPINTVL*1000))
            else:
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 0)
        except socket.error:
            self.log.exception("An error occurred while trying to set a socket option")

    def _get_forward_groups(self):
        """Get the forward groups for this interface"""
        if hasattr(self._forward_groups, 'forward_groups'):
            return self._forward_groups.forward_groups
        else:
            return self._forward_groups

    def _set_forward_groups(self, forward_groups):
        """Set the forward groups used for this interface

        This is used by the forward_groups property. forward_groups may either be an
        integer reprensenting the forwarding groups or an object that has an attribute
        named 'forward_groups'.
        """
        self._forward_groups = forward_groups

    forward_groups = property(_get_forward_groups,_set_forward_groups)

class SNAPtcpServer(EventCallbacks.Registrar):
    """Listens on well known port for connections and dispatches them to handlers"""
    log = logging.getLogger('snap.SNAPtcpServer')
    CONNECTION_TIMEOUT = 90  # seconds

    def __init__(self, scheduler, local_addr):
        """
        Initializes an instance of a SNAPtcpConnection object

        Args:
            scheduler : This SNAP Connect instance scheduler
            local_addr : The SNAP address to use
        """
        self.scheduler = scheduler
        EventCallbacks.Registrar.__init__(self, ('connection_opened', 'connection_closed'))

        self.active_handlers = []
        self._retrys = {}
        self.auth_info = None
        self.ip = None
        self.port = None
        self.server_name = None
        self.server_port = None
        self.tcp_keepalives = False
        self.forward_groups = 0x01
        self.local_addr = local_addr
        self.accepting_sniffers = False

    def accept_sniffer(self):
        """Start allowing remote sniffer connections over TCP"""
        self.accepting_sniffers = True

    def add_handler(self, handler):
        """
        Add the SNAPtcpConnection handler to our active list

        Args:
            handler : A SNAPtcpConnection
        """
        if handler not in self.active_handlers:
            self.active_handlers.append(handler)
            self.scheduler.schedule(0, self._notify_callbacks, 'connection_opened', handler)

    def broadcast(self, data, interface, group=0x01):
        """Broadcast specified data out all active connections"""
        for handler in self.active_handlers:
            if (handler.interface != interface
               and not handler.interface.is_sniffer_server
               and not handler.interface.is_sniffer_client):
                handler.on_broadcast(data, group)

    def _check_connection(self, handler):
        """Check the new connection to make sure it connected, if not close"""
        if isinstance(handler, SNAPtcpConnection) and not handler.authenticated:
            # If not authenticated within timeout then close
            handler.handle_close()

    def close_all(self):
        """Closes all active handlers"""
        # We need a copy here because a handler will remove itself from the servers handler list
        # when it finishes disconnecting
        active_handlers_copy = self.active_handlers[:]

        for handler in active_handlers_copy:
            handler.disconnect()

    def disconnect_connection(self, host, port=None, all=False, retry=False):
        """
        Disconnects the connection to the specified host

        Returns:
            The number of instances disconnected
        """
        count = 0
        _port = port

        for handler in self.active_handlers:
            if port is None:
                if handler.secure:
                    _port = 443
                else:
                    _port = SNAP_IP_PORT
            try:
                if ((handler.initiator and handler._conn == (host, _port)) or
                    (not handler.initiator and handler.socket.getpeername() == (host, _port))):
                    handler.disconnect(retry)
                    count += 1
                    if not all:
                        return count
            except:
                pass

        if not retry:
            for (handler, event) in self._retrys.items():
                if port is None:
                    if handler.secure:
                        _port = 443
                    else:
                        _port = SNAP_IP_PORT

                try:
                    if ((handler.initiator and handler._conn == (host, _port)) or
                        (not handler.initiator and handler.socket.getpeername() == (host, _port))):
                        handler.retry_timeout = 0
                        event.Stop()
                        self._retrys.pop(handler)
                        count += 1
                        if not all:
                            return count
                except:
                    pass

        return count

    def get_connection(self, host, port=None):
        """
        Returns an active SNAPtcpConnection given a host and port

        Args:
            host : The remote host
            port : The remote port
        """
        for handler in self.active_handlers:
            _port = port
            if port is None:
                if handler.secure:
                    _port = 443
                else:
                    _port = SNAP_IP_PORT

            if handler._conn == (host, _port):
                return handler

        return None

    def get_remote_addr(self, addr_pair):
        """
        Returns the SNAP address for the TCP address pair

        Args:
            addr_pair : A TCP (host, port) pair
        """
        for handler in self.active_handlers:
            _port = addr_pair[1]
            if addr_pair[1] is None:
                if handler.secure:
                    _port = 443
                else:
                    _port = SNAP_IP_PORT

            if handler._conn == (addr_pair[0], _port):
                return handler.remote_addr

    def get_stats(self, host, port=None):
        """Returns the bytes sent and received for the specified connection"""
        for handler in self.active_handlers:
            _port = port
            if port is None:
                if handler.secure:
                    _port = 443
                else:
                    _port = SNAP_IP_PORT

            if handler._conn == (host, _port):
                return (handler.total_bytes_sent, handler.total_bytes_recv)

        return None

    def has_connection(self, host):
        """Checks to see if we have a connection to the specified host"""
        for handler in self.active_handlers:
            try:
                if handler.socket.getpeername() == host:
                    return True
            except:
                pass
        return False

    def has_next_hop(self, addr):
        """
        Returns a count of how many handlers have the specified SNAP address

        Args:
            addr : A SNAP address
        """
        cntr = 0
        for handler in self.active_handlers:
            if handler.remote_addr == addr:
                cntr += 1
        return cntr

    def on_conn_err(self, handler, err_no=None):
        """
        Schedules retrying a SNAP TCP connection

        Args:
            handler : A SNAPtcpConnection
            err_no : Optionally a socket error number
        """
        assert isinstance(handler, SNAPtcpConnection)
        if handler.retry_timeout:
            self._retrys[handler] = self.scheduler.schedule(handler.retry_timeout, handler.reconnect)

    def on_new_connection(self, handler):
        """
        Called when a new connection is trying to be established.
        Ensures that the connection is established within a timeout

        Args:
            handler : A SNAPtcpConnection
        """
        SNAPtcpServer.log.debug("New SNAPtcpConnection @ %s" % (handler.interface))
        handler.check_connection_event = self.scheduler.schedule(self.CONNECTION_TIMEOUT, self._check_connection, handler)

    def on_next_hop_addr(self, addr, interface):
        """
        Called when an interface determines the SNAP address of his directly connected neighbor

        Args:
            addr : The directly connect SNAP address
            interface : The interface the address is connected to
        """
        for handler in self.active_handlers:
            if handler.remote_addr == addr:
                if handler.interface == interface:
                    continue  # Keep new connection active
                self.log.debug("Disconnecting %s due to new %s with same next hop" % (handler.interface, interface))
                self.scheduler.schedule(0, handler.disconnect)

    def remove_handler(self, handler):
        """
        Removes a SNAPtcpConnection for the list of active connections

        Args:
            handler : A SNAPtcpConnection
        """
        self.scheduler.schedule(0, self._notify_callbacks, 'connection_closed', handler)
        try:
            self.active_handlers.remove(handler)
        except ValueError:
            pass

    def start_listening(self, auth_info, ip='', port=SNAP_IP_PORT, tcp_keepalives=False, forward_groups=0x01):
        """
        Start listening for SNAP TCP connections

        Args:
            auth_info : The authentication function to call
            ip : The IP address to bind to
            port : The TCP port number to listen on
            tcp_keepalives : Enable TCP level keepalives
            forward_groups : The SNAP multicast forwarding groups to use for this interface
        """
        self.tcp_keepalives = tcp_keepalives
        self.auth_info = auth_info
        self._forward_groups = forward_groups

        self.server_port = self._listen(port, ip)

        # TODO: The following code is not async!
        if not ip:
            try:
                ip = socket.gethostbyname(socket.gethostname())
            except socket.gaierror:
                ip = socket.gethostbyname('localhost')
        try:
            self.server_name = socket.gethostbyaddr(ip)[0]
        except socket.error:
            # Cannot do reverse lookup
            self.server_name = ip       # use the IP address as the "hostname"

        self.ip = ip

    def stop_accepting_sniffer(self, close_existing):
        """Stop accepting remote sniffer connections over TCP

        Args:
            close_existing: If True, close existing sniffer connections
        """
        self.accepting_sniffers = False

        if close_existing:
            for handler in self.active_handlers:
                if handler.interface.is_sniffer_server:
                    handler.disconnect()

    def stop_listening(self):
        """Stop listening for SNAP TCP connections"""
        self.close()

    def _get_forward_groups(self):
        """Get the forward groups for this interface"""
        if hasattr(self._forward_groups, 'forward_groups'):
            return self._forward_groups.forward_groups
        else:
            return self._forward_groups

    def _set_forward_groups(self, forward_groups):
        """Set the forward groups used for this interface

        This is used by the forward_groups property. forward_groups may either be an
        integer reprensenting the forwarding groups or an object that has an attribute
        named 'forward_groups'.
        """
        self._forward_groups = forward_groups

    forward_groups = property(_get_forward_groups,_set_forward_groups)
