# (c) Copyright 2013, Synapse Wireless, Inc.
"""Support for doing SNAP TCP connections using the standard library asyncore"""


import asyncore
import socket
import sys
import errno

from snaplib import interfaces

import snaptcp


class AsyncoreSNAPtcpConnection(snaptcp.SNAPtcpConnection, asyncore.dispatcher):
    def __init__(self, **kwargs):
        snaptcp.SNAPtcpConnection.__init__(self, **kwargs)
        # Do socket stuff after everything is setup
        conn = kwargs['conn']
        if isinstance(conn, socket.SocketType):
            asyncore.dispatcher.__init__(self, conn)
            self.send = self.intf_write = self.interface.write = self.send_data  # Attach our socket send method to the interface write method
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._setup_tcp_keepalive(kwargs['tcp_keepalives'])
        elif len(conn) == 2:
            asyncore.dispatcher.__init__(self)
            self.send = self.intf_write = self.interface.write = self.send_data  # Attach our socket send method to the interface write method
            self._conn = conn
            self.initiator = True
            self.reconnect()
        else:
            raise TypeError("Invalid conn type")

    def handle_connect(self):
        self.closed = False
        self.authenticated = False
        self.connecting = False
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

    def handle_error(self):
        error_type, error_value, nil = sys.exc_info()
        
        if hasattr(error_value,"errno"):
            error = error_value.errno
        else:
            error = error_value[0]

        if error_type == socket.error and error in (errno.EAGAIN, errno.EWOULDBLOCK):
            #If asyncore tries to read from a nonblocking socket and there is no data to read,
            #EAGAIN or EWOULDBLOCK (Errno 11) will be raised. This is a temporary condition that can
            #resolve itself, so we don't want to close our connection at this point.
            pass
        else:
            self.handle_close()

    def handle_expt(self):
        #From EffBot:
        #Called when a connection fails (Windows), or when out-of-band data arrives (Unix)
        #Since we aren't doing OOB

        err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
        if err:
            self.log.error(format_error(err))  # Using error instead of exception since there is not a stack trace
        self.handle_close()

    def handle_read(self):
        self.on_tcp_data(self.recv(8192))

    def handle_write(self):
        bytes_sent = asyncore.dispatcher.send(self, self._write_buffer)
        #self.log.info("handle_write: %s" % (pprint.pformat(self._write_buffer[:bytes_sent]),))
        #self.debug_buff += self._write_buffer[:bytes_sent]
        #self.debug_buff.append(self._write_buffer[:bytes_sent])
        self._write_buffer = self._write_buffer[bytes_sent:]
        self.total_bytes_sent += bytes_sent

    def writable(self):
        if self._write_buffer:
            return True
        elif self._close_after_send:
            self.handle_close()

        return False

    def write(self, data):
        if not self.closed:
            self._write_buffer += data
            #print ord(data[1])
            #pprint.pprint(ord(data[10]))
            asyncore.write(self)  # Try to send now instead of waiting for a poll

        # Yes, there is a chance that your data may never get written since
        # we schedule calling the closed hook after we actually close the
        # connection. So the route table may still have a route and you may try
        # to call this function to send data. Also we may try to broadcast a
        # packet out this interface. I think this is ok since in embedded SNAP
        # you can send to a node that you have a route to and get a RERR back
        # and totally drop a packet that way.


class AsyncoreSnapTcpServer(snaptcp.SNAPtcpServer, asyncore.dispatcher):
    def __init__(self, scheduler, local_addr, **kwargs):
        asyncore.dispatcher.__init__(self)
        snaptcp.SNAPtcpServer.__init__(self, scheduler, local_addr)

    def handle_accept(self):
        try:
            tup = self.accept()
        except socket.error:
            # linux: on rare occasions we get a bogus socket back from
            # accept.  socketmodule.c:makesockaddr complains that the
            # address family is unknown.  We don't want the whole server
            # to shut down because of this.
            return

        try:
            conn, addr = tup
        except TypeError:
            # unpack non-sequence.  this can happen when a read event
            # fires on a listening socket, but when we call accept()
            # we get EWOULDBLOCK, so dispatcher.accept() returns None.
            # Seen on FreeBSD3 and Linux.
            #self.log_info ('warning: server accept() returned %s '
            #               '(EWOULDBLOCK?)' % tup, 'warning')
            return

        handler = AsyncoreSNAPtcpConnection(local_addr=self.local_addr, server=self, conn=conn, auth_info=self.auth_info, tcp_keepalives=self.tcp_keepalives, forward_groups=self._forward_groups)
        self.on_new_connection(handler)

    def _listen(self, port=snaptcp.SNAP_IP_PORT, ip='127.0.0.1'):
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind((ip, port))

        # lower this to 5 if your OS complains
        self.listen(1024)

        host, port = self.socket.getsockname()
        return port


if __name__ == '__main__':
    from apy import EventScheduler

    scheduler = EventScheduler.EventScheduler()
    server = AsyncoreSnapTcpServer(scheduler, '127.0.0.1')
    server.start_listening(lambda realm, username: "public")
    while True:
        scheduler.poll()
        asyncore.poll(0.01)
