# (c) Copyright 2013, Synapse Wireless, Inc.
"""Support for doing SNAP TCP connections using tornado"""


import socket

from tornado.tcpserver import TCPServer
from tornado import iostream

from snaplib import interfaces

import snaptcp

class TornadoSNAPtcpConnection(snaptcp.SNAPtcpConnection):
    def __init__(self, stream=None, **kwargs):
        snaptcp.SNAPtcpConnection.__init__(self, **kwargs)
        conn = kwargs['conn']

        if stream and isinstance(conn, socket.SocketType):
            #This is an incoming TCP connection
            self.stream = stream
            self.socket = self.stream.socket
            self.send = self.intf_write = self.interface.write = self.send_data  # Attach our socket send method to the interface write method
            self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self._setup_tcp_keepalive(kwargs['tcp_keepalives'])
            self.stream.read_until_close(lambda x: None, self._on_data)
        elif len(conn) == 2:
            #This is an outgoing TCP connection
            self.stream = None
            self.send = self.intf_write = self.interface.write = self.send_data  # Attach our socket send method to the interface write method
            self._conn = conn
            self.initiator = True
            self.reconnect()
        else:
            raise TypeError("Invalid conn type")

    def close(self):
        """Close the tornado tcp connection"""
        self.stream.close()

    def _on_data(self, data):
        """Called whenever incoming data is sent to the tornado tcp connection"""
        self.on_tcp_data(data)
        if not self.stream.closed():
            if self._close_after_send:
                self.stream.write(self._write_buffer,self.handle_close)
            else:
                self.stream.write(self._write_buffer)
        self._write_buffer = ""

    def buffered_write(self,data):
        """Add data to be written to the write buffer"""
        self._write_buffer += data
        if self.stream and not self.stream.closed():
            #A tornado iostream has a built in write buffer
            self.stream.write(self._write_buffer)
            self._write_buffer = ""

    def write(self, data):
        """Write to the tornado iostream"""
        if not self.stream.closed():
            self.stream.write(self._write_buffer + data)
        self._write_buffer = ""

    def create_socket(self, family, type):
        """Match asyncore.create_socket behaviour"""
        self.socket = socket.socket(family, type)
        self.stream = iostream.IOStream(self.socket,io_loop=self._server.io_loop)

    def connect(self, address):
        """Match asyncore.connect behaviour"""
        self.stream.connect(address,self.handle_connect)
        self.stream.read_until_close(lambda x: None, self._on_data)
        if self._write_buffer:
            self.stream.write(self._write_buffer)
            self._write_buffer = ""

    def handle_connect(self):
        """Called after the tornado iostream connection is made"""
        self.closed = False
        self.authenticated = False
        self.connecting = False
        self.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)

class TornadoSnapTcpServer(TCPServer, snaptcp.SNAPtcpServer):
    r"""
    A non-blocking, single-threaded SNAP TCP server.
    """
    def __init__(self, scheduler, local_addr, io_loop=None, ssl_options=None, **kwargs):
        snaptcp.SNAPtcpServer.__init__(self, scheduler, local_addr)
        TCPServer.__init__(self, io_loop=io_loop, ssl_options=ssl_options, **kwargs)
        self.io_loop = io_loop

    def close(self):
        self.stop()

    def handle_stream(self, stream, address):
        assert isinstance(stream, iostream.IOStream)
        handler = TornadoSNAPtcpConnection(stream, local_addr=self.local_addr, server=self, conn=stream.socket, auth_info=self.auth_info, tcp_keepalives=self.tcp_keepalives, forward_groups=self._forward_groups)
        self.on_new_connection(handler)

    def _listen(self, port=snaptcp.SNAP_IP_PORT, ip=''):
        self.listen(port, ip)
        for sock in self._sockets.itervalues():
            sock_info = sock.getsockname()
            if len(sock_info) == 2:
                return sock_info[1]
        return ''


if __name__ == '__main__':
    import time
    from tornado import ioloop
    from apy import ioloop_scheduler

    server = TornadoSnapTcpServer(ioloop_scheduler.IOLoopScheduler(), '127.0.0.1')
    server.start_listening(lambda realm, username: "public")
    ioloop.IOLoop.instance().timefunc = time.time
    ioloop.IOLoop.instance().start()
