# (c) Copyright 2012 Synapse Wireless, Inc.
#
# The job of this class is to format (snaplayout) a memory image of the passed in "core file"
# that matches the needs of the various "Boot Helpers".
#
# This class is a "generification" of the original ATmega128RFA1_Layout module.
#

import snaplib.crcUtil

from CoreLayoutError import *

class CoreLayout(object):
    #
    # When instantiating a CoreLayout object, you must supply at least 3 things:
    # 1) A "Core Reader" object, such as pyintelhex.IntelHexReader or pys19.S19
    # 2) A string containing the entire "Hexadecimal-ASCII" file contents (.hex, .s19)
    # ***Note that the supplied contents must be compatible with the supplied coreReader***
    # 3) A "tuple of tuples" specifying the memory segment ranges that should be included
    # in the final image file. Some examples:
    #   ((0x0000,0xFFFF),) - A single segment, 64K in size (Notice that last comma, very important!!!)
    #   ((0x0000,0x03FF), (0x8000,0xFFFF)) - a 1K segment at offset 0, followed by a 32K segment in the upper 32K
    # In the example above, the "core image" would be 33K long.
    # There are two more optional parameters:
    # 4) Another "tuple of tuples" that specifies address ranges that DISQUALIFY the file as valid
    # This is used to detect things like "trying to load ***pre-2.4.32*** ATMEL firmware Over-The-Air
    # 5) A minimum length to be considered a "possible SNAP core"
    def __init__(self, coreReader, contents, segments, invalid_segments=(), minimum_length=0):
        lines = contents.split()
        coreReader.do_verify(lines)

        # temp1 - temp4 will be the pieces of the "OTA image"
        # temp1 - temp3 will make up the 8-byte header
        # temp1 will just be the magic cookie 'SNAP'
        # temp2 will be the length of the actual core image
        # temp3 will be the computed CRC of the actual core image
        # temp4 will be the actual core code

        temp1 = 'SNAP'

        temp4 = ''
        core_records = coreReader.get_data()
        record_count = len(core_records)
        record_index = 0

        segment_index = 0
        low_address = segments[segment_index][0]
        high_address = segments[segment_index][1]
        expected_address = low_address

        while record_index < record_count:
            record = core_records[record_index]

            do_include = False
            numeric_address = record.get_start_address()

            # Keep users from uploading bad stuff...
            if len(invalid_segments) > 0:
                invalid_segment_index = 0
                while invalid_segment_index < len(invalid_segments):
                    if (numeric_address >= invalid_segments[invalid_segment_index][0]) and (numeric_address <= invalid_segments[invalid_segment_index][1]):
                        raise CoreLayoutError("This file is not compatible with Over-The-Air firmware upload!")
                    invalid_segment_index += 1

            # If record data falls within the current segment...
            if (numeric_address >= low_address) and (numeric_address <= high_address):
                do_include = True
            # Otherwise handle segment boundaries...
            elif segment_index < len(segments)-1: # (in other words, there are more segments left)
                # Handle any "holes" in THIS segment of the core file (IOW, "pad out" the image to the end of the segment)
                while expected_address <= high_address:
                    temp4 += '\xFF'
                    expected_address += 1

                segment_index += 1
                low_address = segments[segment_index][0]
                high_address = segments[segment_index][1]
                expected_address = low_address
                # If record data falls within this new current segment...
                if (numeric_address >= low_address) and (numeric_address <= high_address):
                    do_include = True

            if do_include:
                # This is data we DO want to include into the composite image

                # Handle any "holes" in the core file (gaps between records, INSIDE of a segment)
                while expected_address < numeric_address:
                    temp4 += '\xFF'
                    expected_address += 1
                temp4 += record.data
                expected_address += len(record.data)
            record_index += 1

        length = len(temp4)

        # We added the following check to prevent users from trying to upload SNYF firmware over-the-air
        # (The node just ignores it, but silently... by raising this exception we can explain to them what they are doing wrong)
        if length < minimum_length:
            raise CoreLayoutError("This file is too small to be a valid SNAP firmware image!")
        # support for larger flash devices, this specifically the Atmega256RFR2
        # temp2 = chr(length & 0xFF) + chr((length >> 8) & 0xFF) + chr((length >> 16)&0xFF)
        temp2 = chr(length & 0xFF) + chr(length >> 8) 

        crc = snaplib.crcUtil.getCrc(temp4)

        temp3 = chr(crc & 0xFF) + chr(crc >> 8)

        self.composite = temp1 + temp2 + temp3 + temp4

    def getContents(self, space_avail):
        assert(space_avail >= len(self.composite))
        return self.composite
