# (c) Copyright 2007-2012 Synapse Wireless, Inc.
"""Reads an S19 AKA SREC File"""
__docformat__ = "plaintext en"


import binascii


MAX_ADDRESS = 0x1000000


def getNextBlock(addr, blockSize):
    return MAX_ADDRESS - (blockSize * (1 + ( (MAX_ADDRESS-(addr+blockSize)-1) / blockSize) ) )
    #return ((MAX_ADDRESS - (blockSize) *(1 + ((MAX_ADDRESS-((addr)+(blockSize))-1) / (blockSize)))))


class SData(object):
    def __init__(self, startMemLoc, data):
        self.startMemLoc = startMemLoc
        self.data = data

    def getLen(self):
        return len(self.data)

    def getChecksum(self):
        #hexData = "%02x%04x%s" % (len(self.data)+3, self.startMemLoc, binascii.hexlify(self.data))
        #return (~((sum([int(binascii.hexlify(self.data),16) for i in range(0,len(self.data),2)]) + len(hexData) + 3 + (self.startMemLoc & 0xFF) + ((self.startMemLoc>>8) & 0xFF)) & 0xFF)) & 0xFF
        return (~((sum([int(binascii.hexlify(self.data)[i:i+2],16) for i in range(0,len(self.data)*2,2)]) + len(self.data) + 3 + (self.startMemLoc & 0xFF) + ((self.startMemLoc>>8) & 0xFF)) & 0xFF)) & 0xFF

    len = property(getLen)
    checksum = property(getChecksum)

    #
    # This next routine was added to "unify" pyintelhex.py and pys19.py, while at the same
    # time maintaining 100% backwards compatibility with existing clients of those classes.
    #
    def get_start_address(self):
        return self.startMemLoc

class S19(object):
    def __init__(self, byteBoundary=64):
        """
        maxBytes -- Maximum number of bytes to hold
        """
        self._lines = [] #Raw text lines from input file
        self.data = [] #Holds SData object containing continousdata
        self.splitdata = [] #Holds SData objects split up at byteBoundary
        self.byteBoundary = byteBoundary
        self._nextMemLoc = 0
        self._curRecord = None

    def read(self, file):
        self._lines = open(file, 'r').readlines()
        self.verify()

    def verify(self, lines=None):
        if lines is None:
            lines = self._lines

        for line in lines:
            line = line.strip()
            if len(line) < 6 and line[0] == 'S':
                #According to file format there is always at least 6 bytes per line and starts with "S"
                raise Exception("Invalid S19 File")
            if line[1] == '0':
                #descriptive information identifying the following block of S-records.
                pass
            elif line[1] in '123':
                #data record with 1) 16-bit (4 nibble) address, 2) 24-bit (6 nibbble) address, or 3) 32-bit (8 nibble) address 
                _len = int(line[2:4], 16)
                if _len*2 != len(line[4:]):
                    raise Exception("Length of record not equal to what is reported")
                if line[1] == '1':
                    hexData = line[8:-2]
                    memLoc = int(line[4:8], 16)
                elif line[1] == '2':
                    hexData = line[10:-2]
                    memLoc = int(line[4:10], 16)
                if line[1] == '3':
                    hexData = line[12:-2]
                    memLoc = int(line[4:12], 16)
                #Original code only supported S19 records (16-bit/2-byte addresses)
                #if int(line[-2:], 16) != (~((sum([int(hexData[i:i+2],16) for i in range(0,len(hexData),2)]) + _len + (memLoc & 0xFF) + ((memLoc>>8) & 0xFF)) & 0xFF)) & 0xFF:
                expected_checksum = int(line[-2:], 16)
                computed_checksum = sum([int(line[i:i+2],16) for i in range(2,len(line)-2,2)])
                computed_checksum = ~computed_checksum
                computed_checksum &= 0xFF
                if expected_checksum != computed_checksum:
                    raise Exception("Checksum mismatch")

                if self._curRecord and self._curRecord.startMemLoc+self._curRecord.len == memLoc:
                    self._curRecord.data += binascii.unhexlify(hexData)
                else:
                    if self._curRecord:
                        self.data.append(self._curRecord)
                    self._curRecord = SData(memLoc, binascii.unhexlify(hexData))
                pass
            elif line[1] == '7':
                #A termination record for a block of S3 records.
                pass
            elif line[1] == '8':
                #A termination record for a block of S2 records.
                pass
            elif line[1] == '9':
                #A termination record for a block of S1 records.
                pass
            else:
                raise Exception("Unhandled Record Type: %s" % line[1])
        self.data.append(self._curRecord) #Don't forget the last record

        for sdata in self.data:
            assert isinstance(sdata, SData)
            memLoc = sdata.startMemLoc
            while sdata.data:
                nextBoundary = getNextBlock(memLoc, self.byteBoundary)
                getLen = nextBoundary - memLoc
                #print "memLoc: %04X nextBoundary: %04X getLen: %04X" % (memLoc, nextBoundary, getLen)
                self.splitdata.append(SData(memLoc, sdata.data[:getLen]))
                sdata.data = sdata.data[getLen:]
                memLoc = nextBoundary

    #
    # These next two routines were added to "unify" pyintelhex.py and pys19.py,
    # while at the same time maintaining 100% backwards compatibility with existing
    # clients of those classes.
    #
    def get_data(self):
        return self.splitdata
            
    def do_verify(self, lines=None):
        self.verify(lines)
            

if __name__ == '__main__':
    s19 = S19()
    s19.read("\\synapse\\Portal\\trunk\\firmware\\SnapV2.0.25.S19")
    print "file looks good"
