# (c) Copyright 2008-2009 Synapse Wireless, Inc.
"""
Functions to encode and decode SNAP Mesh Routing packets.
Currently only supports decoding Route Request (RREQ) packets,
and building and encoding RREP (Route Reply) packets. These are
all that are currently needed.
"""


from SnapConstants import *
import SnapConstantsHelper
import interfaces
import sequence


MESH_BROADCAST_GROUP = "\x00\x01"

# Mesh defaults
MAX_TIMEOUT = 60000
MIN_TIMEOUT = 1000
NEW_TIMEOUT = 5000
USED_TIMEOUT = 5000
DELETE_TIEMOUT = 10000
RREQ_RETRIES = 3
RREQ_WAIT_TIMEOUT = 500
INITIAL_HOP_LIMIT = 1
MAX_HOP_LIMIT = 5
START_SEQ_NUM = 1
MESH_OVERRIDE = 0
LQ_THRESHOLD = 127


class MeshError(Exception):
    def __init__(self, descr):
        self.descr = descr


class MeshDescriptor(object):
    def __init__(self, crypto=None):
        self.msgId = None
        self.hopLimit = None
        self.target = None
        self.targetSequence = None
        self.targetDistance = None
        self.originator = None
        self.originatorSequence = None
        self.originatorDistance = None
        self.additionalAddresses = []
        self.additionalSeqs = []
        self.incoming_interface = interfaces.Interface()
        self.source = None
        self.destination = None
        self.crypto = crypto

    def get_encoded_data(self):
        return MeshCodec.encode(self, crypto=self.crypto)

    encoded_data = property(get_encoded_data)


class MeshCodec(object):
    sequence_number = 1 # class attribute, not an instance attribute (no self. prefix)

    @staticmethod
    def increment_sequence():
        """Increment our Mesh Routing sequence number, per the DYMO spec"""
        MeshCodec.sequence_number += 1
        if MeshCodec.sequence_number == 65536:
            MeshCodec.sequence_number = 256

    @staticmethod
    def adjustSequenceNumber(meshDesc):
        """See if our Mesh Routing sequence number should be incremented, per the DYMO spec"""
        if meshDesc.targetSequence == 0:
            MeshCodec.increment_sequence()
        elif meshDesc.targetSequence < MeshCodec.sequence_number:
            MeshCodec.increment_sequence()

    @staticmethod
    def build_rrep(descriptor):
        """Build the RREP (Route Reply) based on the original RREQ (Route Request)"""
        assert isinstance(descriptor, MeshDescriptor)
        assert descriptor.msgId == MESH_RREQ
        rrep_desc = MeshDescriptor(crypto=descriptor.crypto)
        rrep_desc.msgId = MESH_RREP
        rrep_desc.hopLimit = descriptor.originatorDistance + 1 # +1 because we have to count ourself as a hop

        rrep_desc.target = descriptor.originator
        rrep_desc.targetSequence = descriptor.originatorSequence
        rrep_desc.targetDistance = descriptor.originatorDistance

        rrep_desc.originator = descriptor.target
        rrep_desc.destination = descriptor.source
        rrep_desc.originatorSequence = MeshCodec.sequence_number
        rrep_desc.originatorDistance = 0

        rrep_desc.source = descriptor.target

        return rrep_desc

    @staticmethod
    def build_rreq(who, source, retries, crypto=None):
        """Build a RREQ (Route Request) message for who we are looking for from our source"""
        assert len(who) == SNAP2_ADDR_LEN
        assert len(source) == SNAP2_ADDR_LEN
        assert retries > 0
        assert retries < 256

        MeshCodec.increment_sequence()

        descriptor = MeshDescriptor(crypto=crypto)
        descriptor.msgId = MESH_RREQ
        descriptor.target = who
        descriptor.hopLimit = retries

        # If we have a route entry we should fill these in
        descriptor.targetSequence = 0
        descriptor.targetDistance = 0

        descriptor.source = descriptor.originator = source
        descriptor.originatorSequence = MeshCodec.sequence_number
        descriptor.originatorDistance = 0

        return descriptor

    @staticmethod
    def encode(descriptor, crypto=None):
        """Encode a Mesh Routing packet for output to a SNAP device"""
        assert isinstance(descriptor, MeshDescriptor)

        pkt = ''
        seq_num = sequence.get_number(descriptor.source)

        if descriptor.msgId == MESH_RREP:
            pid = PID_MESH_PTP
            if crypto and crypto.has_valid_key:
                pid |= 0x80
            pkt += chr(pid)
            pkt += descriptor.destination
            pkt += descriptor.source
            pkt += chr(seq_num)
            pkt += descriptor.msgId
            pkt += chr(descriptor.hopLimit)
            pkt += descriptor.target
            pkt += chr(descriptor.targetSequence >> 8)
            pkt += chr(descriptor.targetSequence & 0xff)
            pkt += chr(descriptor.targetDistance)
            pkt += descriptor.originator
            pkt += chr(descriptor.originatorSequence >> 8)
            pkt += chr(descriptor.originatorSequence & 0xff)
            pkt += chr(descriptor.originatorDistance)
        elif descriptor.msgId == MESH_RREQ or descriptor.msgId == MESH_RERR:
            pid = PID_MESH_BCAST
            if crypto and crypto.has_valid_key:
                pid |= 0x80
            pkt += chr(pid)
            pkt += MESH_BROADCAST_GROUP
            pkt += chr(1) # TTL
            pkt += descriptor.source
            pkt += chr(seq_num)
            pkt += descriptor.msgId
            pkt += chr(descriptor.hopLimit)
            pkt += descriptor.target
            pkt += chr(descriptor.targetSequence >> 8)
            pkt += chr(descriptor.targetSequence & 0xff)
            if descriptor.msgId == MESH_RREQ:
                pkt += chr(descriptor.targetDistance)
                pkt += descriptor.originator
                pkt += chr(descriptor.originatorSequence >> 8)
                pkt += chr(descriptor.originatorSequence & 0xff)
                pkt += chr(descriptor.originatorDistance)
                for addr in descriptor.additionalAddresses:
                    pkt += addr
            elif descriptor.msgId == MESH_RERR:
                assert len(descriptor.additionalAddresses) == len(descriptor.additionalSeqs)
                for i in range(0, len(descriptor.additionalAddresses)):
                    pkt += descriptor.additionalAddresses[i]
                    pkt += chr(descriptor.additionalSeqs[i] >> 8)
                    pkt += chr(descriptor.additionalSeqs[i] & 0xff)
        else:
            raise MeshError('encode function does not yet support message type %c (%d))' % (descriptor.msgId, descriptor.msgId))

        try:
            pkt = "%c%s" % (len(pkt), pkt) # Prepend length
        except OverflowError:
            raise MeshError("Packet is greater than 255 bytes")

        if len(pkt) > SnapConstantsHelper.MAX_PKT_SIZE:
            raise MeshError('Mesh Routing messsage too large to encode (%d bytes, max is %d)' % (len(pkt), SnapConstantsHelper.MAX_PKT_SIZE))

        if crypto and crypto.has_valid_key:
            if descriptor.msgId == MESH_RREQ or descriptor.msgId == MESH_RERR:
                pkt = pkt[:MESH_MULTI_PKT_DATA] + crypto.encrypt(seq_num, pkt[MESH_MULTI_PKT_DATA:])
            elif descriptor.msgId == MESH_RREP:
                pkt = pkt[:UNI_PKT_DATA] + crypto.encrypt(seq_num, pkt[UNI_PKT_DATA:])

        return pkt

    @staticmethod
    def decode(meshPacketString, crypto=None):
        """Decode a MeshDescriptor from a Mesh Routing Packet"""
        meshDesc = MeshDescriptor()

        meshPacket = map(ord,meshPacketString)

        pid = meshPacket[RPC_PID_OFS]
        rxSequence = meshPacket[RPC_SEQ_OFS]

        if crypto and crypto.has_valid_key:
            if pid == (PID_MESH_PTP | 0x80):
                meshPacketString = meshPacketString[:UNI_PKT_DATA] + crypto.decrypt(rxSequence, meshPacketString[UNI_PKT_DATA:])
            elif pid == (PID_MESH_BCAST | 0x80):
                meshPacketString = meshPacketString[:MESH_MULTI_PKT_DATA] + crypto.decrypt(rxSequence, meshPacketString[MESH_MULTI_PKT_DATA:])
            meshPacket = map(ord, meshPacketString)
            meshDesc.crypto = crypto

        if pid == PID_MESH_BCAST or pid == (PID_MESH_BCAST | 0x80):
            meshDesc.msgId = meshPacketString[MBST_MSGID_OFS]
            if meshDesc.msgId == MESH_RREQ or meshDesc.msgId == MESH_RERR:
                # could decode length but don't care
                # could store pid in descriptor but it is implicit in the MsgID
                # could decode Group but should always be 0x0001
                # could decode TTL but should always be 1
                meshDesc.source = meshPacketString[MBST_SRC_OFS:MBST_SRC_OFS + SNAP2_ADDR_LEN]
                # could store rxSequence in descriptor but don't care
                meshDesc.hopLimit = meshPacket[MBST_MAXHOP_OFS]
                meshDesc.target = meshPacketString[MBST_TGT_OFS:MBST_TGT_OFS + SNAP2_ADDR_LEN]
                meshDesc.targetSequence = (meshPacket[MBST_TGT_SEQ_OFS] << 8) + meshPacket[MBST_TGT_SEQ_OFS+1]

                if meshDesc.msgId == MESH_RREQ:
                    meshDesc.originator = meshPacketString[MBST_ORIG_OFS:MBST_ORIG_OFS + SNAP2_ADDR_LEN]
                    meshDesc.originatorSequence = (meshPacket[MBST_ORIG_SEQ_OFS] << 8) + meshPacket[MBST_ORIG_SEQ_OFS+1]
                    meshDesc.originatorDistance = meshPacket[MBST_ORIG_DIST_OFS]
                    meshDesc.targetDistance = meshPacket[MBST_TGT_DIST_OFS]
                    additionalBytes = len(meshPacketString) - MBST_EXTRA_OFS
                    index = MBST_EXTRA_OFS
                    while additionalBytes >= SNAP2_ADDR_LEN:
                        meshDesc.additionalAddresses.append(meshPacketString[index:index + SNAP2_ADDR_LEN])
                        index += SNAP2_ADDR_LEN
                        additionalBytes -= SNAP2_ADDR_LEN
                elif meshDesc.msgId == MESH_RERR:
                    index = MBST_TGT_SEQ_OFS+2
                    additionalBytes = len(meshPacketString) - index
                    while additionalBytes >= SNAP2_ADDR_LEN+2:
                        meshDesc.additionalAddresses.append(meshPacketString[index:index + SNAP2_ADDR_LEN])
                        index += SNAP2_ADDR_LEN
                        meshDesc.additionalSeqs.append((meshPacket[index] << 8) + meshPacket[index+1])
                        index += 2
                        additionalBytes -= SNAP2_ADDR_LEN+2
                    assert len(meshDesc.additionalAddresses) == len(meshDesc.additionalSeqs)
            elif __debug__ and pid != (PID_MESH_BCAST | 0x80): # pragma: no cover
                assert False
        elif pid == PID_MESH_PTP or pid == (PID_MESH_PTP | 0x80):
            meshDesc.msgId = meshPacketString[MPTP_TYP_OFS]
            if meshDesc.msgId == MESH_RREP:
                meshDesc.source = meshPacketString[MPTP_SRC_OFS:MPTP_SRC_OFS + SNAP2_ADDR_LEN]
                meshDesc.hopLimit = meshPacket[MBST_MAXHOP_OFS]
                meshDesc.target = meshPacketString[MBST_TGT_OFS:MBST_TGT_OFS + SNAP2_ADDR_LEN]
                meshDesc.targetSequence = (meshPacket[MBST_TGT_SEQ_OFS] << 8) + meshPacket[MBST_TGT_SEQ_OFS+1]
                meshDesc.targetDistance = meshPacket[MBST_TGT_DIST_OFS]
                meshDesc.originator = meshPacketString[MBST_ORIG_OFS:MBST_ORIG_OFS + SNAP2_ADDR_LEN]
                meshDesc.originatorSequence = (meshPacket[MBST_ORIG_SEQ_OFS] << 8) + meshPacket[MBST_ORIG_SEQ_OFS+1]
                meshDesc.originatorDistance = meshPacket[MBST_ORIG_DIST_OFS]
            elif __debug__ and pid != (PID_MESH_BCAST | 0x80): # pragma: no cover
                assert False, "Unsupported Mesh PTP message ID"
        else:
            return None

        return meshDesc
