# (c) Copyright 2007-2012 Synapse Wireless, Inc.
"""Holds data and gives it to an underlying connection"""


import logging
log = logging.getLogger("snaplib.%s" % (__name__))
import binascii, pprint, types, datetime, sys

import EventCallbacks
import RpcCodec, DataModeCodec
import PacketSerialProtocol
import MinimalMeshHandler
import SnapConstantsHelper
import Globleton, TraceRouteCodec
import MeshCodec
import SnapEncryptions
try:
    import aes
    AES_SUPPORT = True
except (ImportError, SystemError):
    AES_SUPPORT = False


from apy import Deferred
if __debug__:
    import timeit


class QueuedObject(object):
    def __init__(self, obj):
        self.obj = obj
        self.deferred = Deferred.Deferred()


class Connector(EventCallbacks.Registrar):
    """Queues up packets and feeds them to a connection like packet serial or the gateway client"""
    def __init__(self, callsink, pktSerialDriver, pgClient=None, scClient = None, crypto=None, rpc_response_timeout=SnapConstantsHelper.SNAP_RPC_TIMEOUT):
        """
        callsink -- Default callsink to use for calling function from incoming RPCs
        pktSer -- Has information about the serial connection
        pgClient -- XMLRPC channel for sending and receiving RPCs
        scClient -- SNAPtcp counterpart to pgClient
        (gwClient - is a handle to either pgClient or scClient. Client code is not aware we have added a second protocol)
        """
        EventCallbacks.Registrar.__init__(self, ['DataMode', 'MeshRouting', 'TraceRoute', 'FailedResponse', 'NotOpen', 'QueueLengthChange', 'Rpc'])
        self.callsink = callsink
        self.pgClient = pgClient
        self.scClient = scClient
        self.gwClient = pgClient # but see also selectProtocol()
        self.pktSer = pktSerialDriver
        if pgClient is not None:
            self.pgClient.registerRcvCallback(self.gwPktRx)
            self.pgClient.registerTxCallback(self.OnPktTx)
        if scClient is not None:
            self.scClient.registerRcvCallback(self.gwPktRx)
            self.scClient.registerTxCallback(self.OnPktTx)
            self.scClient.registerCallback('DataMode', self.relayDataModeDescriptor)
            self.scClient.registerCallback('TraceRoute', self.relayTraceRouteDescriptor)

        self.queue = [] # List of pending user requests
        self.systemQueue = [] # List of pending system packets (i.e. Mesh)
        self.sentQueue = [] #List of RPC requests that have been sent to the PacketSerialDriver
        self._paused = False
        self.curDeferred = None
        self.waitingOnResponse = None
        self.startedWaiting = datetime.datetime.now()
        self.curTx = None
        self.responseTimeouts = 0
        self.mesh_addresses = []
        self.crypto = None
        self.rpc_response_timeout = rpc_response_timeout

    # Client code like the DataModeHandler.DataModeLogger is not aware of the multiple ways
    # we can now reach remote SNAP nodes. At present, these multiple ways (objects) are hidden
    # inside of this PacketQueue.Connector class. Here we are forwarding a DATA MODE report
    # from the PortalSNAPcomClient to anybody who registered interest in them with THIS object.
    def relayDataModeDescriptor(self, descriptor):
        assert isinstance(descriptor, DataModeCodec.DataModeDescriptor)
        self._notifyCallbacks('DataMode', descriptor)

    # Here we are doing the same thing for 'TraceRoute' events
    def relayTraceRouteDescriptor(self, descriptor):
        assert isinstance(descriptor, TraceRouteCodec.TraceRouteDescriptor)
        if self.waitingOnResponse == descriptor.srcAddr:
            self.waitingOnResponse = None
        self._notifyCallbacks('TraceRoute', descriptor)

    def selectProtocol(self, protocol):
        if protocol == SnapConstantsHelper.PROTOCOL_SNAP_OVER_TCP:
            self.gwClient = self.scClient
        elif protocol == SnapConstantsHelper.PROTOCOL_SNAP_OVER_XMLRPC:
            self.gwClient = self.pgClient
        else:
            self.gwClient = None

    def addAddress(self, address, clear=False):
        if clear:
            del self.mesh_addresses[:]
        if address not in self.mesh_addresses:
            self.mesh_addresses.append(address)

    def checkWaitingResponse(self):
        if self.waitingOnResponse:
            if (datetime.datetime.now() - self.startedWaiting) > datetime.timedelta(milliseconds=self.rpc_response_timeout):
                log.error("Did not get an expected response within timeout period from %s" % binascii.hexlify(self.waitingOnResponse))
                self._notifyCallbacks('FailedResponse', self.waitingOnResponse)
                self.waitingOnResponse = None
                self.responseTimeouts += 1
                self.sendNextPkt()
        else:
            self.sendNextPkt()

    def clear(self):
        """Clear any pending RPCs, normaly called when the serial port is disconnected"""
        for obj in self.queue:
            obj.deferred(False) #Tell everyone that it was not sent sucessfully
        self.queue = []
        #Should we be clearing the system queue too?
        self.waitingOnResponse = None
        self.responseTimeouts = 0
        self.curDeferred = None
        if __debug__:
            self.sentQueue = []
        self._notifyCallbacks('QueueLengthChange', len(self.queue))

    def getPaused(self):
        return self._paused

    def getConnectionName(self):
        if self.pktSer.serial.isOpen:
            #return self.pktSer.serial.getPortstr()
            return self.pktSer.serial.portName() + " Port"
        elif self.gwClient.configured:
            return self.gwClient.gwAddr
        else:
            return "UNKNOWN"

    def gwPktRx(self, rpc):
        """Called from the Portal Gateway Client upon receiving a complete packet"""
        assert isinstance(rpc, RpcCodec.RpcDescriptor)

        if rpc:
            if self.waitingOnResponse is not None and (rpc.srcAddr == self.waitingOnResponse or len(self.waitingOnResponse) == 2):
                self.waitingOnResponse = None
            self.sendNextPkt()

            # TODO 3813 remove this check
            if rpc.funcName == 'dataMode':
                #We have to remake the data mode packet :(
                desc = DataModeCodec.DataModeDescriptor()
                desc.data = rpc.args[0]
                desc.srcAddr = rpc.srcAddr
                desc.dstAddr = rpc.dstAddr

                pkt = DataModeCodec.encode(desc)
                self.tellDataModeCallbacks(pkt)
            else:
                try:
                    rpc(self.callsink)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except:
                    if __debug__:
                        log.exception('An error occurred while trying to call the RPC function %s' % rpc.funcName)
                    pass
        else:
            log.debug("Received unsupported packet")

    def gwPktTx(self):
        if not self._paused and self.gwClient.configured:
            if self.systemQueue:
                obj = self.systemQueue.pop(0)
            elif self.queue and self.curDeferred is None and self.waitingOnResponse is None:
                #log.debug("queue READY")
                obj = self.queue.pop(0)
            else:
                #log.debug("queue NOT ready")
                return
            assert isinstance(obj, QueuedObject)
            data = obj.obj
            self.curDeferred = obj.deferred
            self.waitingOnResponse = None
            if __debug__:
                self.sentQueue.append(obj)
                if len(self.sentQueue) > 100:
                    self.sentQueue.pop(0)
                data.txTime = timeit.default_timer()

            if isinstance(data, RpcCodec.RpcDescriptor):
                obj.deferred.callername = "rpcTx: %s.%s%s" % (binascii.hexlify(data.dstAddr), data.funcName, data.args)
                if data.isMulticast:
                    self.gwClient.sendMcastRpc(data.dstAddr, data.ttl, data.funcName, data.args)
                else:
                    self.gwClient.sendRpc(data.dstAddr, data.funcName, data.expectsResponse, data.args)
                if data.expectsResponse:
                    self.waitingOnResponse = data.dstAddr
                    self.startedWaiting = datetime.datetime.now()
            elif isinstance(data, DataModeCodec.DataModeDescriptor):
                if data.isMulticast:
                    self.gwClient.sendMcastData(data.dstAddr, data.ttl, data.data)
                else:
                    self.gwClient.sendData(data.dstAddr, data.data)
            elif isinstance(data, TraceRouteCodec.TraceRouteDescriptor):
                self.gwClient.sendTraceroute(data.dstAddr, data.sequence)
                self.waitingOnResponse = data.dstAddr
                self.startedWaiting = datetime.datetime.now()
            elif callable(data):
                try:
                    data()
                except (KeyboardInterrupt, SystemExit):
                    raise
                except:
                    log.exception('An error occurred while trying to call a function in the pkt queue')
                self.OnPktTx(True) #Simulate that the "packet" was transmitted by packet serial
            else:
                log.error("A unsupported object type was enqueued: %s" % data.__class__)

            self._notifyCallbacks('QueueLengthChange', len(self.queue))

    def handlePktRx(self, pkt, source_intf):
        """Called from the PacketSerialDriver upon receiving a complete packet"""
        #Check if we are "open", if not silently drop (we have no way of clearing entries)
        if self.pktSer.serial.isOpen:
            pid = ord(pkt[SnapConstantsHelper.RPC_PID_OFS])
            if self.crypto and self.crypto.has_valid_key:
                pid ^= SnapConstantsHelper.PID_ENCRYPTION_MASK
                pkt = SnapEncryptions.decode_pkt(pkt, self.crypto)

            if ((pid == SnapConstantsHelper.PID_RPC) or (pid == SnapConstantsHelper.PID_MULTICAST_RPC)):
                try:
                    rpc = RpcCodec.RpcCodec.decode(pkt)
                except (KeyboardInterrupt, SystemExit):
                    raise
                except:
                    log.exception("Unable to decode RPC packet due to receiving bad serial data")
                    return

                if rpc:
                    if self.waitingOnResponse is not None and (rpc.srcAddr == self.waitingOnResponse or len(self.waitingOnResponse) == 2):
                        self.waitingOnResponse = None
                        self.handlePktTx()

                    try:
                        rpc(self.callsink)
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except RpcCodec.WrongArgCount:
                        if self.callsink.logUnknownFunc:
                            log.warning("Node %s called function %s whose signature does not match" % (binascii.hexlify(rpc.srcAddr), rpc.funcName))
                    except:
                        (ex_type, ex, tb) = sys.exc_info()
                        log.error('An error occurred while trying to call the RPC function %s\nFile %s, at line %i:\n%s' %
                                      (rpc.funcName, tb.tb_frame.f_code.co_filename,tb.tb_lineno, str(ex)))
                else:
                    log.warning("An invalid RPC packet was received")
                    log.info("Bad Packet: %s" % (pprint.pformat(pkt)))
            elif ((pid == SnapConstantsHelper.PID_DM_PTP) or (pid == SnapConstantsHelper.PID_DM_BCAST)):
                desc = DataModeCodec.decode(pkt)
                self._notifyCallbacks('DataMode', desc)
            elif pid in (SnapConstantsHelper.PID_MESH_PTP,
                         SnapConstantsHelper.PID_MESH_BCAST,
                         SnapConstantsHelper.PID_MESH_PTP | SnapConstantsHelper.PID_ENCRYPTION_MASK,
                         SnapConstantsHelper.PID_MESH_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK):
                reply_pkt = MinimalMeshHandler.processMeshPacket(pkt, self.mesh_addresses)
                if reply_pkt:
                    self.rpcReady(reply_pkt, True, True)
            elif (pid == SnapConstantsHelper.PID_TRACE_ROUTE):
                desc = TraceRouteCodec.decode(pkt)
                if desc:
                    assert isinstance(desc, TraceRouteCodec.TraceRouteDescriptor)
                    if __debug__:
                        log.debug(desc)
                    if self.waitingOnResponse is not None and (desc.orig_src_addr == self.waitingOnResponse or len(self.waitingOnResponse) == 2):
                        self.waitingOnResponse = None
                        self.handlePktTx()

                    self._notifyCallbacks('TraceRoute', desc)
            else:
                log.debug("Received unsupported PID")

    def handlePktTx(self):
        """Called to have the PacketSerialDriver send a packet"""
        # Get next piece of data from queue, NOTE: don't check waitingOnResponse here or systemQueue won't get services
        if self.pktSer.ready_to_send() and self.curDeferred is None:
            if self.systemQueue:
                obj = self.systemQueue.pop(0)
            elif not self._paused and self.queue and self.waitingOnResponse is None:
                obj = self.queue.pop(0)
            else:
                return

            assert isinstance(obj, QueuedObject)
            data = obj.obj
            self.curDeferred = obj.deferred
            #print "handlePktTx: curDeferred=%s" % (obj.deferred)
            self.waitingOnResponse = None
            if __debug__:
                self.sentQueue.append(obj)
                if len(self.sentQueue) > 100:
                    self.sentQueue.pop(0)

            try:
                if isinstance(data, RpcCodec.RpcDescriptor):
                    if __debug__:
                        data.txTime = timeit.default_timer()
                    self.pktSer.send_data(RpcCodec.RpcCodec.encode(data, crypto=self.crypto))
                    if data.expectsResponse:
                        self.waitingOnResponse = data.dstAddr
                        self.startedWaiting = datetime.datetime.now()
                        #log.debug("EXPECTS RESPONSE")
                elif isinstance(data, DataModeCodec.DataModeDescriptor):
                    self.pktSer.send_data(DataModeCodec.encode(data, crypto=self.crypto))
                elif isinstance(data, TraceRouteCodec.TraceRouteDescriptor):
                    self.pktSer.send_data(TraceRouteCodec.encode(data, crypto=self.crypto))
                    self.waitingOnResponse = data.dstAddr
                    self.startedWaiting = datetime.datetime.now()
                elif isinstance(data, MeshCodec.MeshDescriptor):
                    self.pktSer.send_data(MeshCodec.MeshCodec.encode(data, crypto=self.crypto))
                elif data.__class__ == types.StringType:
                    self.pktSer.send_data(data)
                elif callable(data):
                    try:
                        data()
                    except (KeyboardInterrupt, SystemExit):
                        raise
                    except:
                        log.exception('An error occurred while trying to call a function in the pkt queue')
                    self.OnPktTx(True) #Simulate that the "packet" was transmitted by packet serial
                else:
                    log.error("A unsupported object type was enqueued: %s" % data.__class__)
            except RpcCodec.RpcError, e:
                log.exception("An error occurred while trying to create the RPC: %s" % (e.descr))
                obj.deferred(False) #Tell everyone that it was not sent sucessfully
                self.curDeferred = None
                #print "handlePktTx: curDeferred=None"

            self._notifyCallbacks('QueueLengthChange', len(self.queue))

    def isConnected(self):
        """Is the underlying connection open"""
        if self.pktSer.serial and self.pktSer.serial.isOpen:
            return True
        else:
            return self.gwClient.configured

    def isSnapComClient(self):
        return self.gwClient == self.scClient and self.isConnected()

    def OnPktTx(self, lastSendOk, tell=True):
        """Called upon sending a complete packet"""
        if self.curDeferred is not None:
            assert isinstance(self.curDeferred, Deferred.Deferred)
            #We need to make a copy because calling the deferred could trigger sending another packet
            #which would change curDeferred which we are about to set to None
            deferred = self.curDeferred

            #We need to set this to None here because the gateway client will call this method
            #more than once since we get a success on tx of a packet and when the response comes in
            self.curDeferred = None

            deferred(lastSendOk)
            #log.debug("OnPktTx: %s" % (deferred.callername))
        elif __debug__:
            log.error("OnPktTx: No deferred")

        if not lastSendOk:
            self.waitingOnResponse = None

        self.sendNextPkt()

    def pendingRpcCount(self):
        """Returns the number of RPCs pending to be sent"""
        return len(self.queue)

    def rpcReady(self, obj, urgent=False, system=False):
        """
        obj -- An object to get queued for sending
        urgent -- If the object should be added to the beginning (default False)
        system -- This queue gets serviced before the other queue and is intended for things like mesh (default False)
        In most cases, enqueues (but does not send) the packet, and returns a deferred object.
        In the special case of "urgent==True and system==True" WILL try to send the packet immediately, but then must return None.
        In other words, you cannot use the "immediate send" and the "deferred" features at the same time.
        """
        # Portal now receives/processes multicasts, which means this can fire with legitimate traffic
        #if __debug__ and obj.__class__ == RpcCodec.RpcDescriptor:
            #assert obj.srcAddr != obj.dstAddr

        newObj = QueuedObject(obj)
        #newObj.deferred.callername = sys._getframe(2).f_code.co_name

        if system:
            queue = self.systemQueue
        else:
            queue = self.queue

        if urgent:
            queue.insert(0, newObj)
        else:
            queue.append(newObj)

        if self.pktSer.serial and not self.pktSer.serial.isOpen and not self.gwClient.configured:
            self._notifyCallbacks('NotOpen')

        self._notifyCallbacks('QueueLengthChange', len(self.queue))

        # #4113 and #2457
        # Allow system+urgent packets to go out immediately, but at the expense of not allowing "deferred" operations on that packet
        # This let's us keep the capability (needed for Mesh Routing, see #4113) but without re-introducing the original #2457 issue.
        if system and urgent:
            self.sendNextPkt()
            # Callers using the "send immediate" capability should not also be trying to use the "deferred" feature.
            # Return a None, so that they will realize the error of their ways if they DO try and specify deferred actions.
            return None
        else:
            # Normal packets don't get sent immediately, but the caller can specify deferred actions on the packet.
            return newObj.deferred

    def sendNextPkt(self):
        if self.gwClient is not None and self.gwClient.configured:
            self.gwPktTx()
        else:
            self.handlePktTx()

    def setPaused(self, value):
        """Change the pause state of the queue"""
        if __debug__:
            log.debug("Changing pause: %s" % str(value))
        self._paused = value
        self.sendNextPkt()

    paused = property(getPaused, setPaused)
