# (c) Copyright 2007-2009 Synapse Wireless, Inc.
"""
Packetized Serial Data Encapsulation, for communication with SNAP devices.
"""


import logging
import datetime
if __debug__:
    import pprint, binascii
    import Snap2Decoder
import random
import string
import re
import pprint

try:
    import multiprocessing

    if multiprocessing.current_process().name != 'Serial':
        log = logging.getLogger("snaplib.%s" % (__name__))
    else:
        log = multiprocessing.get_logger()
except ImportError:
    #Don't want to be dependent on multiprocessing module
    log = logging.getLogger("snaplib.%s" % (__name__))

import EventCallbacks
import SnapConstantsHelper
import interfaces


# Packet Serial Constants
SOH, END, ESC, ESC_SOH, ESC_END, ESC_ESC = (chr(0xC0), chr(0xC1), chr(0xC2), chr(0xE0), chr(0xE1), chr(0xE2))
SOH_END_RE = re.compile(r'\xc0([^\xc1]+)\xc1')

UNACKD_PIDS = (SnapConstantsHelper.PID_DM_BCAST,
               SnapConstantsHelper.PID_DM_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK,
               SnapConstantsHelper.PID_MULTICAST_RPC,
               SnapConstantsHelper.PID_MULTICAST_RPC | SnapConstantsHelper.PID_ENCRYPTION_MASK,
               SnapConstantsHelper.PID_MESH_BCAST,
               SnapConstantsHelper.PID_MESH_BCAST | SnapConstantsHelper.PID_ENCRYPTION_MASK,
               )


class PacketSerialProtocol(EventCallbacks.Registrar):
    # Packet Serial Constants
    escapes = {SOH : ESC + ESC_SOH, END : ESC + ESC_END, ESC : ESC + ESC_ESC}
    unEscapes = {ESC_SOH : SOH, ESC_END : END, ESC_ESC : ESC}
    sendTimeout = datetime.timedelta(milliseconds=250)
    maxRetries = 20

    def __init__(self):
        # Callbacks:  tx(sentOk), rx(dataBuf)
        EventCallbacks.Registrar.__init__(self, ['tx', 'rx'])

        # Transmitter state
        self.txSequence = random.randint(0, 255)

        self.curTxPkt = None
        self.curOrigString = ''
        self.curTxCount = 0
        self._incoming_data = ''

        # Receiver state
        self.state = 'IDLE'
        self.rxPkt = ''
        self.lastRxSequence = None
        self.neverAck = False # We won't ACK and we won't expect an ACK back
        self.check_len = True

        # Statistics
        self.numTxFail = 0
        self.numTxRetries = 0
        self.rxErrors = 0
        self.numRxAcks = 0   # Total ACK packets received
        self.numTxAcked = 0  # Acknowledged transmits (rx ACK matches tx SEQ)
        self.numRxPkts = 0
        self.numProcessErrors = 0

        #Control logging
        self.verboseLogging = False
        self.decode = True
        self.log_raw = False
        self.log_header = True

    def build_pkt(self, string, sequenceNum=None):
        """Build packet (add sequence and checksum)"""

        # Sequence number is supplied only for ACK packets - use normal txSequence otherwise
        if sequenceNum == None:
            sequenceNum = self.txSequence = (self.txSequence + 1) & 0xFF
            #print sequenceNum

        data = chr(sequenceNum) + string

        csum = (~(0xFF & sum(map(ord,data))) + 1) & 0xFF

        data = data + chr(csum)

        # Escape special chars
        data = ''.join(map(self._escape_data, data))

        # Add Framing
        data = SOH + data + END
        return data

    @staticmethod
    def _escape_data(char):
        """Apply special-char escape rules, and return escaped version of char"""
        return PacketSerialProtocol.escapes.get(char, char)

    def flush_write_buff(self):
        """This method should be overwritten by your own instance"""
        raise NotImplementedError()

    def get_intf_name(self):
        """Used for debugging messages"""
        return "Unknown"

    def on_data(self, rx_data, source_interface=interfaces.Interface()):
        """Called from external source upon receiving data"""
        # uncomment the following if you think you are receiving partial packets
        # if __debug__ and self.verboseLogging: log.debug("Received data %s" % (pprint.pformat(string)))
        match = None
        self._incoming_data += rx_data
        #pprint.pprint(rx_data)
        for match in re.finditer(SOH_END_RE, self._incoming_data):
            # I've done timming on using replace like this versus using another re but this appears to be faster (Python 2.6.6 1/20/2012)
            self.rxPkt = match.groups()[0].replace(ESC+ESC_SOH, SOH).replace(ESC+ESC_END, END).replace(ESC+ESC_ESC, ESC)
            try:
                self._process_rx_pkt(source_interface)
            except (SystemExit, KeyboardInterrupt):
                raise
            except:
                # We added this exception handler here in case some other code throws an exception while processing this data
                # If it throws an unhandled exception and we get here we will never get to the lines below where we clear the
                # buffer. Thus we will keep processing this bad packet over and over
                log.exception("An error occurred while trying to process received packet serial data:")
                self.numProcessErrors += 1
        if match:
            self._incoming_data = self._incoming_data[match.end():]

    def _packet_tx(self):
        self.curTxCount += 1
        self.tSend = datetime.datetime.now()
        assert self.curTxPkt
        self.write(self.curTxPkt)
        #pprint.pprint(self.curTxPkt)
        if self.neverAck:
            self.curTxPkt = None

    def _poll(self):
        """Periodic poll hook.  Checks for send timeout, initiates resend"""
        if self.curTxPkt:
            curTime = datetime.datetime.now()
            if curTime - self.tSend > self.sendTimeout:
                self.flush_write_buff()
                if self.curTxCount <= self.maxRetries:
                    if __debug__:
                        log.debug("RETRY #%i: %i" % (self.txSequence, self.curTxCount))
                    self._packet_tx()
                    self.numTxRetries += 1
                else:
                    self.numTxFail += 1
                    self.curTxCount = 0
                    self.curTxPkt = None
                    self.curOrigString = ''
                    self._notifyCallbacks("tx", False)
                    log.warning("SNAP communication timeout")
            return True
        return False

    def _process_rx_pkt(self, source_interface=interfaces.Interface()):
        """Given a packet with sequence prepended and cksum appended, make sense of it!"""
        if len(self.rxPkt) < 2:
            self.rxErrors += 1
            log.debug("Received packet less than 2 bytes")
            return
        elif self.check_len and len(self.rxPkt) > (SnapConstantsHelper.MAX_PKT_SIZE + 2): #+2=packet serial stuff
            self.rxErrors += 1
            log.debug("Received large packet with size %i, max size is %i" % (len(self.rxPkt), SnapConstantsHelper.MAX_PKT_SIZE))
            return

        statedCksum = ord(self.rxPkt[-1])
        self.rxPkt = self.rxPkt[:-1]
        calcSum = (~(0xFF & sum(map(ord,self.rxPkt))) + 1) & 0xFF
        if calcSum != statedCksum:
            self.rxErrors += 1
            log.debug("Received packet with incorrect checksum (%i), got %i" % (statedCksum, calcSum))
            if __debug__ and self.verboseLogging:
                log.debug("BAD Packet: %s" % (pprint.pformat(self.rxPkt)))
            return

        rxSequence = ord(self.rxPkt[0])
        self.rxPkt = self.rxPkt[1:]

        # Zero length packet is an ACK
        if len(self.rxPkt) == 0:
            self.numRxAcks += 1
            # Handle ACK packets, if we have an outstanding tx pkt
            if self.curTxPkt:
                if rxSequence == self.txSequence:
                    if __debug__ and self.verboseLogging: log.debug("Received ACK packet with sequence #%i" % (rxSequence))
                    self.numTxAcked += 1
                    self.curTxCount = 0
                    try:
                        pid = ord(self.curOrigString[SnapConstantsHelper.RPC_PID_OFS])
                    except IndexError:
                        pid = 0
                    #log.debug("pid = %i" % (pid))
                    self.curTxPkt = None
                    self.curOrigString = ''
                    self._notifyCallbacks("tx", True, pid != SnapConstantsHelper.PID_MESH_PTP)
                    if __debug__ and self.verboseLogging:
                        log.debug("NOTIFY about %i" % (rxSequence))
                elif __debug__ and self.verboseLogging:
                    log.debug("Did NOT receive the same sequence number as transmitted (#%i), got #%i" % (self.txSequence, rxSequence))
            elif __debug__ and self.verboseLogging:
                log.debug("Received ACK for sequence #%i and we DON'T have a current TX packet" % (rxSequence))
        else:
            self.numRxPkts += 1

            if __debug__ and self.log_raw:
                log.debug("Received packet #%i: %s" % (rxSequence, pprint.pformat(self.rxPkt)))

            # if it was not an xCast packet
            if len(self.rxPkt) > 2:
                rxPID = ord(self.rxPkt[SnapConstantsHelper.RPC_PID_OFS])
                if (not self.neverAck and rxPID not in UNACKD_PIDS):
                    # Acknowledge receipt of this data
                    self.send_ack(rxSequence)

                if __debug__ and self.decode:
                    (headers, sections) = Snap2Decoder.get_formatted_output(self.rxPkt)
                    log_str = "%s received packet #%i:\n%s" % (self.get_intf_name(), rxSequence, sections)
                    if self.log_header:
                        log_str = "%s received packet #%i:\n%s\n%s" % (self.get_intf_name(), rxSequence, headers, sections)
                    log.debug(log_str)
            else:
                #If the packet is still valid up to this point, ACK it
                #This is Portals way of figuring out if we are talking to a packet serial device
                self.send_ack(rxSequence)

            #if rxSequence > self.lastRxSequence+1:
                #pass
            if rxSequence != self.lastRxSequence:
                self.lastRxSequence = rxSequence
                # Pass the received packet to registrants
                self._notifyCallbacks("rx", self.rxPkt, source_interface)
            elif __debug__ and self.verboseLogging:
                log.debug("Received sequence number MATCHES last sequence number")

    def ready_to_send(self):
        raise NotImplementedError()

    def send_ack(self, sequenceNum):
        ackPkt = self.build_pkt('', sequenceNum)
        self.write(ackPkt)
        if __debug__ and self.verboseLogging: log.debug("send_ack #%i" % (sequenceNum))

    def send_data(self, string):
        """Encode the specified string and write it out the current interface"""
        #You should be calling ready_to_send before calling send_data
        assert self.ready_to_send()
        self.curOrigString = string
        self.curTxPkt = self.build_pkt(string)

        #Logging
        if __debug__:
            if self.log_raw: log.debug("%s packetTx #%i: %s" % (self.get_intf_name(), self.txSequence, pprint.pformat(self.curTxPkt)))
            if self.decode:
                (headers, sections) = Snap2Decoder.get_formatted_output(string)
                log_str = "%s packetTx #%i:\n%s" % (self.get_intf_name(), self.txSequence, sections)
                if self.log_header:
                    log_str = "%s packetTx #%i:\n%s\n%s" % (self.get_intf_name(), self.txSequence, headers, sections)
                log.debug(log_str)

        self._packet_tx()

        return self.txSequence

    def write(self, data):
        """This method should be overwritten by your own instance"""
        raise NotImplementedError()


class PacketSerialDriver(PacketSerialProtocol):
    """Encapsulate data for packetized transfer over a serial interface.
    Usage:
      Register "rx" and "tx" callbacks.  The 'send_data' call may only be made when 'ready_to_send()' returns True.
      This is indicated by the "tx" callback.

    Protocol Definition:
     SOH | Sequence byte| DATA | Checksum byte | END

      DATA field has special characters escaped - i.e. the characters SOH and END never appear in DATA.
        Escape Substitutions:
          SOH => ESC+ESC_SOH
          END => ESC+ESC_END
          ESC => ESC+ESC_ESC

      Checksum is over unescaped Sequence + DATA fields.  Sequence and Checksum bytes are escaped too.

    Acknowledge:
      Correctly received packets will be responded to with a zero-DATA packet matching the sent Sequence number.

    Timeouts / Retries:
      Sender may timeout and resend same sequence number (configurable timeout and resend limit).
    """

    def __init__(self, pyserDrv, pollSource):
        """
        pyserDrv -- A PyserialDriver instance to use for sending and receiving
        pollSource -- A PollSource, with which we can register for periodic polls
        """
        PacketSerialProtocol.__init__(self)
        self.serial = pyserDrv
        self.pollSource = pollSource
        self.serial.registerRxCallback(self.on_data)
        self.pollSource.registerPollCallback(self.handlePoll)

        if __debug__:
            #Used to help wing's autocomplete
            from serialwrapper import PyserialDriver
            isinstance(self.serial, PyserialDriver.PyserialWrapper)

    def flush_write_buff(self):
        """This method should be overwritten by your own instance"""
        return self.serial.flushWriteBuf()

    def get_intf_name(self):
        return self.serial.serial.portstr

    def handlePoll(self):
        if self.serial.isOpen:
            self._poll()

    def ready_to_send(self):
        if self.serial and self.serial.isOpen and self.curTxPkt is None:
            return True
        return False

    def write(self, data):
        return self.serial.write(data)


if __name__ == '__main__':
    import time, binascii
    import RpcCodec, PacketQueue

    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s:%(msecs)03d %(levelname)-8s %(name)-8s %(message)s', datefmt='%H:%M:%S')

    print "Test PacketSerialDriver"

    class MySerial:
        '''Crossover, baby'''
        def __init__(self, pollSrc):
            self.rxRegistry = []
            pollSrc.registerPollCallback(self.poll)

            # Buffers to forward to connected endpoint
            self.pendingBufs = []

            self.intermittentFail = 4
            self.failRate = 5
            self.isOpen = True

        def poll(self):
            for buf in self.pendingBufs:
                self.other.notifyRx(buf)
            self.pendingBufs = []

        def connect(self, other):
            'Connect to another serial endpoint'
            self.other = other

        def notifyRx(self, buf):
            for cb in self.rxRegistry:
                cb(buf)

        def registerRxCallback(self, callback):
            self.rxRegistry.append(callback)

        def write(self, buf):
            self.intermittentFail = (self.intermittentFail + 1) % self.failRate
            if self.intermittentFail == 0:
                buf = buf[:len(buf)/2]

            # Defer forwarding bufs to next poll (don't want reply to happen in context of caller's send() call)
            self.pendingBufs.append(buf)

        def flushWriteBuf(self):
            pass

    class MyPollSource:
        def __init__(self):
            self.registry = []
            #self.pollTimer = wx.PyTimer(self.poll)
            #self.pollTimer.Start(50, wx.TIMER_CONTINUOUS)

        def poll(self):
            for cb in self.registry:
                cb()

        def registerPollCallback(self, handler):
            self.registry.append(handler)

    class PsdCaller:
        def __init__(self, num, psd):
            self.num = num
            self.psd = psd
            psd.registerCallback("tx", self.tx)
            psd.registerCallback("rx", self.rx)
            self.txQueue = []

        def rx(self,buf):
            print "PSD %d: rx = %s" % (self.num, pprint.pformat(buf))

        def tx(self,status):
            print "PSD %d: tx status = %d" % (self.num, int(status))
            if self.txQueue:
                self.psd.send_data(self.txQueue.pop(0))

        def startTx(self, txQueue):
            self.txQueue = txQueue
            self.psd.send_data(self.txQueue.pop(0))


    # Create two serial test-drivers, and cross-connect them
    poller = MyPollSource()
    ser1 = MySerial(poller)
    ser2 = MySerial(poller)
    ser1.connect(ser2)
    ser2.connect(ser1)

    # Instantiate a couple of PSDs, connected to these serial drivers
    psd1 = PacketSerialDriver(ser1, poller)
    psd2 = PacketSerialDriver(ser2, poller)

    # Hook a couple of simple "users" to the PSDs
    caller1 = PsdCaller(1, psd1)
    caller2 = PsdCaller(2, psd2)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # ----- Now let's hook up the RpcCodec, and test some calls -----
    class CallSink:
        def __init__(self, num):
            self.num = num

        class CallProxy(object):
            def __init__(self, cs, name):
                self.fname = name
                self.cs = cs
            def __call__(self, *args):
                print "CallSink%d (from %s) :  %s%s" % (self.cs.num, binascii.hexlify(self.cs.remoteAddr), self.fname, args)

        def __getattr__(self, name):
            return CallSink.CallProxy(self, name)

    rpcCallSink1 = CallSink(1)
    rpcCallSink2 = CallSink(2)

    queue1 = PacketQueue.Connector(rpcCallSink1, psd1)
    queue2 = PacketQueue.Connector(rpcCallSink2, psd2)

    rpcDescr1 = RpcCodec.RpcDescriptor()
    rpcDescr1.srcAddr = '\x12\x34\x56'
    rpcDescr1.dstAddr = '\xAB\xCD\xEF'
    rpcDescr1.funcName = "foo"
    rpcDescr1.args = (1, 2, True, "Baz", None)

    queue1.rpcReady(rpcDescr1)

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Let the simulation begin!\
    while True:
        poller.poll()
        time.sleep(0.050)
