# (c) Copyright 2007-2012 Synapse Wireless, Inc.
from __future__ import with_statement

import logging
log = logging.getLogger('snaplib.ScriptsManager')
import sys, os, types, binascii, bz2, traceback, math

import SnapConstants
from snappy import hooks

import Globleton
import SnappyGen
import crcUtil

from ppcrypto import ARC4


IMAGES_DIR_NAME = 'snappyImages'
SNAPPY_EXPORT_EXTENSION = '.spy'
SNAPPY_FILE_EXTENSION = '.py'
_SNAPPY_EXPORT_KEY = "SynapseExport"


class SNAPpy_sys(object):
  def __init__(self, platform='', version='2.1.0'):
    self.platform = platform
    self.version = version

  def __eq__(self, other):
    if (self.platform == other.platform and
        self.version == other.version):
      return True
    return False

  def __hash__(self):
    return hash(self.platform + self.version)

  def __ne__(self, other):
    if (self.platform != other.platform or
        self.version != other.version):
      return True
    return False

  def __str__(self):
    return "%s %s" % (self.platform, self.version)


class SNAPpy_Info(object):
  def __init__(self):
    self.image_string = None  # Should setting the string clear the exception and vice-versa
    self.err_title = None
    self.exception = None
    self.filepath = None
    self._err_desc = None
    self._err_lineNo = None
    self._err_obj = None
    self._err_fname = None
    self.sGen = None
    self.lint_err_desc = None
    self.lint_err_title = None
    self.lint_lineNo = None
    self.lint_filepath = None
    self.is_spy = False

  def run_snappy_lint(self):
    found = False

    #Check for locals using the same name as a global
    global_var_names = set([obj.name for obj in self.sGen.imageContents.itervalues() if not obj.isFunc and not obj.isCFunc])
    builtin_var_name = set([obj.name for obj in self.sGen.imageContents.itervalues() if obj.isCFunc])
    local_var_names = set(self.sGen.localLines.iterkeys())
    common_var_names = local_var_names.intersection(global_var_names)
    local_vars_named_builtin = local_var_names.intersection(builtin_var_name)
    if len(common_var_names) > 0:
      found = True
      var_name = common_var_names.pop()
      (self.lint_lineNo, self.lint_filepath) = self.sGen.localLines[var_name][0]
      self.lint_err_desc = "The local variable named %s on line %i is also defined as a global" % (var_name, self.lint_lineNo)
      if self.lint_filepath != "<string>":
        self.lint_err_desc += " in file:\n%s" % (self.lint_filepath)
      self.lint_err_title = "Local and Global Name"
    if len(local_vars_named_builtin) > 0:
      found = True
      var_name = local_vars_named_builtin.pop()
      (self.lint_lineNo, self.lint_filepath) = self.sGen.localLines[var_name][0]
      self.lint_err_desc = "The local variable named %s on line %i is also a built-in function name" % (var_name, self.lint_lineNo)
      if self.lint_filepath != "<string>":
        self.lint_err_desc += " in file:\n%s" % (self.lint_filepath)
      self.lint_err_title = "Local and Built-In Name"

    return found

  def get_code_size(self):
    if self.sGen:
      return self.sGen.imageCodeSize

  def get_crc(self):
    if self.sGen:
      return self.sGen.imageCrc
    elif self.is_spy:
      (filename_and_path, extension) = os.path.splitext(self.filepath)
      return get_spy_crc(filename_and_path + '.spy')

  def get_data_size(self):
    if self.sGen:
      return self.sGen.imageDataSize

  def get_err_desc(self):
    if (self._err_desc is None and
        self.exception is not None and
        hasattr(self.exception, 'descr')):
      return self.exception.descr
    return self._err_desc

  def get_err_fname(self):
    if (self._err_fname is None and
        self.exception is not None):
      # Gotta love how many different ways execeptions save the filename
      if hasattr(self.exception, 'fname'):
        return self.exception.fname
      elif hasattr(self.exception, 'filepath'):
        return self.exception.filepath
      elif hasattr(self.exception, 'filename'):
        return self.exception.filename
      elif self.filepath is not None:
        return self.filepath
    return self._err_fname

  def get_err_lineNo(self):
    if (self._err_lineNo is None and
        self.exception is not None and
        hasattr(self.exception, 'lineNo')):
      return self.exception.lineNo
    return self._err_lineNo

  def get_err_obj(self):
    if (self._err_obj is None and
        self.exception is not None and
        hasattr(self.exception, 'obj')):
      return self.exception.obj
    return self._err_obj

  def get_export_string(self):
    if self.image_string is not None:
      imageString = bz2.compress(self.image_string)
      if imageString:
        encoder = ARC4(_SNAPPY_EXPORT_KEY+str(math.pi))
        return encoder.encrypt(imageString)

  def get_extraDict(self):
    extraDict = {}

    if self.filepath is not None:
      extraDict['filepath'] = self.filepath
    if self.err_lineNo is not None:
      extraDict['lineNo'] = self.err_lineNo
    if self.err_obj is not None:
      extraDict['obj'] = self.err_obj
    if self.err_fname is not None:
      extraDict['fname'] = self.err_fname

    return extraDict

  def get_funcs(self, features=SnappyGen.SnappyFeatureSet()):
    funcList = []

    if self.sGen is not None:
      funcList = [obj for obj in self.sGen.imageContents.itervalues() if obj.isFunc or obj.isCFunc]
      funcList.sort(key=lambda x: x.name.upper())
      funcList.sort(key=lambda x: x.funcModule.upper())
    else:
      sGen = SnappyGen.SnappyGen(features)
      sGen.build(imageName='')
      funcList = [obj for obj in sGen.imageContents.itervalues() if obj.isFunc or obj.isCFunc]
      funcList.sort(key=lambda x: x.name)

    return funcList

  def get_module_doc_string(self, file_path):
    if self.sGen is not None:
      if file_path == self.filepath:
        return self.sGen.doc_string

      for (name, module) in self.sGen.dependencies:
        if (module is not None and
            module.__file__ == file_path and
            module.__doc__ is not None):
          return module.__doc__.strip()

    return ""

  def get_warnings(self):
    if self.sGen:
      return self.sGen.warnings
    return []

  def genMapFile(self):
    if (self.exception is None and
        self.sGen is not None and
        self.image_string is not None):
      return self.sGen.genMapFile(Globleton.userFilesPath)
    return False

  def get_size(self):
    if self.sGen:
      return self.sGen.imageSize

  def set_err_desc(self, val):
    self._err_desc = val

  def set_err_fname(self, val):
    self._err_fname = val

  def set_err_lineNo(self, val):
    self._err_lineNo = val

  def set_err_obj(self, val):
    self._err_obj = val

  codeSize = property(get_code_size)
  crc = property(get_crc)
  dataSize = property(get_data_size)
  err_desc = property(get_err_desc, set_err_desc)
  err_lineNo = property(get_err_lineNo, set_err_lineNo)
  err_obj = property(get_err_obj, set_err_obj)
  err_fname = property(get_err_fname, set_err_fname)
  extraDict = property(get_extraDict)
  funcs = property(get_funcs)
  size = property(get_size)
  export_string = property(get_export_string)
  warnings = property(get_warnings)

def exec_file(filename, globals):
  """Run the builtin execfile our way"""
  if os.path.isfile(filename):
    execfile(filename, globals)
  else:
    # Add \npass\n to fix Python issue with having a comment on the last line, see ticket #1632
    exec "%s\npass\n" % (filename.replace('\r\n', '\n')) in globals

def clearCache(newMods, globals):
  """
  Clean up module cache droppings.
  """
  #print newMods
  globals['_sysCacheRemoved'] = {}
  for name,mod in newMods:
    try:
      # Remove from system modules cache but add to a dict to keep a reference
      # See this for more info: http://mail.python.org/pipermail/python-dev/2007-January/070690.html
      # Essentially it's only safe to delete the module once snappy gen is done otherwise deleting the module
      # when a function's func_globals references something in the module creates problems when GC happens
      globals['_sysCacheRemoved'][name] = sys.modules[name]
      del sys.modules[name]
    except:
      pass

    # Delete .pyc file if found
    try:
      if mod.__file__[-4:] == '.pyc':
        os.remove(mod.__file__)
      elif mod.__file__[-3:] == '.py':
        os.remove(mod.__file__ + 'c')
    except:
      pass

def execFileNoCache(filename, globals):
  """
  Run execfile, cleaning up module cache droppings.
  Returns the set of additional dependencies this file required.
  """
  curMods = set(sys.modules.items())
  #print curMods

  #This is to get around having to modify pyinstaller since it caches filenames in directories
  # REMOVED. Obsoleted by new pyinstaller
  #if hasattr(sys, 'frozen') and hasattr(Globleton, 'userFilesPath') and not sys.platform.startswith('darwin'):
  #  import iu
  #  for dirname in iu._os_listdir.func_defaults[0].keys():
  #    if dirname.startswith(Globleton.userFilesPath):
  #      #print "REMOVING: ", dirname
  #      iu._os_listdir.func_defaults[0].pop(dirname, None)

  try:
    exec_file(filename, globals)
  finally:
    newMods = set(sys.modules.items()) - curMods
    clearCache(newMods, globals)

  return newMods

def _addSnappyObjects(globals, sgen):
  """Scrape funcs and global vars from 'globals' dict, and add to snappyGen image"""
  other_globals = dict()

  for (name, val) in globals.iteritems():
    if isinstance(val, types.FunctionType):
      sgen.addGlobal(val)
      other_globals.update(val.func_globals)
    elif type(val) in SnapConstants.PY_TYPES:
      sgen.addGlobal(val, name)
    else:
      pass

  main_set = set(sgen.initObjs.iterkeys())
  if other_globals:
    new_set = set([name for name in other_globals.iterkeys() if isinstance(other_globals[name], types.FunctionType)])
    diff = new_set - main_set
    if diff:
      _addSnappyObjects(dict([(name, other_globals[name]) for name in diff]), sgen)

def _getSnappyImageString(file,
                          sGen=None,
                          name='RawText',
                          featureSet=SnappyGen.SnappyFeatureSet(),
                          snap_sys=SNAPpy_sys()):
  if os.path.isfile(file):
    (name, ext) = os.path.splitext(os.path.basename(file))

  if sGen == None:
    sGen = SnappyGen.SnappyGen(featureSet)

  execGlobals = {'snappyGen': sGen,
                 'types': types,
                 'SnapConstants': SnapConstants,
                 'setHook': sGen.decSetHook,
                 'bindElf': sGen.decBindElf}
  try:
    from synapse import snapsys
    snapsys.platform = str(snap_sys.platform)
    snapsys.version = str(snap_sys.version)
  except ImportError:
    pass

  #Add hooks to globals
  for hook in dir(hooks):
    if hook[0] != '_':
      execGlobals[hook] = getattr(hooks, hook)

  sGen.dependencies = execFileNoCache(file, execGlobals)
  _addSnappyObjects(execGlobals, sGen)
  sGen.build(imageName=name)
  sGen.doc_string = execGlobals.get('__doc__', '').strip()

  return (sGen.getImageString(), sGen)

def getPortalFunctionInfo():
  rtnVals = []

  try:
    for value in Globleton.portalSandbox.itervalues():
      if isinstance(value, types.FunctionType):
        snappyObj = SnappyGen.SnappyGen.SnappyGlobalObject()
        snappyObj.name = value.func_name
        snappyObj.funcDoc = value.func_doc
        snappyObj.isFunc = True
        snappyObj.funcArgs = value.func_code.co_varnames[:value.func_code.co_argcount]
        snappyObj.funcModule = os.path.splitext(os.path.basename(Globleton.scriptsBaseFile))[0]
        snappyObj.funcModulePath = Globleton.scriptsBaseFile
        rtnVals.append(snappyObj)
  except:
    pass

  rtnVals.sort(cmp=lambda x,y: cmp(x.name.lower(), y.name.lower()))

  return rtnVals

def getPortalFunction(funcName):
  return Globleton.portalSandbox.get(funcName)

def getPortalFunctionNames():
  rtnVals = []

  try:
    rtnVals = [value.func_name for value in Globleton.portalSandbox.itervalues() if isinstance(value, types.FunctionType)]
  except:
    pass

  rtnVals.sort()

  return rtnVals

def getSnappyImages(includeExports=True, extra_path=None, add_path=False):
  files = []
  if extra_path is None:
    path = os.path.join(Globleton.userFilesPath, IMAGES_DIR_NAME)
  else:
    path = os.path.join(Globleton.userFilesPath, IMAGES_DIR_NAME, extra_path)

  if os.path.isdir(path):
    for file in os.listdir(path):
      (root, ext) = os.path.splitext(file)
      if ext == SNAPPY_FILE_EXTENSION or (includeExports and ext == SNAPPY_EXPORT_EXTENSION):
        if add_path:
          file = os.path.join(extra_path, file)
        files.append(file)
      #elif os.path.isdir(os.path.join(path, file)):
        #files.extend(getSnappyImages(extra_path=file, add_path=True))

    # Now that we allow the working directory to be changed on the fly, they might have a direcotry with NO files in it
    if len(files) != 0:
        if isinstance(file[-1], str):
          files.sort(key=str.lower)
        else:
          files.sort(key=unicode.lower)

  return files

def getSnappyStringFromExport(exportString):
  encoder = ARC4(_SNAPPY_EXPORT_KEY+str(math.pi))
  exportString = encoder.decrypt(exportString)
  return bz2.decompress(exportString)

def getNameFromString(string):
  return string[3:int(binascii.hexlify(string[2]),16)+3]

def get_spy_crc(path_and_filename):
  """
  Returns the CRC of the passed in SNAPpy script file (.spy)

  path_and_filename : The filename, including directory, of the exported SNAPpy script
  """
  if os.path.isfile(path_and_filename):
    with open(path_and_filename, 'rb') as f:
      try:
        data = getSnappyStringFromExport(f.read())
      except IOError:
        return None # Something happend while decompressing the data trying to get the CRC
    return crcUtil.getCrc(data[:-2])  # I think we are calculating our own CRC here because the value in the export image you have to worry about endianess


_BLANK_SIZE = 0
_BLANK_CODE_SIZE = 0
_BLACK_DATA_SIZE = 0

def calc_blank_sizes():
  global _BLANK_SIZE, _BLANK_CODE_SIZE, _BLACK_DATA_SIZE
  if _BLANK_SIZE == 0:
    try:
      (imgstr, sGen) = _getSnappyImageString('', name='a')
      _BLANK_SIZE = int(binascii.hexlify(imgstr[0:2]), 16)
      _BLANK_CODE_SIZE = sGen.imageCodeSize
      _BLACK_DATA_SIZE = sGen.imageDataSize
      del imgstr
    except:
      if __debug__:
        raise
      pass

def get_blank_size():
  calc_blank_sizes()
  return _BLANK_SIZE

def get_blank_code_size():
  calc_blank_sizes()
  return _BLANK_CODE_SIZE

def get_blank_data_size():
  calc_blank_sizes()
  return _BLACK_DATA_SIZE

BLANK_SIZE = property(get_blank_size)

def get_snappy_info(file,
                    sGen=None,
                    name='RawText',
                    featureSet=SnappyGen.SnappyFeatureSet(),
                    snap_sys=SNAPpy_sys(),
                    map_path=None):
  info = SNAPpy_Info()
  (filename_and_path, extension) = os.path.splitext(file)
  try:
    if os.path.isfile(file):
      info.filepath = file
    elif os.path.isfile(filename_and_path + '.spy'):
      info.is_spy = True
      info.filepath = file
    else:
      info.filepath = os.path.join(Globleton.userFilesPath, IMAGES_DIR_NAME, name + '.py')

    if not info.is_spy:
      (info.image_string, info.sGen) = _getSnappyImageString(file, sGen, name, featureSet, snap_sys)
      if not info.image_string:
        info.exception = SnappyGen.SnappyError("A valid SNAPpy image was unable to be created")
        info.err_title = "Invalid SNAPpy Image"
        info.err_desc = "A valid SNAPpy image was unable to be created"
      elif map_path is not None and os.path.isdir(map_path):
        info.sGen.genMapFile(map_path)
  except SnappyGen.SnappyCatalogError, e:
    info.exception = e
    info.err_title = "Script Error"
    info.err_desc = "An error occurred on line %i while creating the SNAPpy image:\n%s" % (e.lineNo, e.descr)
  except SnappyGen.SnappyObjError, e:
    info.exception = e
    info.err_desc = "An error occurred while creating the SNAPpy image:\n%s" % (e.descr)
    info.err_title = "Object Error"
  except SnappyGen.SnappyCodeError, e:
    #extraDict['filepath'] = file
    info.exception = e
    info.err_title = "Code Error"
    info.err_desc = "Error in script, %s: %s at Line: %d" % (e.fname, e.descr, e.lineNo)
  except SyntaxError, ex:
    if ex.lineno is None:
      ex.lineno = 0 #We need to have a int
    if ex.filename is None or ex.filename == '<string>':
      ex.filename = info.filepath
    info.err_lineNo = ex.lineno #Go figure, this exception uses lower case N

    info.exception = ex
    info.err_desc = "A syntax error on line %i was found in the file %s:\n%s" % (ex.lineno, ex.filename, ex.msg)
    info.err_title = "Syntax Error"
  except (KeyboardInterrupt, SystemExit):
      raise
  except Exception, ex:
    if __debug__:
      log.exception("ScriptsManager Error:")
    last_tb = tb = sys.exc_info()[2]
    while tb.tb_next is not None:
      last_tb = tb
      tb = tb.tb_next

    if (hasattr(sys, 'frozen') and
        (
          (not os.path.isfile(file) and
           not isinstance(ex, NameError))
        or
        isinstance(ex, ImportError)
        )):

      #PyInstaller has extra import hook
      info.err_lineNo = last_tb.tb_lineno
      if hasattr(last_tb, 'tb_frame') and hasattr(last_tb.tb_frame, 'f_code') and hasattr(last_tb.tb_frame.f_code, 'co_filename'):
        info.err_fname = last_tb.tb_frame.f_code.co_filename
      info.err_desc = "An error occurred on line %i while creating the SNAPpy image" % (last_tb.tb_lineno)
    else:
      info.err_lineNo = tb.tb_lineno
      if hasattr(tb, 'tb_frame') and hasattr(tb.tb_frame, 'f_code') and hasattr(tb.tb_frame.f_code, 'co_filename'):
        info.err_fname = tb.tb_frame.f_code.co_filename
      info.err_desc = "An error occurred on line %i while creating the SNAPpy image" % (tb.tb_lineno)

    if hasattr(ex, 'message'):
      info.err_desc += ':\n%s' % ex.message
    info.err_title = "Unable to Create SNAPpy Image"
    info.exception = ex

  return info
