# (c) Copyright 2010-2011, Synapse Wireless, Inc.


import pprint

import SnapConstantsHelper
import RpcCodec
import crcUtil

from snappy import BuiltIn

crypto = None


# From ProtocolId.h
# Protocol IDs
PID_RESERVED = 0x00
PID_DM_BCAST = 0x01
PID_DM_PTP = 0x02
PID_NAK = 0x03
PID_RPC = 0x04
PID_ACK = 0x05 # was PID_RPC_ACK but now there is only ONE type of ACK
PID_MESH_BCAST = 0x06
PID_MESH_PTP = 0x07
PID_RPC_BCAST = 0x08
PID_TRACEROUTE = 0x09

# Keep advancing this as PIDs get added!
PID_MAX_VALUE = 0x09         # Use by MAC layer to filter out invalid PIDs

ENCRYPTED_FLAG = 0x80


# These are IN-MEMORY offsets, not OTA offsets (OTA, all packets start with a 2 byte NID)

# These offsets apply to all SNAPv2 packets
PKT_LENGTH = 0
PKT_PID = 1

# These offsets apply to all UNICAST packets
# For non-forwardable packets, these are the ONLY addresses present.
# You can think of them as the "immediate" dest and src.
# You can also think of them as the "original" src and the "final" dest
# When these fields are re-used by FORWARDABLE packets, these represent
# the immediate source and destination.
UNI_PKT_DEST_HI = 2
UNI_PKT_DEST_MID = 3
UNI_PKT_DEST_LOW = 4

UNI_PKT_SRC_HI = 5
UNI_PKT_SRC_MID = 6
UNI_PKT_SRC_LOW = 7

UNI_PKT_SEQ = 8

UNI_PKT_DATA = 9

# You can think of FWD (forwardable) packet headers as "extending" the non-FWD ones.
# These *additional* offsets apply to all FORWARDABLE UNICAST packets
# This address represents the FINAL destination for the packet
FWD_UNI_PKT_FINAL_DEST_HI = 9
FWD_UNI_PKT_FINAL_DEST_MID = 10
FWD_UNI_PKT_FINAL_DEST_LOW = 11
# This address represents the ORIGINAL source of the packet
FWD_UNI_PKT_ORIG_SRC_HI = 12
FWD_UNI_PKT_ORIG_SRC_MID = 13
FWD_UNI_PKT_ORIG_SRC_LOW = 14

FWD_UNI_PKT_DATA = 15


# These offsets apply to all MULTICAST packets except MESH Broadcasts
MULTI_PKT_DEST_MSB = 2
MULTI_PKT_DEST_LSB = 3

MULTI_PKT_TTL = 4

MULTI_PKT_SRC_HI = 5
MULTI_PKT_SRC_MID = 6
MULTI_PKT_SRC_LOW = 7

MULTI_PKT_SEQ = 8

# Mesh Routing's broadcast packet header ends here
MESH_MULTI_PKT_DATA = 9
# Most broadcast/multicast headers keep going...

MULTI_PKT_ORIG_SRC_HI = 9
MULTI_PKT_ORIG_SRC_MID = 10
MULTI_PKT_ORIG_SRC_LOW = 11

MULTI_PKT_DATA = 12


# Multicast Groups are specified as a bit mask. Broadcast is the least significant bit
#typedef U16 McastGroup;
BROADCAST_GROUP = 0x0001


# SNAPv2 uses three-byte addressing
ADDRESS_SIZE = 3

ADDRESS_HI = 0
ADDRESS_MID = 1
ADDRESS_LOW = 2

MAC_ADDR_LENGTH = 8
NETWORK_ID_LENGTH = 2
CHANNEL_LENGTH = 1

# RPC constants
FWD_UNI_RPC_ARGC_OFFSET = (FWD_UNI_PKT_DATA)
FWD_UNI_RPC_ARGV_OFFSET = (FWD_UNI_PKT_DATA + 1)

MCAST_RPC_ARGC_OFFSET = (MULTI_PKT_DATA)
MCAST_RPC_ARGV_OFFSET = (MULTI_PKT_DATA + 1)


# Multicast Data Mode constants
MCAST_DM_PAYLOAD_OFFSET = (MULTI_PKT_DATA)

# RREQ and RREP messages have the same general structure
# The spec refers to both of these as RMs (Routing Messages)
# RM message offsets (relative to xxx_PKT_DATA)
MSG_ID = 0
HOP_LIMIT = 1
TARGET_HI = 2
TARGET_MID = 3
TARGET_LOW = 4
TARGET_SEQ_MSB = 5
TARGET_SEQ_LSB = 6
TARGET_DIST = 7
ORIGINATOR_HI = 8
ORIGINATOR_MID = 9
ORIGINATOR_LOW = 10
ORIG_SEQ_MSB = 11
ORIG_SEQ_LSB = 12
ORIGINATOR_DIST = 13

BASE_RREQ_MSG_SIZE = 14
BASE_RERR_MSG_SIZE = 7

TRACEROUTE_MSG_ID = 0
TRACEROUTE_SEQ = 1
TRACEROUTE_TIME_MSB = 2 # outbound timestamp, inbound elapsed time
TRACEROUTE_TIME_LSB = 3 # the rest of the time field
TRACEROUTE_ADDRESS = 4
TRACEROUTE_LQ = 7

dm_payload_offset_map = {PID_DM_BCAST : MCAST_DM_PAYLOAD_OFFSET, PID_DM_PTP : FWD_UNI_PKT_DATA}

def get_decoded_pkt_sections(pkt, decrypted=False):
    assert len(pkt) > 3
    pkt_pid = ord(pkt[PKT_PID])

    #Used in ALL packets
    headers = ['Len', 'Type']
    sections = ['%i' % (ord(pkt[0])), '%s' % (SnapConstantsHelper.getPidString(pkt_pid))]

    # Common Packet section
    if (len(pkt) > 9 and
        (pkt_pid == PID_MESH_BCAST or
        pkt_pid == (PID_MESH_BCAST | ENCRYPTED_FLAG) or
        pkt_pid == PID_RPC_BCAST or
        pkt_pid == (PID_RPC_BCAST | ENCRYPTED_FLAG) or
        pkt_pid == PID_DM_BCAST or
        pkt_pid == (PID_DM_BCAST | ENCRYPTED_FLAG))):
        headers.extend(('Group', 'TTL', 'Source', 'Seq'))
        sections.extend(('%02X.%02X' % (ord(pkt[MULTI_PKT_DEST_MSB]), ord(pkt[MULTI_PKT_DEST_LSB])),
                         '%2i' % (ord(pkt[MULTI_PKT_TTL])),
                         '%02X.%02X.%02X' % (ord(pkt[MULTI_PKT_SRC_HI]), ord(pkt[MULTI_PKT_SRC_MID]), ord(pkt[MULTI_PKT_SRC_LOW])),
                         '%02X' % (ord(pkt[MULTI_PKT_SEQ])),
                         ))
        if crypto and crypto.has_valid_key and pkt_pid & ENCRYPTED_FLAG:
            pkt_pid ^= ENCRYPTED_FLAG
            if not decrypted:
                if pkt_pid == PID_MESH_BCAST:
                    pkt = (pkt[:MESH_MULTI_PKT_DATA] + crypto.decrypt(ord(pkt[MULTI_PKT_SEQ]), pkt[MESH_MULTI_PKT_DATA:]))[:ord(pkt[0])+1]
                elif pkt_pid in (PID_RPC_BCAST, PID_DM_BCAST):
                    pkt = (pkt[:MULTI_PKT_DATA] + crypto.decrypt(ord(pkt[MULTI_PKT_SEQ]), pkt[MULTI_PKT_DATA:]))[:ord(pkt[0])+1]
                else:
                    assert False
    elif (len(pkt) > 9 and
          (pkt_pid == PID_ACK or
          pkt_pid == PID_NAK or
          pkt_pid == PID_RPC or
          pkt_pid == (PID_RPC | ENCRYPTED_FLAG) or
          pkt_pid == PID_MESH_PTP or
          pkt_pid == (PID_MESH_PTP | ENCRYPTED_FLAG) or
          pkt_pid == PID_DM_PTP or
          pkt_pid == (PID_DM_PTP | ENCRYPTED_FLAG) or
          pkt_pid == PID_TRACEROUTE or
          pkt_pid == (PID_TRACEROUTE | ENCRYPTED_FLAG))):
        headers.extend(('Dst Addr', 'Src Addr', 'Seq'))
        sections.extend(('%02X.%02X.%02X' % (ord(pkt[UNI_PKT_DEST_HI]), ord(pkt[UNI_PKT_DEST_MID]), ord(pkt[UNI_PKT_DEST_LOW])),
                         '%02X.%02X.%02X' % (ord(pkt[UNI_PKT_SRC_HI]), ord(pkt[UNI_PKT_SRC_MID]), ord(pkt[UNI_PKT_SRC_LOW])),
                         '%02X' % (ord(pkt[UNI_PKT_SEQ]))
                         ))
        if crypto and crypto.has_valid_key and pkt_pid & ENCRYPTED_FLAG:
            pkt_pid ^= ENCRYPTED_FLAG
            if not decrypted:
                if pkt_pid == PID_MESH_PTP:
                    pkt = (pkt[:UNI_PKT_DATA] + crypto.decrypt(ord(pkt[UNI_PKT_SEQ]), pkt[UNI_PKT_DATA:]))[:ord(pkt[0])+1]
                elif pkt_pid in (PID_DM_PTP, PID_RPC, PID_TRACEROUTE):
                    pkt = (pkt[:FWD_UNI_PKT_DATA] + crypto.decrypt(ord(pkt[UNI_PKT_SEQ]), pkt[FWD_UNI_PKT_DATA:]))[:ord(pkt[0])+1]
                else:
                    assert False

    # Specific packet sections
    if (pkt_pid == PID_MESH_BCAST or
        pkt_pid == PID_MESH_PTP):
        headers.extend(('Msg', 'Hop Limit'))
        sections.extend(('%c' % (chr(ord(pkt[MESH_MULTI_PKT_DATA + MSG_ID]))),
                         '%2i' % (ord(pkt[MESH_MULTI_PKT_DATA + HOP_LIMIT])),
                         ))
        additional_addrs = 'None'
        if pkt_pid == PID_MESH_BCAST:
            if chr(ord(pkt[MESH_MULTI_PKT_DATA + MSG_ID])) == 'Q':
                if len(pkt) > MESH_MULTI_PKT_DATA + BASE_RREQ_MSG_SIZE:
                    i = 0
                    additional_addrs = ''
                    for c in pkt[MESH_MULTI_PKT_DATA + BASE_RREQ_MSG_SIZE:]:
                        additional_addrs += '%02X' % (ord(c))
                        if i < 2:
                            additional_addrs += '.'
                            i += 1
                        else:
                            additional_addrs += ' '
                            i = 0

        if (chr(ord(pkt[MESH_MULTI_PKT_DATA + MSG_ID])) == 'Q' or
            chr(ord(pkt[MESH_MULTI_PKT_DATA + MSG_ID])) == 'P'):
            headers.extend(('Target', 'Seq', 'Dist', 'Originator', 'Seq', 'Dist', 'Additional'))
            sections.extend(('%02X.%02X.%02X' % (ord(pkt[MESH_MULTI_PKT_DATA + TARGET_HI]), ord(pkt[MESH_MULTI_PKT_DATA + TARGET_MID]), ord(pkt[MESH_MULTI_PKT_DATA + TARGET_LOW])),
                             '%04X' % ((ord(pkt[MESH_MULTI_PKT_DATA + TARGET_SEQ_MSB]) << 8) + ord(pkt[MESH_MULTI_PKT_DATA + TARGET_SEQ_LSB])),
                             '%2i' % (ord(pkt[MESH_MULTI_PKT_DATA + TARGET_DIST])),
                             '%02X.%02X.%02X' % (ord(pkt[MESH_MULTI_PKT_DATA + ORIGINATOR_HI]), ord(pkt[MESH_MULTI_PKT_DATA + ORIGINATOR_MID]), ord(pkt[MESH_MULTI_PKT_DATA + ORIGINATOR_LOW])),
                             '%04X' % ((ord(pkt[MESH_MULTI_PKT_DATA + ORIG_SEQ_MSB]) << 8) + ord(pkt[MESH_MULTI_PKT_DATA + ORIG_SEQ_LSB])),
                             '%2i' % (ord(pkt[MESH_MULTI_PKT_DATA + ORIGINATOR_DIST])),
                             additional_addrs,
                             ))
        elif chr(ord(pkt[MESH_MULTI_PKT_DATA + MSG_ID])) == 'E':
            additional_addrs = 'None'
            addrs_start = MESH_MULTI_PKT_DATA + BASE_RERR_MSG_SIZE
            if len(pkt) > addrs_start:
                additional_addrs = ''
                for i in range(addrs_start, len(pkt), ADDRESS_SIZE+2):
                    additional_addrs += "%02X.%02X.%02X:%04X " % (ord(pkt[i]),
                                                                  ord(pkt[i+1]),
                                                                  ord(pkt[i+2]),
                                                                  (ord(pkt[i+3]) << 8) + ord(pkt[i+4]))

            headers.extend(('Unreachable Addr', 'Seq', 'Additional'))
            sections.extend(('%02X.%02X.%02X' % (ord(pkt[MESH_MULTI_PKT_DATA + TARGET_HI]), ord(pkt[MESH_MULTI_PKT_DATA + TARGET_MID]), ord(pkt[MESH_MULTI_PKT_DATA + TARGET_LOW])),
                             '%04X' % ((ord(pkt[MESH_MULTI_PKT_DATA + TARGET_SEQ_MSB]) << 8) + ord(pkt[MESH_MULTI_PKT_DATA + TARGET_SEQ_LSB])),
                             additional_addrs,
                             ))
    elif (pkt_pid == PID_RPC or
          pkt_pid == PID_DM_PTP or
          pkt_pid == PID_TRACEROUTE):
        headers.extend(('Final Dst', 'Orig Src'))
        sections.extend(('%02X.%02X.%02X' % (ord(pkt[FWD_UNI_PKT_FINAL_DEST_HI]), ord(pkt[FWD_UNI_PKT_FINAL_DEST_MID]), ord(pkt[FWD_UNI_PKT_FINAL_DEST_LOW])),
                         '%02X.%02X.%02X' % (ord(pkt[FWD_UNI_PKT_ORIG_SRC_HI]), ord(pkt[FWD_UNI_PKT_ORIG_SRC_MID]), ord(pkt[FWD_UNI_PKT_ORIG_SRC_LOW])),
                         ))
        argc_offset = FWD_UNI_RPC_ARGC_OFFSET
        dm_payload_offset = FWD_UNI_PKT_DATA
    elif (len(pkt) > 12 and
          (pkt_pid == PID_RPC_BCAST or
          pkt_pid == PID_DM_BCAST)):
        headers.append("Orig Src")
        sections.append('%02X.%02X.%02X' % (ord(pkt[MULTI_PKT_ORIG_SRC_HI]), ord(pkt[MULTI_PKT_ORIG_SRC_MID]), ord(pkt[MULTI_PKT_ORIG_SRC_LOW])))
        argc_offset = MCAST_RPC_ARGC_OFFSET
        dm_payload_offset = MCAST_DM_PAYLOAD_OFFSET

    if (pkt_pid == PID_RPC or
        (pkt_pid == PID_RPC_BCAST and len(pkt) > 12)):
        headers.extend(('#Args', 'Func Type', 'Func Val'))
        sections.extend(('%i' % (ord(pkt[argc_offset])),
                         '%i (%s)' % (ord(pkt[argc_offset+1]), SnapConstantsHelper.get_arg_type_string(ord(pkt[argc_offset+1]))),
                         ))

        next_arg_offset = argc_offset+4
        crc_loc = argc_offset+((ord(pkt[argc_offset])+1)*3)+1

        int_val = ord(pkt[argc_offset+2]) << 8 | ord(pkt[argc_offset+3])
        int_val = ((int_val + 32768) % 65536) - 32768 #Convert to signed int
        str_len = 0
        if ord(pkt[argc_offset+1]) == SnapConstantsHelper.TC_CFUNC:
            func_str = '%2i' % (int_val)

            for (key, value) in BuiltIn.__cfuncs__.iteritems():
                if value == int_val:
                    func_str += ' (%s)' % (key)
                    break

            sections.append(func_str)
        elif ord(pkt[argc_offset+1]) == SnapConstantsHelper.TC_STRING:
            # Matching the style of the existing args error checking down below
            try:
                str_offset = argc_offset + 1 + int_val
                str_len = ord(pkt[str_offset])
                func_str = str(pkt[str_offset + 1 : str_offset + 1 + str_len ])
                sections.append(func_str)
                crc_loc += str_len+1
            except IndexError:
                sections.append('?')
        else:
            sections.append('?')

        try:
            for i in range(0, ord(pkt[argc_offset])):
                headers.extend(('Arg Type', 'Arg Val'))
                arg_type = ord(pkt[next_arg_offset])
                int_val = ord(pkt[next_arg_offset+1]) << 8 | ord(pkt[next_arg_offset+2])
                int_val = ((int_val + 32768) % 65536) - 32768 #Convert to signed int
                next_arg_offset += 3

                sections.append('%i (%s)' % (arg_type, SnapConstantsHelper.get_arg_type_string(arg_type)))

                if arg_type == SnapConstantsHelper.TC_NONE:
                    sections.append('0 (None)')
                elif arg_type == SnapConstantsHelper.TC_BOOL:
                    sections.append('%i (%s)' % (int_val, bool(int_val)))
                elif arg_type == SnapConstantsHelper.TC_INT:
                    sections.append('%i' % (int_val))
                elif arg_type == SnapConstantsHelper.TC_STRING:
                    str_offset = argc_offset + 1 + int_val
                    str_len = ord(pkt[str_offset])
                    arg_str = pprint.pformat(pkt[str_offset + 1 : str_offset + 1 + str_len ])
                    sections.append(arg_str)
                    crc_loc += str_len+1
                elif arg_type == SnapConstantsHelper.TC_CFUNC:
                    val = "%i" % (int_val)
                    for (key, value) in BuiltIn.__cfuncs__.iteritems():
                        if value == int_val:
                            val += ' (%s)' % (key)
                            break
                    sections.append('%s' % (val))
                else:
                    sections.append('?')

            if crc_loc < len(pkt):
                payload = pkt[argc_offset:]
                if len(payload) >= 3:
                    payload = payload[:len(payload)-2]
                    crc16 = crcUtil.getSnapCrc(payload)
                    headers.append("RPC CRC")
                    sections.append("%04X (%04X)" % (ord(pkt[crc_loc]) << 8 | ord(pkt[crc_loc+1]), crc16))
        except IndexError:
            # If we get here than the packet is malformed
            for i in range(0, len(headers)-len(sections)):
                sections.append('?')
    elif (pkt_pid == PID_DM_PTP or
          (pkt_pid == PID_DM_BCAST and len(pkt) > 12)):
        headers.append('Data')
        sections.append(pprint.pformat(pkt[dm_payload_offset_map[pkt_pid]:])[1:-1]) #Remove single qoutes that pprint adds
    elif pkt_pid == PID_TRACEROUTE:
        headers.extend(('Msg', 'Seq', 'Time'))
        sections.extend(('%c' % (pkt[FWD_UNI_PKT_DATA + TRACEROUTE_MSG_ID]),
                         '%02X' % (ord(pkt[FWD_UNI_PKT_DATA + TRACEROUTE_SEQ])),
                         '%i' % (ord(pkt[FWD_UNI_PKT_DATA + TRACEROUTE_TIME_MSB]) << 8 | ord(pkt[FWD_UNI_PKT_DATA + TRACEROUTE_TIME_LSB])),
                         ))

        record_start = FWD_UNI_PKT_DATA + TRACEROUTE_ADDRESS
        records_len = len(pkt) - record_start
        total_records = records_len / SnapConstantsHelper.TR_RECORD_SIZE

        for i in range(0, total_records):
            headers.extend(('Address', 'LQ'))
            sections.extend(('%02X.%02X.%02X' % (ord(pkt[record_start]), ord(pkt[record_start+1]), ord(pkt[record_start+2])),
                             '%02X' % (ord(pkt[record_start+3])),
                             ))
            record_start += SnapConstantsHelper.TR_RECORD_SIZE

    assert len(headers) == len(sections)
    return (headers, sections)

def get_formatted_output(pkt, decrypted=False):
    (headers, sections) = get_decoded_pkt_sections(pkt, decrypted=decrypted)

    header_str = '|'
    section_str = '|'
    for i in range(0, len(headers)):
        header_str += ' %s |' % (headers[i].center(len(sections[i]), ' '))
        section_str += ' %s |' % (sections[i].center(len(headers[i]), ' '))

    return (header_str, section_str)

def print_decoded_pkt(pkt, show_header=True, decrypted=False):
    (header_str, section_str) = get_formatted_output(pkt, decrypted=decrypted)
    if show_header:
        print header_str
    print section_str