# (c) Copyright 2007-2012 Synapse Wireless, Inc.
"""Constants defined for SNAP and the SNAPpy virtual machine"""


import types
from nvparams import *
from snappy.hooks import *
from sysInfo import SI_NO_ENCRYPTION, SI_AES128_ENCRYPTION, SI_SNAP_ENCRYPTION


# SNAP Protocol IDs
PID_RESERVED      =   0x00
PID_DM_BCAST      =   0x01
PID_DM_PTP        =   0x02
PID_NACK          =   0x03
PID_RPC           =   0x04
PID_ACK           =   0x05
PID_MESH_BCAST    =   0x06
PID_MESH_PTP      =   0x07
PID_MULTICAST_RPC =   0x08
PID_TRACE_ROUTE   =   0x09

PID_ENCRYPTION_MASK = 0x80


# Snappy TypeCodes
TC_NONE   = 0       # val is ignored
TC_INT    = 1       # val is U16, mutable in-place
TC_STRING = 2       # val is: (msb=0) 15-bit offset into Blob, OR
                    #         (msb=1) 15-bit raw pointer into RAM
TC_FUNC   = 3       # val is U16 offset into Blob
TC_CFUNC  = 4       # val is index in pyCFuncTbl[]
TC_BOOL   = 5       # val is U16: nonzero if True, zero if False
TC_TUPLE  = 6       # val is same as string
TC_UNBOUND = 0xFF

TCM_BLOBOFS = 0x40  # mask: val is offset into Blob
TCM_DYNAMIC = 0x80  # mask: val is dynamically allocated


PY_TYPES = (types.NoneType, types.IntType, types.StringType, types.FunctionType, types.BooleanType, types.TupleType)


# VM Status Codes
# Status reported via tellVmStat(status, arg) RPC call - used in Image Management
VM_RESET      = 0
VM_IMG_ERASE  = 1
VM_WR_BLK     = 2    # arg is the last block-offset written
VM_INIT       = 3    # arg is nonzero on success, zero on failure
VM_NVREAD     = 4    # arg is NV object, hi-byte of status is NV ID
VM_INFO       = 10    # arg is Image Name (Pstr), hi-byte of status is LQ
VM_VER        = 6    # #.#.# (date)
VM_NET        = 7    # arg is NetID, hi-byte of status is channel
VM_SPACE      = 8
VM_SCAN       = 9
VM_NAME       = 5
VM_DATA_SPACE = 11
# We aren't doing debugging yet, got ahead of myself
#VM_STEPOVER   = VM_DATA_SPACE + 1
#VM_STEPINTO   = VM_STEPOVER + 1
#VM_DEBUG_RUN  = VM_STEPINTO + 1


TELL_VM_WR_BLK_ERR = -1
TELL_VM_WR_BLK_REFUSED_LOW_POWER = -2
TELL_VM_WR_BLK_UNSUPPORTED = -3

HOOK_DESCRIPTIONS = {0: 'Startup',
                     1: 'GPIN',
                     2: '100ms Timer',
                     3: 'STDIN',
                     4: 'STDOUT',
                     5: 'RPC Sent',
                     6: '10ms Timer',
                     7: '1ms Timer',
                     8: '1s Timer',
                     9: 'SNAPcom Opened',
                     10: 'SNAPcom Closed',
                     HOOK_TRACEROUTE: 'Traceroute',
                     HOOK_SERIAL_CLOSE: 'Serial Closed',
                     HOOK_SERIAL_OPEN: 'Serial Opened',
                     }


# These are Over-The-Air (OTA) offsets
NID_OFS = 0
LEN_OFS = 2
PID_OFS = 3
SEQ_OFS = 10

# According to the S-buffer code it's 123; Doesn't include NID or over the air things like CRC
MAX_PKT_SIZE  = 123 #This includes the length byte however the length byte doesn't include itself

# TODO: Refactor the snot out of everything. This is a quick kludge.
_originalMaxPktSize = MAX_PKT_SIZE
def packetCRCEnabled(enable):
    if enable:
        MAX_PKT_SIZE = _originalMaxPktSize
    else:
        MAX_PKT_SIZE = _originalMaxPktSize - 2

SNAP2_ADDR_LEN = 3 #Length of SNAP version 2 addresses = last 3 bytes of MAC address

# These are in-memory offsets

# Offsets for RPC Packets
RPC_LENGTH_OFS    = 0
RPC_PID_OFS       = 1
RPC_HOPADDR_OFS   = 2  #This is the next hop address for the packet
RPC_TTL_OFS       = 4  #Used in mcast
RPC_SRCADDR_OFS   = 5
RPC_SEQ_OFS       = 8
RPC_DSTADDR_OFS   = 9  # Unicast only!
RPC_ORIG_SRC_OFS  = 12 # Unicast only!
MRPC_ORIG_SRC_OFS = 9  # Multicast only!
RPC_NARGS_OFS     = 15 # Unicast only!
RPC_FUNC_OFS      = 16 # Unicast only!
MRPC_NARGS_OFS    = 12 # Multicast only!
MRPC_FUNC_OFS     = 13 # Multicast only!

# Offsets for ACK Packets

ACK_DST_OFS = 2
ACK_SRC_OFS = 5
ACK_SEQ_OFS = 8

# Offsets for Point-To-Point Mesh Routing Packets (RREP)

MPTP_DST_OFS = 2
MPTP_SRC_OFS = 5
MPTP_TYP_OFS = 9

# Offsets for Broadcast Mesh Routing Packets (RREQ, RERR)

MBST_GRP_OFS = 2
MBST_SRC_OFS = 5

MBST_MSGID_OFS = 9

# Offsets for AES
FWD_UNI_PKT_DATA = 15
UNI_PKT_DATA = 9
MULTI_PKT_DATA = 12
MESH_MULTI_PKT_DATA = 9

# (values for the MSGID field)
MESH_RREQ = 'Q'
MESH_RREP = 'P'
MESH_RERR = 'E'

MBST_MAXHOP_OFS = 10

MBST_TGT_OFS = 11
MBST_TGT_SEQ_OFS = 14
MBST_TGT_DIST_OFS = 16

MBST_ORIG_OFS = 17
MBST_ORIG_SEQ_OFS = 20
MBST_ORIG_DIST_OFS = 22

MBST_EXTRA_OFS = 23

# Offsets for Point-To-Point Data Mode Packets

DMPTP_DST_OFS = 2
DMPTP_SRC_OFS = 5
DMPTP_SEQ_OFS = 8
DMPTP_FINALDST_OFS = 9
DMPTP_ORIG_SRC_OFS = 12


# Offsets for Broadcast Data Mode Packets

DMBAST_TTL_OFS      = 4 #Used in mcast
DMBCAST_ORIG_SRC_OFS = 9 # Multicast only!

# Offsets for "Trace Route" (Point-To-Point) Packets

TR_DST_OFS = 2
TR_SRC_OFS = 5
TR_SEQ_OFS = 8
TR_FINALDST_OFS = 9
TR_ORIG_SRC_OFS = 12
TR_MSGID_OFS = 15
TR_SEQUENCE_OFS = 16
TR_TIMESTAMP_MSB_OFS = 17
TR_TIMESTAMP_LSB_OFS = 18
TR_RECORDS_OFS = 19
# Remainder of message is a variable number of records, each consisting of...
TR_ADDRESS_LEN = 3
TR_LQ_LENGTH = 1
TR_RECORD_SIZE = 4

# (values for the MSGID field)
TR_REQ = 'Q'
TR_REP = 'P'


#See nvparams.py for IDs

NV_PARAM_STRINGS = {NV_MAC_ADDR_ID: "MAC Address",
                    NV_NETWORK_ID: "Network ID",
                    NV_CHANNEL_ID: "Channel",
                    NV_GROUP_INTEREST_MASK_ID: "Multi-cast Processed Groups",
                    NV_GROUP_FORWARDING_MASK_ID: "Multi-cast Forwarded Groups",
                    NV_MFG_DATE_ID: "Manufacturing Date",
                    NV_DEVICE_NAME_ID: "Device Name",
                    NV_SYSTEM_ERROR_ID: "Panic Code",
                    NV_DEVICE_TYPE_ID: "Device Type",
                    NV_FEATURE_BITS_ID: "Feature Bits",
                    NV_DEFAULT_UART_ID: "Default UART",
                    NV_UART_DM_TIMEOUT_ID: "Buffering Timeout",
                    NV_UART_DM_THRESHOLD_ID: "Buffering Threshold",
                    NV_UART_DM_INTERCHAR_ID: "Inter-character Timeout",
                    NV_CARRIER_SENSE_ID: "Carrier Sense",
                    NV_COLLISION_DETECT_ID: "Collision Detect",
                    NV_COLLISION_AVOIDANCE_ID: "Collision Avoidance",
                    NV_SNAP_MAX_RETRIES_ID: "Radio Unicast Retries",
                    NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID: "Mesh Routing Maximum Timeout",
                    NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID: "Mesh Routing Minimum Timeout",
                    NV_MESH_ROUTE_NEW_TIMEOUT_ID: "Mesh Routing New Timeout",
                    NV_MESH_ROUTE_USED_TIMEOUT_ID: "Mesh Routing Used Timeout",
                    NV_MESH_ROUTE_DELETE_TIMEOUT_ID: "Mesh Routing Delete Timeout",
                    NV_MESH_RREQ_TRIES_ID: "Mesh Routing RREQ Retries",
                    NV_MESH_RREQ_WAIT_TIME_ID: "Mesh Routing RREQ Wait Timeout",
                    NV_MESH_INITIAL_HOPLIMIT_ID: "Mesh Routing Initial Hop Limit",
                    NV_MESH_MAX_HOPLIMIT_ID: "Mesh Routing Maximum Hop Limit",
                    NV_MESH_SEQUENCE_NUMBER_ID: "Mesh Sequence Number",
                    NV_MESH_OVERRIDE_ID: "Mesh Override",
                    NV_MESH_PENALTY_LQ_THRESHOLD_ID: "Mesh Penalty LQ Threshold",
                    NV_MESH_REJECTION_LQ_THRESHOLD_ID: "Mesh Rejection LQ Threshold",
                    NV_SNAP_LQ_THRESHOLD_ID: "Radio LQ Threshold",
                    NV_SNAPPY_CRC_ID: "Image CRC",
                    NV_SYS_PLATFORM_ID: "Platform",
                    NV_AES128_ENABLE_ID: "Encryption Type",
                    NV_AES128_KEY_ID: "Encryption Key",
                    NV_LOCKDOWN_FLAGS_ID: "Lockdown Bitmask",
                    NV_ALT_RADIO_TRIM_ID: "Radio Trim",
                    NV_VENDOR_SETTINGS_ID: "Vendor Config Bits",
                    NV_CS_CD_LEVEL_ID: "Radio CS/CD Threshold"}


NV_CAT_DEBUG = 0
NV_CAT_NETWORK = 1
NV_CAT_DEVICE = 2
NV_CAT_MULTICAST = 3
NV_CAT_DATAMODE = 4
NV_CAT_MESH = 5
NV_CAT_SECURITY = 6

# Also order of tabs in dialog
NV_CAT_STRINGS = {NV_CAT_DEBUG: "Debug",
                  NV_CAT_NETWORK: "Network",
                  NV_CAT_DEVICE: "Device",
                  NV_CAT_MULTICAST: "Multi-cast",
                  NV_CAT_DATAMODE: "UART",
                  NV_CAT_MESH: "Mesh",
                  NV_CAT_SECURITY: "Security"}

NV_PARAM_CATS = {NV_MAC_ADDR_ID: NV_CAT_DEVICE,
                 NV_NETWORK_ID: NV_CAT_NETWORK,
                 NV_CHANNEL_ID: NV_CAT_NETWORK,
                 NV_SNAP_MAX_RETRIES_ID: NV_CAT_NETWORK,
                 NV_SNAP_LQ_THRESHOLD_ID: NV_CAT_NETWORK,
                 NV_GROUP_INTEREST_MASK_ID: NV_CAT_MULTICAST,
                 NV_GROUP_FORWARDING_MASK_ID: NV_CAT_MULTICAST,
                 NV_MFG_DATE_ID: NV_CAT_DEBUG,
                 NV_DEVICE_NAME_ID: NV_CAT_DEVICE,
                 NV_SYSTEM_ERROR_ID: NV_CAT_DEBUG,
                 NV_DEVICE_TYPE_ID: NV_CAT_DEVICE,
                 NV_FEATURE_BITS_ID: NV_CAT_DEVICE,
                 NV_DEFAULT_UART_ID: NV_CAT_DEVICE,
                 NV_UART_DM_TIMEOUT_ID: NV_CAT_DATAMODE,
                 NV_UART_DM_THRESHOLD_ID: NV_CAT_DATAMODE,
                 NV_UART_DM_INTERCHAR_ID: NV_CAT_DATAMODE,
                 NV_CARRIER_SENSE_ID: NV_CAT_MULTICAST,
                 NV_COLLISION_DETECT_ID: NV_CAT_MULTICAST,
                 NV_COLLISION_AVOIDANCE_ID: NV_CAT_MULTICAST,
                 NV_MESH_ROUTE_AGE_MAX_TIMEOUT_ID: NV_CAT_MESH,
                 NV_MESH_ROUTE_AGE_MIN_TIMEOUT_ID: NV_CAT_MESH,
                 NV_MESH_ROUTE_NEW_TIMEOUT_ID: NV_CAT_MESH,
                 NV_MESH_ROUTE_USED_TIMEOUT_ID: NV_CAT_MESH,
                 NV_MESH_ROUTE_DELETE_TIMEOUT_ID: NV_CAT_MESH,
                 NV_MESH_RREQ_TRIES_ID: NV_CAT_MESH,
                 NV_MESH_RREQ_WAIT_TIME_ID: NV_CAT_MESH,
                 NV_MESH_INITIAL_HOPLIMIT_ID: NV_CAT_MESH,
                 NV_MESH_MAX_HOPLIMIT_ID: NV_CAT_MESH,
                 NV_MESH_SEQUENCE_NUMBER_ID: NV_CAT_MESH,
                 NV_MESH_OVERRIDE_ID: NV_CAT_MESH,
                 NV_MESH_PENALTY_LQ_THRESHOLD_ID: NV_CAT_MESH,
                 NV_MESH_REJECTION_LQ_THRESHOLD_ID: NV_CAT_MESH,
                 NV_SNAPPY_CRC_ID: NV_CAT_DEBUG,
                 NV_SYS_PLATFORM_ID: NV_CAT_DEVICE,
                 NV_AES128_ENABLE_ID: NV_CAT_SECURITY,
                 NV_AES128_KEY_ID: NV_CAT_SECURITY,
                 NV_LOCKDOWN_FLAGS_ID: NV_CAT_SECURITY,
                 NV_ALT_RADIO_TRIM_ID: NV_CAT_DEVICE,
                 NV_VENDOR_SETTINGS_ID: NV_CAT_DEVICE,
                 NV_CS_CD_LEVEL_ID: NV_CAT_MULTICAST}

NV_SNAP_COMM_CATS = (
                    #NV_CAT_DEVICE, # <- currently disabled because no fields apply
                    NV_CAT_MULTICAST,
                    NV_CAT_MESH,
                    #NV_CAT_SECURITY # <- currently disabled because no fields apply
                    )
NV_SNAP_COMM_DISABLED_IDS = (NV_DEFAULT_UART_ID,
                             NV_SYS_PLATFORM_ID,
                             NV_ALT_RADIO_TRIM_ID,
                             NV_CARRIER_SENSE_ID,
                             NV_COLLISION_AVOIDANCE_ID,
                             NV_COLLISION_DETECT_ID,
                             NV_MESH_PENALTY_LQ_THRESHOLD_ID,
                             NV_MESH_REJECTION_LQ_THRESHOLD_ID,
                             NV_CS_CD_LEVEL_ID,
                             NV_MAC_ADDR_ID, # This could be reinstated if it was made to work
                             NV_DEVICE_NAME_ID, # this could be reinstated once "Portal renaming" fixed (see #4164)
                             NV_DEVICE_TYPE_ID, # this could be reinstated if it was made to do something
                             NV_VENDOR_SETTINGS_ID,
                             NV_FEATURE_BITS_ID,
                             NV_LOCKDOWN_FLAGS_ID,
                             NV_CRYPTO_TYPE, # These next two could be reinstarted if additional work was done
                             NV_CRYPTO_KEY, # (see TRAC #4154)
                             )


ERROR_MESSAGES = {1: "Operation not defined",
                  2: "Unsupported Opcode",
                  3: "Unresolved dependency",
                  4: "Incompatible types",
                  5: "Target not callable",
                  6: "Unbound local variable",
                  7: "Bad global index",
                  8: "Exceeded the maximum block stack",
                  9: "Exceeded the maximum frame stack",
                  10: "Exceeded the maximum object stack",
                  11: "Invalid function argument",
                  12: "Unsubscriptable object",
                  13: "Invalid subscript",
                  14: "Local stack exceeded",
                  15: "Bad constant index",
                  16: "Allocation reference underflow",
                  17: "Allocation reference overflow",
                  18: "Allocation failed",
                  19: "Unsupported type",
                  20: "Exceeded Maximum Packet Size",
                  21: "Exceeded Maximum String Size"}

# SwitchBoard "D"ata "S"ource parameters
DS_NULL = 0
DS_UART1 = 1
DS_UART2 = 2
DS_TRANSPARENT = 3
DS_STDIO = 4
DS_ERROR = 5
DS_PACKET_SERIAL = 6

DEFAULT_SNAP_NETADDR = '\x00\x00\xFF'


# Radios
RADIO_802_15_4 = 0 # what SNAP first ran on
RADIO_NONE = 1
RADIO_900_DSSS = 2
RADIO_868 = 3
RADIO_POWERLINE = 4
RADIO_900_FH = 5 # Frequency Hopping
RADIO_WIFI = 6 # Could be A, B, G, N, etc.
RADIO_SIMULATED = 7
RADIO_802_15_4_900_212B = 8


# CPUs
CPU_MC9S08GT60A = 0
CPU_8051 = 1
CPU_MC9S08QE = 2
CPU_COLDFIRE = 3
CPU_ARM7 = 4
CPU_ATMEGA = 5
CPU_ATMEGA128RFA1 = 5
CPU_SI1000 = 6
CPU_X86 = 7
CPU_UNKNOWN = 8
CPU_SPARC_LEON = 9
CPU_ARM_CORTEX_M3 = 10
# As defined in SysInfo.h: #define CPU_CPUX            11 // For example, as used inside the MSP430 series (Using the vendor's naming scheme here)
CPU_ATMEGA128RFR2 =  12
CPU_ATMEGA1284RFR2 = 13


cpus = [CPU_MC9S08GT60A, CPU_8051, CPU_MC9S08QE, CPU_COLDFIRE, CPU_ARM7, CPU_ATMEGA128RFA1, CPU_SI1000, CPU_X86, CPU_ARM_CORTEX_M3, CPU_ATMEGA128RFR2, CPU_ATMEGA1284RFR2]


# Platform Vendors
SYNAPSE = 0
PANASONIC = 1
FREESCALE = 2
CEL = 3
ATMEL = 4
SILICON_LABS = 5
SIMULATION = 6
PC = 7
YITRAN = 8

# Platform strings
if __debug__:
    platforms = ['RF100', 'RF200', 'RF266', 'RF300', 'RF301', 'SS200', 'SM200', 'SM300', 'SM301', 'SM700', 'PAN4561_SE', 'PAN4555_SE', 'ZIC2410_SE', 'Portal', 'Simulated', 
                 'SM222','SM220']
else:
    platforms = ['RF100', 'RF200', 'RF266', 'RF300', 'RF301', 'SS200', 'SM200', 'SM300', 'SM301', 'SM700', 'PAN4561_SE', 'PAN4555_SE', 'ZIC2410_SE', 'Portal', 'Simulated',
                 'SM222','SM220']


# Platform enumerations
RF_ENGINE = 0 # what SNAP first ran on
PAN4555 = 1 # first port to Panasonic hardware, this is now deprecated
FLEXIS_EVB = 2 # deprecated, we now use a firmware build across a family of hardware
CEL_ZIC2410 = 3 # still valid
MC1321x_EVK = 4 # deprecated
MC1321X = 5 # still valid
ATmega128RFA1 = 6
SNAPCOM = 7
SI100X_DSSS = 8
MC1322X = 9
IT700 = 10 # NOTE! - unfinished port
SI100X_KADEX = SI100X_FHSS = 11
CEL_ZIC2411 = 12 # NOTE! - unfinished port
RF300 = 13 # Si100x with Aux Memory
RF200_PFM = 14 # ATmega128RFA1 with I2C and SPI moved
SM300 = 15 # Si1000 with Aux Memory, in a SMT built by Panasonic
SM301 = 16 # Si1002 with Aux Memory, in a SMT built by Panasonic
SM200_PFM = 17 # ATmega128RFA1 with I2C and SPI moved, inverted PA, in a SMT built by Panasonic
RN_G2C547 = 18 # Wi-Fi part acquired from G2 by Roving Networks - unfinished port!
RF266_PFM = 19 # 20-pin module (not a SNAP Engine) using ATmega128RFA1 chipset (PA moved again!)
STM32W108xB = 20 # First SNAP port to a STMicroelectronics chip
SIMULATED = 23
RF400_PFM = 24 # RF200_PFM using an external 915Mhz radio
SM222_PFM = 25
ATmega128RFR2_PFM = 26
SM220UF1_PFM = 27
ATmega1284RFR2_PFM = 28
PLATFORM_UNKNOWN = 29 # Keep this one at the end of the enumeration!

UNLIMITED_REBOOTS = 0x7FFF

#Interface types
INTF_TYPE_UNKNOWN = 0
INTF_TYPE_802154 = 1
INTF_TYPE_SERIAL = 2
INTF_TYPE_SILABS_USB = 3
INTF_TYPE_ETH = 4
INTF_TYPE_PROXY = 5
INTF_TYPE_SNAPSTICK = 6
INTF_TYPE_BTLE_SERIAL = 7


XCAST_DUPLICATE_SEQUENCE_WINDOW = 3
