# (c) Copyright 2008-2010, Synapse Wireless, Inc.


import logging, os, binascii
log = logging.getLogger("snaplib.%s" % (__name__))

import SnapConstantsHelper


SNAPPY_CHUNK_HEADER_LEN = 28 # Amount of packet taken up RPC packet excluding the data chunk. Leaves 2 bytes for RPC_CRC
SNAPPY_UPLOAD_TIMEOUT = 5 #s
SNAPPY_CHUNK_RETRIES = 3

SNAPPY_PROGRESS_ERASE = 1
SNAPPY_PROGRESS_UPLOAD = 2
SNAPPY_PROGRESS_COMPLETE = 3
SNAPPY_PROGRESS_ERROR = 4
SNAPPY_PROGRESS_CANCELED = 5
SNAPPY_PROGRESS_TIMEOUT = 6
SNAPPY_PROGRESS_WRITE_ERROR = 7
SNAPPY_PROGRESS_WRITE_REFUSED = 8
SNAPPY_PROGRESS_UNSUPPORTED = 9

SNAPPY_MCAST_ERASE_DELAY = 5
SNAPPY_MCAST_CHUNK_DELAY = 4
SNAPPY_MCAST_GROUP = "\x00\x03"
SNAPPY_MCAST_TTL = 2


class AlreadyInProgressError(Exception):
    pass


class SnappyUpload(object):
    def __init__(self, netAddr, image, rpcFunc, eventScheduler, finishedCallback=None, mcast=None):
        self.finishedCallbacks = []
        self.progressCallbacks = []
        self._currSnappyChunk = 0
        self._snappyChunkRetries = 0
        self.rpcFunc = rpcFunc
        self.eventScheduler = eventScheduler
        self._uploadImage = self._origUploadImage = image
        self.netAddr = netAddr
        self.canceled = False
        self._snappyProgressDlg = None
        self._snappyUploadTimer = None
        self._timeoutEvent = None
        self._inErase = True
        self._waiting_on_reboot = False
        self.mcast = mcast

        # Determine the maximum size of a chunk that can be written
        # This is necessary for cases where the MAX_PKT_SIZE has been altered
        self.chunk_size = SnapConstantsHelper.MAX_PKT_SIZE - SNAPPY_CHUNK_HEADER_LEN

        if finishedCallback is not None:
            self.registerFinishedCallback(finishedCallback)

        if self.mcast:
            deferred = self.mcast(SNAPPY_MCAST_GROUP, SNAPPY_MCAST_TTL, 'eraseImage')
        else:
            deferred = self.rpcFunc(netAddr, 'eraseImage')
        deferred.addCallbacks(self.onEraseImageSent)

    def cancel(self):
        if self._timeoutEvent:
            self._timeoutEvent.Stop()
        self.canceled = True
        self.tellFinished(SNAPPY_PROGRESS_CANCELED)

    def onEraseImageSent(self, success):
        if success:
            if self.mcast:
                self.eventScheduler.schedule(SNAPPY_MCAST_ERASE_DELAY, self.snappyEraseSuccess)
            else:
                self._timeoutEvent = self.eventScheduler.schedule(SNAPPY_UPLOAD_TIMEOUT, self.onUploadTimeout)
                assert self._timeoutEvent
            log.debug("Expecting to upload %i bytes" % (len(self._uploadImage) + self.chunk_size))
        else:
            self.onUploadTimeout()

    def on_reboot_callback(self):
        if self._timeoutEvent:
            self._timeoutEvent.Stop()
        self._waiting_on_reboot = False
        self.eventScheduler.schedule(0, self.tellFinished, SNAPPY_PROGRESS_COMPLETE)

    def onSnappyChunkSent(self, success):
        if success:
            if self.mcast:
                self.eventScheduler.schedule(SNAPPY_MCAST_CHUNK_DELAY, self.snappyChunkAck, self._currSnappyChunk)
            else:
                self._timeoutEvent = self.eventScheduler.schedule(SNAPPY_UPLOAD_TIMEOUT, self.onUploadTimeout)
                assert self._timeoutEvent
        else:
            self.onUploadTimeout()

    def onUploadTimeout(self):
        if self._snappyChunkRetries > SNAPPY_CHUNK_RETRIES:
            if self._waiting_on_reboot:
                log.error("Node %04X did not notify us of rebooting" % (int(binascii.hexlify(self.netAddr), 16)))
                self.on_reboot_callback()
            else:
                self.inSnappyUpload = False

                log.error("Node %04X failed uploading SNAPpy image" % (int(binascii.hexlify(self.netAddr), 16)))

                if self._snappyProgressDlg:
                    self._snappyProgressDlg.Destroy()
                    self._snappyProgressDlg = None

                self.tellFinished(SNAPPY_PROGRESS_TIMEOUT)
        else:
            self._snappyChunkRetries += 1
            if self._inErase and not self.canceled:
                deferred = self.rpcFunc(self.netAddr, 'eraseImage')
                deferred.addCallbacks(self.onEraseImageSent)
            elif self._waiting_on_reboot:
                self.rpcFunc(self.netAddr, 'callback', 'su_recvd_reboot', 'reboot')
                self._timeoutEvent = self.eventScheduler.schedule(SNAPPY_UPLOAD_TIMEOUT, self.onUploadTimeout)
            elif not self.canceled:
                if __debug__:
                    log.debug("Retrying chunk %i" % (self._currSnappyChunk))
                self.sendNextSnappyChunk()
                self.tellProgress(self._currSnappyChunk)
        return False

    def registerFinishedCallback(self, callback):
        assert callable(callback)
        if callable(callback):
            self.finishedCallbacks.append(callback)

    def registerProgressCallback(self, callback):
        assert callable(callback)
        if callable(callback):
            self.progressCallbacks.append(callback)

    def sendNextSnappyChunk(self):
        if self.canceled:
            return

        if self.mcast:
            deferred = self.mcast(SNAPPY_MCAST_GROUP, SNAPPY_MCAST_TTL, 'writeChunk', self._currSnappyChunk, self._uploadImage[:self.chunk_size])
        else:
            deferred = self.rpcFunc(self.netAddr, 'writeChunk', self._currSnappyChunk, self._uploadImage[:self.chunk_size])

        deferred.addCallbacks(self.onSnappyChunkSent)
        if self._timeoutEvent:
            self._timeoutEvent.Stop()

    def snappyChunkAck(self, start):
        if self.canceled:
            return

        if (start & 0xFFFF) == self._currSnappyChunk:
            self._uploadImage = self._uploadImage[self.chunk_size:]
            self._currSnappyChunk += self.chunk_size
            self._snappyChunkRetries = 0
            self.tellProgress(start & 0xFFFF)
        elif ((start == SnapConstantsHelper.TELL_VM_WR_BLK_ERR) or
              (start == SnapConstantsHelper.TELL_VM_WR_BLK_REFUSED_LOW_POWER) or
              (start == SnapConstantsHelper.TELL_VM_WR_BLK_UNSUPPORTED)):
            if start == SnapConstantsHelper.TELL_VM_WR_BLK_ERR:
                log.error("Node %04X failed uploading SNAPpy image, could not write chunk" % (int(binascii.hexlify(self.netAddr), 16)))
                self.tellFinished(SNAPPY_PROGRESS_WRITE_ERROR)
            elif start == SnapConstantsHelper.TELL_VM_WR_BLK_UNSUPPORTED:
                log.error("Node %04X refused uploading SNAPpy image, uploading unsupported" % (int(binascii.hexlify(self.netAddr), 16)))
                self.tellFinished(SNAPPY_PROGRESS_UNSUPPORTED)
            else:
                log.error("Node %04X refused uploading SNAPpy image, operating voltage too low" % (int(binascii.hexlify(self.netAddr), 16)))
                self.tellFinished(SNAPPY_PROGRESS_WRITE_REFUSED)

            if self._snappyProgressDlg:
                self._snappyProgressDlg.Destroy()
                self._snappyProgressDlg = None

            self._timeoutEvent.Stop()
            return
        else:
            log.debug("%04X: Got ACK for chunk %i expected %i" % (int(binascii.hexlify(self.netAddr), 16), start & 0xFFFF, self._currSnappyChunk))
            if self._snappyChunkRetries > SNAPPY_CHUNK_RETRIES:
                self.tellFinished(SNAPPY_PROGRESS_ERROR)
                self._snappyChunkRetries += 1
                self._timeoutEvent.Stop()
                return
            elif self._currSnappyChunk > (start & 0xFFFF):
                return #Let normal timeout occurr

        if self._timeoutEvent:
            self._timeoutEvent.Stop()

        if len(self._uploadImage) > 0:
            self.sendNextSnappyChunk()
        else:
            if self.mcast:
                self.mcast(SNAPPY_MCAST_GROUP, SNAPPY_MCAST_TTL, 'reboot')
            else:
                self.rpcFunc(self.netAddr, 'callback', 'su_recvd_reboot', 'reboot')
                self._timeoutEvent = self.eventScheduler.schedule(SNAPPY_UPLOAD_TIMEOUT, self.onUploadTimeout)
                self._waiting_on_reboot = True

    def snappyEraseSuccess(self):
        self._inErase = False
        self.sendNextSnappyChunk()

    def tellFinished(self, result):
        for callback in self.finishedCallbacks:
            try:
                callback(self, result)
            except:
                log.exception("An error occurred while notifiying a callback of a SNAPpy upload being finished")

    def tellProgress(self, chunk):
        for callback in self.progressCallbacks:
            try:
                callback(self, chunk)
            except:
                log.exception("An error occurred while notifiying a callback of a SNAPpy Upload progress")


class SnappyUploadMgr(object):
    def __init__(self, eventScheduler, rpcFunc, supportsCompile=False):
        self._uploads = {}
        self.eventScheduler = eventScheduler
        self.rpcFunc = rpcFunc
        self.supportsCompile = supportsCompile
        self._addCallbacks = []
        self._massUploadCallback = None
        self._massUploadAddrs = []
        self._massUploadResults = []

    def _nextMassUpload(self, uploadImage, rpcFunc):
        netAddr = self._massUploadAddrs.pop(0)
        self._uploads[netAddr] = SnappyUpload(netAddr, uploadImage, rpcFunc, self.eventScheduler, self.onFinished)
        #Do we need to call people registered in addCallbacks?

    def onFinished(self, upload, result):
        assert isinstance(upload, SnappyUpload)
        addr = self._uploads.pop(upload.netAddr, 0)
        if self._massUploadCallback is not None:
            self._massUploadResults.append(result)
            if len(self._massUploadAddrs):
                self._nextMassUpload(upload._origUploadImage, upload.rpcFunc)
            else:
                self._massUploadCallback(self._massUploadResults)
                self._massUploadCallback = None

    def onTellVmStat(self, remote_addr, status, arg):
        status_code = status & 0xFF

        if status_code == SnapConstantsHelper.VM_IMG_ERASE:
            upload = self._uploads.get(remote_addr, None)
            if upload:
                upload.snappyEraseSuccess()
        elif status_code == SnapConstantsHelper.VM_WR_BLK:
            upload = self._uploads.get(remote_addr, None)
            if upload:
                upload.snappyChunkAck(arg)

    def on_recvd_reboot(self, remote_addr):
        upload = self._uploads.get(remote_addr, None)
        if upload:
            upload.on_reboot_callback()

    def registerAddCallback(self, callback):
        assert callable(callback)
        if callable(callback):
            self._addCallbacks.append(callback)

    def startUpload(self, netAddr, image, rpcFunc=None):
        if self._uploads.has_key(netAddr) or netAddr in self._massUploadAddrs:
            raise AlreadyInProgressError("There is currently a SNAPpy upload already in progress")

        if rpcFunc is None:
            rpcFunc = self.rpcFunc

        self._uploads[netAddr] = SnappyUpload(netAddr, image, rpcFunc, self.eventScheduler, self.onFinished)
        for callback in self._addCallbacks:
            try:
                callback(self._uploads[netAddr])
            except:
                log.exception("An error occurred while notifying a callback of a new SNAPpy upload")

        return self._uploads[netAddr]

    def startMcastUpload(self, image, mcastRpcFunc):
        if self._uploads.has_key(SNAPPY_MCAST_GROUP) or SNAPPY_MCAST_GROUP in self._massUploadAddrs:
            raise AlreadyInProgressError("There is currently a SNAPpy upload already in progress")

        self._uploads[SNAPPY_MCAST_GROUP] = SnappyUpload(SNAPPY_MCAST_GROUP, image, None, self.eventScheduler, self.onFinished, mcast=mcastRpcFunc)
        for callback in self._addCallbacks:
            try:
                callback(self._uploads[SNAPPY_MCAST_GROUP])
            except:
                log.exception("An error occurred while notifying a callback of a new SNAPpy upload")

        return self._uploads[SNAPPY_MCAST_GROUP]

    def massUpload(self, addrs, image, callback):
        if len(self._massUploadAddrs) > 0 or len(self._uploads):
            raise AlreadyInProgressError("There is currently a SNAPpy upload already in progress")

        if rpcFunc is None:
            rpcFunc = self.rpcFunc

        self._massUploadAddrs = addrs
        self._nextMassUpload(image, rpcFunc)

        self._massUploadResults = []
        self._massUploadCallback = callback
