# (c) Copyright 2009-2012 Synapse Wireless, Inc.
"""
Functions to encode and decode SNAP "Trace Route" packets
"""
__docformat__ = "plaintext en"


import logging, binascii
log = logging.getLogger("snaplib.%s" % (__name__))

import SnapConstantsHelper
import SnapEncryptions
import sequence # This is for the SNAP sequence number

class TraceRouteEncodeError(Exception):
    def __init__(self, descr):
        self.descr = descr


# Data object
class TraceRouteRecord():
    def __init__(self, address, linkQuality):
        self.address = address
        self.linkQuality = linkQuality

    def __str__(self):
        return "%s:%i" % (binascii.hexlify(self.address), self.linkQuality)


class TraceRouteDescriptor(object):
    tr_sequence = 0 # this is for the "app layer" sequence number that is ALSO in the TR packet

    def next_tr_sequence(self):
        TraceRouteDescriptor.tr_sequence += 1
        if TraceRouteDescriptor.tr_sequence > 255:
            TraceRouteDescriptor.tr_sequence = 0
        return TraceRouteDescriptor.tr_sequence

    def __init__(self, crypto=None):
        self.srcAddr = None    # 3-byte string
        self.dstAddr = None    # 3-byte string
        self.msgId = ''
        self.sequence = None # This is the SNAP sequence number that all packets have
        self.elapsed = 0
        self.records = []
        self.isMulticast = False # Used to help re-use some of the RPC code
        self.final_dst_addr = None
        self.orig_src_addr = None
        self.crypto = crypto
        # The intent of tr_sequence was that you would use it to ensure the TR 'R'eply
        # packet got associated to the correct TR re'Q'uest packet. As of 01/30/2012,
        # this field has never been actually used, and up through 01/30/2012 was
        # always 0.
        self.tr_sequence = None # This is like an "app layer" sequence number

    def __str__(self):
        if self.tr_sequence is None:
            self.tr_sequence = self.next_tr_sequence()
        rtn_val = "%s%d=" % (self.msgId, self.tr_sequence)
        for record in self.records:
            rtn_val += "%s;" % (record)
        return "%s %ims" % (rtn_val, self.elapsed)

    def get_encoded_elapsed(self):
        return "%c%c" % ((self.elapsed>>8) & 0xFF, self.elapsed & 0xFF)

    encoded_elapsed = property(get_encoded_elapsed)

    def get_encoded_data(self):
        return encode(self, crypto=self.crypto)

    encoded_data = property(get_encoded_data)

    #def get_dst_addr(self):
        #return self.dstAddr

    #final_dst_addr = get(get_dst_addr)

def decode(pkt, crypto=None):
    """
    Decodes a Trace Route packet from a SNAP device

    pkt -- raw data packet from packet serial
    """
    desc = TraceRouteDescriptor(crypto=crypto)
    pkt = SnapEncryptions.decode_pkt(pkt, crypto)

    desc.sequence = ord(pkt[SnapConstantsHelper.TR_SEQ_OFS])
    msgId = pkt[SnapConstantsHelper.TR_MSGID_OFS]
    if msgId != SnapConstantsHelper.TR_REQ and msgId != SnapConstantsHelper.TR_REP:
        return None
    desc.msgId = msgId
    desc.tr_sequence = ord(pkt[SnapConstantsHelper.TR_SEQUENCE_OFS])
    desc.elapsed = ord(pkt[SnapConstantsHelper.TR_TIMESTAMP_MSB_OFS])
    desc.elapsed <<= 8
    desc.elapsed += ord(pkt[SnapConstantsHelper.TR_TIMESTAMP_LSB_OFS])


    desc.srcAddr = pkt[SnapConstantsHelper.RPC_SRCADDR_OFS:SnapConstantsHelper.RPC_SRCADDR_OFS+3]
    desc.orig_src_addr = pkt[SnapConstantsHelper.TR_ORIG_SRC_OFS:SnapConstantsHelper.TR_ORIG_SRC_OFS+3]

    desc.dstAddr = pkt[SnapConstantsHelper.RPC_HOPADDR_OFS:SnapConstantsHelper.RPC_HOPADDR_OFS+3]
    desc.final_dst_addr = pkt[SnapConstantsHelper.TR_FINALDST_OFS:SnapConstantsHelper.TR_FINALDST_OFS+3]

    remainingLength = len(pkt) - SnapConstantsHelper.TR_RECORDS_OFS
    remainingRecords = remainingLength / SnapConstantsHelper.TR_RECORD_SIZE

    index = SnapConstantsHelper.TR_RECORDS_OFS
    while remainingRecords > 0:
        address = pkt[index:index+SnapConstantsHelper.TR_ADDRESS_LEN]
        index += SnapConstantsHelper.TR_ADDRESS_LEN
        linkQuality = ord(pkt[index])
        index += 1

        record = TraceRouteRecord(address, linkQuality)
        desc.records.append(record)

        remainingRecords -= 1

    # The node this message just came through also counts as part of the route...
    address = pkt[SnapConstantsHelper.TR_SRC_OFS:SnapConstantsHelper.TR_SEQ_OFS]
    linkQuality = 0 # Packet Serial does not have a real LQ
    record = TraceRouteRecord(address, linkQuality)
    desc.records.append(record)

    return desc

def encode(desc, crypto=None):
    """
    Encode a Trace Route request packet for output to a SNAP device

    desc -- A TraceRouteDescriptor object
    """
    assert isinstance(desc, TraceRouteDescriptor)
    final_dst_addr = desc.final_dst_addr
    if final_dst_addr is None:
        final_dst_addr = desc.dstAddr

    pid = SnapConstantsHelper.PID_TRACE_ROUTE

    if desc.sequence is not None:
        txSequence = desc.sequence
    else:
        txSequence = sequence.get_number(desc.srcAddr)

    if desc.tr_sequence is not None:
        trSequence = desc.tr_sequence
    else:
        trSequence = desc.next_tr_sequence()

    # PID | DstAddr | SrcAddr | Sequence | Final DstAddr | Orig SrcAddr | Msg Id | Sequence | Elapsed Time
    pkt = "%c%s%s%c%s%s%s%c%s" % (pid, desc.dstAddr, desc.srcAddr, txSequence, final_dst_addr, desc.orig_src_addr, desc.msgId, trSequence, desc.encoded_elapsed)

    for record in desc.records:
        assert isinstance(record, TraceRouteRecord)
        pkt += record.address
        pkt += "%c" % (record.linkQuality)

    if len(pkt) > SnapConstantsHelper.MAX_PKT_SIZE:
        raise TraceRouteEncodeError('The data is too large to encode (%d bytes, max is %d)' % (len(pkt), SnapConstantsHelper.MAX_PKT_SIZE))
    else:
        try:
            pkt = "%c%s" % (len(pkt), pkt) # Prepend length
        except OverflowError:
            raise TraceRouteEncodeError("Packet is greater than 255 bytes")

    return SnapEncryptions.encode_pkt(pkt, crypto)
