# (c) Copyright 2007, Synapse


import RpcCodec


class VirtualEndpointProxyFunction(object):
  """
  Represents a function on a remote SNAP device
  """
  def __init__(self, rpcDesc, rpcPktQueue):
    """
    rpcDesc -- The RpcDescriptor object to pass to the RpcCodec when called
    rpcPktQueue -- The RpcPktQueue to notify when the remote function is called
    """
    self.rpcPktQueue = rpcPktQueue
    self.rpcDesc = rpcDesc
    assert isinstance(rpcDesc, RpcCodec.RpcDescriptor)

  def __call__(self, *args, **kwargs):
    if self.rpcPktQueue:
      self.rpcDesc.args = args
      urgent = False
      if kwargs.has_key('__urgent') and kwargs['__urgent']:
        urgent = True
      deferred = self.rpcPktQueue.rpcReady(self.rpcDesc, urgent=urgent)
      if __debug__:
        deferred.addCallbacks(self.rpcDesc.OnSend)
      return deferred


class VirtualEndpointNetAddr(object):
  """
  Represents a remote SNAP device at the specified address
  """
  def __init__(self, srcAddr, addr, rpcPktQueue, expectsResponse=False):
    """
    srcAddr -- The source network address
    addr -- The network address that this object represents
    rpcPktQueue -- The PacketQueue to notify when the remote function is called
    expectsResponse  -- A response is expected
    """
    self.rpcDesc = RpcCodec.RpcDescriptor(expectsResponse=expectsResponse)
    self.rpcDesc.srcAddr = srcAddr
    self.rpcDesc.dstAddr = addr
    self.rpcPktQueue = rpcPktQueue

  def __getattr__(self, name):
    self.rpcDesc.funcName = name
    return VirtualEndpointProxyFunction(self.rpcDesc, self.rpcPktQueue)


class VirtualEndpointMulticast(object):
  """
  Represents a SNAP multicast group
  """
  def __init__(self, srcAddr, group, ttl, rpcPktQueue, expectsResponse=False):
    """
    srcAddr -- The source network address
    group -- The destination multicast group
    ttl -- The time to live
    rpcPktQueue -- The PacketQueue to notify when the remote function is called
    expectsResponse  -- A response is expected
    """
    self.rpcDesc = RpcCodec.RpcDescriptor(expectsResponse=expectsResponse)
    self.rpcDesc.srcAddr = srcAddr
    self.rpcDesc.dstAddr = group
    self.rpcDesc.ttl = ttl
    self.rpcPktQueue = rpcPktQueue
    self.rpcDesc.isMulticast = True

  def __getattr__(self, name):
    self.rpcDesc.funcName = name
    return VirtualEndpointProxyFunction(self.rpcDesc, self.rpcPktQueue)


class VirtualEndpoint(object):
  """
  A virtual SNAP endpoint that connects between Portal and a SNAP packet serial driver
  """
  def __init__(self, srcAddr, callSink, rpcPktQueue):
    """
    srcAddr -- The source network address of this endpoint
    callSink -- An object that will get called with responses,
                and have the property remoteAddr set
    rpcPktQueue -- The PacketQueue to notify when the remote function is called
    """
    self.srcAddr = srcAddr
    self.callSink = callSink
    self.rpcPktQueue = rpcPktQueue
    self.connectedNetAddr = None

  def netAddr(self, addr=None, srcAddr=None, expectsResponse=False):
    """Returns a VirtualEndpointNetAddr object for the requested addr"""
    if addr is None:
      if self.connectedNetAddr is None:
        raise Exception("The Virtual Endpoint has not been connected to a node")
      addr = self.connectedNetAddr
    if srcAddr is None:
      srcAddr = self.srcAddr
    return VirtualEndpointNetAddr(srcAddr, addr, self.rpcPktQueue, expectsResponse=expectsResponse)

  def multicast(self, group, ttl=5, srcAddr=None, expectsResponse=False):
    if srcAddr is None:
      srcAddr = self.srcAddr
    return VirtualEndpointMulticast(srcAddr, group, ttl, self.rpcPktQueue, expectsResponse=expectsResponse)

class _testCallSink(object):
  def __init__(self):
    object.__init__(self)


if __name__ == '__main__':
  #This really only tests that the syntax works like we expect
  virtEp = VirtualEndpoint(1, _testCallSink(), None)
  virtEp.netAddr('\x12\x34\x56').foo(1, 2, 3, 5)
  virtEp.multicast('\xFF\xFF').foo(1, 2, 3, 5)
