# (c) Copyright 2010, Synapse Wireless, Inc.
"""
Cipher stealing en/decryption using AES-128.

greg.luterman@synapse-wireless.com

Message handling depends on the length of the plaintext. Plaintext can either be:

1. Less than 1 block long
   Plaintext must be padded with zeros to a whole block length before
   encryption.

2. A multiple of whole blocks
   Each block may be encrypted without any special processing.

3. More than one block, but not evenly divisible into blocks.
   We must use ciphertext stealing to handle the leftover bytes.
"""


from Crypto.Cipher import AES


class CipherStealingAES128(object):
    """Wrapper around pycrypto's AES128 to include cipher stealing."""
    key   = None      # The actual key
    ivKey = None      # Pre-processed data to expedite work

    def __init__(self, key, enabled=True):
        # This is *NOT* the en/decryption key. This key is used to calculate the
        # IV only.
        self.key = key
        self.enabled = enabled

    def _calcKeys(self, sequenceNumber):
        # Calculate an IV based on sequenceNumber
        iv = self.ivKey.encrypt(chr(sequenceNumber) * AES.block_size)

        # Create the decryption key
        cryptKey = AES.new(self._key, AES.MODE_CBC, iv)

        return iv, cryptKey

    def decrypt(self, sequenceNumber, cipherText):
        iv, decryptKey = self._calcKeys(sequenceNumber)

        ctLen = len(cipherText)
        if ctLen < AES.block_size:
            # We should never get something that is less then the block size
            return cipherText

        # Do we have an uneven number of blocks?
        if ctLen % AES.block_size:
            # Yes, we will need to use ciphertext stealing

            # We can decrypt normally up to the second to last whole block
            normalLen = ((ctLen / AES.block_size) - 1) * AES.block_size
            if normalLen:
                plainText = decryptKey.decrypt(cipherText[:normalLen])
            else:
                plainText = ""

            # Perform an AES pass on the last whole block
            block = cipherText[normalLen : normalLen + AES.block_size]
            block = self.ivKey.decrypt(block)

            # Ciphertext steal the last few bytes
            lastBytes = cipherText[normalLen + AES.block_size : ]
            lbLen = len(lastBytes)
            newLastBytes = ""
            for i in xrange(lbLen):
                newLastBytes += chr(ord(lastBytes[i]) ^ ord(block[i]))
            block = lastBytes + block[lbLen:]

            # Complete decryption
            plainText += decryptKey.decrypt(block) + newLastBytes

        else:
            # No, we can just decrypt it as is
            plainText = decryptKey.decrypt(cipherText)

        return plainText

    def encrypt(self, sequenceNumber, plainText):
        iv, encryptKey = self._calcKeys(sequenceNumber)

        # Is the plaintext less than one block in length?
        ptLen = len(plainText)
        if ptLen < AES.block_size:
            # Yes, we will need to pad the text to make it a whole block long
            plainText += "\0" * (AES.block_size - ptLen)

        # Do we have an uneven number of blocks?
        ptLen = len(plainText)
        if ptLen % AES.block_size:
            # Yes, we will need to use ciphertext stealing

            # We can encrypt normally up to the second to last whole block
            normalLen = ((ptLen / AES.block_size) - 1) * AES.block_size
            if normalLen:
                cipherText = encryptKey.encrypt(plainText[:normalLen])
            else:
                cipherText = ""

            # Encrypt the last whole block
            block = plainText[normalLen : normalLen + AES.block_size]
            block = encryptKey.encrypt(block)

            # Ciphertext steal the last few bytes
            lastBytes = plainText[normalLen + AES.block_size : ]
            lbLen = len(lastBytes)
            newLastBytes = block[:lbLen]
            newPrefix = ""
            for i in xrange(lbLen):
                newPrefix += chr(ord(lastBytes[i]) ^ ord(block[i]))
            block = newPrefix + block[lbLen:]

            # Perform an AES pass on the last whole block
            block = self.ivKey.encrypt(block)
            cipherText += block + newLastBytes

        else:
            # No, we can just encrypt it as is
            cipherText = encryptKey.encrypt(plainText)

        return cipherText

    def get_has_valid_key(self):
        """Is AES enabled and do we have a valid key"""
        return self.enabled and hasattr(self, '_key') and len(self._key) == AES.block_size

    def set_key(self, value):
        if value and len(value) == AES.block_size:
            self._key = value
            self.ivKey = AES.new(value, AES.MODE_ECB)
        else:
            self._key = ""

    key = property(None, set_key)
    has_valid_key = property(get_has_valid_key)
