# (c) Copyright 2012, Synapse Wireless Inc.


import hashlib
import bz2
import logging
import binascii

from snaplib import EventCallbacks


TIMEOUT_VAL = 5 # seconds
BLOCK_SIZE = 96

log = logging.getLogger(__name__)


def send_data(rpc_func, dest_addr, source, identifier, compress=True):
    if hasattr(source, 'read'):
        data = f.read()
    else:
        data = source
    # This library just handles the transfer it is up to you to actually give us data to send

    if compress:
        data = bz2.compress(data)
   
    hasher = hashlib.md5()
    hasher.update(data)
    data_len = len(data)
    while data:
        rpc_func(dest_addr, 'd', data[:BLOCK_SIZE])
        data = data[BLOCK_SIZE:]

    # Signal completion of xfer
    rpc_func(dest_addr, 'd_complete', identifier, hasher.digest(), True)
    log.debug("Sent %i bytes of '%s' data with MD5 of %s" % (data_len, identifier, hasher.hexdigest()))


class DataReceiver(EventCallbacks.Registrar):
    def __init__(self, scheduler, source_addr_func):
        EventCallbacks.Registrar.__init__(self, events=['timeout', 'progress', 'decompress_error']) # We add events on-demand
        self.data = {} # For now we will hold incoming data in memory per sender address
        self.scheduler = scheduler
        self._timeout_events = {} # {snap_addr: event}
        self.timeouts_count = 0
        self.source_addr_func = source_addr_func

    def init_transfer(self, source_addr):
        self.data[source_addr] = ''
        event = self._timeout_events.get(source_addr, None)
        if event:
            event.Stop()

    def notify_on_transfer(self, identifier, callback):
        if not self._registeredEvents.has_key(identifier):
            self._add_event_to_registrar(identifier)
        self.register_callback(identifier, callback)

    def on_data(self, data):
        source_addr = self.source_addr_func()

        total_data = self.data.setdefault(source_addr, '')
        self.data[source_addr] = total_data + data

        event = self._timeout_events.get(source_addr, None)
        if event:
            event.Stop()

        self._timeout_events[source_addr] = self.scheduler.schedule(TIMEOUT_VAL, self._on_timeout, source_addr)
        self._notify_callbacks('progress', len(self.data[source_addr]), None)

    def on_data_complete(self, identifier, md5, compressed=True):
        source_addr = self.source_addr_func()

        if self._registeredEvents.has_key(identifier):
            data = None
            equal = False
            hasher = hashlib.md5()
            hasher.update(self.data.setdefault(source_addr, ''))

            if hasher.digest() == md5:
                equal = True

            if compressed:
                try:
                    data = bz2.decompress(self.data.setdefault(source_addr, ''))
                except IOError:
                    log.debug("Unable to decompress data; MD5 equality is %s" % (equal,))
                    self._notify_callbacks('decompress_error', self.data, source_addr)
            else:
                data = self.data.setdefault(source_addr, None)

            if data is not None:
                log.debug("Received %i bytes of '%s' data with from %s MD5 of %s" % (len(data), identifier, binascii.hexlify(source_addr), hasher.hexdigest()))
                self._notify_callbacks(identifier, data, equal, source_addr)

        self.init_transfer(source_addr)

    def _on_timeout(self, source_addr):
        self.init_transfer(source_addr)
        self.timeouts_count += 1
        self._notify_callbacks('timeout', source_addr)
