from array import array
from struct import pack, unpack, calcsize


# see /usr/include/elf.h for more details

EI_NIDENT = 16
EI_CLASS = 4
EI_DATA = 5
EI_VERSION = 6
EI_MAG = "\x7fELF"

ELFCLASS32 = 1
ELFCLASS64 = 2

EV_CURRENT = '\x01'


VARS = [
  ('Half', 'e_type'),         # Object file type
  ('Half', 'e_machine'),      # Architecture
  ('Word', 'e_version'),      # Object file version
  ('Addr', 'e_entry'),        # Entry point virtual address
  ('Off',  'e_phoff'),        # Program header table file offset
  ('Off',  'e_shoff'),        # Section header table file offset
  ('Word', 'e_flags'),        # Processor-specific flags
  ('Half', 'e_ehsize'),       # ELF header size in bytes
  ('Half', 'e_phentsize'),    # Program header table entry size
  ('Half', 'e_phnum'),        # Program header table entry count
  ('Half', 'e_shentsize'),    # Section header table entry size
  ('Half', 'e_shnum'),        # Section header table entry count
  ('Half', 'e_shstrndx')      # Section header string table index
]

Elf32_Shdr = [
  ('Word', 'sh_name'),
  ('Word', 'sh_type'),
  ('Word', 'sh_flags'),
  ('Addr', 'sh_addr'),
  ('Off',  'sh_offset'),
  ('Off',  'sh_size'),
  ('Word', 'sh_link'),
  ('Word', 'sh_info'),
  ('Word', 'sh_addralign'),
  ('Word', 'sh_entsize'),
]

class ElfException(Exception):
    pass


class Sections(object):
    def __init__(self, header):
        assert isinstance(header, Header)
        self.header = header
        self.sections = []
        self.section_names = []

    def read(self, file_handle):
        assert isinstance(file_handle, file)
        fmt = self.header.endian + ''.join(self.header.typmap[typ] for typ, name in Elf32_Shdr)

        file_handle.seek(self.header.header['e_shoff'])
        for i in range(0, self.header.header['e_shnum']):
            sect = file_handle.read(self.header.header['e_shentsize'])
            self.sections.append(unpack(fmt, sect))

        file_handle.seek(self.sections[self.header.header['e_shstrndx']][4])
        string_data = file_handle.read(self.sections[self.header.header['e_shstrndx']][5])
        self.section_names = string_data.split('\x00') #They are null terminated strings
        #for (sh_name, sh_type, sh_flags, sh_addr, sh_offset, sh_size, sh_link, sh_info, sh_addralign, sh_entsize) in self.sections:
            #pass
        pass

    def get(self, file_handle, name):
        assert isinstance(file_handle, file)

        index = 0
        for section_name in self.section_names:
            if name == section_name:
                file_handle.seek(self.sections[index][4])
                return file_handle.read(self.sections[index][5])
            index += 1

class Header(object):
    def __init__(self):
        self.ident = {}
        self.header = {}
        self.typmap = {}
        self.endian = ''

    def read(self, fi):
        #Start by reading in EI_NIDENT number of bytes
        self.ident_data = fi.read(EI_NIDENT)
        #Check the "magic number"
        if self.ident_data[:len(EI_MAG)] != EI_MAG:
            raise ElfException("File header does not contain magic number")

        if self.ident_data[EI_CLASS] == '\x01':
            ptr = 'L'
        elif self.ident_data[EI_CLASS] == '\x02':
            ptr = 'Q'
        else:
            raise ElfException("Invalid ELF class found")

        if self.ident_data[EI_DATA] == '\x01':
            self.endian = '<'
        elif self.ident_data[EI_DATA] == '\x02':
            self.endian = '>'
        else:
            raise ElfException("Invalid data encoding found")

        if self.ident_data[EI_VERSION] != EV_CURRENT:
            raise ElfException("Invalid version found")

        self.typmap = {'Half': 'H', 'Word': 'L', 'Addr': ptr, 'Off': ptr}

        self.fmt = self.endian + ''.join(self.typmap[typ] for typ, name in VARS)
        self.header_data = fi.read( calcsize(self.fmt) )

        self.set_attrs()

    def set_attrs(self):
        #self.e_ident = array('c', self.header_data)
        tmp = unpack(self.fmt, self.header_data)
        for i, (typ, name) in enumerate(VARS):
            #setattr(self, name, tmp[i])
            self.header[name] = tmp[i]

    def display(self):
        print 'ELF Header:'
        #print '    %s' % self.e_ident
        #for (name, val) in self.header.items():
        for typ, name in VARS:
            #val = getattr(self, name)
            val = self.header[name]
            if typ == 'Addr':
                val = hex(val)
            print '    %-11s = %12s (%s)' % (name, val, typ)


def get_code_string(path):
    eh = Header()

    fi = open(path, 'rb')
    eh.read(fi)

    #eh.display()

    sects = Sections(eh)
    sects.read(fi)
    data = sects.get(fi, '.text')

    fi.close()

    return data


if __name__ == '__main__':
    import sys
    import pprint

    pprint.pprint(get_code_string(sys.argv[1]))
