"""
Copyright (c) 2010, Synapse Wireless Inc.

Pure Python Crypto Libraries

Licensed under the Eclipse Public License, Version 1.0
http://www.eclipse.org/legal/epl-v10.html
"""


class ARC4(object):
    def __init__(self, key):
        if len(key) <= 0 or len(key) > 256:
            raise ValueError("The key must be between at least 1 character in length and no more than 256")
        self.__key = [ord(x) for x in key]

    def encrypt(self, data):
        # Originally ported from C code at http://en.wikipedia.org/wiki/RC4
        output = []
        s = range(256)
        key_length = len(self.__key)

        # key-scheduling algorithm (KSA)
        j = 0
        for i in range(256):
            j = (j + self.__key[i % key_length] + s[i]) & 255
            s[i], s[j] = s[j], s[i] # swap

        # pseudo-random generation algorithm (PRGA)
        i = j = 0
        for character in data:
            i = (i + 1) & 255
            j = (j + s[i]) & 255
            s[i], s[j] = s[j], s[i]
            output.append(chr(ord(character) ^ (s[(s[i] + s[j]) & 255])))

        return ''.join(output)

    decrypt = encrypt # ARC4 is a symmetric encryption algorithm


if __name__ == '__main__':
    tests = [("Key", "Plaintext"),
             ("Wiki", "pedia"),
             ("Secret", "Attack at dawn")]

    for key, plaintext in tests:
        algo = ARC4(key)
        encrypted_text = algo.encrypt(plaintext)
        #print encrypted_text

        unencrypted_text = algo.decrypt(encrypted_text)
        #print unencrypted_text

        assert unencrypted_text == plaintext

    import timeit

    test1 = timeit.Timer(setup="""
import ppcrypto
algo = ppcrypto.ARC4("Secret")
""", stmt="""
encrypted_text = algo.encrypt("Attack at dawn")
d=algo.decrypt(encrypted_text)
assert d == "Attack at dawn"
""")

    print "%.2f usec/pass" % (1000000 * test1.timeit(number=10000)/10000)
    #print test1.timeit(number=1)
