#!/usr/bin/python

"""
Packs the contents of an SREG file with Huffman encoding and returns it as a
string appropriate for writing into a Kaibab module.
"""

# The output string is treated as one huge number, with the LSB of the first
# byte as the LSB of the stream. The MSB of the last byte in the string is the
# sequence's MSB. Multi-bit values are encoded consistently in this scheme.
# Their LSB is written into the stream first and their MSB is written last.

# The following sequences are written into the stream. There are no gaps between
# each entry:
#     Identifying constant
#     Timestamp
#     Huffman definition table
#     Zero or more data blocks
#     Execution entry point
#     EOF sentinel

# The identifying constant is:
#     16-bit: 0xFFFE
# Since the maximum allowed script length (which would be located here if this
# portion of Flash holds a script instead of core code) is 0xFFFF, this large
# number is a good (quick) indication that valid code is present.

# The timestamp is the Unix-time of the upload:
#     32-bit: timestamp
# This value will not "roll over" until 2106.

# The Huffman definition table is list of 1 to 511 Huffman definitions. Each
# definition is either a branch:
#     1-bit: 0
# or a leaf:
#     1-bit: 1
#     8-bits: decoded value
# The size of the Huffman table is not explicitly listed. It is, however, simple
# to tell where the table ends by inspection of the content. If the unpacker
# detects more than 511 entries (this would happen, for example, if the data was
# all 0's), then the compressed file is invalid and should be ignored.

# Each data block contains the following:
#     2-bit:  0
#     24-bit: start address for the unpacked data to be stored
#     24-bit: number of expanded bytes in this block (n)
#     packed data (n entries of various length)

# The execution entry point contains the following:
#     2-bit: 2
#     24-bit: execution entry point

# The EOF sentinel is as follows:
#     2-bit: 3
#     0 to 7 bits of padding (0's) to byte-align the stream
#     32-bit: CRC-32 of all bytes in the stream up to this point (including the
#             padding)

from binascii import crc32, unhexlify
from time import time
import struct

# Constants:
COMPRESSED_ID   = 0xFFFE
UNCOMPRESSED_ID = 0xFFFD
INT_WIDTH       = 24
MARK_DATA_BLOCK = 0
MARK_ENTRY_PT   = 2
MARK_EOF        = 3

class Node(object):
    "A node in the Huffman tree."
    Char  = None
    Total = None
    Shift = None
    Add   = None
    Children = None
    def __init__(self, X, Y):
        """
Create a node. X and Y can be other nodes or a character and number of
occurrences, respectively.
"""
        if isinstance(X, Node):
            self.Children, self.Total = [X, Y], X.Total + Y.Total
        else:
            self.Char, self.Total = X, Y

    def AssignCodes(self, Dict={}, Shift=0, Add=0):
        """
        This function must be called on the table root once all the nodes have
        been added. It returns a character-indexed dictionary to speed encoding.
        """
        self.Shift, self.Add = Shift, Add
        if self.Children:
            self.Children[0].AssignCodes(Dict, Shift + 1, Add)
            self.Children[1].AssignCodes(Dict, Shift + 1, Add | (1 << Shift))
        else:
            Dict[self.Char] = self
        return Dict

    def DumpCodes(self, String):
        "Write the node and children to a BitString."
        if self.Children:
            String.Dump(1, 0) # Branch
            self.Children[0].DumpCodes(String)
            self.Children[1].DumpCodes(String)
        else:
            String.Dump(1, 1) # Leaf
            String.Dump(8, ord(self.Char))

    def DumpCode(self, String):
        "Write the compressed code (not the node) to a BitString."
        String.Dump(self.Shift, self.Add)

class BitString(object):
    "A string that is built bit(s) at a time."
    String = None
    Accum  = None
    Bits   = None
    def __init__(self):
        self.String, self.Accum, self.Bits = "", 0, 0

    def Dump(self, Shift, Add=0):
        "Add one or more bits to the string."
        self.Accum = self.Accum | (Add << self.Bits)
        self.Bits += Shift
        while self.Bits >= 8:
            self.String += chr(self.Accum & 0xFF)
            self.Bits -= 8
            self.Accum >>= 8

    def PadAlign(self):
        "Add enough bits (0's) to byte-align the stream."
        if self.Bits:
            self.Dump(8 - self.Bits)

    def __str__(self):
        return self.String

class HuffmanCompressor(object):
    OccDict  = None
    XrefDict = None
    def __init__(self):
        self.OccDict = {}

    def AccumOccurrences(self, BinaryString):
        """
        Count and accummulate the occurrences of each value in the string. Can
        be called on multiple strings before dumping the compressor into a
        string.
        """
        for C in BinaryString:
            self.OccDict[C] = self.OccDict.get(C, 0) + 1

    def DumpTableToString(self, String):
        "Output the Huffman definition table to the given BitString."
        # Create a list of all the characters that appear in the file
        NodeList = [Node(C, N) for C, N in self.OccDict.iteritems()]

        # Combine pairs until only a single "root" exists
        while len(NodeList) > 1:
            # Sort the list to find the two smallest items
            NodeList.sort(lambda X, Y:  cmp(Y.Total, X.Total))

            # Replace the two smallest nodes with a branch
            NodeList.append(Node(NodeList.pop(), NodeList.pop()))

        # The root is the lone, remaining node.
        Table = NodeList[0]

        # Build a cross-reference table.
        self.XrefDict = Table.AssignCodes()

        # Dump the table into the string
        Table.DumpCodes(String)

    def CompressToString(self, Addr, InString, OutString):
        "Add a data record to the string."
        OutString.Dump(2, MARK_DATA_BLOCK)
        OutString.Dump(INT_WIDTH, Addr)
        OutString.Dump(INT_WIDTH, len(InString))
        for C in InString:
            self.XrefDict[C].DumpCode(OutString)

    def DumpEPToString(self, ExecPt, String):
        "Add an execution entry point to the string."
        String.Dump(2, MARK_ENTRY_PT)
        String.Dump(INT_WIDTH, ExecPt)

    def DumpEOFToString(self, String):
        String.Dump(2, MARK_EOF)
        String.PadAlign()
        #print 'crc=%08X' % (crc32(str(String)) & 0xFFFFFFFF)
        String.Dump(32, crc32(str(String)) & 0xFFFFFFFF)

def ParseSRec(String):
    Blocks = []
    BlockAddr, BlockData, ExecAddr = None, "", None

    # Loop over lines in file
    for S in String.split("\n"):
        if S:
            # Parse line
            Type = S[:2]
            Content = unhexlify(S[4:-2])

            # Header record?
            if Type == "S0":
                pass

            # Data record?
            elif Type == "S2":
                # Parse the address and data
                Addr = struct.unpack(">L", "\0" + Content[:3])[0]
                Data = Content[3:]

                # First data record?
                if BlockAddr is None:
                    # Yes, we'll start here
                    BlockAddr = Addr

                # Are we starting a new block?
                if Addr != (BlockAddr + len(BlockData)):
                    # Yes, save the block we had been working on and re-init
                    Blocks.append((BlockAddr, BlockData))
                    BlockAddr, BlockData = Addr, ""

                # Append record
                BlockData += Data

            # Entry point record?
            elif Type == "S8":
                # Parse the entry point
                ExecAddr = struct.unpack(">L", "\0" + Content)[0]

            # Other?
            else:
                raise LookupError("Unknown command type: %r" % Type)

    # Append any incomplete block
    Blocks.append((BlockAddr, BlockData))

    return Blocks, ExecAddr

def CompressSRec(String):
    "Compress the contents of an SRec file."
    Blocks, ExecAddr = ParseSRec(String)

    # Build a compressor over all the data blocks
    Compressor = HuffmanCompressor()
    for Addr, Data in Blocks:
        Compressor.AccumOccurrences(Data)

    # Save the ID constant and timestamp
    String = BitString()
    String.Dump(16, COMPRESSED_ID)
    String.Dump(32, int(time()))

    # Create the Huffman table
    Compressor.DumpTableToString(String)

    # Compress the blocks
    for Addr, Data in Blocks:
        Compressor.CompressToString(Addr, Data, String)

    # Append the entry point
    Compressor.DumpEPToString(ExecAddr, String)

    # Add a CRC
    Compressor.DumpEOFToString(String)

    # Return the results
    return str(String)

def EncodeSRec(String):
    "Encode the contents of an SRec file without compressing it."
    Blocks, ExecAddr = ParseSRec(String)

    # Save the ID constant and timestamp
    RetVal = struct.pack("<HL", UNCOMPRESSED_ID, int(time()))

    # Save the data
    for Addr, Data in Blocks:
        RetVal += chr(MARK_DATA_BLOCK) + struct.pack("<L", Addr)[:3] + \
            struct.pack("<L", len(Data))[:3] + Data

    # Append the entry point
    RetVal += chr(MARK_ENTRY_PT) + struct.pack("<L", ExecAddr)[:3]

    # Add a CRC
    RetVal += chr(MARK_EOF)
    #print 'crc=%08X' % (crc32(str(RetVal)) & 0xFFFFFFFF)
    RetVal += struct.pack("<L", crc32(str(RetVal)) & 0xFFFFFFFF)

    # Return the results
    return RetVal

if __name__ == "__main__":
    SRec = """
S00D0000626C696E6B2E7372656307
S21440803C4800A0E3800480E38018A0E3101080E59D
S21440804CF42AA0E3902F82E3E037A0E3083080E5E3
S21440805C403DA0E30C3080E50230A0E10000A0E1FA
S21440806C013053E2FCFFFF1A001080E50230A0E11D
S21440807C0000A0E1013053E2FCFFFF1AF1FFFFEADB
S80440803CAF
"""
    Output = CompressSRec(SRec)
    print "".join(["\\x%02X" % ord(i) for i in Output])
