"""Generic over the air upgrader for SNAP node firmware

This content originally lived in /flasher/OtaUpgrader/ but was split out into a common file
that could be shared between SNAP Connect and Portal.

See the OtaUpgrader docstring for instructions on how to subclass and tie into other interfaces.

"""
from __future__ import with_statement
import binascii
import os
try:
    import bz2
except (ImportError, SystemError):
    #For platforms where the bz2 package isn't built, decompressed files can still be used
    bz2 = None
import logging
import SnapConstantsHelper

from SnappyUploader import AlreadyInProgressError

log = logging.getLogger("OtaUpgrader")

##########################
# OTA Upgrader Constants
##########################
OTA_RPC_HEADER_LEN = 28 # Amount of packet taken up RPC packet excluding the data chunk. Leaves 2 bytes for RPC_CRC
OTA_UPLOAD_TIMEOUT = 5 # timeout in seconds
OTA_CHUNK_RETRIES = 3

#Progress enumerations
OTA_PROGRESS_ERASE = 1
OTA_PROGRESS_UPLOAD = 2
OTA_PROGRESS_COMPLETE = 3
OTA_PROGRESS_ERROR = 4
OTA_PROGRESS_CANCELED = 5
OTA_PROGRESS_TIMEOUT = 6
OTA_PROGRESS_WRITE_ERROR = 7
OTA_PROGRESS_WRITE_REFUSED = 8
OTA_PROGRESS_UNSUPPORTED = 9

#Error enumerations
OTA_ERROR_BAD_FILENAME = "OTA_ERROR_BAD_FILENAME"
OTA_ERROR_IMCOMPATIBLE_IMAGE = "OTA_ERROR_IMCOMPATIBLE_IMAGE"
OTA_ERROR_UNABLE_TO_OPEN = "OTA_ERROR_UNABLE_TO_OPEN"
OTA_ERROR_FLASHER = "OTA_ERROR_FLASHER"

class OtaUpgrader(object):
    """Generic OTA Firmware upgrader for a SNAP node

    Usage:
        First, you must register layous for different CPUs in the initialization function. Ex:
            def __init__(self):
                super(OtaUpgrader, self).__init__()

                #Supported Layouts (CPU, Layout Class)
                self.register_layout(SnapConstantsHelper.CPU_ARM7, MC1322x_Layout.MC1322xLayout)
                ...

        The following functions must be implemented by a child class:
            start_upload(self, snaplayout)
            handle_error(self, error_type, arg)
            upload_progress(self, value, meta=None, max_progress=None)
            upload_complete(self, addr, result, meta=None)

        The following functions must be set as callbacks for receiving RPC:
            'tellVmStat': OtaUpgrader.onTellVmStat
            'getBankSpace': OtaUpgrader.onGetBankSpace
    """
    def __init__(self, scheduler, rpc_func):
        """Construct an OtaUpgrader

        Args:
            scheduler: Used for scheduling events that should happen asynchronously
                       Expected signature is scheduler.schedule(time, func, *args)
            rpc_func: Function used for making RPC calls.
                      Expected signature is rpc_func(addr, function, *args)
                      This RPC function should return a deferred object that calls
                      the callback when the message has actually sent.
        """
        self.scheduler = scheduler
        self.rpc_func = rpc_func

        self._layouts = {} # Registered Layouts {cpu: layout_class}
        self._uploads = {} # Current Uploads { address: OtaUpgrade }
        self._addCallbacks = [] # Callbacks for when an upgrade is started

    def register_layout(self, cpu, layout):
        """Register a snaplayout for a given CPU and platform

        Args:
            cpu: The CPU type (See SnapConstantsHelper.CPU_*)
            snaplayout: The snaplayout class to see (Constructor should accept decompressed firmware image)
        """
        self._layouts[cpu] = layout

    def registerAddCallback(self, callback):
        """Register a callback for after an upgrade has started"""
        assert callable(callback)
        if callable(callback):
            self._addCallbacks.append(callback)

    def upgrade_firmware(self, addr, file_path, cpu=None, platform=None):
        """Start an over the air firmware upgrade.

        Programming may take up to 2 minutes after the transfer completes.
        Do not remove power to the device during this time.

        Args:
            addr: The address of the node to upgrade
            file_path: The path to the firmware image
            cpu: The CPU type (See SnapConstantsHelper.CPU_*)
            platform: The platform (See SnapConstantsHelper.*_PFM)

        If the CPU and Platform are None then the node type will not be checked

        Returns:
            Returns a firmware upgrade object.
            You may call cancel() on this object to attempt to cancel the firmware upgrade
            If the update cannot be canceled this will return None, this could happen if
            multiple nodes are being upgraded.
        """
        try:

            file_name = os.path.split(file_path)[1]
            file_ext = os.path.splitext(file_path)[1]

            # Determine the cpu and platform based on the filename
            file_cpu, file_platform = self._get_file_architecture(file_name)

            if file_cpu is None:
                #Filename did not include enough info to determine the platform
                self.handle_error(addr, OTA_ERROR_BAD_FILENAME, file_name)
                return

            #if caller does not provide which cpu or platform to use, don't check it
            if (cpu is not None) or (platform is not None):
                #Always check the CPU matches, but only check the platform if we need to for the file type
                if (file_cpu != cpu) or (file_platform is not None and file_platform != platform):
                    #Image File is not compatible with provided architecture for the Node
                    self.handle_error(addr, OTA_ERROR_IMCOMPATIBLE_IMAGE, file_name)
                    return

            #Based on the cpu, set the correct snaplayout from the decompressed file contents
            layout_class = self._get_layout(file_cpu)

            #We know the file is valid, so retrieve the contents
            try:
                with open(file_path, 'rb') as f:
                    file_text = f.read()

                #*.sfi files are compressed
                if file_ext == '.sfi':
                    file_text = bz2.decompress(file_text)
            except:
                self.handle_error(addr, OTA_ERROR_UNABLE_TO_OPEN, file_path)
                return

            #Parse the file contents based on the layout for different processors
            try:
                layout = layout_class(file_text)
            except Exception, ex:
                self.handle_error(addr, OTA_ERROR_FLASHER, ex)
                return

            return self.start_upload(addr, layout) #This user function should call startOtaUpgrade

        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception, ex:
            self.handle_error(addr, OTA_ERROR_FLASHER, ex)
            return

    ###############################################
    # Functions that must be tied to RPC responses
    ###############################################
    def onGetBankSpace(self, remoteAddr, freeSpace):
        """Callback that must be called on receipt of 'getBankSpace' rpc"""
        assert(remoteAddr in self._uploads)
        self._uploads[remoteAddr].setBankSpace(freeSpace & 0xFFFF)

    def onTellVmStat(self, remoteAddr, status, arg):
        """Callback that must be registered for receipt of 'tellVmStat' rpc"""
        statusCode = status & 0xFF

        if statusCode == SnapConstantsHelper.VM_IMG_ERASE:
            try:
                upload = self._uploads.get(remoteAddr)
                upload.eraseSuccess()
            except:
                pass
        elif statusCode == SnapConstantsHelper.VM_WR_BLK:
            try:
                upload = self._uploads.get(remoteAddr)
                upload.chunkAck(arg)
            except:
                pass

    def startOtaUpgrade(self, addr, layout, meta=None):
        """Called by start_upload from the child class to begin an upgrade

        Args:
            addr: The address of the Node that is starting an upgrade
            snaplayout: The snaplayout class passed to start_upload
            meta: Additional meta data related to this upgrade. This will be passed to upload_complete if provided
        """
        if addr in self._uploads:
            raise AlreadyInProgressError("There is currently a firmware upload already in progress")

        upgrade = OtaUpgrade(addr, layout, self.rpc_func, self.scheduler, self, self._upload_complete, meta=meta)

        self._uploads[addr] = upgrade

        for callback in self._addCallbacks:
            try:
                callback(upgrade)
            except:
                log.exception("An error occurred while notifying a callback of a new firmware upload")

        return upgrade

    #########################################################
    # Functions that must be implemented in the base class
    #########################################################
    def start_upload(self, addr, layout):
        """Kick off the upload process. This function needs to call startOtaUpgrade 1 or more times

        Return:
            This function should return an object with a 'cancel()' function that can be used to stop the upload
        """
        raise NotImplementedError("start_upload must call startOtaUpgrade in the OtaUpgrader derived class")

    def handle_error(self, error_type, arg=None):
        """Handle errors that occur during over the air upgrades

        This must be implemented in subclasses of the OtaUpgrader

        The following errors can occur with the given argument:
            handle_error(OTA_ERROR_BAD_FILENAME, filename) - The image filename did not include the platform (Ex: RF200 or SS200)
            handle_error(OTA_ERROR_IMCOMPATIBLE_IMAGE, filename) - The image File is not compatible with provided architecture for the Node
            handle_error(OTA_ERROR_UNABLE_TO_OPEN, filepath) - An error occurred while attempting to open the given file
            handle_error(OTA_ERROR_FLASHER, exception) - An error occurred in the flasher
        """
        raise NotImplementedError("Error handling must be performed by the OtaUpgrader derived class")

    def upload_complete(self, addr, result, meta=None):
        """Called on completion of an OTA firmware upgrade

        Args:
            addr: Address of the node that has had the upload completed
            result: Result code. See OTA_PROGRESS_* enumerations in generic_upgrader.py
            meta: Additional meta that was passed to startOtaUpgrade

        The following results could appear:
            OTA_PROGRESS_CANCELED
            OTA_PROGRESS_TIMEOUT
            OTA_PROGRESS_ERROR
            OTA_PROGRESS_WRITE_ERROR
            OTA_PROGRESS_WRITE_REFUSED
            OTA_PROGRESS_UNSUPPORTED
            OTA_PROGRESS_COMPLETE
        """
        raise NotImplementedError("The upload completion callback must be implemented by the OtaUpgrader derived class")

    def upload_progress(self, addr, value, meta=None, max_progress=None):
        """Called on upload progress

        Args:
            addr: Address of the node that is being upgraded
            value: The current progress of the upload.
                   This will be called initially when the progress
            meta: Any meta data associated with the upload in progress
            max_progress: On the initial call to upload progress this will hold the value for an upload at 100% progress
        """
        raise NotImplementedError("The upload progress callback must be implemented by the OtaUpgrader derived class")

    #####################
    # Private functions
    #####################
    def _get_file_architecture(self, filename):
        """Return the (file_cpu, file_platform) for the given filename

        If the type cannot be determined then it will be set to None
        """
        file_cpu, file_platform = None, None
        if "MC1322x" in filename or "SM700" in filename:
            file_cpu = SnapConstantsHelper.CPU_ARM7
        elif "STM32W108" in filename:
            file_cpu = SnapConstantsHelper.CPU_ARM_CORTEX_M3
        # ATmega family
        elif "SM222" in filename:
            file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFR2
            file_platform = SnapConstantsHelper.SM222_PFM
        elif "ATmega128RFR2" in filename:
            file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFR2
            file_platform = SnapConstantsHelper.ATmega128RFR2_PFM
        elif "ATmega1284RFR2" in filename:
            file_cpu = SnapConstantsHelper.CPU_ATMEGA1284RFR2
            file_platform = SnapConstantsHelper.ATmega1284RFR2_PFM
        elif "SM220" in filename:
            file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFA1
            file_platform = SnapConstantsHelper.SM220UF1_PFM
        elif "SS200" in filename or "RF200" in filename or "RF266" in filename or "SM200" in filename or "ATmega128RFA1" in filename:
            file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFA1
        # TODO: @khc consider full specification for platforms in release 2.5. Present implementation allows this last elif files to be loaded
        #            onto any platform that has the CPU_ATMEGA128RFA1 which could cause it to become non-responsive after an OTA upgrade.
        #elif "SM200" in filename:
            #file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFA1
            #file_platform = SnapConstantsHelper.SM200_PFM#elif "SM220" in filename:
        #elif "SS200" in filename or "RF200" in filename:
            #file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFA1
            #file_platform = SnapConstantsHelper.RF200_PFM
        #elif "RF266" in filename:
            #file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFA1
            #file_platform = SnapConstantsHelper.RF266_PFM
        #elif "ATmega128RFA1" in filename:
            #file_cpu = SnapConstantsHelper.CPU_ATMEGA128RFA1
            #file_platform = SnapConstantsHelper.ATmega128RFA1

        return file_cpu, file_platform

    def _get_layout(self, cpu):
        """Return the correct snaplayout class given the cpu

        This depends on having layouts registered with the register_layout() function

        Args:
             cpu: The CPU type (See SnapConstantsHelper.CPU_*)

        Returns:
            The snaplayout class for the given CPU or None if a suitable snaplayout was not registered
        """
        return self._layouts.get(cpu, None)

    def _upload_complete(self, addr, result, meta=None):
        """Called on upload complete

        This peforms cleanup before calling into the subclass version of the complete function
        """
        self._uploads.pop(addr, 0)
        self.upload_complete(addr, result, meta)


class OtaUpgrade(object):
    """Represent and perform an OTA firmware upgrade"""
    def __init__(self, addr, layout, rpc_func, scheduler, upgrader, finished_callback=None, meta=None):
        """Construct an Ota Upgrade object to perform the upgrade

        Args:
            addr: The SNAP address of the Node to upgrade
            snaplayout: The snaplayout class for the given architecture
            rpc_func: The function used for sending RPCs and returning a deferred object
            scheduler: The scheduler for events
            finished_callback: Callback called on completion with (addr, result, meta=None)
            meta: Any additional meta data to pass to the finished callback
        """
        self.addr = addr
        self.layout = layout
        self.rpc_func = rpc_func
        self.scheduler = scheduler
        self.upgrader = upgrader
        self.meta = meta

        # Determine the maximum size of a chunk that can be written
        # This is necessary for cases where the MAX_PKT_SIZE has been altered
        self.chunk_size = SnapConstantsHelper.MAX_PKT_SIZE - OTA_RPC_HEADER_LEN

        self._finished_callbacks = []

        if finished_callback is not None:
            self.registerFinishedCallback(finished_callback)

        self.canceled = False

        self._currChunk = 0
        self._chunkRetries = 0
        self._max_progress = 0 #The largest number that can be reported by upload_progress

        self._timeoutEvent = None
        self._inGetInfo = True
        self._inErase = False
        self._image = None

        #Kick off the upgrade by getting Bank Free Space (16)
        deferred = self.rpc_func(self.addr, 'callback', 'getBankSpace', 'getInfo', 16)
        deferred.addCallbacks(self.onGetBankSpaceSent)

    def cancel(self):
        """Cancel an OTA upgrade"""
        if self._timeoutEvent:
            self._timeoutEvent.Stop()

        if not self.canceled:
            # Prevent multiple calls from cancel to keep activating the progress callback
            self.canceled = True
            self.tellFinished(OTA_PROGRESS_CANCELED)

    def onUploadTimeout(self):
        """Called after a stage of the upload has timed out"""
        if self._chunkRetries > OTA_CHUNK_RETRIES:
            log.error("Node %s failed uploading firmware image" % (binascii.hexlify(self.addr)))
            self.tellFinished(OTA_PROGRESS_TIMEOUT)
        else:
            self._chunkRetries += 1
            if self.canceled:
                pass
            elif self._inGetInfo:
                deferred = self.rpc_func(self.addr, 'callback',
                    'getBankSpace', 'getInfo', 16)
                deferred.addCallbacks(self.onGetBankSpaceSent)
            elif self._inErase:
                deferred = self.rpc_func(self.addr, 'eraseImage')
                deferred.addCallbacks(self.onEraseImageSent)
            elif not self.canceled:
                if __debug__:
                    log.debug("Retrying chunk %i" % (self._currChunk))
                self.sendNextChunk()
                self.upgrader.upload_progress(self.addr, (self._currChunk / self.chunk_size) + 2, self.meta)
        return False

    def onGetBankSpaceSent(self, success):
        """Callback after construction has SENT the RPC to get the bank space"""
        if success:
            log.debug("Fetching free bank space with getInfo")
            self._timeoutEvent = self.scheduler.schedule(OTA_UPLOAD_TIMEOUT,
                self.onUploadTimeout)
            assert self._timeoutEvent
        else:
            self.onUploadTimeout()

    def registerFinishedCallback(self, callback):
        """Register a callback in the format of (addr, result, meta=None) for completion of upgrade"""
        assert callable(callback)
        if callable(callback):
            self._finished_callbacks.append(callback)

    def setBankSpace(self, freeSpace):
        """Called by the manager after the node makes an RPC to tell the bank space"""
        if self._timeoutEvent:
            self._timeoutEvent.Stop()

        self._inGetInfo = False
        self._inErase = True
        self._image = self.layout.getContents(freeSpace)

        image_size = len(self._image)

        self._max_progress = ((image_size - 1) / self.chunk_size) + 4
        self.upgrader.upload_progress(self.addr, 1, self.meta, self._max_progress)

        log.debug("Expecting to upload %i bytes" % image_size)

        deferred = self.rpc_func(self.addr, 'eraseImage')
        deferred.addCallbacks(self.onEraseImageSent)

    def onEraseImageSent(self, success):
        """Callback after sending RPC to erase the image on the node"""
        if success:
            self._timeoutEvent = self.scheduler.schedule(OTA_UPLOAD_TIMEOUT,
                self.onUploadTimeout)
            assert self._timeoutEvent
        else:
            self.onUploadTimeout()

    def eraseSuccess(self):
        """Called after an image has been successfully erased from the managers onTellVmStat function"""
        self._inErase = False

        self.upgrader.upload_progress(self.addr, 2, self.meta)

        self.sendNextChunk()

    def sendNextChunk(self):
        """Send the next chunk of firmware data"""
        if self.canceled:
            return

        deferred = self.rpc_func(self.addr, 'writeChunk', self._currChunk, self._image[:self.chunk_size])
        deferred.addCallbacks(self.onChunkSent)

        if self._timeoutEvent:
            self._timeoutEvent.Stop()

    def onChunkSent(self, success):
        """Called when RPC for writing a chunk has sent"""
        if success:
            self._timeoutEvent = self.scheduler.schedule(OTA_UPLOAD_TIMEOUT,
                self.onUploadTimeout)
            assert self._timeoutEvent
        else:
            self.onUploadTimeout()

    def chunkAck(self, start):
        """Called when the manager receives a tellVmStat to ack a chunck receipt"""
        if self.canceled:
            return

        posStart = start & 0xFFFF
        if posStart == self._currChunk:
            self._image = self._image[self.chunk_size:]
            self._currChunk += self.chunk_size
            self._chunkRetries = 0
            self.upgrader.upload_progress(self.addr, (posStart / self.chunk_size) + 2, self.meta)
        elif ((start == SnapConstantsHelper.TELL_VM_WR_BLK_ERR) or
              (start == SnapConstantsHelper.TELL_VM_WR_BLK_REFUSED_LOW_POWER) or
              (start == SnapConstantsHelper.TELL_VM_WR_BLK_UNSUPPORTED)):
            if start == SnapConstantsHelper.TELL_VM_WR_BLK_ERR:
                log.error("Node %04X failed uploading firmware image, could not write chunk" %
                    (int(binascii.hexlify(self.addr), 16)))
                self.tellFinished(OTA_PROGRESS_WRITE_ERROR)
            elif start == SnapConstantsHelper.TELL_VM_WR_BLK_UNSUPPORTED:
                log.error("Node %04X refused uploading firmware image, uploading unsupported" %
                    (int(binascii.hexlify(self.addr), 16)))
                self.tellFinished(OTA_PROGRESS_UNSUPPORTED)
            else:
                log.error("Node %04X refused uploading firmware image, operating voltage too low" %
                    (int(binascii.hexlify(self.addr), 16)))
                self.tellFinished(OTA_PROGRESS_WRITE_REFUSED)

            self._timeoutEvent.Stop()
            return
        else:
            log.debug("%04X: Got ACK for chunk %i expected %i" %
                (int(binascii.hexlify(self.addr), 16), posStart,
                self._currChunk))
            if self._chunkRetries > OTA_CHUNK_RETRIES:
                self.tellFinished(OTA_PROGRESS_ERROR)
                self._chunkRetries += 1
                self._timeoutEvent.Stop()
                return
            elif self._currChunk > posStart:
                return #Let normal timeout occurr

        if self._timeoutEvent:
            self._timeoutEvent.Stop()

        if len(self._image) > 0:
            self.sendNextChunk()
        else:
            self.rpc_func(self.addr, 'reboot')
            self.tellFinished(OTA_PROGRESS_COMPLETE)

    def tellFinished(self, result):
        """On completion, call all registered callbacks with the result"""
        self.upgrader.upload_progress(self.addr, self._max_progress, self.meta)

        for callback in self._finished_callbacks:
            try:
                callback(self.addr, result, self.meta)
            except:
                log.exception("An error occurred while notifiying a callback of a firmware upload being finished")
